package org.quintet.backbeat.dataportal;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.quintet.backbeat.BackBeatException;
import org.quintet.backbeat.io.InputConduit;
import org.quintet.backbeat.io.OutputConduit;


import java.util.logging.Logger;
import java.util.logging.Level;

/**
 * A <code>ActiveDataPortal</code> optionally requests data from an external data source and
 * optionally returns the data to the <code>ActiveDataPoral</code>'s client.
 * An <code>ActivateDataPortal</code>'s activities are determined by its <code>InputConduit</code>
 * and its <code>OutputConduit</code>; if <code>outputConduit</code> is not null then
 * <code>ActiveDataSource</code> transmits a data request, and if <code>inputConduit</code>
 * is not null then <code>ActiveDataSource</code> garners a response and returns it to
 * the caller.
 *
 * @author Peter Thaggard
 * @version     1
 * @see org.quintet.backbeat.dataportal.DataPortal
 * @see org.quintet.backbeat.io.InputConduit
 * @see org.quintet.backbeat.io.OutputConduit
 */

public class ActiveDataPortal
    implements DataPortal {


  /**All diagnostic messages are sent to this Logger**/
  private static Logger log = Logger.getLogger("org.quintet.backbeat.dataportal");

  /**Responses are gathered via <code>inputConduit</code>.  If <code>inputConduit</code>
   * is null then <code>ActiveDataPortal</code> makes no attempt to garner a response.
   */

  protected InputConduit inputConduit;

  /**Requests are transmitted via <code>outputConduit</code>.  If <code>outputConduit</code>
   * is null then <code>ActiveDataPortal</code> makes no attempt to transmit a request.
   */

  protected OutputConduit outputConduit;



  private static void closeResources(InitialContext ic){
    if(ic != null){
      try{
        ic.close();
      }catch(Throwable eatIt){}
    }
  }

  /**
   * Construct a new <code>ActiveDataPortal</code>.  Construction is largely the act of
   * constructing an <code>InputConduit</code> and an <code>OutputConduit</code>, both
   * of which can be null.  Both conduit's compositions are derived from JNDI entries, so
   * this constructor uses <code>InitialContext</code>.
   * @throws DataPortalException
   */
  public ActiveDataPortal() throws DataPortalException {
    String inputConduitClassName = null;
    String outputConduitClassName = null;

    InitialContext ic = null;

    try {
      ic = new InitialContext();
    }
    catch (NamingException ne) {
    	String err = "Unable to obtain a JNDI InitialContext.  " +
    	this.getClass().getName() + " must reside with a J2EE server's JVM.";
    	log.log(Level.SEVERE,err,ne);
      throw new DataPortalException(err,ne);
    }
    if (ic != null) {
    	final String INPUT_CONDUIT_REF=
    		"java:comp/env/input-conduit-class-name";
      try {
        inputConduitClassName =
            (String) ic.lookup(INPUT_CONDUIT_REF);
        if(null == inputConduitClassName || inputConduitClassName.equals("null")){
          inputConduitClassName = null;
        }
      }
      catch (NamingException ne) {
    	String warn = "Unable to find " + INPUT_CONDUIT_REF +
    	"in JNDI, setting inputConduitClassName to null";
    	log.log(Level.WARNING,warn,ne);
        inputConduitClassName = null;
      }
      final String OUTPUT_CONDUIT_REF=
    	  "java:comp/env/output-conduit-class-name";
      try {
        outputConduitClassName =
            (String) ic.lookup(OUTPUT_CONDUIT_REF);
        if(null == outputConduitClassName || outputConduitClassName.equals("null")){
          outputConduitClassName = null;
        }
      }
      catch (NamingException ne) {
    	String warn = "Unable to find " + OUTPUT_CONDUIT_REF +
    	"in JNDI, setting outputConduitClassName to null";
    	log.log(Level.WARNING,warn,ne);
        outputConduitClassName = null;
      }

    }

    try {
      if (inputConduitClassName != null && !inputConduitClassName.equals("")) {
        try{
        	Class inputConduitClass = Class.forName(inputConduitClassName);
        	inputConduit = (InputConduit) inputConduitClass.newInstance();
        }catch(ClassNotFoundException ce){
        	String err = "Unable to find InputConduit class \""+
        	inputConduitClassName + "\"";
        	log.log(Level.SEVERE,err,ce);
        	throw new DataPortalException(err,ce);
        }
      }
      if (outputConduitClassName != null && !outputConduitClassName.equals("")) {
    	  try{
    		  Class outputConduitClass = Class.forName(outputConduitClassName);
    		  outputConduit = (OutputConduit) outputConduitClass.newInstance();
    	  }catch(ClassNotFoundException ce){
          	String err = "Unable to find OutputConduit class \""+
        	outputConduitClassName + "\"";
        	log.log(Level.SEVERE,err,ce);
        	throw new DataPortalException(err,ce);
    	  }
      }

      if (inputConduit != null) {
        this.setInputConduit(inputConduit);
      }
      if (outputConduit != null) {
        this.setOutputConduit(outputConduit);
      }
    }
    catch (InstantiationException ie) {
      String err = "InstantiationException: " + ie;
      log.log(Level.SEVERE,err,ie);
      throw new DataPortalException(err,ie);
    }
    catch (IllegalAccessException ile) {
      String err = "IllegalAccessException:" + ile;
      log.log(Level.SEVERE,err,ile);
      throw new DataPortalException(err,ile);
    }finally{
      closeResources(ic);
    }
  }

  /**
      Access method for the inputConduit property.
      @return   the current value of the inputConduit property
   */
  public InputConduit getInputConduit() {
    return this.inputConduit;
  }

  /**
      Sets the value of the inputConduit property.
      @param aInputConduit the new value of the inputConduit property
   */
  public void setInputConduit(InputConduit aInputConduit) {
    inputConduit = aInputConduit;
  }

  /**
      Access method for the outputConduit property.
      @return   the current value of the outputConduit property
   */
  public OutputConduit getOutputConduit() {
    return this.outputConduit;
  }

  /**
      Sets the value of the outputConduit property.
      @param aOutputConduit the new value of the outputConduit property
   */
  public void setOutputConduit(OutputConduit aOutputConduit) {
    outputConduit = aOutputConduit;
  }



  /**
   *  Transmit a request (if <code>outputConduit</code> is not null) and then garner a
   *  response (if <code>inputConduit</code> is not null).  If a response is garnered then
   *  return it to the caller. If there is no response then return <code>context</code>.
   *  @param context - The context in which the data are to be obtained; generally,
   *  the context will be some sort of request, say for a stock quote.
   *  @return java.lang.Object
   *  @throws org.quintet.backbeat.dataportal.DataPortalException
   *  @roseuid 4017FD5503D3
   */
  public Object getData(Object context) throws DataPortalException {
    if (context == null) {
      String warn = "context is null, therefore returning null";
      log.warning(warn);
      return context;
    }

    Object obj = context;
    try {
      if (this.getOutputConduit() != null) {
        obj = this.getOutputConduit().writeData(context);
        if (obj != null) {
          return obj;
        }
      }
      if (this.getInputConduit() != null) {
        obj = this.getInputConduit().getData(context);
      }
    }
    catch (BackBeatException ie) {
      String err = "caught an BackBeatException:"+ie;
      Throwable orig = ie.getOriginalException();
      if(orig != null){
        log.log(Level.SEVERE,err,orig);
        throw new DataPortalException(err,orig);
      }else{
        log.log(Level.SEVERE,err,ie);
        throw new DataPortalException(err,ie);
      }
    }
    return obj;
  }
}
