package org.quintet.backbeat.io;

import javax.naming.InitialContext;
import java.io.File;
import javax.mail.Transport;
import javax.activation.DataHandler;
import javax.mail.Session;
import javax.activation.FileDataSource;
import javax.mail.MessagingException;
import javax.naming.NamingException;
import javax.mail.internet.MimeMultipart;
import javax.mail.Multipart;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeBodyPart;
import javax.mail.BodyPart;
import java.util.StringTokenizer;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.lang.reflect.Method;
import org.quintet.backbeat.BackBeatException;


public class EmailOutputConduit
    implements OutputConduit, ForwardConduit {
  private static final String MAIL_SESSION_KEY =
      "java:comp/env/email-session-ref";
  private static final String FROM_ADDR_KEY = "java:comp/env/from-addr";
  private static final String TO_ADDR_KEY = "java:comp/env/to-addr";
  private static final String TO_ADDR_GETTER_METHOD_KEY =
      "java:comp/env/to-addr-getter-method";
  private static final String CC_ADDR_KEY = "java:comp/env/cc-addr";
  private static final String SUBJECT_LINE_KEY = "java:comp/env/subject-line";
  private static final String MESSAGE_BODY_KEY = "java:comp/env/message-body";
  private String fromAddr;
  private String toAddrList;
  private String ccAddrList;
  private String subject;
  private String msgBody;
  private static final String TO_ADDR_GETTER_TARGET_KEY =
      "java:comp/env/to-addr-getter-target";

  /**
      all diagnostic messages are sent to this Logger*
   */
  private static Logger log = Logger.getLogger("org.quintet.backbeat.io");

  /**
      My <code>Session</code> in which I communicate with the mail server
   */
  private Session mailSession = null;

  /**
      The <code>Method</code> via which <code>EmailOutputConduit</code> obtains its email recipient's address.
   */
  private Method toAddrGetter = null;

  /**
      <code>EmailOutputConduit</code> obtains the "to address" by invoking <code>toAddrGetter</code> upon an instance of this class
   */
  private Class getterTargetClass = null;


  /**
      Constructor for EmailOutputConduit


      @throws org.quintet.backbeat.BackBeatException
      @roseuid 4119127800A7
   */
  public EmailOutputConduit() throws BackBeatException {
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }
    catch (NamingException ne) {
      String fatal =
          "Unable to obtain JNDI InitialContext(is the App Server running?)";
      log.log(Level.SEVERE,fatal, ne);
      throw new BackBeatException(fatal, ne);
    }

    try {
      mailSession = (Session) ic.lookup(MAIL_SESSION_KEY);
    }
    catch (NamingException ne) {
      String err = "unable to find mail session " + MAIL_SESSION_KEY +
          " in JNDI.";
      log.log(Level.SEVERE,err, ne);
      throw new BackBeatException(err, ne);
    }

    try {
      fromAddr = (String) ic.lookup(FROM_ADDR_KEY);
    }
    catch (NamingException ne) {
      String err = FROM_ADDR_KEY + " not found in JNDI:";
      log.log(Level.SEVERE,err, ne);
      throw new BackBeatException(err, ne);
    }

    String getterTargetClassName = null;
    String toAddrGetterMethodName = null;
    try {
      toAddrList = (String) ic.lookup(TO_ADDR_KEY);
      getterTargetClass = this.getClass();
      getterTargetClassName = getterTargetClass.getName();
      toAddrGetterMethodName = "getToAddrList";
    }
    catch (NamingException ne) {
      String info = TO_ADDR_KEY +
          " not found in JNDI, trying to-addr-getter-method";
      log.info(info);
      try {
        toAddrGetterMethodName = (String) ic.lookup(TO_ADDR_GETTER_METHOD_KEY);
      }
      catch (NamingException noGetter) {
        String err = "Neither " + TO_ADDR_KEY + " nor " +
            TO_ADDR_GETTER_METHOD_KEY + " specified";
        log.log(Level.SEVERE,err, noGetter);
        throw new BackBeatException(err, noGetter);
      }

      try {
        getterTargetClassName = (String) ic.lookup(TO_ADDR_GETTER_TARGET_KEY);
      }
      catch (NamingException noTargetClass) {
        String err = TO_ADDR_GETTER_TARGET_KEY + " not found in JNDI";
        log.log(Level.SEVERE,err, noTargetClass);
        throw new BackBeatException(err, noTargetClass);
      }
    }
    try {
      setGetterTargetClass(Class.forName(getterTargetClassName));
    }
    catch (ClassNotFoundException ex) {
      String fatal = "Unable to load Class " + getterTargetClassName;
      log.log(Level.SEVERE,fatal, ex);
      throw new BackBeatException(fatal, ex);
    }

    try {
      setToAddrGetter(getGetterTargetClass().getMethod(toAddrGetterMethodName, (Class[])null));
    }
    catch (SecurityException ex1) {
      String fatal = "SecurityException when getting 'to address' method " +
          toAddrGetterMethodName;
      log.log(Level.SEVERE,fatal, ex1);
      throw new BackBeatException(fatal, ex1);
    }
    catch (NoSuchMethodException ex1) {
      String fatal = "NoSuchMethodException when getting 'to address' method " +
          toAddrGetterMethodName;
      log.log(Level.SEVERE,fatal, ex1);
      throw new BackBeatException(fatal, ex1);
    }

    try {
      ccAddrList = (String) ic.lookup(CC_ADDR_KEY);
    }
    catch (NamingException ne) {
      String err = CC_ADDR_KEY + " not found in JNDI:";
      log.warning(err);
      ccAddrList = "";
    }
    try {
      subject = (String) ic.lookup(SUBJECT_LINE_KEY);
    }
    catch (NamingException ne) {
      String err = SUBJECT_LINE_KEY + " not found in JNDI:";
      log.log(Level.SEVERE,err, ne);
      throw new BackBeatException(err, ne);
    }
    try {
      msgBody = (String) ic.lookup(MESSAGE_BODY_KEY);
    }
    catch (NamingException ne) {
      log.warning("no MESSAGE-BODY found");
    }
    // add a new line (so that attachment is properly aligned)
    if (msgBody == null) {
      msgBody = "";
    }
    msgBody = msgBody + "\n\n";
  }

  /**
      Access method for the fromAddr property.

      @return   the current value of the fromAddr property
   */
  public String getFromAddr() {
    return fromAddr;
  }

  /**
      Sets the value of the fromAddr property.

      @param aFromAddr the new value of the fromAddr property
   */
  public void setFromAddr(String aFromAddr) {
    fromAddr = aFromAddr;
  }

  /**
      Access method for the toAddrList property.

      @return   the current value of the toAddrList property
   */
  public String getToAddrList() {
    return toAddrList;
  }

  /**
      Sets the value of the toAddrList property.

      @param aToAddrList the new value of the toAddrList property
   */
  public void setToAddrList(String aToAddrList) {
    toAddrList = aToAddrList;
  }

  /**
      Access method for the mailSession property.

      @return   the current value of the mailSession property
   */
  public Session getMailSession() {
    return mailSession;
  }

  /**
      Sets the value of the mailSession property.

      @param aMailSession the new value of the mailSession property
   */
  public void setMailSession(Session aMailSession) {
    mailSession = aMailSession;
  }

  /**
      Access method for the toAddrGetter property.

      @return   the current value of the toAddrGetter property
   */
  public Method getToAddrGetter() {
    return toAddrGetter;
  }

  /**
      Sets the value of the toAddrGetter property.

      @param aToAddrGetter the new value of the toAddrGetter property
   */
  public void setToAddrGetter(Method aToAddrGetter) {
    toAddrGetter = aToAddrGetter;
  }

  /**
      Access method for the getterTargetClass property.

      @return   the current value of the getterTargetClass property
   */
  public Class getGetterTargetClass() {
    return getterTargetClass;
  }

  /**
      Sets the value of the getterTargetClass property.

      @param aGetterTargetClass the new value of the getterTargetClass property
   */
  public void setGetterTargetClass(Class aGetterTargetClass) {
    getterTargetClass = aGetterTargetClass;
  }

  /**
      Identifies the type of data sent. It has to be either a String, File or
      Multipart. An exception is thrown if the data is not one of the above
      types.The information that is required to send out a email, like the
      From address list, To address list, CC address list, Subject of the
   message and the body of the message are read from the environment entries.
      @param data - Content for the email to be sent. It can be a String, File
      or Multipart
      @return java.lang.Object
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 4119127800A8
   */
  public Object writeData(Object data) throws BackBeatException {
    return sendEmail(data);
  }

  /**
      @param data
      @return Object
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 418ADCA900B7
   */
  public Object forwardData(Object data) throws BackBeatException {
    return sendEmail(data);
  }

  /**
      @param data
      @return Object
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 418ADCA900FD
   */
  public Object sendEmail(Object data) throws BackBeatException {

    if (data == null) {
      String err = "Received a null data object";
      log.log(Level.SEVERE,err);
      throw new BackBeatException(err);
    }

    String err = null;
    MimeMessage mailMsg = null;
    mailMsg = new MimeMessage(mailSession);
    //set the from address
    try {
      mailMsg.setFrom(new InternetAddress(getFromAddr()));
    }
    catch (MessagingException ex) {
      err = "MessagingException";
      log.log(Level.SEVERE,err, ex);
      throw new BackBeatException(err, ex);
    }

    //use the toAddrGetter method to obtain the "to address"
    setToAddrList(getToAddress(data));

    //Check to see if TO Address has more than one name. If so add each
    //name to the TO address list, else add that one name
    if (toAddrList.indexOf(",") != -1) {
      StringTokenizer st = new StringTokenizer(toAddrList, ",");
      String toAddr = null;
      while (st.hasMoreTokens()) {
        toAddr = st.nextToken();
        toAddr = toAddr.trim();
        try {
          mailMsg.addRecipient(
              javax.mail.Message.RecipientType.TO,
              new InternetAddress(toAddr));
        }
        catch (MessagingException ex1) {
          err = "MessagingException";
          log.log(Level.SEVERE,err);
          throw new BackBeatException(err, ex1);
        }
      }
    }
    else {
      try {
        mailMsg.addRecipient(
            javax.mail.Message.RecipientType.TO,
            new InternetAddress(toAddrList));
      }
      catch (MessagingException ex2) {
        err = "MessagingException";
        log.log(Level.SEVERE,err, ex2);
        throw new BackBeatException(err, ex2);
      }
    }

    //Check to see if CC Address has more than one name. If so add each
    //name to the CC address list, else add that one name
    if (ccAddrList != null && (ccAddrList.equals("") == false) &&
        ccAddrList.indexOf(",") != -1) {
      StringTokenizer st = new StringTokenizer(ccAddrList, ",");
      String ccAddr = null;
      while (st.hasMoreTokens()) {
        ccAddr = st.nextToken();
        ccAddr = ccAddr.trim();
        try {
          mailMsg.addRecipient(
              javax.mail.Message.RecipientType.CC,
              new InternetAddress(ccAddr));
        }
        catch (MessagingException ex3) {
          err = "MessagingException";
          log.log(Level.SEVERE,err, ex3);
          throw new BackBeatException(err, ex3);
        }
      }
    }
    else {
      try {
        if (ccAddrList != null && (ccAddrList.equals("") == false)) {
          mailMsg.addRecipient(
              javax.mail.Message.RecipientType.CC,
              new InternetAddress(ccAddrList));
        }
      }
      catch (MessagingException ex4) {
        log.log(Level.SEVERE,"MessagingException", ex4);
        throw new BackBeatException(err, ex4);
      }
    }


  //set the subject
  try {
    mailMsg.setSubject(subject);
  }
  catch (MessagingException ex5) {
    log.log(Level.SEVERE,"MessagingException", ex5);
    throw new BackBeatException(err, ex5);
  }
  //set the sent date
  try {
    mailMsg.setSentDate(new java.util.Date());
  }
  catch (MessagingException ex5) {
    log.log(Level.SEVERE,"MessagingException", ex5);
    throw new BackBeatException(err, ex5);
  }

  //check to see if data is a String. If so, create a message as
  //plain text
  if (data instanceof String) {
    err = "sending email with text message:  "
        + data + " (via email) to " + toAddrList; ;
    log.info(err);
    try {
      mailMsg.setContent(msgBody + data, "text/plain");
    }
    catch (MessagingException ex5) {
      log.log(Level.SEVERE,"MessagingException", ex5);
      throw new BackBeatException(err, ex5);
    }
  }
  //check to see if data is a File. If so, create a message as
  //Multipart with the file as an attachment.
  else if (data instanceof File) {
    err = "sending file " + ( (File) data).getName() + " (via email) to " +
        toAddrList;
    log.info(err);
    File file = (File) data;
    Multipart multipart = new MimeMultipart();

    // add the text message
    BodyPart messageBodyPart = new MimeBodyPart();
    try {
      messageBodyPart.setText(msgBody);
    }
    catch (MessagingException ex5) {
      log.log(Level.SEVERE,"MessagingException", ex5);
      throw new BackBeatException(err, ex5);
    }
    try {
      multipart.addBodyPart(messageBodyPart);
    }
    catch (MessagingException ex5) {
      log.log(Level.SEVERE,"MessagingException", ex5);
      throw new BackBeatException(err, ex5);
    }

    //attach the attachment
    BodyPart fileBodyPart = new MimeBodyPart();
    FileDataSource source = new FileDataSource(file);
    try {
      fileBodyPart.setDataHandler(new DataHandler(source));
    }
    catch (MessagingException ex5) {
      log.log(Level.SEVERE,"MessagingException", ex5);
      throw new BackBeatException(err, ex5);
    }
    try {
      fileBodyPart.setFileName(file.getName());
    }
    catch (MessagingException ex5) {
      log.log(Level.SEVERE,"MessagingException", ex5);
      throw new BackBeatException(err, ex5);
    }
    try {
      multipart.addBodyPart(fileBodyPart);
    }
    catch (MessagingException ex5) {
      log.log(Level.SEVERE,"MessagingException", ex5);
      throw new BackBeatException(err, ex5);
    }

    //Put parts in message
    try {
      mailMsg.setContent(multipart);
    }
    catch (MessagingException ex5) {
      log.log(Level.SEVERE,"MessagingException", ex5);
      throw new BackBeatException(err, ex5);
    }
  }
  //check to see if data is a Multipart. If so, create a message as
  //Multipart.
  else if (data instanceof Multipart) {
    err = "Sending email with multipart message: (via email) to " +
        toAddrList;
    log.info(err);
    try {
      mailMsg.setContent( (Multipart) data);
    }
    catch (MessagingException ex5) {
      log.log(Level.SEVERE,"MessagingException", ex5);
      throw new BackBeatException(err, ex5);
    }
  }
  else if (data instanceof org.quintet.backbeat.io.Email) {
    Email email = (Email) data;
    try {
      mailMsg.setContent(msgBody + email.getBody(), "text/plain");
    }
    catch (MessagingException ex5) {
      log.log(Level.SEVERE,"MessagingException", ex5);
      throw new BackBeatException(err, ex5);
    }

  }
  //If data is none of the above throw an error
  else {
    err = "Expected a String, File or a Multipart, but received a " +
        data.getClass().getName();
    log.log(Level.SEVERE,err);
    throw new BackBeatException(err);
  }
  try {
    Transport.send(mailMsg);
  }
  catch (MessagingException ex5) {
    log.log(Level.SEVERE,"MessagingException", ex5);
    throw new BackBeatException(err, ex5);
  }
  return mailMsg;
}

/**
    Invoke the <code>toAddrGetter</code> method to obtain the email's primary
    recipient(s).
 @param inTargetObject - The data payload upon which the getter method might be
    called; if <code>getterTargetClass.equals(this.getClass())</code> then
 <code>toAddrGetter</code> will be invoked upon <code>this</code>, otherwise it
    will be invoked up <code>inTargetObject</code>
    @return String
    @throws org.quintet.backbeat.BackBeatException
    @roseuid 4219FAE0016E
 */
private String getToAddress(Object inTargetObject) throws BackBeatException {
  Object target = null;
  if (this.getGetterTargetClass().equals(this.getClass())) {
    target = this;
  }
  else {
    target = inTargetObject;
  }

  try {
    Object answer = getToAddrGetter().invoke(target, (Object[])null);
    return answer.toString();
  }
  catch (InvocationTargetException ex) {
    String err = "InvocationTargetException when invoking " +
        target.getClass().getName() + "." +
        getToAddrGetter().getDeclaringClass().getName();
    Throwable cause = ex.getCause();
    log.log(Level.SEVERE,err, cause);
    throw new BackBeatException(err, ex);
  }
  catch (IllegalArgumentException ex) {
    String err = "IllegalArgumentException when invoking " +
        target.getClass().getName() + "." +
        getToAddrGetter().getDeclaringClass().getName();
    log.log(Level.SEVERE,err, ex);
    throw new BackBeatException(err, ex);

  }
  catch (IllegalAccessException ex) {
    String err = "IllegalAccessException when invoking " +
        target.getClass().getName() + "." +
        getToAddrGetter().getDeclaringClass().getName();
    log.log(Level.SEVERE,err, ex);
    throw new BackBeatException(err, ex);
  }
}
}
