package org.quintet.backbeat.io;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.logging.Logger;
import java.util.logging.Level;
import org.quintet.backbeat.BackBeatException;

/**
 Submit a request by invoking a "target method" (</code>targetMethod</code>)
 upon a "target instance" (<code>targetInstance</code>) of <code>Class</code>
 <code>targetClass</code>.  The target method and the target class are specified
 via JNDI entries <code>target-method-name</code> and
 <code>target-class-name</code> respectively.
 * @author Peter Thaggard
 */
public class MethodOutputConduit
    implements OutputConduit {

  /**
   The name by which to find this <code>MethodOutputConduit</code>'s target method
   name in JNDI
   */
  private static final String TARGET_METHOD_NAME_KEY =
      "java:comp/env/target-method-name";

  /**
   The name by which to find this <code>MethodOutputConduit</code>'s target class
      name in JNDI
   */
  private static final String TARGET_CLASS_NAME_KEY =
      "java:comp/env/target-class-name";

  /**
   The name by which the <code>targetObjectIsTransient</code> flag is found in
      JNDI.  If the flag's value is not found in JNDI then it is set to
      <code>false</code>
   */
  private static final String TARGET_OBJECT_IS_TRANSIENT_KEY =
      "java:comp/env/target-object-is-transient";

  private static final String DATA_OBJECT_IS_TARGET_KEY = "java:comp/env/data-object-is-target";
  /**
      This flag indicates whether this <code>MethodOutputConduit</code> should
      instantiate the target object at system startup time
   (<code>targetObjectIsTransient = false</code>) or if it should re-instantiate
   it every time <code>writeData</code> is invoked (<code>targetObjectIsTransient
      = true</code>).
   */
  private boolean targetObjectIsTransient = false;

  /**
      The sink to which all <code>MethodOutputConduit</code> send their logging
      messages
   */
  private static Logger log = Logger.getLogger("org.quintet.backbeat");
  private Method output;
  private Object target;

  /**
   * If true then the object passed to <code>writeData</code> is the target upon
   * which the <code>output</code> method is invoked
   */
  private boolean dataObjectIsTarget = false;

  /**
      The <code>Method</code> to which this <code>MethodOutputConduit</code> forwards requests
   */
  private Method targetMethod = null;

  /**
      The <code>Class</code> that implements <code>targetMethod</code>
   */
  private Class targetClass = null;

  /**
      The object upon which <code>targetMethod</code> is invoked.
   */
  private Object targetObject = null;

  public void closeResources(InitialContext ic){
  if(ic != null){
    try{
      ic.close();
    }catch(Throwable eatIt){}
  }
 }

  /**
      @param inTargetObject - The object upon which this
      <code>MethodOutputConduit</code> will invoke methods
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 4213B41300B5
   */
  public MethodOutputConduit(Object inTargetObject) throws BackBeatException {
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }
    catch (NamingException ne) {
      String err = "Unable to obtain JNDI initial context -- probably means that App Server is down";
      log.log(Level.SEVERE,err, ne);
      throw new BackBeatException(err, ne);
    }

    String targetMethodName;
    try {
      targetMethodName = (String) ic.lookup(TARGET_METHOD_NAME_KEY);
    }
    catch (NamingException ne) {
      String err = "no " + TARGET_METHOD_NAME_KEY + " found in JNDI context.  " +
          "Please check your deployment descriptor";
      log.log(Level.SEVERE,err, ne);
      closeResources(ic);
      throw new BackBeatException(err, ne);
    }

    setTargetObjectIsTransient(false);

    setTargetClass(inTargetObject.getClass());
    setTargetObject(inTargetObject);

    //find the designated method within targetClass.  The method must take
    //exactly one argument, and that argument must be a java.lang.Object
    try {
      setTargetMethod(getTargetClass().getDeclaredMethod(targetMethodName,
          new Class[] {java.lang.Object.class}));
    }
    catch (SecurityException ex1) {
      String err = "SecurityException when trying to find method " +
          targetMethodName + " in class " + getTargetClass().getName();
      log.log(Level.SEVERE,err, ex1);
      closeResources(ic);
      throw new BackBeatException(err, ex1);
    }
    catch (NoSuchMethodException ex1) {
      String err = "no method " + targetMethodName +
          "(java.lang.Object) found in class " + getTargetClass().getName();
      log.log(Level.SEVERE,err, ex1);
      closeResources(ic);
      throw new BackBeatException(err, ex1);
    }
    closeResources(ic);
  }

  /**
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 4202665F02B7
   */
  public MethodOutputConduit() throws BackBeatException {
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }
    catch (NamingException ne) {
      String err = "Unable to obtain JNDI initial context -- probably means that App Server is down";
      log.log(Level.SEVERE,err, ne);
      throw new BackBeatException(err, ne);
    }

    String targetMethodName;
    try {
      targetMethodName = (String) ic.lookup(TARGET_METHOD_NAME_KEY);
    }
    catch (NamingException ne) {
      String err = "no " + TARGET_METHOD_NAME_KEY + " found in JNDI context.  " +
          "Please check your deployment descriptor";
      log.log(Level.SEVERE,err, ne);
      closeResources(ic);
      throw new BackBeatException(err, ne);
    }

    String targetClassName;
    try {
      targetClassName = (String) ic.lookup(TARGET_CLASS_NAME_KEY);
    }
    catch (NamingException ne) {
      String err = "no " + TARGET_CLASS_NAME_KEY + " found in JNDI context.  " +
          "Please check your deployment descriptor.";
      log.log(Level.SEVERE,err, ne);
      closeResources(ic);
      throw new BackBeatException(err, ne);
    }

    String transientFlagAsString;
    try {
      transientFlagAsString = (String) ic.lookup(TARGET_OBJECT_IS_TRANSIENT_KEY);
    }
    catch (NamingException ex2) {
      String info = TARGET_OBJECT_IS_TRANSIENT_KEY +
          " not specified, setting to false";
      log.info(info);
      transientFlagAsString = "false";
    }

    if ("true".equalsIgnoreCase(transientFlagAsString)) {
      setTargetObjectIsTransient(true);
    }
    else {
      setTargetObjectIsTransient(false);
    }

    String dataObjectIsTargetString;
    try{
      dataObjectIsTargetString = (String)ic.lookup(DATA_OBJECT_IS_TARGET_KEY);
    }catch(NamingException ne){
      String info = "data-object-is-target not found in JNDI, assuming 'false'";
      log.info(info);
      dataObjectIsTargetString = "false";
    }

    this.dataObjectIsTarget = new Boolean(dataObjectIsTargetString).booleanValue();

    //instantiate the designated class
    try {
      setTargetClass(Class.forName(targetClassName));
    }
    catch (ClassNotFoundException ex) {
      String err = "No class named " + targetClassName +
          " found in CLASSPATH.  " +
          "Please check your CLASSPATH and your deployment descriptor and try again";
      log.log(Level.SEVERE,err, ex);
      closeResources(ic);
      throw new BackBeatException(err, ex);
    }

    //find the designated method within targetClass.  The method must take
    //exactly one argument, and that argument must be a java.lang.Object
    try {
      setTargetMethod(getTargetClass().getDeclaredMethod(targetMethodName,
          new Class[] {java.lang.Object.class}));
    }
    catch (SecurityException ex1) {
      String err = "SecurityException when trying to find method " +
          targetMethodName + " in class " + targetClassName;
      log.log(Level.SEVERE,err, ex1);
      closeResources(ic);
      throw new BackBeatException(err, ex1);
    }
    catch (NoSuchMethodException ex1) {
      String err = "no method " + targetMethodName +
          "(java.lang.Object) found in class " + targetClassName;
      log.log(Level.SEVERE,err, ex1);
      closeResources(ic);
      throw new BackBeatException(err, ex1);
    }

    //we have now instantiated both the target class and the target method.  If
    //targetObjectIsTransient == false then go ahead and instantiate the target object now.
    if (targetObjectIsTransient == false && dataObjectIsTarget == false) {
      try {
        setTargetObject(getTargetClass().newInstance());
      }
      catch (IllegalAccessException ex3) {
        String err =
            "IllegalAccessException when instantiating an object of type " +
            targetClassName;
        log.log(Level.SEVERE,err, ex3);
        closeResources(ic);
        throw new BackBeatException(err, ex3);
      }
      catch (InstantiationException ex3) {
        String err =
            "InstantiationException when instantiating an object of type " +
            targetClassName;
        log.log(Level.SEVERE,err, ex3);
        closeResources(ic);
        throw new BackBeatException(err, ex3);
      }
    }
    else {
      setTargetObject(null);
    }
    closeResources(ic);
  }

  /**
      Determines if the targetObjectIsTransient property is true.

   @return   <code>true<code> if the targetObjectIsTransient property is true
   */
  public boolean getTargetObjectIsTransient() {
    return targetObjectIsTransient;
  }

  /**
      Sets the value of the targetObjectIsTransient property.

      @param aTargetObjectIsTransient the new value of the targetObjectIsTransient property
   */
  public void setTargetObjectIsTransient(boolean aTargetObjectIsTransient) {
    targetObjectIsTransient = aTargetObjectIsTransient;
  }

  /**
      Access method for the output property.

      @return   the current value of the output property
   */
  public Method getOutput() {
    return output;
  }

  /**
      Sets the value of the output property.

      @param aOutput the new value of the output property
   */
  public void setOutput(Method aOutput) {
    output = aOutput;
  }

  /**
      Access method for the target property.

      @return   the current value of the target property
   */
  public Object getTarget() {
    return target;
  }

  /**
      Sets the value of the target property.

      @param aTarget the new value of the target property
   */
  public void setTarget(Object aTarget) {
    target = aTarget;
  }

  /**
      Access method for the targetMethod property.

      @return   the current value of the targetMethod property
   */
  public Method getTargetMethod() {
    return targetMethod;
  }

  /**
      Sets the value of the targetMethod property.

      @param aTargetMethod the new value of the targetMethod property
   */
  public void setTargetMethod(Method aTargetMethod) {
    targetMethod = aTargetMethod;
  }

  /**
      Access method for the targetClass property.

      @return   the current value of the targetClass property
   */
  public Class getTargetClass() {
    return targetClass;
  }

  /**
      Sets the value of the targetClass property.

      @param aTargetClass the new value of the targetClass property
   */
  public void setTargetClass(Class aTargetClass) {
    targetClass = aTargetClass;
  }

  /**
      Access method for the targetObject property.

      @return   the current value of the targetObject property
   */
  public Object getTargetObject() {
    return targetObject;
  }

  /**
      Sets the value of the targetObject property.

      @param aTargetObject the new value of the targetObject property
   */
  public void setTargetObject(Object aTargetObject) {
    targetObject = aTargetObject;
  }

  /**
      Invoke <code>target</code>'s <code>output</code> method and supply
      <code>data</code> as its actual parameter.
   @param data - The data to be sent to <code>output</code> as its one-and-only
      argument
      @return Object
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 401A84FC00B5
   */
  public Object writeObject(Object data) throws BackBeatException {
    return null;
  }

  /**
      @roseuid 401AA1640028
   */
  public void close() {

  }

  /**
      @param data
      @return Object
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 42025F8C0161
   */
  public Object writeData(Object data) throws BackBeatException {
    //write data by passing data to targetMethod of targetObject
    if (data == null) {
      //I suppose that this is ok, 'though probably unlikely, so log it
      String info =
          "MethodOutputConduit.writeData was passed a null data object";
      log.info(info);
    }

    //if target object is supposed to be instantiated transiently then do so now
    if (targetObjectIsTransient && dataObjectIsTarget == false) {
      try {
        setTargetObject(getTargetClass().newInstance());
      }
      catch (IllegalAccessException ex3) {
        String err =
            "IllegalAccessException when instantiating an object of type " +
            getTargetClass().getName();
        log.log(Level.SEVERE,err, ex3);
        throw new BackBeatException(err, ex3);
      }
      catch (InstantiationException ex3) {
        String err =
            "InstantiationException when instantiating an object of type " +
            getTargetClass().getName();
        log.log(Level.SEVERE,err, ex3);
        throw new BackBeatException(err, ex3);
      }
    }else if(dataObjectIsTarget){
      setTargetObject(data);
    }
    Object target = getTargetObject();
    if (target == null) {
      //jeepers, no target object.  That is a real shame.
      String err = "Target object is null, cannot fulfill request";
      log.log(Level.SEVERE,err);
      throw new BackBeatException(err);
    }
    try {
      return getTargetMethod().invoke(target, new Object[] {data});
    }
    catch (IllegalAccessException iae) {
      String err = "IllegalAccessException when invoking method " +
          getTargetMethod().getName() + " of class " + getTargetClass().getName();
      log.log(Level.SEVERE,err, iae);
      throw new BackBeatException(err, iae);
    }
    catch (InvocationTargetException ite) {
      Throwable orig = ite.getCause();
      String err = "method " + getTargetMethod().getName() + " of class " +
          getTargetClass().getName() + " threw an exception: " + orig.toString();
      log.log(Level.SEVERE,err, orig);
      throw new BackBeatException(err, ite);
    }
  }
}
