package org.quintet.backbeat.io;



import javax.naming.InitialContext;
import javax.naming.NamingException;

import javax.ejb.EJBHome;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import java.util.logging.Logger;
import java.util.logging.Level;

import org.quintet.backbeat.BackBeatException;

/**
 *<code>SessionBeanMethodConduit</code> invokes methods on Session Beans indirectly via reflection.
 *It first fetches a home interface (designated by the home-interface-ref deployment descriptor entry),
 *and then uses that home interface to create a session bean instance.  Finally it invokes a method, designated by
 *the method-name deployment descriptor entry.
 <p>
 *@author Peter Thaggard
 *@version 1.0
 *@since Quintet 2002_1
 **/

public class SessionBeanMethodConduit
    implements  ForwardConduit{


  private static Logger log = Logger.getLogger("org.quintet.backbeat");

  private static String HOME_INTERFACE_ENV_ENTRY =
      "java:comp/env/ejb/home-interface-ref";
  private static String METHOD_ENV_ENTRY =
      "java:comp/env/method-name";

  /**<code>SessionBeanMethodConduit</code> stores an instance of the home interface with which it collaborates**/
  private EJBHome homeInterface = null;

  /**<code>SessionBeanMethodConduit</code> stores an instance of the target session bean's component interface**/
  private Object target = null;

  /**<code>conduitMethod</code> stores the create <code>Method</code> that it calls on <code>homeInterface</code>**/
  Method conduitMethod = null;
  Method createMethod = null;


  public void closeResources(InitialContext ic){
    if(ic != null){
      try{
        ic.close();
      }catch(Throwable eatIt){}
    }
  }

  public SessionBeanMethodConduit() throws org.quintet.backbeat.BackBeatException {
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }catch(NamingException ne){
      String err = "unable to obtain initial context";
      log.log(Level.SEVERE,err,ne);
      throw new BackBeatException(err,ne);
    }

    try{
      Object homeInterfaceObj = ic.lookup(HOME_INTERFACE_ENV_ENTRY);
      log.info("Home Interface instanceof EJBHome?" + (homeInterfaceObj instanceof EJBHome));
      log.info("Home Interface instanceof EJBLocalHome?" + (homeInterfaceObj instanceof javax.ejb.EJBLocalHome));
      homeInterface = (EJBHome) homeInterfaceObj;
    }catch(NamingException ne){
      String err = ":unable to find " + HOME_INTERFACE_ENV_ENTRY+" in JNDI:" + ne;
      log.log(Level.SEVERE,err,ne);
      closeResources(ic);
      throw new BackBeatException(err,ne);
    }

      Class homeInterfaceClass = homeInterface.getClass();
      try {
        createMethod = homeInterfaceClass.getMethod("create", (Class[])null);
      }
      catch (SecurityException ex) {
        String err = "SecurityException:"+ex;
        log.log(Level.SEVERE,err,ex);
        closeResources(ic);
        throw new BackBeatException(err,ex);
      }
      catch (NoSuchMethodException ex) {
        String err = "NoSuchMethodException:"+ex;
        log.log(Level.SEVERE,err,ex);
        closeResources(ic);
        throw new BackBeatException(err,ex);
      }

      //we now have the target's home interface and the Method to invoke thereon
      //in order to create the actual target
      Class targetClass = createMethod.getReturnType();
      try {
        target = createMethod.invoke(homeInterface, (Object[])null);
      }
      catch (InvocationTargetException ex1) {
        String err = ":InvocationTargetException:"+ex1;
        log.log(Level.SEVERE,err,ex1.getCause());
        closeResources(ic);
        throw new BackBeatException(err,ex1);
      }
      catch (IllegalArgumentException ex1) {
        String err = ":IllegalArgumentException:"+ex1;
        log.log(Level.SEVERE,err,ex1);
        closeResources(ic);
        throw new BackBeatException(err,ex1);
      }
      catch (IllegalAccessException ex1) {
        String err = ":IllegalAccessException:"+ex1;
        log.log(Level.SEVERE,err,ex1);
        closeResources(ic);
        throw new BackBeatException(err,ex1);
      }

      //we now have the actual target Session Bean remote interface.  Time to get the
      //method to invoke thereon.
      String conduitMethodName = null;
      try {
        conduitMethodName = (String)ic.lookup(METHOD_ENV_ENTRY);
      }
      catch (NamingException ex2) {
        String err = ":unable to find JNDI entry " + METHOD_ENV_ENTRY;
        log.log(Level.SEVERE,err,ex2);
        closeResources(ic);
        throw new BackBeatException(err,ex2);
      }
      try {
        conduitMethod = targetClass.getMethod(conduitMethodName,
                                              new Class[] {java.lang.Object.class});
      }
      catch (SecurityException ex3) {
        String err = ":SecurityException:"+ex3;
        log.log(Level.SEVERE,err,ex3);
        closeResources(ic);
        throw new BackBeatException(err,ex3);
      }
      catch (NoSuchMethodException ex3) {
        String err = ":NoSuchMethodException:"+ex3;
        log.log(Level.SEVERE,err,ex3);
        closeResources(ic);
        throw new BackBeatException(err,ex3);
      }
      closeResources(ic);
  }

  public Object forwardData(Object pk) throws BackBeatException {
    try {
      return conduitMethod.invoke(target, new Object[] {pk});
    }
    catch (InvocationTargetException ite) {
        ite.printStackTrace();
      throw new BackBeatException(
          "SessionBeanMethodConduit:forwardData:InvocationTargetException:" + ite, ite);
    }
    catch (IllegalAccessException iae) {
        iae.printStackTrace();
      throw new BackBeatException(
          "SessionBeanMethodConduit:forwardData:IllegalAccessException:" + iae, iae);
    }
  }
}
