package org.quintet.backbeat.io;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import org.quintet.backbeat.BackBeatException;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;

/**
 * <code>LocalSessionBeanMethodConduit30</code> pushes data (in either the
 * "output conduit" or "forward conduit" roles) to a downstream EJB 3.0 style
 * session bean.
 * <p>Copyright: Copyright (c) 2008</p>
 *
 * <p>Company: 6th Sense Analytics</p>
 *
 * @author Peter Thaggard
 * @version 1.0
 */
public class SessionBeanMethodConduit30 implements OutputConduit,ForwardConduit {

    private static final Logger log = Logger.getLogger("org.quintet.backbeat");

    /**
     * The <code>target</code> is the downstream component that will accept
     * "pushed" data from this conduit.
     */
    private Object target = null;

    /**
     * The <code>Method</code> via which data are "pushed" to <code>target</code>
     */
    private Method conduitMethod = null;

    public SessionBeanMethodConduit30() throws BackBeatException{
        //grab the target session bean from JNDI
        InitialContext ic = null;
        try{
            ic = new InitialContext();
        }catch(NamingException ne){
            String err = "Unable to obtain a JNDI InitialContext.  "+
                         "This object can ONLY run within a J2EE container's "+
                         "JVM.";
            log.log(Level.SEVERE,err,ne);
            throw new IllegalStateException(err,ne);
        }
        final String LOCAL_SESSION_BEAN_REF =
            "java:comp/env/local-session-bean-ref";
        try{
            target = ic.lookup(LOCAL_SESSION_BEAN_REF);
        }catch(NamingException ne){
            String err = "Unable to find \""+
                         LOCAL_SESSION_BEAN_REF + "\" in JNDI.  "+
                         "Please check your deployment descriptor.";
            log.log(Level.SEVERE,err,ne);
            throw new IllegalStateException(err,ne);
        }

        //find the target method specified in the deployment descriptor

        //we now have the actual target Session Bean remote interface.  Time to get the
        //method to invoke thereon.
        final String METHOD_ENV_ENTRY = "java:comp/env/method-name";
        Class targetClass = target.getClass();
        String conduitMethodName = null;
        try {
          conduitMethodName = (String)ic.lookup(METHOD_ENV_ENTRY);
        }
        catch (NamingException ex2) {
          String err = ":unable to find JNDI entry " + METHOD_ENV_ENTRY;
          log.log(Level.SEVERE,err,ex2);
        }

        //finally, we need to get the parameter types expected by the method
        Class[] methodParmTypes = null;

        methodParmTypes = getMethodParameterTypes(ic);



        try {
          conduitMethod = targetClass.getMethod(conduitMethodName,methodParmTypes);
        }
        catch (SecurityException ex3) {
          String err = ":SecurityException:"+ex3;
          log.log(Level.SEVERE,err,ex3);
          closeResources(ic);
          throw new BackBeatException(err,ex3);
        }
        catch (NoSuchMethodException ex3) {
          String err = ":NoSuchMethodException:"+ex3;
          log.log(Level.SEVERE,err,ex3);
          closeResources(ic);
          throw new BackBeatException(err,ex3);
        }
      closeResources(ic);
    }

    private static Map<String,Class> primitives = null;
    static{
        primitives = new HashMap<String,Class>();
        primitives.put("boolean",Boolean.TYPE);
        primitives.put("byte",Byte.TYPE);
        primitives.put("char",Character.TYPE);
        primitives.put("double",Double.TYPE);
        primitives.put("float",Float.TYPE);
        primitives.put("int",Integer.TYPE);
        primitives.put("long",Long.TYPE);
    }
    private Class[] getMethodParameterTypes(InitialContext ic) throws BackBeatException{
        final String PARM_TYPES_KEY = "java:comp/env/method-parameter-types";
        String parmTypeNames = null;
        try{
            parmTypeNames = (String)ic.lookup(PARM_TYPES_KEY);
        }catch(NamingException ne){
            String warn = "no parameter types specified, assuming no-parm method";
            log.log(Level.INFO,warn);
            return (Class[])null;
        }
        List<Class> parmClasses = new ArrayList<Class>();
        StringTokenizer lex = new StringTokenizer(parmTypeNames,",");
        while(lex.hasMoreTokens()){
            String className = lex.nextToken();
            //if it has no dot then it's a primitive type
            if(!(className.contains("."))){
                Class primitiveType = primitives.get(className);
                if(primitiveType == null){
                    String err = "\""+className+"\" is an unknown primitive";
                    log.log(Level.SEVERE,err);
                    throw new BackBeatException(err);
                }
                parmClasses.add(primitiveType);
            }else{
                //its an actual class
                try{
                    Class clazz = Class.forName(className);
                    parmClasses.add(clazz);
                }catch(Throwable t){
                    String err = t.getClass().getName() +
                                 " when trying to load class \""+
                                 className + "\"";
                    log.log(Level.SEVERE,err,t);
                    throw new BackBeatException(err,t);
                }
            }
        }
        if(parmClasses.size() == 0)
            return (Class[])null;
        Class[] finalAnswer = new Class[parmClasses.size()];
        parmClasses.toArray(finalAnswer);
        return finalAnswer;
    }
    /**
     * forwardData
     *
     * @param data Object
     * @return Object
     * @throws BackBeatException
     * @todo Implement this org.quintet.backbeat.io.ForwardConduit method
     */
    public Object forwardData(Object data) throws BackBeatException {
        return writeData(data);
    }

    public Object writeData(Object data) throws BackBeatException{
        if(data == null){
            String info = "unable to process null objects";
            log.log(Level.INFO,info);
            return data;
        }
        try{
            return conduitMethod.invoke(this.target,new Object[]{data});
        }catch(Throwable t){
            String err = t.getClass().getName() +
                         "when invoking " + target.getClass().getName()+
                         "." + conduitMethod.getName() + "(" +
                         data.getClass().getName();
            log.log(Level.SEVERE,err);
            throw new BackBeatException(err,t);
        }
    }

    public void closeResources(InitialContext ic){
      if(ic != null){
        try{
          ic.close();
        }catch(Throwable eatIt){}
      }
  }
}
