package org.quintet.backbeat.io;

import java.io.Serializable;

import javax.ejb.EJBException;
import javax.jms.JMSException;
import javax.jms.ObjectMessage;
import javax.jms.Session;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

import java.util.logging.Logger;
import java.util.logging.Level;
import org.quintet.backbeat.BackBeatException;
import javax.transaction.*;

public class TopicForwardConduit
    implements ForwardConduit {
  private static final Logger log = Logger.getLogger("org.quintet.backbeat.io");
  private static final String CONNECTION_FACTORY_REF =
            "java:comp/env/quintet-connection-factory-ref";

  private static final String TOPIC_REF =
      "java:comp/env/jms/quintet-notification-topic-ref";
private static final String TX_MANAGER_NAME = "java:appserver/TransactionManager";
  /**
      Builds topic connections for me
   */
  private TopicConnectionFactory connectionFactory = null;

  /**
      The topic to which I publish response notifications
   */
  private Topic forwardTopic = null;

  private TransactionManager tm = null;

  /**
   * The <code>messageCategory</code> is an optional
   * property that will be added as a JMS message property to all
   * outgoing messages.  It is useful for JMS message filtering.
   */
  private String messageCategory = null;

  public TopicForwardConduit(String topicName, String connectionFactoryName) throws BackBeatException{
      InitialContext ic = null;
      try {
        ic = new InitialContext();
      }
      catch (NamingException ne) {
        String fatal =
            "Unable to obtain a JNDI InitialContext(is the AppServer running?)";
        log.log(Level.SEVERE,fatal, ne);
        throw new BackBeatException(fatal, ne);
      }

      try {
        TopicConnectionFactory f = (TopicConnectionFactory) ic.lookup(
            connectionFactoryName);
        setConnectionFactory(f);
      }
      catch (NamingException ne) {
        String fatal =
            "No Topic Connection Factory configured in app server";
        log.log(Level.SEVERE,fatal, ne);
        throw new BackBeatException(fatal, ne);
      }

      try {
        forwardTopic = (Topic) ic.lookup(topicName);

      }
      catch (NamingException ne) {
        String fatal = "No topic ref named '" + topicName + "' found in JNDI";
        log.log(Level.SEVERE,fatal, ne);
        throw new BackBeatException(fatal, ne);
      }

      try {
    	  tm = (TransactionManager) ic.lookup(TX_MANAGER_NAME);
          if(tm == null){
              String err = "tm (TransactionManager) is NULL, "+
                           "many exceptions may result";
              log.log(Level.SEVERE,err);
          }
      } catch ( NamingException nex ) {
          String fatal = "Error grabbing TransactionManager " + TX_MANAGER_NAME + "' from JNDI";
          log.log(Level.SEVERE,fatal, nex);
          throw new BackBeatException(fatal, nex);
      }

      messageCategory="";
  }
  /**
   * Construct a new TopicForwardConduit that will be wired to send messages
   * to the <code>Topic</code> named <code>topicName</code>.
   * @param topicName String
   * @throws BackBeatException
   */
  public TopicForwardConduit(String topicName) throws BackBeatException{
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }
    catch (NamingException ne) {
      String fatal =
          "Unable to obtain a JNDI InitialContext(is the AppServer running?)";
      log.log(Level.SEVERE,fatal, ne);
      throw new BackBeatException(fatal, ne);
    }

    try {
      TopicConnectionFactory f = (TopicConnectionFactory) ic.lookup(
          CONNECTION_FACTORY_REF);
      setConnectionFactory(f);
    }
    catch (NamingException ne) {
      String fatal =
          "No Topic Connection Factory configured in app server";
      log.log(Level.SEVERE,fatal, ne);
      throw new BackBeatException(fatal, ne);
    }

    try {
      forwardTopic = (Topic) ic.lookup(topicName);

    }
    catch (NamingException ne) {
      String fatal = "No topic ref named '" + topicName + "' found in JNDI";
      log.log(Level.SEVERE,fatal, ne);
      throw new BackBeatException(fatal, ne);
    }

    final String MESSAGE_CATEGORY_REF="java:comp/env/message-category";
    try{
      messageCategory=(String)ic.lookup(MESSAGE_CATEGORY_REF);
    }catch(NamingException ne){
      //it's ok, you don't absolutely have to set message-category
    }
  }
  /**
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 42163EA5035F
   */
  public TopicForwardConduit() throws BackBeatException {
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }
    catch (NamingException ne) {
      String fatal =
          "Unable to obtain a JNDI InitialContext(is the AppServer running?)";
      log.log(Level.SEVERE,fatal, ne);
      throw new BackBeatException(fatal, ne);
    }

    try {
      TopicConnectionFactory f = (TopicConnectionFactory) ic.lookup(
          CONNECTION_FACTORY_REF);
      setConnectionFactory(f);
    }
    catch (NamingException ne) {
      String fatal =
          "No Topic Connection Factory configured in deployment descriptor";
      log.log(Level.SEVERE,fatal, ne);
      throw new BackBeatException(fatal, ne);
    }

    try {
      forwardTopic = (Topic) ic.lookup(TOPIC_REF);

    }
    catch (NamingException ne) {
      String fatal = "No topic ref named '" + TOPIC_REF + "' found in JNDI";
      log.log(Level.SEVERE,fatal, ne);
      throw new BackBeatException(fatal, ne);
    }
    final String MESSAGE_CATEGORY_REF="java:comp/env/message-category";
    try{
      messageCategory=(String)ic.lookup(MESSAGE_CATEGORY_REF);
    }catch(NamingException ne){
      //it's ok, you don't absolutely have to set message-category
    }
  }

  /**
      Access method for the connectionFactory property.

      @return   the current value of the connectionFactory property
   */
  public TopicConnectionFactory getConnectionFactory() {
    return connectionFactory;
  }

  /**
      Sets the value of the connectionFactory property.

      @param aConnectionFactory the new value of the connectionFactory property
   */
  public void setConnectionFactory(TopicConnectionFactory aConnectionFactory) {
    connectionFactory = aConnectionFactory;
  }

  /**
      Access method for the forwardTopic property.

      @return   the current value of the forwardTopic property
   */
  public Topic getForwardTopic() {
    return forwardTopic;
  }

  /**
      Sets the value of the forwardTopic property.

      @param aForwardTopic the new value of the forwardTopic property
   */
  public void setForwardTopic(Topic aForwardTopic) {
    forwardTopic = aForwardTopic;
  }

  /**
      @roseuid 42163E68002D
   */
  private void sendMessage() {

  }

  /**
   *  This method calls <code>forwardData</code> with a category based on the class field <code>messageCategory</code>.
      @param data The data to forward, must implement <code>java.io.Serializable</code>
      @return Object
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 42163E9001BB
   */
  public Object forwardData(Object data) throws BackBeatException {
      try{
          return this.forwardData(data, this.messageCategory, false);
      }catch(Throwable t){
          String err = t.getClass().getName() +
                       "from forwardData(data,messageCategory,useTX)";
          log.log(Level.SEVERE,err,t);
          throw new BackBeatException(err,t);
      }
  }

  /**
   *
      @param data The data to forward, must implement <code>java.io.Serializable</code>
      @param category String category set as the message property MESSAGE_CATEGORY, optional
      @return Object
      @throws org.quintet.backbeat.BackBeatException
      @roseuid 42163E9001BB
   */
  public Object forwardData(Object data, String category, boolean useTx) throws BackBeatException {
    if (data == null) {
      String warn = "Sorry, I cannot forward a null object";
      log.warning(warn);
      return null;
    }
    String nonNullCat = category == null ? "No-Category":category;
    if (! (data instanceof Serializable)) {
      String err = "Sorry, I cannot forward non-Serializable objects";
      log.log(Level.SEVERE,err);
      throw new BackBeatException(err,
          new java.io.NotSerializableException(data.
          getClass().
          getName()));
    }

    TopicConnection connection = null;
    TopicSession topicSession = null;
    TopicPublisher publisher = null;

    ObjectMessage forwardedMessage = null;

    Transaction currentTx = null;
    try {
    	if ( useTx == false ) {
                try{
                    if(tm == null){
                        //TransactionManager went away, get a new one

                        String err = "TransactionManager is NULL, cannot use "+
                                     "it to suspend current transaction, "+
                                     "will try to get a new one";
                        log.log(Level.SEVERE,err);
                        getNewTransactionManager();
                    }else{
                        if (tm.getStatus() == Status.STATUS_NO_TRANSACTION) {
                            String warn = "No Transaction to suspend, leaving " +
                                          "well enough alone";
                            log.log(Level.INFO, warn);
                        } else {
                            currentTx = tm.suspend();
                        }
                    }
                }catch(Throwable noSuspend){
                    String err = noSuspend.getClass().getName() +
                                 "indicates unable to suspend current transaction, "+
                                 "proceeding with forwarding work anyway.";
                    log.log(Level.SEVERE,err);
                }
    	}
      //create a connection
      connection = getConnectionFactory().createTopicConnection();

      int ackType = ( useTx  ? Session.SESSION_TRANSACTED : Session.AUTO_ACKNOWLEDGE );
      //create a session
      topicSession = connection.createTopicSession(useTx, ackType);

      //create a publisher
      publisher = topicSession.createPublisher(forwardTopic);

      //start the connection
      connection.start();
      //create a message
      Serializable messageObj = (Serializable) data;
      forwardedMessage = topicSession.createObjectMessage("What The Heck?");
      forwardedMessage.setJMSDeliveryMode(javax.jms.DeliveryMode.NON_PERSISTENT);
      forwardedMessage.setObject(messageObj);
      if(category != null){
        forwardedMessage.setStringProperty("MESSAGE_CATEGORY",category);
      }
      publisher.publish(forwardedMessage);
      return data;
    }
    catch (JMSException je) {
    	log.log(Level.SEVERE,"Linked Exception is: " + je.getLinkedException(), je.getLinkedException());
      throw new EJBException("Error in sending etopic: " + je.getMessage());
    } catch ( Exception e ) {
    	log.log(Level.SEVERE, "Error while forwarding data: " + data, e);
    	throw new BackBeatException("Error while forwarding data: " + data, e);
    }
    finally {
      try {
        if (publisher != null) {
          publisher.close();
        }
        if (topicSession != null) {
          topicSession.close();
        }
        if (connection != null) {
          connection.close();
        }
        if ( currentTx != null ) {
        	try {
                    if(tm != null)
        		tm.resume(currentTx);
        	} catch ( Exception ex ) {}
        }
      }
      catch (Throwable je) {
      }
    }
  }

  private void getNewTransactionManager(){
      InitialContext ic = null;
      try{
          ic = new InitialContext();
      }catch(NamingException ne){
          String err = "getNewTransactionManager is unable to obtain a "+
                       "JNDI context, therefore cannot obtain TXManager";
          log.log(Level.SEVERE,err,ne);
          return;
      }

      try{
          tm = (TransactionManager)ic.lookup(this.TX_MANAGER_NAME);
      }catch(Throwable t){
          String err = t.getClass().getName() + " from ic.lookup("+
                       TX_MANAGER_NAME + ")";
          log.log(Level.SEVERE,err,t);
      }finally{
          if(ic != null){
              try{
                  ic.close();
              }catch(Throwable eatIt){}
          }
      }
  }
}
