package org.quintet.backbeat.supplier;

import javax.naming.NamingException;
import javax.naming.InitialContext;
import java.sql.Timestamp;
import java.util.logging.Logger;
import java.util.logging.Level;
import javax.ejb.*;
import javax.jms.*;

/**
 * <code>AsynchSupplier</code> is the base MDB through which all Dealer price
 * requests are generated and transmitted.  It conforms to the Supplier Pattern.
 * @author PThaggard
 * @version 1.1
 * @since UPS Release 2004.2
 * @see org.quintet.backbeat.supplier.Supplier
 */
public class AsynchSupplier
    extends Supplier implements MessageDrivenBean, MessageListener {
  static final long serialVersionUID = 200501001;
  /**
   * Attributes declaration
   */
  private String processDescription = "No Description Available";

  /**
   * The name of the "vendor" associated with this <code>AsynchSupplier</code>  The vendor
   *  is usually taken to mean the organization that supplies data to the Supplier.
   */
  private String vendorID = null;
  private Logger log = Logger.getLogger("org.quintet.backbeat");
  private MessageDrivenContext context;

  /**
   * @roseuid 42162B4F005A
   * @J2EE_METHOD  --  AsynchSupplier
   */
  public AsynchSupplier() {

  }

  /**
   * @roseuid 42162B4F0061
   * @J2EE_METHOD  --  onMessage
   * Called by the bean's container when a message has arrived for the bean to service.
   * It contains the business logic that handles the processing of the message. Only
   * message-driven beans can asynchronously receive messages. Session and entity beans
   * are not permitted to be JMS MessageListener.
   * @param message
   */
  public void onMessage(Message message) {
    Timestamp now = new Timestamp(System.currentTimeMillis());
    log.info(now + ":Beginning Process " + processDescription);

    ObjectMessage objMessage = (ObjectMessage) message;
    Object objMsg = null;
    try {
      objMsg = objMessage.getObject();
      if (processDescription.equals("Asynchronously aggregate Observations")){
        log.log(Level.FINE,"aggregating");
      }
    }
    catch (JMSException ex1) {
      String err = "JMSException, setting rollback only";
      log.log(Level.SEVERE,err, ex1);
      return;
    }
    Object answer = doIt(objMsg);
    log.info(now + ":Successfully completed Process " + processDescription);
  }

  /**
   * @roseuid 42162B4F007D
   * @J2EE_METHOD  --  doIt
   */
  public java.lang.Object doIt(Object requestContext) {
    try {
      return acquireInformation(requestContext);
    }
    catch (SupplierException se) {
      //REMEMBER:  the rule around here is "only the thrower logs an exception"
      String err = processDescription +
          ":SupplierException, requestContext is " + requestContext.toString();
      log.log(Level.SEVERE,err);
      return null;
    }
  }

  /**
   * @roseuid 42162B4F007F
   * @J2EE_METHOD  --  ejbRemove
   * A container invokes this method before it ends the life of the message-driven object.
   * This happens when a container decides to terminate the message-driven object. This
   * method is called with no transaction context.
   */
  public void ejbRemove() {

  }


  public void closeResources(InitialContext ic){
    if(ic != null){
      try{
        ic.close();
      }catch(Throwable eatIt){}
    }
  }


  /**
   * @roseuid 42162B4F0087
   * @J2EE_METHOD  --  setMessageDrivenContext
   * Set the associated message-driven context. The container calls this method after the
   * instance creation. The enterprise Bean instance should store the reference to the
   * context object in an instance variable. This method is called with no transaction
   * context.
   * @param inContext
   * @throws javax.ejb.EJBException
   */
  public void setMessageDrivenContext(MessageDrivenContext inContext) throws
      EJBException {
    context = inContext;
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }
    catch (NamingException ne) {
      String err = "unable to obtain an InitialContext:";
      log.log(Level.SEVERE,err, ne);
      throw new EJBException(err, ne);
    }

    try {
      processDescription = (String) ic.lookup(
          "java:comp/env/process-description");
    }
    catch (NamingException ne) {
      log.info(
          "process-description not found in JNDI, defaulting to '<unknown process>'");
      processDescription = "<unknown process>";
    }

    try {
      vendorID = (String) ic.lookup("java:comp/env/dealer-id");
    }
    catch (NamingException ne) {
      String err =
          ":no dealer ID available at java:comp/env/dealer-id, using UNKNOWN";
      log.log(Level.FINEST,err);
      vendorID = "UNKNOWN";
    }

    log.info(
        "configuring requester for Supplier " + vendorID);
    try {
      EJBSupplierConfigurator.configure(this);
    }
    catch (SupplierException ex) {
      String err = ":SupplierException while configurating " + vendorID;
      log.log(Level.SEVERE,err);
      closeResources(ic);
      throw new EJBException(err, ex);
    }
    closeResources(ic);
  }

  /**
   * @roseuid 42162B4F00F6
   * @J2EE_METHOD  --  ejbCreate
   * This <code>ejbCreate</code> method is required by the EJB spec.
   */
  public void ejbCreate() {

  }

  /**
   * @return   the current value of the vendorID property
   * Access method for the vendorID property
   */
  public String getvendorID() {
    return vendorID;
  }

  /**
   * @param vendorID the new value of the vendorID property
   * Sets the value of the vendorID property
   */
  public void setvendorID(String vendorID) {
    this.vendorID = vendorID;
  }
}
