package org.quintet.backbeat.supplier;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.quintet.chambers.EntityRetriever;
import org.quintet.backbeat.dataportal.DataPortal;
import org.quintet.backbeat.io.ForwardConduit;
import org.quintet.backbeat.transformer.Transformer;
import org.quintet.backbeat.validation.Validator;
import java.util.logging.Logger;
import java.util.logging.Level;

/**
 <code>EJBSupplierConfigurator</code> Reads deployment descriptor entries and
 uses them to configure <code>inSupplier</code>.
 */
public class EJBSupplierConfigurator {
  private static final String DATA_PORTAL_CLASS_NAME_ENV_ENTRY =
      "java:comp/env/data-portal-class-name";
  private static final String RETRIEVER_CLASS_NAME_ENV_ENTRY =
      "java:comp/env/request-retriever-class-name";
  private static final String REQUEST_CONDUIT_CLASS_NAME_ENV_ENTRY =
      "java:comp/env/request-conduit-class-name";
  private static final String RESPONSE_CONDUIT_CLASS_NAME_ENV_ENTRY =
      "java:comp/env/response-conduit-class-name";
  private static final String REQUEST_TRANSFORMER_CLASS_NAME_ENV_ENTRY =
      "java:comp/env/outbound-transformer-class-name";
  private static final String RESPONSE_TRANSFORMER_CLASS_NAME_ENV_ENTRY =
      "java:comp/env/inbound-transformer-class-name";
  private static final String RESPONSE_VALIDATOR_CLASS_NAME_ENV_ENTRY =
      "java:comp/env/response-validator-class-name";
  private static final String FORWARD_CONDUIT_CLASS_NAME_ENV_ENTRY =
      "java:comp/env/forward-conduit-class-name";

  private static final Logger log = Logger.getLogger("org.quintet.backbeat");
  /**
      @roseuid 412204C403B5
   */
  public EJBSupplierConfigurator() {

  }

  public static void closeResources(InitialContext ic){
    if(ic != null){
      try{
        ic.close();
      }catch(Throwable eatIt){}
    }
  }
  /**
      Read JNDI entries and use them to configure <code>inSupplier</code>


      @param inSupplier - The <code>Supplier</code> being configured
      @throws org.quintet.backbeat.supplier.SupplierException
      @roseuid 412203A20157
   */
  public static void configure(Supplier inSupplier) throws SupplierException {
    log.log(Level.FINE,"EJBSupplierConfigurator:  starting");
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }
    catch (NamingException ex) {
      String err = "NamingException when trying to get a JNDI InitialContext";
      log.log(Level.SEVERE,err, ex);
      throw new SupplierException(err, ex);
    }

    //-----------------  configure DataPortal ----------------------
    log.log(Level.FINE,"Configuring DataPortal");
    String dataPortalClassName = null;
    try {
      dataPortalClassName = (String) ic.lookup(
          DATA_PORTAL_CLASS_NAME_ENV_ENTRY);
    }
    catch (NamingException ex1) {
      String err = "NamingException when looking up '" +
          DATA_PORTAL_CLASS_NAME_ENV_ENTRY + "'";
      log.log(Level.SEVERE,err, ex1);
      closeResources(ic);
      throw new SupplierException(err, ex1);
    }
    log.log(Level.FINE,"Data Port Class Name:  '" + dataPortalClassName + "'");
    if (dataPortalClassName != null && !dataPortalClassName.equals("") &&
        !dataPortalClassName.equals("null")) {
      Class dataPortalClass = null;
      try {
        log.log(Level.FINE,
            "data Portal Class Name is NOT null, therefore trying to load the class");
        dataPortalClass = Class.forName(dataPortalClassName);
      }
      catch (ClassNotFoundException ex2) {
        String fatal = "ClassNotFoundException when doing Class.forName on '" +
            dataPortalClassName + "'";
        log.log(Level.SEVERE,fatal, ex2);
        closeResources(ic);
        throw new SupplierException(fatal, ex2);
      }
      try {
        log.log(Level.FINE,"doing a newInstance on " + dataPortalClass.getName());
        DataPortal myDataPortal = (DataPortal) dataPortalClass.newInstance();
        inSupplier.setDataPortal(myDataPortal);
      }
      catch (IllegalAccessException ex3) {
        String fatal = "IllegalAccessException when calling '" +
            dataPortalClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex3);
        closeResources(ic);
        throw new SupplierException(fatal, ex3);
      }
      catch (InstantiationException ex3) {
        String fatal = "InstantiationException when calling '" +
            dataPortalClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex3);
        closeResources(ic);
        throw new SupplierException(fatal, ex3);
      }
      catch(IllegalArgumentException iae){
        String fatal = "iae";
        log.log(Level.SEVERE,fatal,iae);
        closeResources(ic);
        throw new SupplierException(fatal,iae);
      }
    }
    else {
      log.log(Level.FINE,
          "EJBSupplierConfiguratorBean:WARNING:setting data portal to null");
      inSupplier.setDataPortal(null);
    }

    //------------------------ configure request retriever --------------------------------
    log.log(Level.FINE,"Configuring Request Retriever");
    String retrieverClassName = null;
    try {
      retrieverClassName = (String) ic.lookup(RETRIEVER_CLASS_NAME_ENV_ENTRY);
    }
    catch (NamingException ne) {
      String fatal = "NamingException when looking up '" +
          RETRIEVER_CLASS_NAME_ENV_ENTRY + "'";
      log.log(Level.SEVERE,fatal, ne);
      closeResources(ic);
      throw new SupplierException(fatal, ne);
    }
    log.log(Level.FINE,"retriever class name is '" + retrieverClassName + "'");
    if (retrieverClassName != null && !retrieverClassName.equals("") &&
        !retrieverClassName.equals("null")) {
      Class retrieverClass = null;
      try {
        log.log(Level.FINE,"invoking '" + retrieverClassName + ".forName()'");
        retrieverClass = Class.forName(retrieverClassName);
      }
      catch (ClassNotFoundException ex) {
        String fatal = "ClassNotFoundException from '" + retrieverClassName +
            ".forName()'";
        log.log(Level.SEVERE,fatal, ex);
        closeResources(ic);
        throw new SupplierException(fatal, ex);
      }
      try {
        log.log(Level.FINE,"Invoking '" + retrieverClassName + ".newInstance()'");
        EntityRetriever _retriever = (EntityRetriever) retrieverClass.
            newInstance();
        inSupplier.setRetriever(_retriever);
      }
      catch (IllegalAccessException ex) {
        String fatal = "IllegalAccessException when invoking '" +
            retrieverClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex);
        closeResources(ic);
        throw new SupplierException(fatal, ex);
      }
      catch (InstantiationException ex) {
        String fatal = "InstantiationException when invoking '" +
            retrieverClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex);
        closeResources(ic);
        throw new SupplierException(fatal, ex);
      }
    }
    else {
      log.log(Level.FINE,"setting retriever to NULL");
      inSupplier.setRetriever(null);
    }
    //---------------------- configure request transformer -------------------------------
    log.log(Level.FINE,"Configuring Request Transformer");
    String requestTransformerClassName = null;
    try {
      requestTransformerClassName = (String) ic.lookup(
          REQUEST_TRANSFORMER_CLASS_NAME_ENV_ENTRY);
    }
    catch (NamingException ex4) {
      String fatal = "NamingException when looking up '" +
          REQUEST_TRANSFORMER_CLASS_NAME_ENV_ENTRY + "'";
      log.log(Level.SEVERE,fatal, ex4);
      closeResources(ic);
      throw new SupplierException(fatal, ex4);
    }
    log.log(Level.FINE,"requestTransformerClassName is '" + requestTransformerClassName +
              "'");
    if (requestTransformerClassName != null &&
        !requestTransformerClassName.equals("") &&
        !requestTransformerClassName.equals("null")) {
      Class requestTransformerClass = null;
      try {
        log.log(Level.FINE,"invoking '" + requestTransformerClassName + ".forName()");
        requestTransformerClass = Class.forName(requestTransformerClassName);
      }
      catch (ClassNotFoundException ex5) {
        String fatal = "ClassNotFoundException from '" +
            requestTransformerClassName + ".forName()'";
        log.log(Level.SEVERE,fatal, ex5);
        closeResources(ic);
        throw new SupplierException(fatal, ex5);
      }
      try {
        log.log(Level.FINE,"invoking '" + requestTransformerClassName +
                  ".newInstance()'");
        Transformer requestTransformer = (Transformer) requestTransformerClass.
            newInstance();
        inSupplier.setOutboundTransformer(requestTransformer);
      }
      catch (IllegalAccessException ex6) {
        String fatal = "IllegalAccessException when invoking '" +
            requestTransformerClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex6);
        closeResources(ic);
        throw new SupplierException(fatal, ex6);
      }
      catch (InstantiationException ex6) {
        String fatal = "InstantiationException when invoking '" +
            requestTransformerClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex6);
        closeResources(ic);
        throw new SupplierException(fatal, ex6);
      }
    }
    else {
      inSupplier.setOutboundTransformer(null);
    }
    //------------------------- configure response transformer --------------------------------------------
    log.log(Level.FINE,"Configuring Response Transformer");
    String responseTransformerClassName = null;
    try {
      responseTransformerClassName = (String) ic.lookup(
          RESPONSE_TRANSFORMER_CLASS_NAME_ENV_ENTRY);
    }
    catch (NamingException ex7) {
      String fatal = "NamingException when looking up '" +
          RESPONSE_TRANSFORMER_CLASS_NAME_ENV_ENTRY + "'";
      log.log(Level.SEVERE,fatal, ex7);
      closeResources(ic);
      throw new SupplierException(fatal, ex7);
    }
    log.log(Level.FINE,"Response Transformer Class Name is '" +
              responseTransformerClassName + "'");
    if (responseTransformerClassName != null &&
        !responseTransformerClassName.equals("") &&
        !responseTransformerClassName.equals("null")) {
      Class responseTransformerClass = null;
      try {
        log.log(Level.FINE,"invoking '" + responseTransformerClassName + ".forName()'");
        responseTransformerClass = Class.forName(responseTransformerClassName);
      }
      catch (ClassNotFoundException ex8) {
        String fatal = "ClassNotFoundException when invoking '" +
            responseTransformerClassName + ".forName()'";
        log.log(Level.SEVERE,fatal, ex8);
        closeResources(ic);
        throw new SupplierException(fatal, ex8);
      }
      try {
        log.log(Level.FINE,"invoking '" + responseTransformerClassName +
                  ".newInstance()'");
        Transformer responseTransformer = (Transformer)
            responseTransformerClass.newInstance();
        inSupplier.setInboundTransformer(responseTransformer);
      }
      catch (IllegalAccessException ex9) {
        String fatal = "IllegalAccessException when invoking '" +
            responseTransformerClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex9);
        closeResources(ic);
        throw new SupplierException(fatal, ex9);
      }
      catch (InstantiationException ex9) {
        String fatal = "InstantiationException when invoking '" +
            responseTransformerClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex9);
        closeResources(ic);
        throw new SupplierException(fatal, ex9);
      }
    }
    else {
      inSupplier.setInboundTransformer(null);
    }
    //-----------------  configure ForwardConduit ----------------------
    log.log(Level.FINE,"Configuring Forwarding Conduit");
    String forwardConduitClassName = null;
    try {
      forwardConduitClassName = (String) ic.lookup(
          FORWARD_CONDUIT_CLASS_NAME_ENV_ENTRY);
    }
    catch (NamingException ex1) {
      String fatal = "Unable to lookup '" +
          FORWARD_CONDUIT_CLASS_NAME_ENV_ENTRY + "'";
      log.log(Level.SEVERE,fatal, ex1);
    }
    log.log(Level.FINE,"Forward Conduit Class Name is '" +
              FORWARD_CONDUIT_CLASS_NAME_ENV_ENTRY + "'");
    if (forwardConduitClassName != null && !forwardConduitClassName.equals("") &&
        !forwardConduitClassName.equals("null")) {
      Class forwardConduitClass = null;
      try {
        log.log(Level.FINE,"invoking '" + forwardConduitClassName + ".forName()'");
        forwardConduitClass = Class.forName(forwardConduitClassName);
      }
      catch (ClassNotFoundException ex2) {
        String fatal = "ClassNotFoundException when invoking '" +
            forwardConduitClassName + ".forName()'";
        log.log(Level.SEVERE,fatal, ex2);
        closeResources(ic);
        throw new SupplierException(fatal, ex2);
      }
      try {
        log.log(Level.FINE,"invoking '" + forwardConduitClassName + ".newInstance()'");
        ForwardConduit myForwardConduit =
            (ForwardConduit) forwardConduitClass.newInstance();
        inSupplier.setForwardConduit(myForwardConduit);
      }
      catch (IllegalAccessException ex3) {
        String fatal = "IllegalAccessException when invoking '" +
            forwardConduitClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex3);
        closeResources(ic);
        throw new SupplierException(fatal, ex3);
      }
      catch (InstantiationException ex3) {
        String fatal = "InstantiationException when invoking '" +
            forwardConduitClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex3);
        closeResources(ic);
        throw new SupplierException(fatal, ex3);
      }
    }
    else {
      log.log(Level.FINE,
          "EJBSupplierConfiguratorBean:WARNING:setting forward conduit to null");
      inSupplier.setForwardConduit(null);
    }
    //------------------------ configure validator --------------------------
    log.log(Level.FINE,"Configuring Validator");
    String validatorClassName = null;
    try {
      validatorClassName = (String) ic.lookup(
          RESPONSE_VALIDATOR_CLASS_NAME_ENV_ENTRY);
    }
    catch (NamingException ex10) {
      String fatal = "unable to lookup '" +
          RESPONSE_VALIDATOR_CLASS_NAME_ENV_ENTRY + "'";
      log.log(Level.SEVERE,fatal, ex10);
      closeResources(ic);
      throw new SupplierException(fatal, ex10);
    }
    log.log(Level.FINE,"validator class is '" + validatorClassName + "'");
    if (validatorClassName != null && !validatorClassName.equals("") &&
        !validatorClassName.equals("null")) {
      Class validatorClass = null;
      try {
        log.log(Level.FINE,"invoking '" + validatorClassName + ".forName()'");
        validatorClass = Class.forName(validatorClassName);
      }
      catch (ClassNotFoundException ex11) {
        String fatal = "ClassNotFoundException when invoking '" +
            validatorClassName + ".forName()'";
        log.log(Level.SEVERE,fatal, ex11);
        closeResources(ic);
        throw new SupplierException(fatal, ex11);
      }
      try {
        log.log(Level.FINE,"invoking '" + validatorClassName + ".newInstance()'");
        Validator validator = (Validator) validatorClass.newInstance();
        inSupplier.setValidator(validator);
      }
      catch (IllegalAccessException ex12) {
        String fatal = "IllegalAccessException when invoking '" +
            validatorClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex12);
        closeResources(ic);
        throw new SupplierException(fatal, ex12);
      }
      catch (InstantiationException ex12) {
        String fatal = "InstantiationException when invoking '" +
            validatorClassName + ".newInstance()'";
        log.log(Level.SEVERE,fatal, ex12);
        closeResources(ic);
        throw new SupplierException(fatal, ex12);
      }
    }
    else {
      inSupplier.setValidator(null);
    }
    log.log(Level.FINE,"finished configuring");
    closeResources(ic);
  }
}
