package org.quintet.backbeat.supplier;

import org.quintet.chambers.EntityRetriever;
import org.quintet.chambers.RetrieverException;
import org.quintet.backbeat.BackBeatException;
import org.quintet.backbeat.dataportal.DataPortal;
import org.quintet.backbeat.io.ForwardConduit;
import org.quintet.backbeat.transformer.Transformer;
import org.quintet.backbeat.transformer.TransformerException;
import org.quintet.backbeat.validation.ValidationException;
import org.quintet.backbeat.validation.Validator;

/**
<code>Supplier</code> requests and obtains data from an external
data source
 */
public abstract class Supplier
{

   /**
    * <code>outputTransformer</code> converts requests from internal canonical form (most likely
    *  an entity bean) to a externally-palatable form such as XML or CSV.
    */
   private Transformer outboundTransformer = null;

   /**
   <code>inputTransformer</code> converts an external format, such as XML or CSV, into an internal canonical form, such as an entity bean instance.
    */
   private Transformer inboundTransformer = null;

   /**
   <code>dataPortal</code> is the mechanism by which this <code>Supplier</code> sends requests, obtains responses, or both
    */
   private DataPortal dataPortal = null;

   /**
   <code>forwardConduit</code> is the mechanism by which this
   <code>Supplier</code> forwards requests
    */
   private ForwardConduit forwardConduit = null;

   /**
   <code>retriever</code> is the mechanism by which this <code>Supplier</code> reconstitutes a full request from the
   request's primary key
    */
   private EntityRetriever retriever = null;
   private Validator validator;

   /**
   @roseuid 40FE4EF70254
    */
   public Supplier()
   {

   }

   /**
   Access method for the outboundTransformer property.

   @return   the current value of the outboundTransformer property
    */
   public Transformer getOutboundTransformer()
   {
    return outboundTransformer;
   }

   /**
   Sets the value of the outboundTransformer property.

   @param aOutboundTransformer the new value of the outboundTransformer property
    */
   public void setOutboundTransformer(Transformer aOutboundTransformer)
   {
      outboundTransformer = aOutboundTransformer;
   }

   /**
   Access method for the inboundTransformer property.

   @return   the current value of the inboundTransformer property
    */
   public Transformer getInboundTransformer()
   {
    return inboundTransformer;
   }

   /**
   Sets the value of the inboundTransformer property.

   @param aInboundTransformer the new value of the inboundTransformer property
    */
   public void setInboundTransformer(Transformer aInboundTransformer)
   {
      inboundTransformer = aInboundTransformer;
   }

   /**
   Access method for the dataPortal property.

   @return   the current value of the dataPortal property
    */
   public DataPortal getDataPortal()
   {
      return dataPortal;
   }

   /**
   Sets the value of the dataPortal property.

   @param aDataPortal the new value of the dataPortal property
    */
   public void setDataPortal(DataPortal aDataPortal)
   {
      dataPortal = aDataPortal;
   }

   /**
   Access method for the forwardConduit property.

   @return   the current value of the forwardConduit property
    */
   public ForwardConduit getForwardConduit()
   {
      return forwardConduit;
   }

   /**
   Sets the value of the forwardConduit property.

   @param aForwardConduit the new value of the forwardConduit property
    */
   public void setForwardConduit(ForwardConduit aForwardConduit)
   {
       forwardConduit = aForwardConduit;
   }

   /**
   Access method for the validator property.

   @return   the current value of the validator property
    */
   public Validator getValidator()
   {
      return validator;
   }

   /**
   Sets the value of the validator property.

   @param aValidator the new value of the validator property
    */
   public void setValidator(Validator aValidator)
   {
      validator = aValidator;
   }

   /**
   Retrieve the request that this <code>Supplier</code> is currently processing.
   Typically the client sends an entity's primary key, and
   <code>retrieveRequest</code>'s job is to fetch the entity identified by that
   primary key.
   @param requestIdentifier - The unique ID of the request being retrieved
   @return java.lang.Object

   @throws org.quintet.backbeat.supplier.SupplierException
   @roseuid 40FE4B1800EF
    */
   protected Object retrieveRequest(Object requestIdentifier) throws SupplierException
   {
		EntityRetriever fetch = getRetriever();
		if(fetch == null)
			return requestIdentifier;
		else{
			try{
				return fetch.retrieve(requestIdentifier);
			}catch(RetrieverException re){
				throw new SupplierException("Supplier:retrieveRequest:RetrieveException:"+re.getMessage(),re);
			}
		}
   }

   /**
   Transform a request from <code>internalForm</code> to some
   externally-acceptable form.  Return the externally-acceptable form to the
   caller.
   @param internalForm - The object being transformed
   @return java.lang.Object
   @throws org.quintet.backbeat.supplier.SupplierException
   @roseuid 40FE4BCB020F
    */
   protected Object transformRequest(Object internalForm) throws SupplierException
   {
    Transformer xform = getOutboundTransformer();
	 if(xform == null)
	 	return internalForm;
    try {
      return xform.transform(internalForm);
    }
    catch (TransformerException ex) {
      throw new SupplierException("Supplier:transformRequest:TransformerException:"+ex.getMessage(),ex);
    }
   }

   /**
   Send an externally-palatable request object to its intended recipient via the
   <code>dataPortal</code>.  If the <code>dataPortal</code> returns a non-null
   response (indicating that it forwarded the request AND received a response as
   one atomic operation) then return the response to the caller, otherwise return
   null.
   @param externalForm - The externally-palatable request object being sent
   @return java.lang.Object
   @throws org.quintet.backbeat.supplier.SupplierException
   @roseuid 40FE4C7D018C
    */
   protected Object forwardRequest(Object externalForm) throws SupplierException
   {
    DataPortal portal = getDataPortal();
	 if(portal == null)
	 	return externalForm;
    try {
      //REMEMBER:  if the portal manages to get a response immediately
      //           then it will return that response; else it'll return null
      return portal.getData(externalForm);
    }
    /*
    catch (IOException ex) {
      throw new SupplierException("Supplier:forwardRequest:IOException:"+ex,ex);
    }
}*/

    catch (BackBeatException backbeatEx) {
      backbeatEx.printStackTrace();
      throw new SupplierException("Supplier:forwardRequest:IOException:"+backbeatEx.getMessage(),backbeatEx);
    }

   }

   /**
   Receive a response via the <code>dataPortal</code>.  If this
   <code>Supplier</code>'s <code>dataPortal</code> is of an active nature (that
   is, it sends requests and receives responses as one atomic transaction) then
   <code>receiveReponse</code> will never be invoked by
   <code>acquireInformation</code>; <code>receiveResponse</code> is only called if
   <code>forwardRequest</code> returns <code>null</code>.
   If a <code>dataPortal</code> doesn't send requests (a read-only portal, in
   essence), then <code>receiveResponse</code> can be expected to be invoked.
   @param inContext - The context in which a response is being solicited; this is
   typically some sort of request object.
   @return java.lang.Object
   @throws org.quintet.backbeat.supplier.SupplierException
   @roseuid 40FE4D20026F
    */
   protected Object receiveResponse(Object inContext) throws SupplierException
   {
    DataPortal portal = getDataPortal();
	 if(portal == null)
	 	return inContext;
    try {
      return portal.getData(inContext);
    }
    /*
    catch (IOException ex) {
      throw new SupplierException("Supplier:receiveResponse:IOException:"+ex,ex);
    }
*/
    catch (BackBeatException ex) {
     throw new SupplierException("Supplier:receiveResponse:IOException:"+ex.getMessage(),ex);
   }

   }

   /**
   Transform <code>externalForm</code> to the accepted internal form and then
   return the internal form to the caller.
   For example, an XML document (the external form) might be passed into
   <code>transformResponse</code>, and it might convert it to a
   <code>PriceResponse</code> entity instance.
   @param externalForm - The external form being transformed to an internal form
   @return java.lang.Object
   @throws org.quintet.backbeat.supplier.SupplierException
   @roseuid 40FE4D7A00D7
    */
   protected Object transformResponse(Object externalForm) throws SupplierException
   {
    Transformer xform = getInboundTransformer();
	 if(xform == null)
	 	return externalForm;
    try {
      return xform.transform(externalForm);
    }
    catch (TransformerException ex) {
      throw new SupplierException("Supplier:transformResponse:TransformerException:"+ex.getMessage(),ex);
    }
   }

   /**
   Transmit <code>internalForm</code> to its ultimate recipient.  The "ultimate
   recipient" might be the original caller of this <code>Supplier</code>, or it
   might be some third party such as a JMS Topic.
   If <code>forwardResponse</code> actually does forward a response to
   someone/something/somewhere, then it should also return the thing that it
   forwarded.  The <code>acquireInformation</code> method (which is the only
   method that should ever call <code>forwardResponse</code> will return
   <code>forwardResponse</code>'s return value to whomever called
   <code>acquireInformation</code> in the first place.
   @param internalForm - The internally-palatable response being forwarded
   @return java.lang.Object
   @roseuid 40FE4E1603E0
    */
    protected Object forwardResponse(Object internalForm) throws SupplierException
    {
        ForwardConduit conduit = getForwardConduit();

        if (conduit == null || internalForm == null || internalForm instanceof Void )
        {
            return internalForm;
        }

        try
        {
            return conduit.forwardData(internalForm);
        }
        catch (BackBeatException backbeatEx)
        {
            backbeatEx.printStackTrace();
            throw new SupplierException(
                "Supplier:forwardResponse:IOException:" + backbeatEx.getMessage(),
                backbeatEx);
        }
    }

   /**
   The <code>acquireInformation</code> method is <code>Supplier</code>'s main
   processing loop.  It performs the following steps:<br>
   <li>Obtain caller's request
   <li>Transform request to external format
   <li>Forward request to external data source
   <li>Receive response from external data source
   <li>Transform response into internal format
   <li> Forward response to caller
   Concrete <code>Supplier</code>s can choose to implement any number of these
   steps as no-ops; for example, a passive acquirer of data might choose to not
   implement the "obtain caller's request" operation.
   It is possible for several of these steps to be combined into one.  For
   example, if <code>DataPortal</code>'s <code>getData</code> method both accepts
   a request and returns a response, then the "forward request to external data
   source" and "receive request from external data source" operations are executed
   as one atomic operation.
   @param inRequestID - The ID of the request that goaded this
   <code>Supplier</code> into action
   @return java.lang.Object
   @throws org.quintet.backbeat.supplier.SupplierException
   @roseuid 40FE55FA02BA
    */
   public Object acquireInformation(Object inRequestID) throws SupplierException
   {
    Object originalRequest = retrieveRequest(inRequestID);
    Object externalRequestForm = transformRequest(originalRequest);
    Object externalResponseForm = forwardRequest(externalRequestForm);
    if(externalResponseForm == null)
      externalResponseForm = receiveResponse(originalRequest);

    Object internalResponseForm = transformResponse(externalResponseForm);
    Object validationResult = null;
    try{
      validationResult = validate(internalResponseForm);
      return forwardResponse(internalResponseForm);
    }catch(ValidationException ve){
      throw new SupplierException("Supplier:acquireInformation:validation failed:"+ve.getMessage());
    }
   }

   /**
   @return org.quintet.chambers.EntityRetriever
   @roseuid 4121F36E02CD
    */
   public EntityRetriever getRetriever()
   {
		return retriever;
   }

   /**
   @param inValue
   @roseuid 4121F36E02EB
    */
   public void setRetriever(EntityRetriever inValue)
   {
		retriever = inValue;
   }

   /**
   Validate <code>inObject</code> by invoking this <code>Supplier</code>'s
   <code>validator</code>.  Return the validated object if validation succeeds.
   Propagate <code>ValidationException</code> if validation fails.
   @param inObject - The object being validated
   @return Object
   @throws org.quintet.backbeat.validation.ValidationException
   @roseuid 4121F76302B5
    */
   public Object validate(Object inObject) throws ValidationException
   {
    Validator vlad = getValidator();
    if(vlad == null)
      return inObject;
    else
      return vlad.validate(inObject);
   }
}
