package org.quintet.backbeat.transformer;

import java.beans.Introspector;
import java.beans.BeanInfo;
import java.beans.PropertyDescriptor;
import java.util.Collection;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import java.lang.reflect.Method;
import java.beans.IntrospectionException;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.lang.reflect.*;

/**
 <code>JavaBeanCollectionToMapCollection</code> accepts a
 <code>Collection</code> of JavaBeans-compliant objects.  It converts each such
 JavaBean into a <code>Map</code> whose keys are attribute names and whose
 values are attribute values.
 */
public class JavaBeanCollectionToMapCollection
    implements Transformer {
  private static Logger log = Logger.getLogger("org.quintet.backbeat");
  ;

  /**
      @roseuid 420277ED01A1
   */
  public JavaBeanCollectionToMapCollection() {

  }

  /**
      <code>JavaBeanCollectionToMapCollection</code> accepts a
   <code>Collection</code> of JavaBeans-compliant objects.  It converts each such
   JavaBean into a <code>Map</code> whose keys are attribute names and whose
      values are attribute values.
      This method returns a <code>Collection</code> containing the transformed
      <code>Map</code>s.
   @param inRawData - The <code>Collection</code> of JavaBeans-compliant objects
      that will be transformed into <code>Map</code>s
      @return Object
      @throws TransformerException
      @roseuid 420277ED01AB
   */
  public Object transform(Object inRawData) throws TransformerException {
    if (inRawData == null) {
      return null;
    }
    Collection input = null;
    if ( (inRawData instanceof Collection) == false) {
      String warn = "Expected a java.util.Collection, got a " +
          inRawData.getClass().getName() + ".  Wrapping it in a Collection on your behalf.";
      log.warning(warn);
      input = new ArrayList(1);
      input.add(inRawData);
    }else{
      input = (Collection) inRawData;
    }
    Collection output = new ArrayList(input.size());
    for (Iterator inputIter = input.iterator(); inputIter.hasNext(); ) {
      Object currBean = inputIter.next();
      if (currBean == null) {
        output.add(new HashMap());
        continue;
      }
      try {
        BeanInfo beanInfo = Introspector.getBeanInfo(currBean.getClass());
        if (beanInfo == null) {
          output.add(new HashMap());
          continue;
        }
        PropertyDescriptor[] properties = beanInfo.getPropertyDescriptors();
        if (properties == null || properties.length == 0) {
          output.add(new HashMap());
          continue;
        }
        Map propertyAsMap = new HashMap(properties.length + 1);
        //propertyAsMap.put("class-name", currBean.getClass().getName());
        for (int propIdx = 0; propIdx < properties.length; propIdx++) {
          PropertyDescriptor currProperty = properties[propIdx];
          String propertyName = currProperty.getName();
          if(propertyName.equals("class")){
            continue;//not a good idea to try to flatten a java.lang.Class
          }
          Method getter = currProperty.getReadMethod();
          if (getter == null) {
            propertyAsMap.put(propertyName, null);
          }
          else {
            Class declaredIn = getter.getDeclaringClass();
            log.info ("method " + getter.getName() + "declared in " + declaredIn.getName());
            if(declaredIn.equals(currBean.getClass()) == false){
              String info = getter.getName() +" not declared in " + currBean.getClass().getName() + ", skipping....";
              log.info(info);
              continue;
            }
            try {
              Object propertyValue = getter.invoke(currBean, (Object[])null);
              //do not put unserializable objects into the Map
              if(propertyValue instanceof java.io.Serializable){
                propertyAsMap.put(propertyName, propertyValue);
              }else{
                log.info("object of class " + propertyValue + " is not Serializable, skipping");
              }
            }
            catch (InvocationTargetException ex1) {
              Throwable orig = ex1.getCause();
              String err = "Method " + getter.getName() + " of class " +
                  currBean.getClass().getName() + " threw an exception:  " +
                  orig;
              log.log(Level.SEVERE,err, orig);
              throw new TransformerException(err, ex1);
            }
            catch (IllegalArgumentException ex1) {
              String err = "IllegalArgumentException when invoking " +
                  currBean.getClass().getName() + "." + getter.getName();
              log.log(Level.SEVERE,err, ex1);
              throw new TransformerException(err, ex1);
            }
            catch (IllegalAccessException ex1) {
              String err = "IllegalArumentException when invoking " +
                  currBean.getClass().getName() + "." + getter.getName();
              log.log(Level.SEVERE,err, ex1);
              throw new TransformerException(err, ex1);

            }
          }
        }
        output.add(propertyAsMap);
      }
      catch (IntrospectionException ex) {
        String err = "Unable to introspect object of class " + currBean.getClass().getName();
        log.log(Level.SEVERE,err,ex);
        throw new TransformerException(err,ex);
      }
    }
    return output;
  }
}
