package org.quintet.backbeat.transformer;

import java.lang.reflect.Method;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.lang.reflect.*;

/**
 * <code>ObjectWrapperTransformer</code> "transforms" an object by wrapping it
 * inside another object.  The following &lt;env-entry&gt;s determine what gets
 * wrapped, what it gets wrapped into, and what method is invoked to achieve the
 * wrapping:<br>
 * <li>wrapper-class-name indicates the name of the wrapper class
 * <li>wrapper-class-setter-method-name indicates the name of the wrapper class
 * method to call to "inject" the wrapped object into the wrapper object
 * <li>wrapper-class-setter-arg-class-name indicates the type of object that will be
 * passed to the wrapper-class-setter-method<br>
 * <p>Copyright: Copyright (c) 2006</p>
 *
 * <p>Company: 6th Sense Analytics</p>
 *
 * @author Peter Thaggard
 * @version 1.0
 */
public class ObjectWrapperTransformer implements Transformer {

  private static final Logger log = Logger.getLogger("org.quintent.backbeat");

  /**
   * The <code>wrapperClass</code> attribute houses the <code>Class</code> of
   * the wrapper class.  It is used to reflectively construct new
   * wrapper class instances.
   */
  private Class wrapperClass = null;

  /**
   * The <code>wrapperClassSetter</code> attribute houses the <code>Method</code>
   * that is invoked to "inject" a wrapped object into a wrapper object.
   */
  private Method wrapperClassSetter = null;

  /**
   * The <code>wrapperSetterArgClass</code> attribute houses the <code>Class</code>
   * of the objects that will be wrapped by this <code>ObjectWrapperTransformer</code>.
   */
  private Class wrapperSetterArgClass = null;

  public void closeResources(InitialContext ic){
    if(ic != null){
      try{
        ic.close();
      }catch(Throwable eatIt){}
    }
  }

  public ObjectWrapperTransformer() {}

  private void j2eeInit(){
    InitialContext ic = null;
    try{
      ic = new InitialContext();
    }catch(NamingException ne){
      String err = "Unable to obtain a JNDI InitialContext -- "+
          "either the app server is down or you are using this class "+
          "outside a J2EE app server's JVM.  This class is only usable "+
          "withing a J2EE app server.";
      log.log(Level.SEVERE,err,ne);
      throw new IllegalStateException(err,ne);
    }

    final String WRAPPER_CLASS_NAME_ENTRY =
        "java:comp/env/wrapper-class-name";
    String wrapperClassName = null;
    try{
      wrapperClassName = (String)ic.lookup(WRAPPER_CLASS_NAME_ENTRY);
      setWrapperClass(Class.forName(wrapperClassName));
    }catch(NamingException ne){
      String err = "Unable to find \"" + WRAPPER_CLASS_NAME_ENTRY +
          "\" in JNDI.  Please check the deployment descriptor.";
      log.log(Level.SEVERE,err,ne);
      closeResources(ic);
      throw new IllegalStateException(err,ne);
    }catch(ClassNotFoundException cnfe){
      String err = "Unable to load a classs named \"" + wrapperClassName +
          "\".  Please check your build script to make sure the class is "+
          "being included in the application's JAR/WAR/EAR.";
      log.log(Level.SEVERE,err,cnfe);
      closeResources(ic);
      throw new IllegalStateException(err,cnfe);
    }


    final String WRAPPER_CLASS_SETTER_ARG_CLASS_NAME_ENTRY =
        "java:comp/env/wrapper-class-setter-arg-class-name";
    String setterArgClassName = null;
    try{
      setterArgClassName = (String)ic.lookup(WRAPPER_CLASS_SETTER_ARG_CLASS_NAME_ENTRY);
      setWrapperSetterArgClass(Class.forName(setterArgClassName));
    }catch(NamingException ne){
      String err = "Unable to find \"" +
          WRAPPER_CLASS_SETTER_ARG_CLASS_NAME_ENTRY +
          "\" in JNDI.  Please check the deployment descriptor.";
      log.log(Level.SEVERE,err,ne);
      closeResources(ic);
      throw new IllegalStateException(err,ne);
    }catch(ClassNotFoundException cnfe){
      String err = "Unable to load a classs named \"" + setterArgClassName +
          "\".  Please check your build script to make sure the class is "+
          "being included in the application's JAR/WAR/EAR.";
      log.log(Level.SEVERE,err,cnfe);
      closeResources(ic);
      throw new IllegalStateException(err,cnfe);
    }


    final String WRAPPER_CLASS_SETTER_METHOD_NAME_ENTRY =
        "java:comp/env/wrapper-class-setter-method-name";
    String setterName = null;
    try{
      setterName = (String)ic.lookup(WRAPPER_CLASS_SETTER_METHOD_NAME_ENTRY);
      Class o = java.lang.Object.class;
      Method m = wrapperClass.getMethod(setterName,new Class[]{o});
      setWrapperClassSetter(m);
    }catch(NamingException ne){
      String err = "Unable to find \"" +
          WRAPPER_CLASS_SETTER_METHOD_NAME_ENTRY +
          "\" in JNDI.  Please check the deployment descriptor.";
      log.log(Level.SEVERE,err,ne);
      closeResources(ic);
      throw new IllegalStateException(err,ne);
    }catch(NoSuchMethodException cnfe){
      String err = "Unable to load a method named \"" +
          wrapperClass.getName() + "." + setterName +
          "\".  Please check your build script to make sure the class is "+
          "being included in the application's JAR/WAR/EAR.";
      log.log(Level.SEVERE,err,cnfe);
      closeResources(ic);
      throw new IllegalStateException(err,cnfe);
    }
    closeResources(ic);
  }

  /**
   * Transform <code>inRawData</code> to a new manifestation and return the new
   * manifestation to the client.
   *
   * @param inRawData - The raw data to be transformed to a new manifestation
   * @return java.lang.Object
   * @throws TransformerException
   * @todo Implement this org.quintet.backbeat.transformer.Transformer method
   */
  public Object transform(Object inRawData) throws TransformerException {
    //Preconditions
    //   Precondition 1 -- inRawData IS NOT NULL
    if(null == inRawData){
      String err = "Precondition 1 failed -- inRawData IS NULL";
      log.log(Level.WARNING,err);
      return null;
    }

    //  Precondition 2 -- inRawData is an instance, subclass, or implementer
    //  of wrapperSetterArgClass
    Class rawDataClass = inRawData.getClass();
    if(getWrapperSetterArgClass().isAssignableFrom(rawDataClass) == false){
      String err = "Precondition 2 failed -- expected a class that is assignable from " +
          getWrapperSetterArgClass().getName() + ", found a " +
          rawDataClass.getName();
      log.log(Level.WARNING,err);
      throw new TransformerException(err);
    }
    //End of Preconditions


    //Construct a new Wrapper instance
    Object wrapper = null;
    try {
      wrapper = getWrapperClass().newInstance();
    } catch (Exception ex) {
      String err = ex.getClass().getName() + " from "+
          getWrapperClass().getName() + "::" + getWrapperClass() + "()";
      log.log(Level.SEVERE,err,ex);
      throw new TransformerException(err,ex);
    }

    //invoke the wrapper setter method to "inject" the wrapped object into
    //the wrapper object
    try {
      this.getWrapperClassSetter().invoke(wrapper, new Object[] {inRawData});
    } catch (Exception ex){
      String err = ex.getClass().getName() + " from "+
          getWrapperClass().getName() + "::" +
          getWrapperClassSetter().getName() + "()";
      log.log(Level.SEVERE,err,ex);
      throw new TransformerException(err,ex);
    }

    return wrapper;

  }

  /**
   * Mutator for the <code>wrapperClass</code> attribute.
   * The <code>wrapperClass</code> attribute houses the <code>Class</code> of
   * the wrapper class.  It is used to reflectively construct new
   * wrapper class instances.
   * @param wrapperClass Class
   */
  public void setWrapperClass(Class wrapperClass) {
    this.wrapperClass = wrapperClass;
  }

  public void setWrapperClassName(String inName){
    Class _class = null;
    try{
      _class = Class.forName(inName);
      setWrapperClass(_class);
    }catch(Throwable t){
      String err = t.getClass().getName() +
          " from Class.forName(\""+inName+"\")";
      log.log(Level.SEVERE,err,t);
      throw new IllegalStateException(err,t);
    }
  }

  /**
   * Mutator for the <code>wrapperClassSetter</code> attribute.
   * The <code>wrapperClassSetter</code> attribute houses the <code>Method</code>
   * that is invoked to "inject" a wrapped object into a wrapper object.
   * @param wrapperClassSetter Method
   */
  public void setWrapperClassSetter(Method wrapperClassSetter) {
    this.wrapperClassSetter = wrapperClassSetter;
  }

  public void setWrapperClassSetterName(String inName){
    Method meth = null;
    Class setterClass = null;
    Class _setterArgClass = null;
    try{
      setterClass = this.getWrapperClass();
      if(setterClass == null){
        String err = "wrapper class not set yet";
        log.log(Level.SEVERE,err);
        throw new IllegalStateException(err);
      }
      _setterArgClass = this.getWrapperSetterArgClass();
      if(_setterArgClass == null){
        String err = "wrapper setter arg class not set yet";
        log.log(Level.SEVERE,err);
        throw new IllegalStateException(err);
      }
      meth = setterClass.getMethod(inName,new Class[]{_setterArgClass});
      setWrapperClassSetter(meth);
    }catch(NoSuchMethodException nsme){
      String err = "No Such Method " + setterClass.getName() +
          "."+inName+"("+_setterArgClass.getName() + ")";
      log.log(Level.SEVERE,err,nsme);
      throw new IllegalStateException(err,nsme);
    }
  }
  /**
   * Mutator for the <code>wrapperSetterArgClass</code> attribute.
   * The <code>wrapperSetterArgClass</code> attribute houses the <code>Class</code>
   * of the objects that will be wrapped by this <code>ObjectWrapperTransformer</code>.
   * @param wrapperSetterArgClass Class
   */
  public void setWrapperSetterArgClass(Class wrapperSetterArgClass) {
    this.wrapperSetterArgClass = wrapperSetterArgClass;
  }

  public void setWrapperSetterArgClassName(String inName){
    Class _class = null;
    try{
      _class = Class.forName(inName);
      setWrapperSetterArgClass(_class);
    }catch(Throwable t){
      String err = t.getClass().getName() +
          " from Class.forName(\""+inName+"\")";
      log.log(Level.SEVERE,err,t);
      throw new IllegalStateException(err,t);
    }
  }
  /**
   * Accessor for the <code>wrapperClass</code> attribute.
   * The <code>wrapperClass</code> attribute houses the <code>Class</code> of
   * the wrapper class.  It is used to reflectively construct new
   * wrapper class instances.
   * @return Class
   */
  public Class getWrapperClass() {
    return wrapperClass;
  }

  /**
   * Accessor for the <code>wrapperClassSetter</code> attribute.
   * The <code>wrapperClassSetter</code> attribute houses the <code>Method</code>
   * that is invoked to "inject" a wrapped object into a wrapper object.
   * @return Method
   */
  public Method getWrapperClassSetter() {
    return wrapperClassSetter;
  }


  /**
   * Accessor for the <code>wrapperSetterArgClass</code> attribute.
   * The <code>wrapperSetterArgClass</code> attribute houses the <code>Class</code>
   * of the objects that will be wrapped by this <code>ObjectWrapperTransformer</code>.
   * @return Class
   */
  public Class getWrapperSetterArgClass() {
    return wrapperSetterArgClass;
  }
}
