package org.quintet.backbeat.transformer;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.rmi.RemoteException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.ejb.CreateException;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;

import org.quintet.chambers.EntityException;
import org.quintet.chambers.EntityFactory;
import org.quintet.chambers.EntityFactoryHome;

/**
  *Transform an XML String into a graph of object instances and relationships.
  *<code>XMLToEntityTransformer</code> is a thin veneer atop <code>EntityFactory</code>,
  * which in turn is a thin veneer atop <code>Monk</code>
  @see org.quintet.chambers.EntityFactory
  @see org.quintet.monk.Monk
 */
public class XMLToEntityTransformer
    implements Transformer {


  private static final Logger log = Logger.getLogger("org.quintet.backbeat");
  static {

  }

  /**
      The ejb-ref name by which we look up <code>EntityFactoryHome</code>
   */
  private static final String ENTITY_FACTORY_REF_NAME =
      "java:comp/env/ejb/entity-factory-ref";

  private static final String ENTITY_BLUEPRINTS_KEY =
      "java:comp/env/entity-blueprints-file";

  /**
      <code>XMLToEntityTransformer</code> delegates its work to <code>EntityFactory</code>
   */
  private EntityFactory entityFactory;

  /**
      <code>XMLToEntityTransformer</code> fetches its <code>entityFactory</code> via a call to <code>entityFactoryHome</code>.
   */
  private EntityFactoryHome entityFactoryHome;

  /**
      @throws TransformerException
      @roseuid 412B7B940370
   */
  public XMLToEntityTransformer() throws TransformerException {
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }
    catch (NamingException ne) {
      String err = ":unable to obtain JNDI InitialContext";
      log.log(Level.SEVERE,err, ne);
      throw new TransformerException(err, ne);
    }
    InputStream entityBlueprintsStream = null;
    String entityBlueprintsName=null;
    try{
      entityBlueprintsName = (String)ic.lookup(ENTITY_BLUEPRINTS_KEY);
    }catch(NamingException ne){
      String fatal = "Unable to find an environment variable named '" + ENTITY_BLUEPRINTS_KEY + "'";
      log.log(Level.SEVERE,fatal,ne);
      if(ic != null){
        try{
          ic.close();
        }catch(Throwable eatIt){}
      }
      throw new TransformerException(fatal,ne);
    }

    Object eFactoryHomeObject = null;
    try {
      eFactoryHomeObject = ic.lookup(ENTITY_FACTORY_REF_NAME);
      if(ic != null){
        try{
          ic.close();
        }catch(Throwable eatIt){}
      }
    }
    catch (NamingException ne) {
      String err = ":unable to find JNDI entry " + ENTITY_FACTORY_REF_NAME;
      log.log(Level.SEVERE,err, ne);
      if(ic != null){
        try{
          ic.close();
        }catch(Throwable eatIt){}
      }
      if(entityBlueprintsStream != null){
        try{
          entityBlueprintsStream.close();
        }catch(Throwable eatIt){}
      }
      throw new TransformerException(err, ne);
    }
    try {
      EntityFactoryHome eFactoryHome = (EntityFactoryHome) PortableRemoteObject.
          narrow(eFactoryHomeObject, EntityFactoryHome.class);
      entityFactory = eFactoryHome.create(entityBlueprintsName);
    }
    catch (CreateException ce) {
      String err =  ":unable to create EntityFactory";
      log.log(Level.SEVERE,err, ce);
      throw new TransformerException(err, ce);
    }
    catch (RemoteException re) {
      String err =  ":RemoteException";
      log.log(Level.SEVERE,err, re);
      throw new TransformerException(err, re);
    }finally{
      if(entityBlueprintsStream != null){
        try{
          entityBlueprintsStream.close();
        }catch(Throwable eatIt){}
      }
    }
  }

  /**
      Access method for the entityFactory property.

      @return   the current value of the entityFactory property
   */
  public EntityFactory getEntityFactory() {
    return entityFactory;
  }

  /**
      Sets the value of the entityFactory property.

      @param aEntityFactory the new value of the entityFactory property
   */
  public void setEntityFactory(EntityFactory aEntityFactory) {
    entityFactory = aEntityFactory;
  }

  /**
      Access method for the entityFactoryHome property.

      @return   the current value of the entityFactoryHome property
   */
  public EntityFactoryHome getEntityFactoryHome() {
    return entityFactoryHome;
  }

  /**
      Sets the value of the entityFactoryHome property.

      @param aEntityFactoryHome the new value of the entityFactoryHome property
   */
  public void setEntityFactoryHome(EntityFactoryHome aEntityFactoryHome) {
    entityFactoryHome = aEntityFactoryHome;
  }

  /**
      @param inRawData
      @return java.lang.Object
      @throws org.quintet.backbeat.transformer.TransformerException
      @roseuid 412B7AFF03B2
   */
  public Object transform(Object inRawData) throws TransformerException {
    if (inRawData == null) {
      String warn = ":null input, therefore returning null";
      log.info(warn);
      return null;
    }
    String xmlString = null;
    //The inRawData object can be either a String or an InputStream.
    //If it is an InputStream then we should NOT close the InputStream, as
    //that is the caller's resonsibility.
    boolean iShouldCloseIt = false;
    if (inRawData instanceof java.lang.String) { //XML in String, not file name
      xmlString = (String) inRawData;
    }else if (inRawData instanceof java.io.InputStream){
      log.info("Received an inputstream, loading its contents into a String");
      iShouldCloseIt = false;
      xmlString = getStringFromInputStream((InputStream)inRawData);
    }
    else {
      String err = ":expected a String or InputStream, " +
          "received a " + inRawData.getClass().getName();
      log.log(Level.SEVERE,err);
      throw new TransformerException(err);
    }

    Object answer = null;
    try {
      answer = entityFactory.convertToFullEntity(xmlString);
      if (answer instanceof java.io.Serializable) {
        return answer;
      }
      else {
        log.warning("Object returned from EntityFactory is not Serializable, "+
                    "therefore returning null to caller");
        return null;
      }
    }
    catch (EntityException ee) {
      String err = ":EntityException";
      log.log(Level.SEVERE,err, ee);
      Throwable original = ee.getOriginalException();
      if (original != null) {
        log.log(Level.SEVERE,err + ":original exception is ", original);
        throw new TransformerException(err, original);
      }
      else {
        throw new TransformerException(err, ee);
      }
    }
    catch (RemoteException ex) {
      String err = ":RemoteException";
      log.log(Level.SEVERE,err, ex);
      throw new TransformerException(err, ex);
    }catch(Throwable t){
        String err = t.getClass().getName() +
                     "when trying to convert XML to an Entity";
        log.log(Level.SEVERE,err,t);
        throw new TransformerException(err,t);
    }
  }

  private String getStringFromInputStream(InputStream is) throws TransformerException{
    StringBuffer answer = new StringBuffer();
    InputStreamReader isr = null;
    try {
      isr = new java.io.InputStreamReader(is);
    } catch (Throwable ex) {
      String err = "\"UTF-8\" is an unsupported encoding";
      log.log(Level.SEVERE,err,ex);
      throw new TransformerException(err,ex);
    }
    int numBytes = 0;
    try{
      numBytes = is.available();
    }catch(IOException ioe){
      String err = "IOException when invoking is.available()";
      log.log(Level.SEVERE,err,ioe);
      throw new TransformerException(err,ioe);
    }

    try{
      char[] contents = new char[numBytes];
      isr.read(contents);
      /*for(int charIdx=0;charIdx < contents.length;charIdx++){
        if(contents[charIdx] >= 127)
          contents[charIdx] = 32;
      }*/
      answer.append(contents);
      return answer.toString();
    }catch(IOException ioe){
      String err = "IOException when invoking isr.read(char[])";
      log.log(Level.SEVERE,err,ioe);
      throw new TransformerException(err,ioe);
    }
  }
}
