package org.quintet.backbeat.transformer;

import org.quintet.monk.Monk;
import org.quintet.monk.MonkException;
import org.quintet.monk.MonkConfigurationException;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.IOException;


/**
 * <code>XMLToPOJOTransformer</code> transforms POJO entities from their
 * XML manifestations into their concomitant Java manifestations.  It does so
 * by utilizing Monk; <code>XMLToPOJOTransformer</code> is essentially a
 * <b>Backbeat</b> adapter to <code>Monk</code>.
 *
 * <p>Copyright: Copyright (c) 2006</p>
 *
 *
 * @author Peter Thaggard
 * @version 1.0
 */
public class XMLToPOJOTransformer implements Transformer {
  private static final Logger log = Logger.getLogger("org.quintet.backbeat");

  private static final String ENTITY_BLUEPRINTS_KEY = "java:comp/env/entity-blueprints-file";
  private Monk monk = null;

  public XMLToPOJOTransformer() {
    //first find the name of the blueprints file that describes how
    //monk will transcribe POJOs on my behalf.
    InitialContext ic = null;
    try{
      ic = new InitialContext();
    }catch(NamingException ne){
      String err = "Unable to obtain a JNDI InitialContext.  Either the "+
          "App Server is down or (more likely) you are trying to use " +
          "XMLToPOJOTransformer outside the App Server, which is not possible."+
          "  If you wish to transform XML into POJO instances outside an App "+
          "server then just use Monk (org.quintet.monk.Monk) directly.";
      log.log(Level.SEVERE,err,ne);
      throw new IllegalStateException(err,ne);
    }

    String blueprintsFileName = null;
    try{
      blueprintsFileName = (String)ic.lookup(ENTITY_BLUEPRINTS_KEY);
    }catch(NamingException ne){
      String err = "Unable to find an env-ref named '" + ENTITY_BLUEPRINTS_KEY+
          " in my JNDI context.  Please check the deployment descriptor.";
      log.log(Level.SEVERE,err,ne);
      throw new IllegalStateException(err,ne);
    }finally{
      if(ic != null){
        try{
          ic.close();
        }catch(Throwable eatIt){}
      }
    }

    //create our Monk to perform transcription duties
    try{
      monk = Monk.createMonk(blueprintsFileName);
    }catch(MonkException me){
      String err = "Unable to create a Monk via Monk.createMonk(\""+
          blueprintsFileName + "\")";
      log.log(Level.SEVERE,err,me);
      throw new IllegalStateException(err,me);
    }
  }

  /**
   * Transform <code>inRawData</code> to a new manifestation and return the new
   * manifestation to the client.
   *
   * @param inRawData - The raw data to be transformed to a new manifestation
   * @return java.lang.Object
   * @throws TransformerException
   * @todo Implement this org.quintet.backbeat.transformer.Transformer method
   */
  public Object transform(Object inRawData) throws TransformerException {
    if (inRawData == null) {
      String warn = ":null input, therefore returning null";
      log.info(warn);
      return null;
    }
    //I accept either a String containing a file name or an inputstream
    String xmlFileString = null;
    InputStream xmlStream = null;
    //If the caller passed in an InputStream then I am NOT responsible for
    //closing it.  If the caller passed in a file name (as a String) then
    //I AM responsible for both opening and closing the InputStream.
    boolean iShouldCloseIt = false;
    if (inRawData instanceof java.lang.String) { //XML file name in String
      xmlFileString = (String) inRawData;
      iShouldCloseIt = true;
      try{
        xmlStream = new FileInputStream(xmlFileString);
      }catch(IOException ioe){
        String err = "Unable to create a FileInputStream from the file name \""+
            inRawData + "\".  Returning Null to the caller.";
        log.log(Level.WARNING,err,ioe);
        throw new TransformerException(err,ioe);
      }
    }else if (inRawData instanceof java.io.InputStream){
      xmlStream = (InputStream)inRawData;
      iShouldCloseIt = false;
    }
    else {
      String err = ":expected an InputStream or a String containing a "+
          "file name, received a " + inRawData.getClass().getName();
      log.log(Level.SEVERE,err);
      throw new TransformerException(err);
    }

    Object answer =null;
    try{
      answer = monk.create(xmlStream);
      return answer;
    }catch(MonkException me){
      String err = "MonkException when transcribing XML to POJO";
      log.log(Level.WARNING,err,me);
      throw new TransformerException(err,me);
    }finally{
      if(iShouldCloseIt && xmlStream != null){
        try{
          xmlStream.close();
        }catch(IOException eatIt){}
      }
    }
  }
}
