package org.quintet.backbeat.transformer;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;
import java.io.StringWriter;
import java.io.FileReader;
import java.io.BufferedReader;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.io.*;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.io.IOException;

public class XMLToXMLTransformer
    implements org.quintet.backbeat.transformer.Transformer {
  private static final Logger log = Logger.getLogger("org.quintet.backbeat");

  private String styleSheetFileName;
  private String xmlDoc;
  private Transformer trans;
  private TransformerFactory factory;


  private static final String STYLE_SHEET_ENV_ENTRY =
      "java:comp/env/style-sheet-file-name";

  /**Initialize this <code>XMLToXMLTransformer</code> with configuration information
   * obtained from JNDI
   */
  public XMLToXMLTransformer() throws TransformerException {
    try {
      InitialContext ic = new InitialContext();
      setStyleSheetFileName( (String) ic.lookup(STYLE_SHEET_ENV_ENTRY));
      log.finest("style sheet file is " + getStyleSheetFileName());
      factory = TransformerFactory.newInstance();
    }
    catch (NamingException ne) {
      String err = "NamingException when constructing a TransformerFactory "+
          "with JNDI entry \"" + STYLE_SHEET_ENV_ENTRY + "\" -- " +
          "check the deployment descriptor.";
      log.log(Level.SEVERE,err,ne);
      throw new TransformerException(err,ne);
    }
  }

  public void setStyleSheetFileName(String sheet) {
    styleSheetFileName = sheet;
  }

  public void setXMLDocument(String doc) {
    xmlDoc = doc;
  }

  public String getStyleSheetFileName() {
    return styleSheetFileName;
  }

  public String getXMLDocument() {
    return xmlDoc;
  }

  /**use the style sheet identified by <code>styleSheetFileName</code> to transformer
   * <code>xmlDoc</code>'s XML manifestation
   * @return String containing new XML manifestation of <code>xmlDoc</code>'s contents
   * @exception javax.xml.transform.TransformerException
   */

  public Object transform(Object inRawData) throws TransformerException {
    if(inRawData == null)
      return null;
    StringWriter stringWriter = new StringWriter();
    log.log(Level.FINEST,"pk is an instance of " + inRawData.getClass().getName());
    StreamSource originalFmt = null;
    FileReader styleFR = null;
    BufferedReader styleBR = null;
    StreamSource styleSource = null;
    StreamResult transformedFmt = null;
    try {
      if (inRawData instanceof java.lang.String){  //inRawData contains file's name
        originalFmt = new StreamSource((String)inRawData);
      }else if (inRawData instanceof java.io.File){
        originalFmt = new StreamSource((File)inRawData);
      }else if (inRawData instanceof java.io.Reader){
        originalFmt = new StreamSource((Reader)inRawData);
      }else if (inRawData instanceof java.io.InputStream){
        originalFmt = new StreamSource((InputStream)inRawData);
      }else{
        String err = "Expected a File, Reader, or InputStream, got a " +
            originalFmt.getClass().getName();
        log.log(Level.WARNING,err);
        throw new TransformerException(err);
      }
      styleFR = new FileReader(getStyleSheetFileName());
      styleBR = new BufferedReader(styleFR);
      styleSource = new StreamSource(styleBR);
      trans = factory.newTransformer(styleSource);
      transformedFmt = new StreamResult(stringWriter);
      trans.transform(originalFmt, transformedFmt);
      StringBuffer answer = stringWriter.getBuffer();
      if (answer != null) {
        return answer.toString();
      }
      else {
        return null;
      }
    }
    catch (FileNotFoundException ex) {
      String err = "FileNotFoundException when using style sheet file \"" +
          getStyleSheetFileName() + "\"";
      log.log(Level.SEVERE,err,ex);
      throw new org.quintet.backbeat.transformer.TransformerException(err,ex);
    }
    catch (javax.xml.transform.TransformerException te) {
      String err = "javax.xml.transform.TransformerException when "+
          "transforming object via stylesheet \"" +
          getStyleSheetFileName() + "\"";
      log.log(Level.SEVERE,err,te);
      throw new org.quintet.backbeat.transformer.TransformerException(err, te);
    }finally{
      if(stringWriter != null){
        try{
          stringWriter.close();
          stringWriter = null;
        }catch(IOException ioe){}
      }
      if(transformedFmt != null){
        try{
          OutputStream os = transformedFmt.getOutputStream();
          if(os != null){
            os.close();
            os = null;
          }
          transformedFmt = null;
        }catch(IOException ioe){}
      }
      if(trans != null){
        trans = null;
      }
      if(styleSource != null){
        try{
          InputStream is = styleSource.getInputStream();
          if(is != null)
            is.close();
        }catch(IOException ioe){}
      }
      if(styleBR != null){
        try{
          styleBR.close();
        }catch(IOException ioe){}
      }
      if(styleFR != null){
        try{
          styleFR.close();
        }catch(IOException ioe){}
      }
      if(originalFmt != null){
        try{
          InputStream is = originalFmt.getInputStream();
          if(is != null)
            is.close();
        }catch(IOException ioe){}
      }
    }

  }
}
