package org.quintet.beans;

import java.beans.IntrospectionException;
import java.beans.BeanInfo;
import java.beans.PropertyDescriptor;

import java.lang.reflect.Method;

import java.util.Map;
import java.util.HashMap;

import java.util.logging.Logger;
import java.util.logging.Level;

/**<code>org.quintet.beans.Introspector</code> fixes a bug in the JDK 1.4_2
 * <code>java.beans.Introspector</code>; the bug causes accessible getters to be flagged
 * as inaccessible, thereby causing an <code>IllegalAccessException</code> to be thrown
 * when the method is invoked.  For example, even though it is perfectly legal to access
 * <code>java.util.Map.Entry.getKey()</code> via a reflective <code>Method</code>, the
 * following code snippet throws an <code>IllegalAccessException</code>:<br>
 * <code>
 * Map map = new HashMap();<br>
 * map.put("key1","value1");<br>
 * Iterator entryIter = map.entrySet().iterator();<br>
 * Map.Entry firstEntry = (Map.Entry)entryIter.next();<br>
 * Method getter = null;<br>
 * BeanInfo beanInfo = Introspector.getBeanInfo(firstEntry.getClass());<br>
 * PropertyDescriptor prop1 = beanInfo.getPropertyDescriptors()[0];<br>
 * getter = prop1.getReadMethod();<br>
 * //here is where the IllegalAccessException gets thrown<br>
 * getter.invoke(firstEntry,null);<br>
 * </code><br>
 * <code>org.quintet.beans.Introspector</code> fixes this problem by silently setting
 * all read methods' <code>accessible</code> flag to <code>true</code>.  In all other respects
 * this class is nothing but an adapter for the standard <code>java.beans.Introspector</code>.<br>
 * @author Peter J Thaggard, ADC
 * @see java.beans.Introspector
 */

public class Introspector {

  private static Map cache = new HashMap();

  private static final Logger log = Logger.getLogger("org.quintet.beans");

  private Introspector() {}

  /**
   * Adapter for the <code>java.beans.Introspector.getBeanInfo(Class)</code> method.
   * @param c Class
   * @throws IntrospectionException
   * @return BeanInfo with all read methods' <code>accessible</code> flag = <code>true</code>
   */
  public static BeanInfo getBeanInfo(Class c) throws IntrospectionException {
    BeanInfo answer = (BeanInfo) cache.get(c);
    if (answer != null) {
      return answer;
    }

    answer = java.beans.Introspector.getBeanInfo(c);

    PropertyDescriptor[] properties = answer.getPropertyDescriptors();
    for (int idx = 0; idx < properties.length; idx++) {
      Method getter = properties[idx].getReadMethod();
      if(getter != null)
        getter.setAccessible(true);
      Class propertyClass = properties[idx].getPropertyType();
/*      Package propertyPkg = propertyClass.getPackage();
      if(propertyPkg == null){
        String warn = "class " + propertyClass.getName() +" has no package info";
        log.warn(warn);
        packageName = "";
        continue;
      }else{
        String packageName = propertyPkg.getName();
      }*/
 /*     if("java.lang".equals(packageName)){
        //this is a fundamental type, so change its getter to be "toString"
        log.info("changing read method to 'toString' for class " + propertyClass.getName());
        Method toString = null;
        try {
          toString = properties[idx].getPropertyType().getMethod("toString", null);
          properties[idx].setReadMethod(toString);
          properties[idx].setWriteMethod(null);
        }
        catch (SecurityException ex) {
          String err = "SecurityException when trying to find toString in " + properties[idx].getPropertyType().getName();
          log.error(err);
        }
        catch (NoSuchMethodException ex) {
          String err = "NoSuchMethodException when trying to find toString in " + properties[idx].getPropertyType().getName();
          log.error(err);
        }

      }*/
    }
    cache.put(c, answer);
    return answer;
  }
}
