package org.quintet.chambers;

import java.lang.reflect.Method;
import java.util.Collection;

import javax.ejb.EJBLocalHome;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.persistence.EntityManager;

import java.util.logging.Logger;
import java.util.logging.Level;

/**
 *<code>DynaQueryRetriever</code> invokes EJB entity bean finder methods indirectly via reflection.
 *Its behavior is determined by JNDI environment entries:<br>
 *<li><code>entity-retriever-home-interface</code> is the entity bean home interface that
 * will be invoked by this <code>DynaQueryRetriever</code>.  This environment
 * entry is an <code>ejb-ref</code>, not an <code>env-entry</code>
 *<li><code>entity-retriever-finder-method</code> is the finder method to be invoked
 *<li><code>entity-retriever-pk-type</code> is the type of the primary key that will be passed to the finder
 *<br>
 *<code>DynaQueryRetriever</code> is capable of processing both single primary keys and
 *<code>Collection</code>s; if a single pk is sent to <code>retrieve</code> then a single
 *entity instance will be returned, but if a <code>Collection</code> of PKs is sent to
 *<code>retrieve</code> then a <code>Collection</code> of PKs will be returned.
 *@author Peter Thaggard
 *@version 2.0
 *@since 2002
 **/

public class DynaQueryRetriever
    implements EntityRetriever {

  /**
   * The place where I send diagnostics messages
   */
  private static final Logger log = Logger.getLogger("org.quintet.chambers");

  private static String HOME_INTERFACE_ENV_ENTRY =
      "java:comp/env/ejb/home-interface-name";

  /**<code>DynaQueryRetriever</code> stores an instance of the home interface with which it collaborates**/
  private EJBLocalHome homeInterface = null;

  /**<code>DynaQueryRetriever</code> stores the create <code>Method</code> that it calls on <code>homeInterface</code>**/
  Method finder = null;

  EntityManager em = null;

  /**
   * Construct a new <code>DynaQueryRetriever</code>.  All up-front JNDI lookups and
   * <code>Method</code> fetching is done in this constructor.
   * @throws RetrieverException
   */
  public DynaQueryRetriever() throws RetrieverException {
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }catch(NamingException ne){
      String err = "Unable to obtain a JNDI InitialContext";
      log.log(Level.SEVERE,err,ne);
      throw new RetrieverException(err,ne);
    }

    Class homeInterfaceClass = null;
    try{
      homeInterface = (EJBLocalHome) ic.lookup(DynaQueryRetriever.
                                               HOME_INTERFACE_ENV_ENTRY);
      homeInterfaceClass = homeInterface.getClass();
    }catch(NamingException ne){
      String err = "NamingException when looking up " + HOME_INTERFACE_ENV_ENTRY;
      log.log(Level.SEVERE,err,ne);
      throw new RetrieverException(err,ne);
    }
  }

  /**
   * Retrieve the entity bean instance(s) associated with <code>pk</code>.  If <code>pk</code>
   * is a <code>Collection</code> then return a <code>Collection</code> of entities, otherwise
   * return a single entity.
   * @param pk Object -- either a single PK or a <code>Collection</code> of PKs
   * @throws RetrieverException
   * @return Object -- either a single entity or a <code>Collection</code> of entities
   */
  public Object retrieve(Object query) throws RetrieverException {
    if(query == null){
      String err = "Unable to find entities via a null query";
      log.log(Level.SEVERE,err);
      throw new RetrieverException(err);
    }

	if(!(query instanceof String)){
		String err = "Expected a String, received a '" + query.getClass().getName() + "'";
		log.log(Level.SEVERE,err);
		throw new RetrieverException(err);
	}

	String queryStr = (String)query;

	return dynaQuery(queryStr);
  }

  /**
   * Retrieve the entity beans who meet the criteria specified in
   * <code>ejbQLQuery</code>, return them as a <code>Collection</code>.
   * @param ejbQLQuery String -- an EJB-QL query against the abstract schema associated
   * with this <code>DynaQueryRetriever</code>'s entity domain
   * @return Collection of entity bean instances
   * @throws RetrieverException
   */
  public Collection dynaQuery(String ejbQLQuery) throws RetrieverException{
    javax.persistence.Query query = em.createQuery(ejbQLQuery);
    try {
      return query.getResultList();
    }
    catch (Exception ex) {
      String err = "FinderException when executing query [" + ejbQLQuery + "] against home interface '" + this.HOME_INTERFACE_ENV_ENTRY + "'";
      log.log(Level.SEVERE,err,ex);
      throw new RetrieverException(err,ex);
    }
  }
}
