package org.quintet.chambers;

import org.quintet.monk.Monk;
import org.quintet.monk.Blueprints;
import org.quintet.monk.MonkException;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Collection;
import java.util.Iterator;
import java.lang.reflect.Modifier;
import java.lang.reflect.Method;
import java.util.logging.Logger;
import java.util.logging.Level;

/**
 * <code>EBlueprints</code> are <code>Blueprints</code> that contain references to EJBs.
 * @author Peter Thaggard
 * @version 1.0
 * @see org.quintet.monk.Blueprints
 * @see org.quintet.chambers.EntityFactory
 */
public class EBlueprints
    extends Blueprints {

  private static final Logger log = Logger.getLogger("org.quintet.chambers");

  public EBlueprints() {
  }

  private static Monk eBootstrapFactory = null;

  static {
    try {
      log.log(Level.FINEST,"Eblueprints static initializer");
      eBootstrapFactory = Monk.createMonk(EBlueprints.class);
    }
    catch (MonkException ofe) {
      String ow = "Unable to create bootstrap Monk";
      log.log(Level.SEVERE,ow, ofe);
      throw new java.lang.Error(ow, ofe);
    }
  }

  private static void closeResources(InitialContext ic){
    if(ic != null){
      try{
        ic.close();
      }catch(Throwable eatIt){}
    }
  }
  /**
     @param blueprintsFile
     @return org.quintet.monk.Blueprints
     @throws org.quintet.monk.MonkException
     @roseuid 426923A400A2
   */
  public static final EBlueprints buildEBlueprints(InputStream blueprintsFile) throws
      MonkException {
    InitialContext ctx = null;
    try{
      ctx = new InitialContext();
    }catch(NamingException ne){
      String fatal = "Unable to obtain a JNDI InitialContext";
      log.log(Level.SEVERE,fatal,ne);
      throw new MonkException(fatal,ne);
    }
    EBlueprints blueprints = new EBlueprints();
    //turn the blueprintsFile's contents into some
    //java objects (instances of ClassMetaData, Builder, and Attribute)
    Collection blueprintsCollection = (Collection) eBootstrapFactory.create(
        blueprintsFile);
    if (blueprintsCollection == null || blueprintsCollection.size() == 0) {
      String err = ":no blueprints available";
      log.log(Level.SEVERE,err);
      closeResources(ctx);
      throw new MonkException(err);
    }
    for (Iterator blueprintIter = blueprintsCollection.iterator();
         blueprintIter.hasNext(); ) {
      EClassMetaData blueprint = (EClassMetaData) blueprintIter.next();
      String xmlName = blueprint.getXmlName();
      EBuilder b = blueprint.getEBuilder();
      Class[] argClasses = b.getFormalArgumentClasses();
      Object targetObj = null;
      if (b.isCtor()) {
        if (!b.getIsDeferred()) {
          try {
            Class targetClass = Class.forName(b.getTargetName());
            b.setTarget(targetClass);
          }
          catch (ClassNotFoundException CNFE) {
            String err =
                ":Unable to find class mentioned in blueprints:" +
                b.getTargetName() + ":" + CNFE;
            log.log(Level.SEVERE,err);
            closeResources(ctx);
            throw new MonkException(err, CNFE);
          }
        }
      }
      else {
        if (b.getJndiName() == null) {
          //this is a non-EJB factory class
          Class factoryClass = null;
          if (b.getIsDeferred() == false) {
            try {
              factoryClass = Class.forName(b.getTargetName());
              //if the class is abstract then we obviously cannot instantiate it.
              //However, the blueprints may have specified a static method of an
              //abstract class, so we have to deal with that possibility
              int factoryClassModifiers = factoryClass.getModifiers();
              if (Modifier.isAbstract(factoryClassModifiers)) {
                log.info(b.getTargetName() +
                         " is abstract, therefore its method must be static");
                b.setAbstractFactoryClass(true);
                b.setStaticMethod(true);
                targetObj = factoryClass;
              }
              else {
                targetObj = factoryClass.newInstance();
              }
            }
            catch (IllegalAccessException ex1) {
              String err = ":ERROR:IllegalAccessException when "
                  + "trying to instantiate a " + b.getTargetName() + ":" +
                  ex1;
              log.log(Level.SEVERE,err);
              closeResources(ctx);
              throw new MonkException(err, ex1);
            }
            catch (InstantiationException ex1) {
              String err = ":ERROR:InstantiationException when "
                  + "trying to instantiate a " + b.getTargetName() + ":" +
                  ex1;
              log.log(Level.SEVERE,err);
              closeResources(ctx);
              throw new MonkException(err, ex1);
            }
            catch (ClassNotFoundException ex) {
              String err =
                  ":ERROR:unable to find factory class " + b.getTargetName() +
                  ":" + ex;
              log.log(Level.SEVERE,err);
              closeResources(ctx);
              throw new MonkException(err, ex);
            }
            b.setTarget(targetObj);
            try {
              if (b.getAbstractFactoryClass() == true) {
                Method ctor = ( (Class) targetObj).getMethod(b.
                    getFactoryMethodName(), argClasses);
                b.setFactoryMethod(ctor);
                b.setStaticMethod(true); //it has to be static
              }
              else {
                Method ctor = targetObj.getClass().getMethod(b.
                    getFactoryMethodName(), argClasses);
                b.setFactoryMethod(ctor);
                if (Modifier.isStatic(ctor.getModifiers())) {
                  b.setStaticMethod(true);
                }
              }
            }
            catch (SecurityException ex2) {
              String err = ":unable to find method named " +
                  b.getFactoryMethodName() + " in class " +
                  targetObj.getClass() +
                  ":" + ex2;
              log.log(Level.SEVERE,err,ex2);
              closeResources(ctx);
              throw new MonkException(err);
            }
            catch (NoSuchMethodException ex2) {
              String err = ":unable to find method named " +
                  b.getFactoryMethodName() + " in class " +
                  targetObj.getClass() +
                  ":" + ex2;
              log.log(Level.SEVERE,err,ex2);
              closeResources(ctx);
              throw new MonkException(err);
            }
          }
        }
        else {
          try {
            targetObj = ctx.lookup(b.getJndiName());
            if (targetObj == null) {
              String err = "JNDI returned a NULL object for \"" +
                  b.getJndiName() + "\"";
              log.log(Level.SEVERE, err);
              closeResources(ctx);
              throw new MonkException(err);
            }
            else {
              b.setTarget(targetObj);
            }
            Method ctor = targetObj.getClass().getMethod(b.
                getFactoryMethodName(), argClasses);
            b.setFactoryMethod(ctor);
          }
          catch (NamingException ne) {
            String err = "NamingException when looking up \"" +
                b.getJndiName() + "\"";
            log.log(Level.SEVERE,err,ne);
            MonkException ofe = new MonkException(err,ne);
            ofe.setClassMetaData(blueprint);
            closeResources(ctx);
            throw ofe;
          }
          catch (NoSuchMethodException NSME) {
            String err = "NoSuchMethodException when trying to find method \""+
                targetObj.getClass().getName()+"."+b.getFactoryMethodName()+"\"";
            log.log(Level.SEVERE,err,NSME);
            MonkException ofe = new MonkException(err,NSME);
            ofe.setClassMetaData(blueprint);
            ofe.setErrantAttributeName(b.getFactoryMethodName());
            closeResources(ctx);
            throw ofe;
          }
        }
      }

      log.log(Level.FINEST,"Eblueprint entry:  " + xmlName + "==>" + blueprint.getSubjectName());
      blueprints.put(xmlName, blueprint);
    }
    closeResources(ctx);
    return blueprints;
  }

  /**
     @param fileName
     @return org.quintet.monk.Blueprints
     @throws org.quintet.monk.MonkException
     @roseuid 4269249A016E
   */
  public static final Blueprints buildEBlueprints(String fileName) throws
      MonkException {
    FileInputStream fis = null;
    try {
      fis = new FileInputStream(fileName);
      EBlueprints answer = EBlueprints.buildEBlueprints(fis);
      return answer;
    }
    catch (FileNotFoundException ex) {
      String err = "unable to open blueprints file named '" + fileName + "'";
      log.log(Level.SEVERE,err, ex);
      throw new MonkException(err, ex);
    }
    finally {
      if (fis != null) {
        try {
          fis.close();
        }
        catch (Throwable t) {}
      }
    }
  }

  /**
     @param forClass
     @param fileName
     @return org.quintet.monk.Blueprints
     @throws org.quintet.monk.MonkException
     @roseuid 426924B601FB
   */
  public static final Blueprints buildEBlueprints(Class forClass,
                                                  String fileName) throws
      MonkException {
    InputStream str = null;
    str = forClass.getResourceAsStream(fileName);
    return EBlueprints.buildEBlueprints(str);
  }

}
