package org.quintet.chambers;

import java.util.Collection;
import java.util.StringTokenizer;
import java.util.logging.Logger;
import java.util.logging.Level;

import org.quintet.miles.Activator;
import org.quintet.miles.AggregationException;
import org.quintet.miles.DN;
import org.quintet.miles.DNNode;
import org.quintet.miles.AggregationAttribute;
import javax.ejb.FinderException;

import org.quintet.backbeat.supplier.LocalSessionSupplier;
import org.quintet.backbeat.supplier.LocalSessionSupplierHome;
import org.quintet.backbeat.supplier.SupplierException;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import javax.ejb.CreateException;
import javax.ejb.FinderException;

import org.quintet.QuintetException;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.*;



public class EJBQLActivator
    extends Activator {

  private Calendar cal = new GregorianCalendar();




  enum EJBQLOperator{
    IN("|"),
    LIKE(".*"),
    EQUALS("=");

    private String symbol;
    EJBQLOperator(String sym){
      symbol = sym;
    }

    public static EJBQLOperator fromSymbol(String sym){
      for(EJBQLOperator op:values()){
        if(sym.contains(op.symbol))
          return op;
      }
      return EQUALS;
    }
  }


  private static final Logger log = Logger.getLogger("org.quintet.chambers");

  /**
   * <code>EJBQLActivator</code> uses a Quintet <code>LocalSessionSupplier</code>
   * to actual perform the dynamic EJB-QL Query upon the correct
   * <code>EJBLocalHome</code>.  The <code>supplierName</code> contains the
   * full name by which the concomitant <code>Supplier</code> can be retrieved
   * from JNDI.
   */
  private String supplierName = null;

  /**
   * The <code>supplier</code> attribute houses the <code>LocalSessionSupplier</code>
   * instance which this <code>EJBQLActivator</code> uses to perform EJB-QL
   * queries.
   */
  private LocalSessionSupplier supplier = null;

  /**
   * The <code>baseQuery</code> attribute houses the raw uninstantiated EJB-QL
   * query that will form the basis of all queries performed by this
   * <code>EJBQLActivator</code>.  It is "raw" and "uninstantiated" because it
   * contains formal arguments of the form <i>?&lt;dn-node-number&gt;</i>, and
   * the query is meaningless until these formal arguments have been replaced
   * with actual arguments gleaned from user-supplier DN queries.
   */
  private String baseQuery = null;


  public EJBQLActivator(String inSubjectName) throws IllegalStateException{
    super(inSubjectName);
  }

  public EJBQLActivator(){
    super();
  }


  /**
   * "Activate" the set of <code>Object</code>s designated by <code>setSpec</code>.
   * Activation is the act of retrieving the designated objects from some persistent store.
   * @param setSpec -- the DN describing the set of <code>Object</code>s to be activated
   * @return Collection of activated <code>Object</code>s
   * @throws AggregationException
   */
  public Collection activate(DN setSpec) throws QuintetException {
    //first instantiate my EJBQL with the setSpec's RHSs
      String instantiatedQuery = instantiateQuery(setSpec);
      Object rawAnswer = getSupplier().acquireInformation(instantiatedQuery);
      return (Collection)rawAnswer;
  }

  private String instantiateQuery(DN dnQuery) throws EntityException{
    //sample raw query:  SELECT o from ObservationSchema(o), IN (principal.commits) sub0 where ?0 and ?1
    //sample dn to instantiate this:  uid=todd|peter,defect-type=BASE_CM_ACTIVITY|ENHANCEMENT|BUG
    //results:  SELECT o from ObservationSchema(o), IN(principal.commits) sub0 where o.prinicipal.loginID in ('todd','peter') and sub0.commit.defectType in ('BASE|ENH|BUG')
    int idx=0;
    StringBuffer answerBuf = new StringBuffer(getBaseQuery());
    String answer = answerBuf.toString();
    for(DNNode<String,String> node:dnQuery.getNodes()){
      String RHS = node.RHS;
      String LHS = node.LHS;
      AggregationAttribute aa = getDnAnnotations().get(LHS);
      //if LHS's annotation specifies that the attribute is a
      //"date-expressed-as-milliseconds" then undertake special processing...
      if( null != aa) {
          if( aa.isTimeInMillis()) {
              RHS = adjustTimeBasedRHS(node,aa);
          } else {
//            examine the LHS to determine the operator to use (e.g., LIKE, =, IN)
              EJBQLOperator operator = EJBQLOperator.fromSymbol(RHS);

              //adjust the RHS to conform to EJB-QL syntax
              //e.g. day=01/01/2006|01/02/2006 becomes IN('01/01/2006','01/02/2006')
              RHS = adjustRHSSyntax(operator, aa, RHS);
          }
      }

      String formalArg = "?" + idx++ ;
      int formalArgPos = answer.indexOf(formalArg);
      if(formalArgPos >= 0){
        answer = answer.substring(0,formalArgPos-1) + " " + RHS + answer.substring(formalArgPos + formalArg.length());
      }
    }
    log.log(Level.WARNING,"generated the following EJB-QL:  " + answer);
    //eliminate any unnecessary query terms (i.e., any "and 1=1" parts)
    String finalAnswer = eliminateUnnecessaryTerms(answer);
    return finalAnswer;
  }

  private String eliminateUnnecessaryTerms(String originalQuery){
    int victimPos = -1;
    do{
      victimPos = originalQuery.indexOf("and 1=1");
      if(victimPos >= 0){
        originalQuery = originalQuery.substring(0,victimPos-1) + " " + originalQuery.substring(victimPos + 7);
      }
    }while(victimPos >= 0);
    return originalQuery;
  }

  private String adjustTimeBasedRHS(DNNode<String,String> node, AggregationAttribute aa){
    //we need to convert the LHS (day, week, month) into a Pair<first-milli-of-first-day,last-millis-of-last-day>
    String LHS = node.LHS;
    if(null == LHS){
      String err = "encountered a NULL LHS, indicating a malformed DN Query";
      log.log(Level.SEVERE,err);
      throw new IllegalStateException(err);
    }

    String whereClauseName = aa.whereClauseName();
    if(null == whereClauseName || 0 == whereClauseName.length()){
      String err = "no 'whereClauseName' specified in meta-data annotatation " +
                   "for DN LHS of '" + node.LHS + "'";
      log.log(Level.SEVERE,err);
      throw new IllegalStateException(err);
    }

    if(LHS.startsWith("day")){
      return adjustDayRHS(node,aa);
    }else if(LHS.startsWith("week")){
      return adjustWeekRHS(node,aa);
    }else if(LHS.startsWith("month")){
      return adjustMonthRHS(node,aa);
    }else{
      String err = "expected one of [day,week,month], found '" + LHS + "'";
      log.log(Level.SEVERE,err);
      throw new IllegalStateException(err);
    }
  }

  private String adjustDayRHS(DNNode<String,String> node, AggregationAttribute aa){
    DateFormat fmt = new SimpleDateFormat("MM/dd/yyyy");
    String whereClauseName = aa.whereClauseName();
    StringBuffer answer = new StringBuffer( "(" + whereClauseName + " BETWEEN ");
    String RHS = node.RHS;
    StringTokenizer lex = new StringTokenizer(RHS,"|");
    while(lex.hasMoreTokens()){
      String token = lex.nextToken();
      try {
        Date theDay = fmt.parse(token);
        cal.setTime(theDay);
        cal.set(Calendar.HOUR,0);
        cal.set(Calendar.MINUTE,0);
        cal.set(Calendar.SECOND,0);
        cal.set(Calendar.MILLISECOND,0);
        long firstMilli = cal.getTime().getTime();
        cal.set(Calendar.HOUR_OF_DAY,23);
        cal.set(Calendar.MINUTE,59);
        cal.set(Calendar.SECOND,59);
        cal.set(Calendar.MILLISECOND,999);
        long lastMilli = cal.getTime().getTime();
        answer.append( firstMilli + " AND " + lastMilli );
        if(lex.hasMoreTokens()){
          answer.append(" OR " + whereClauseName + " BETWEEN ");
        }else{
          answer.append(")");
        }
      }
      catch (ParseException ex) {
        String err = "Error parsing the \"date\" '" + RHS + "'";
        log.log(Level.SEVERE,err,ex);
        throw new IllegalStateException(err,ex);
      }
    }
    return answer.toString();
  }

  private String adjustWeekRHS(DNNode<String,String> node, AggregationAttribute aa){
    DateFormat fmt = new SimpleDateFormat("MM/dd/yyyy");
    String whereClauseName = aa.whereClauseName();
    StringBuffer answer = new StringBuffer( "(" + whereClauseName + " BETWEEN ");
    String RHS = node.RHS;
    StringTokenizer lex = new StringTokenizer(RHS,"|");
    while(lex.hasMoreTokens()){
      String token = lex.nextToken();
      try {
        Date theDay = fmt.parse(token);
        cal.setTime(theDay);
        cal.set(Calendar.DAY_OF_WEEK,Calendar.SUNDAY);
        cal.set(Calendar.HOUR,0);
        cal.set(Calendar.MINUTE,0);
        cal.set(Calendar.SECOND,0);
        cal.set(Calendar.MILLISECOND,0);
        long firstMilli = cal.getTime().getTime();
        cal.set(Calendar.DAY_OF_WEEK,Calendar.SATURDAY);
        cal.set(Calendar.HOUR_OF_DAY,23);
        cal.set(Calendar.MINUTE,59);
        cal.set(Calendar.SECOND,59);
        cal.set(Calendar.MILLISECOND,999);
        long lastMilli = cal.getTime().getTime();
        answer.append( firstMilli + " AND " + lastMilli );
        if(lex.hasMoreTokens()){
          answer.append(" OR " + whereClauseName + " BETWEEN ");
        }else{
          answer.append(")");
        }
      }
      catch (ParseException ex) {
        String err = "Error parsing the \"date\" '" + RHS + "'";
        log.log(Level.SEVERE,err,ex);
        throw new IllegalStateException(err,ex);
      }
    }
    return answer.toString();
  }

  private String adjustMonthRHS(DNNode<String,String> node, AggregationAttribute aa){
    DateFormat fmt = new SimpleDateFormat("MM/dd/yyyy");
    String whereClauseName = aa.whereClauseName();
    StringBuffer answer = new StringBuffer( "(" + whereClauseName + " BETWEEN ");
    String RHS = node.RHS;
    StringTokenizer lex = new StringTokenizer(RHS,"|");
    while(lex.hasMoreTokens()){
      String token = lex.nextToken();
      try {
        Date theDay = fmt.parse(token);
        cal.setTime(theDay);
        cal.set(Calendar.DAY_OF_MONTH,1);
        cal.set(Calendar.HOUR,0);
        cal.set(Calendar.MINUTE,0);
        cal.set(Calendar.SECOND,0);
        cal.set(Calendar.MILLISECOND,0);
        long firstMilli = cal.getTime().getTime();
        cal.add(Calendar.MONTH,1);
        cal.set(Calendar.HOUR_OF_DAY,0);
        cal.set(Calendar.MINUTE,0);
        cal.set(Calendar.SECOND,0);
        cal.set(Calendar.MILLISECOND,0);
        cal.add(Calendar.MILLISECOND,-1);
        long lastMilli = cal.getTime().getTime();
        answer.append( firstMilli + " AND " + lastMilli );
        if(lex.hasMoreTokens()){
          answer.append(" OR " + whereClauseName + " BETWEEN ");
        }else{
          answer.append(")");
        }
      }
      catch (ParseException ex) {
        String err = "Error parsing the \"date\" '" + RHS + "'";
        log.log(Level.SEVERE,err,ex);
        throw new IllegalStateException(err,ex);
      }
    }
    return answer.toString();
  }

  private static final String oneEqualsOne = "1=1";

  private String adjustRHSSyntax(EJBQLOperator operator, AggregationAttribute aa, String origRHS){
    StringBuffer answer = new StringBuffer();
    answer.append(aa.whereClauseName() + " ");
    if(operator.equals(EJBQLOperator.IN)){
      answer.append(emitINClause(operator,aa,origRHS));
    }else if(operator.equals(EJBQLOperator.LIKE)){
      //if aa.requiresQuotes == false then origRHS is a NUMBER, which means that it cannot
      //be part of a LIKE term.  Therefore we will eliminate such terms from the expression;
      //this is legal, because terms like "projectid=.*" have no impact whatsoever on
      //the results returned from the query.
      if(aa.requiresQuotes() == false){
        return oneEqualsOne;
      }else{
        answer.append(emitLIKEClause(operator, aa, origRHS));
      }
    }else{
      answer.append(emitEQUALSClause(operator,aa,origRHS));
    }
    return answer.toString();
  }

  private String emitINClause(EJBQLOperator operator, AggregationAttribute aa, String origRHS){
    StringBuffer answer = new StringBuffer("IN(");
    StringTokenizer lex = new StringTokenizer(origRHS,"|");
    while(lex.hasMoreTokens()){
      String currToken = lex.nextToken();
      String quote = "";
      if(aa.requiresQuotes())
        quote = "'";
      answer.append(quote + currToken + quote);
      if(lex.hasMoreTokens()){
        answer.append(",");
      }
    }
    answer.append(")");
    return answer.toString();
  }


  private static final String isNotNull = " IS NOT NULL ";

  private String emitLIKEClause(EJBQLOperator operator, AggregationAttribute aa, String origRHS){
    StringBuffer answer = new StringBuffer("LIKE ");
    int starPos = origRHS.indexOf(".*");
    if(starPos >= 0){
      answer.append("'" + origRHS.substring(0,starPos) + "%" + origRHS.substring(starPos+2) + "'");
    }
    return answer.toString();
  }

  private String emitEQUALSClause(EJBQLOperator operator, AggregationAttribute aa, String origRHS){
    return " = " + origRHS;
  }

  private String getOperator(DNNode<String,String> node){
    String lhs = node.LHS;
    if(lhs.contains(".*"))
      return " LIKE ";
    if(lhs.contains("|"))
      return " IN ";
    return " = ";
  }
  /**
   * Create a verbatim copy of this <code>EJBQLActivator</code>
   * @return Object
   * @throws CloneNotSupportedException
   */
  public Object clone() throws CloneNotSupportedException {
    // TODO Auto-generated method stub
    return super.clone();
  }

  /*#java.util.StringTokenizer Dependency_Link*/

  public void closeResources(InitialContext ic){
    if(ic != null){
      try{
        ic.close();
      }catch(Throwable eatIt){}
    }
  }
  /**
   * Designate the name of the <code>Supplier</code> that this <code>EJBQLActivator</code>
   * will utilize to perform EJB-QL queries.
   * <code>EJBQLActivator</code> uses a Quintet <code>LocalSessionSupplier</code>
   * to actual perform the dynamic EJB-QL Query upon the correct
   * <code>EJBLocalHome</code>.  The <code>supplierName</code> contains the
   * full name by which the concomitant <code>Supplier</code> can be retrieved
   * from JNDI.
   */
  public void setSupplierName(String supplierName) {
    this.supplierName = supplierName;
    InitialContext ic = null;
    try{
      ic = new InitialContext();
    }catch(NamingException ne){
      String fatal = "Unable to obtain an JNDI Initial Context -- " + "I need to be running within a J2EE container, buddy";
      log.log(Level.SEVERE,fatal,ne);
      throw new IllegalStateException(fatal,ne);
    }

    LocalSessionSupplierHome ssHome = null;
    try{
      ssHome = (LocalSessionSupplierHome)ic.lookup(this.supplierName);
    }catch(NamingException ne){
      String fatal = "Unable to find a LocalSessionSupplier named '" + this.supplierName + "' in JNDI.  Check the blueprints file and the deployment descriptor";
      log.log(Level.SEVERE,fatal,ne);
      closeResources(ic);
      throw new IllegalStateException(fatal,ne);
    }

    try{
      setSupplier(ssHome.create());
    }catch(CreateException ce){
      String fatal = "Unable to create a LocalSessionSupplier instance";
      log.log(Level.SEVERE,fatal,ce);
      closeResources(ic);
      throw new IllegalStateException(fatal,ce);
    }
    closeResources(ic);
  }

  /**
   * Designate <code>supplier</code> to be this <code>EJBQLActivator</code>'s
   * concomitant <code>LocalSessionSupplier</code>.
   * The <code>supplier</code> attribute houses the <code>LocalSessionSupplier</code>
   * instance which this <code>EJBQLActivator</code> uses to perform EJB-QL
   * queries.
   * @param supplier LocalSessionSupplier
   */
  public void setSupplier(LocalSessionSupplier supplier) {
    this.supplier = supplier;

  }

  /**
   * Designate <code>inQuery</code> to be this <code>EJBQLActivator</code>'s
   * base query.  The <code>baseQuery</code> attribute houses the raw uninstantiated EJB-QL
   * query that will form the basis of all queries performed by this
   * <code>EJBQLActivator</code>.  It is "raw" and "uninstantiated" because it
   * contains formal arguments of the form <i>?&lt;dn-node-number&gt;</i>, and
   * the query is meaningless until these formal arguments have been replaced
   * with actual arguments gleaned from user-supplier DN queries.
   * @param baseQuery String
   */
  public void setBaseQuery(String inQuery) {
    this.baseQuery = inQuery;
  }

  /**
   * Retrieve the name of the <code>Supplier</code> that this <code>EJBQLActivator</code>
   * will utilize to perform EJB-QL queries.
   * <code>EJBQLActivator</code> uses a Quintet <code>LocalSessionSupplier</code>
   * to actual perform the dynamic EJB-QL Query upon the correct
   * <code>EJBLocalHome</code>.  The <code>supplierName</code> contains the
   * full name by which the concomitant <code>Supplier</code> can be retrieved
   * from JNDI.
   */
  public String getSupplierName() {
    return supplierName;
  }

  /**
   * Retrieve this <code>EJBQLActivator</code>'s concomitant <code>LocalSessionSupplier</code>.
   * The <code>supplier</code> attribute houses the <code>LocalSessionSupplier</code>
   * instance which this <code>EJBQLActivator</code> uses to perform EJB-QL
   * queries.
   */
  public LocalSessionSupplier getSupplier() {
    return supplier;
  }

  /**
   * Retrieve this <code>EJBQLActivator</code>'s base query.
   * The <code>baseQuery</code> attribute houses the raw uninstantiated EJB-QL
   * query that will form the basis of all queries performed by this
   * <code>EJBQLActivator</code>.  It is "raw" and "uninstantiated" because it
   * contains formal arguments of the form <i>?&lt;dn-node-number&gt;</i>, and
   * the query is meaningless until these formal arguments have been replaced
   * with actual arguments gleaned from user-supplier DN queries.
   * @return String
   */
  public String getBaseQuery() {
    return baseQuery;
  }

  /*#org.quintet.miles.DNNode Dependency_Link1*/

  /*#org.quintet.miles.DN Dependency_Link2*/
  /*#weblogic.ejb.QueryLocalHome Dependency_Link3*/



}
