package org.quintet.chambers;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.ejb.EJBLocalHome;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.persistence.EntityManager;
import javax.persistence.Query;

import java.util.logging.Logger;
import java.util.logging.Level;

/**
 *<code>EntityBeanRetriever</code> invokes EJB entity bean finder methods indirectly via reflection.
 *Its behavior is determined by JNDI environment entries:<br>
 *<li><code>entity-retriever-home-interface</code> is the entity bean home interface that
 * will be invoked by this <code>EntityBeanRetriever</code>.  This environment
 * entry is an <code>ejb-ref</code>, not an <code>env-entry</code>
 *<li><code>entity-retriever-finder-method</code> is the finder method to be invoked
 *<li><code>entity-retriever-pk-type</code> is the type of the primary key that will be passed to the finder
 *<br>
 *<code>EntityBeanRetriever</code> is capable of processing both single primary keys and
 *<code>Collection</code>s; if a single pk is sent to <code>retrieve</code> then a single
 *entity instance will be returned, but if a <code>Collection</code> of PKs is sent to
 *<code>retrieve</code> then a <code>Collection</code> of PKs will be returned.
 *@author Peter Thaggard
 *@version 2.0
 *@since 2002
 **/

public class EntityBeanRetriever
    implements EntityRetriever {

  /**
   * The place where I send diagnostics messages
   */
  private static final Logger log = Logger.getLogger("org.quintet.chambers");

  private static String HOME_INTERFACE_ENV_ENTRY =
      "java:comp/env/ejb/home-interface-name";
  private static String FINDER_METHOD_ENV_ENTRY =
      "java:comp/env/entity-retriever-finder-interface-name";
  private static String PK_CLASS_NAME_ENV_ENTRY =
      "java:comp/env/entity-retriever-pk-class-name";



  /**<code>EntityBeanRetriever</code> stores an instance of the home interface with which it collaborates**/
  private EJBLocalHome homeInterface = null;

  /**<code>EntityBeanRetriever</code> stores the create <code>Method</code> that it calls on <code>homeInterface</code>**/
  Method finder = null;

  EntityManager em = null;

  public void closeResources(InitialContext ic){
    if(ic != null){
      try{
        ic.close();
      }catch(Throwable eatIt){}
    }
  }
  /**
   * Construct a new <code>EntityBeanRetriever</code>.  All up-front JNDI lookups and
   * <code>Method</code> fetching is done in this constructor.
   * @throws RetrieverException
   */
  public EntityBeanRetriever() throws RetrieverException {
    InitialContext ic = null;
    try {
      ic = new InitialContext();
    }catch(NamingException ne){
      String err = "Unable to obtain a JNDI InitialContext";
      log.log(Level.SEVERE,err,ne);
      throw new RetrieverException(err,ne);
    }


    Class homeInterfaceClass = null;
    try{
      homeInterface = (EJBLocalHome) ic.lookup(EntityBeanRetriever.
                                               HOME_INTERFACE_ENV_ENTRY);
      homeInterfaceClass = homeInterface.getClass();
    }catch(NamingException ne){
      String err = "NamingException when looking up " + HOME_INTERFACE_ENV_ENTRY;
      log.log(Level.SEVERE,err,ne);
      closeResources(ic);
      throw new RetrieverException(err,ne);
    }
      String finderMethodName = null;
      try{
        finderMethodName = (String) ic.lookup(EntityBeanRetriever.
          FINDER_METHOD_ENV_ENTRY);
      }catch(NamingException ne){
        String err = "Unable to lookup finder method via key '" + FINDER_METHOD_ENV_ENTRY + "'";
        log.log(Level.SEVERE,err,ne);
        closeResources(ic);
        throw new RetrieverException(err,ne);
      }

      String pkClassName = null;
      try{
      pkClassName = (String) ic.lookup(EntityBeanRetriever.
                                              PK_CLASS_NAME_ENV_ENTRY);
      Class pkClass = Class.forName(pkClassName);
      finder = homeInterfaceClass.getMethod("findByPrimaryKey",
                                            new Class[] {pkClass});
    }
    catch (NamingException ne) {
      String err = "Unable to find JNDI entry '" + PK_CLASS_NAME_ENV_ENTRY + "'";
      log.log(Level.SEVERE,err,ne);
      closeResources(ic);
      throw new RetrieverException(err,ne);
    }
    catch (ClassNotFoundException cnfe) {
      String err = "Unable to find class '" + pkClassName;
      log.log(Level.SEVERE,err,cnfe);
      closeResources(ic);
      throw new RetrieverException(err,cnfe);
    }catch(NoSuchMethodException NSME){
      String err = "unable to find method " + homeInterfaceClass.getName() + "." + finderMethodName;
      log.log(Level.SEVERE,err,NSME);
      closeResources(ic);
      throw new RetrieverException(err,NSME);
    }
    closeResources(ic);
  }

  /**
   * Retrieve the entity bean instance(s) associated with <code>pk</code>.  If <code>pk</code>
   * is a <code>Collection</code> then return a <code>Collection</code> of entities, otherwise
   * return a single entity.
   * @param pk Object -- either a single PK or a <code>Collection</code> of PKs
   * @throws RetrieverException
   * @return Object -- either a single entity or a <code>Collection</code> of entities
   */
  public Object retrieve(Object pk) throws RetrieverException {
    //the "pk" might actually be a collection of primary keys.
    //if so then loop through the collection and retrieve each concomitant entity
    if(pk == null){
      String err = "Unable to find entities via a null primary key";
      log.log(Level.SEVERE,err);
      throw new RetrieverException(err);
    }

    Collection pks = null;
    if(!(pk instanceof Collection)){
      log.fine("Received a single PK, wrapping it in a collection");
      pks = new ArrayList(1);
      pks.add(pk);
    }else{
      pks = (Collection)pk;
    }

    Collection answers = new ArrayList(pks.size());
    for(Iterator pkIter = pks.iterator(); pkIter.hasNext();){
      try {
        Object currAnswer = finder.invoke(homeInterface, new Object[] {pkIter.next()});
        answers.add(currAnswer);
      }
      catch (InvocationTargetException ite) {
        Throwable orig = ite.getCause();
        if(orig == null)
          orig = ite;
        String err = "InvocationTargetException (wrapping a " + orig.getClass().getName() + ")";
        log.log(Level.SEVERE,err,orig);
        throw new RetrieverException(err,orig);
      }
      catch (IllegalAccessException iae) {
        String err = "IllegalAccessException";
        log.log(Level.SEVERE,err,iae);
        throw new RetrieverException(err,iae);
      }
    }
    if(answers.size() == 1){
      return answers.iterator().next();
    }else{
      return answers;
    }
  }

  /**
   * Retrieve the entity beans who meet the criteria specified in
   * <code>ejbQLQuery</code>, return them as a <code>Collection</code>.
   * @param ejbQLQuery String -- an EJB-QL query against the abstract schema associated
   * with this <code>EntityBeanRetriever</code>'s entity domain
   * @return Collection of entity bean instances
   * @throws RetrieverException
   */
  public Collection dynaQuery(String ejbQLQuery) throws RetrieverException{
    Query query = em.createQuery(ejbQLQuery);
    try {
      return query.getResultList();
    }
    catch (Exception ex) {
      String err = "FinderException when executing query [" + ejbQLQuery + "] against home interface '" + this.HOME_INTERFACE_ENV_ENTRY + "'";
      log.log(Level.SEVERE,err,ex);
      throw new RetrieverException(err,ex);
    }
  }
}
