package org.quintet.chambers;

import java.io.IOException;
import java.io.InputStream;
import java.io.NotSerializableException;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.ejb.CreateException;
import javax.ejb.EJBException;
import javax.ejb.SessionBean;
import javax.ejb.SessionContext;
import javax.ejb.PrePassivate;
import javax.ejb.PostActivate;
import javax.jms.JMSException;
import javax.jms.ObjectMessage;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.persistence.Entity;
import javax.persistence.Id;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.quintet.monk.ClassMetaData;
import org.quintet.monk.Monk;
import org.quintet.monk.MonkConfigurationException;
import org.quintet.monk.MonkException;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 *
 * <p>Title: EntityFactoryBean</p>
 * <p>Description: <code>EntityFactoryBean</code> is the implementation portion of the
 * <code>EntityFactory</code> session bean.  <code>EntityFactory</code> is a session bean wrapper
 * that allows <code>Monk</code> to create local entity bean instances.  <code>EntityFactory</code>'s
 * actions are driven by its <code>blueprints.xml</code> file, which describes how to convert XML documents into
 * java objects, including POJOs and entity bean instances.  <code>blueprints.xml</code> must be housed in the
 * same place where the class loader finds <code>EntityFactoryBean</code>; for example, if <code>EntityFactoryBean.class</code>
 * is located in a JAR file, then <code>blueprints.xml</code> must be co-located with it within that JAR file's
 * directory structure.</p>
 * The blueprints.xml file is converted (by Monk, naturally) into a collection of  <code>ClassMetaData</code>
 * instances.  Each <code>ClassMetaData</code> instance describes how to convert exactly one XML element (e.g., "instrument-price")
 * into its corresponding java instance (e.g., "InstrumentPrice").  After loading its blueprints, <code>EntityFactory</code>'s job
 * becomes one of delegating work to <code>Monk</code>, which knows how to apply a set of blueprints to a XML document to
 * yield a graph of java instances.
 * <code>EntityFactory</code> has exactly one public method, convertToEntity.
 * @author Peter Thaggard
 * @version 2.1
 * @see org.quintet.monk.Monk
 * @see org.quintet.monk.ClassMetaData
 * @see blueprints.xml
 */
public class EntityFactoryBean implements SessionBean {
    static final long serialVersionUID = 200501001;

    @PostActivate() public void reset() throws EJBException{

    }
    /**
     * The <code>Monk</code> that I use to transcribe XML into entity bean instances
     */
    private Monk monk = null;


    /**The name of the env-entry containing my blueprints' URL
     *
     */
    private static final String BLUEPRINTS_URL_KEY =
            "java:comp/env/blueprints-url";


    /**our logger**/
    private static final Logger log = Logger.getLogger("org.quintet.chambers");

    /**<code>EntityFactoryBean</code> is capable of sending errant XML nodes to an exception
     * queue.  It gains access to these exception queues via the following JMS Queue
     * Connection Factory.  This Queue Connection Factory must be declared in <code>EntityFactory</code>'s
     * deployment descriptor as a resource-ref
     */

    private static final String QUEUE_CONNECTION_FACTORY_NAME =
            "java:comp/env/jms/exception-queue-connection-factory";

    /**
     <code>EntityFactory</code> uses <code>exceptionQueueConnectionFactory</code> to
     * create <code>QueueConnection</code>s.  The <code>QueueConnection</code>s are then used to send errant XML to exception queues.
     */
    private  QueueConnectionFactory exceptionQueueConnectionFactory;

    /**
     * <code>EntityFactoryBean</code> maintains a list of all error destinations mentioned in blueprints.xml;
     * there can potentially be one destination per xml element described in blueprints.xml.
     */
    private Map exceptionDestinations = new HashMap();

    /**
     * <code>EntityFactory</code> uses <code>exceptionSession</code> to obtain
     * <code>QueueSender</code>s for propagating exceptions
     */
    private  QueueSession exceptionSession;

    /**
     * The collection of <code>ClassMetaData</code> instances to be used as blueprints for
     * constructing entity instances
     */
    private  EBlueprints blueprints = null;

    /**
        The context in which this <code>EntityFactory</code> was constructed.
     */
    private SessionContext context;

    /**
     * <code>EntityFactoryBean</code> is part of a session bean, therefore it must
     * comply with the EJB spec, which says that all EJBs must contain a public
     * no-argument constructor.
        @roseuid 407FF87302D6
     */
    public EntityFactoryBean() {

    }

    private void closeResources(InitialContext ic) {
        if (ic != null) {
            try {
                ic.close();
            } catch (Throwable eatIt) {}
        }

        if (exceptionSession != null) {
            try {
                exceptionSession.close();
            } catch (Throwable eatIt) {}
        }
    }

    /**
        Access method for the exceptionQueueConnectionFactory property.

     @return   the current value of the exceptionQueueConnectionFactory property
     */
    public QueueConnectionFactory getExceptionQueueConnectionFactory() {
        return exceptionQueueConnectionFactory;
    }

    /**
        Sets the value of the exceptionQueueConnectionFactory property.

        @param aExceptionQueueConnectionFactory the new value of the exceptionQueueConnectionFactory property
     */
    public void setExceptionQueueConnectionFactory(QueueConnectionFactory
            aExceptionQueueConnectionFactory) {
        exceptionQueueConnectionFactory = aExceptionQueueConnectionFactory;
    }

    /**
        Access method for the exceptionDestinations property.

        @return   the current value of the exceptionDestinations property
     */
    public Map getExceptionDestinations() {
        return exceptionDestinations;
    }

    /**
        Sets the value of the exceptionDestinations property.

        @param aExceptionDestinations the new value of the exceptionDestinations property
     */
    public void setExceptionDestinations(Map aExceptionDestinations) {
        exceptionDestinations = aExceptionDestinations;
    }

    /**
        Access method for the exceptionSession property.

        @return   the current value of the exceptionSession property
     */
    public QueueSession getExceptionSession() {
        return exceptionSession;
    }

    /**
        Sets the value of the exceptionSession property.

        @param aExceptionSession the new value of the exceptionSession property
     */
    public void setExceptionSession(QueueSession aExceptionSession) {
        exceptionSession = aExceptionSession;
    }

    /**
        Access method for the blueprints property.

        @return   the current value of the blueprints property
     */
    public Map getBlueprints() {
        return blueprints;
    }

    /**
        Sets the value of the blueprints property.

        @param aBlueprints the new value of the blueprints property
     */
    public void setBlueprints(EBlueprints aBlueprints) {
        blueprints = aBlueprints;
    }

    public Serializable convertToFullEntity(String inXMLString) throws
            EntityException {

        java.io.ByteArrayInputStream bs = null;
        try {
            //construct a DOM document builder factory
            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.
                    newInstance();
            /*wrap the passed-in string in a StringBufferInputStream, thereby making
              it appear like an InputStream to the DOM document builder*/
            bs = new java.io.ByteArrayInputStream(new String(inXMLString.trim().getBytes(), "UTF-8").getBytes());
            /*construct a DOM document builder from the document builder factory*/
            DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
            docBuilder.setErrorHandler(new org.quintet.monk.MonkErrorHandler());
            /*ask the doc builder to convert the passed-in XML into a DOM document*/
            Document doc = docBuilder.parse(bs);
            if (bs != null) {
                try {
                    bs.close();
                } catch (Throwable eatIt) {}
            }
            /*normalize the document*/
            doc.getDocumentElement().normalize();

            //turn the tree into some java objects

            Object answer = monk.create(doc);
            try {
                /*if the answer returned from Monk is Serializable then
                  return it to the caller*/
                Serializable serializableAnswer = null;
                if (answer instanceof java.io.Serializable) {
                    serializableAnswer = (Serializable) answer;
                } else {
                    String err = answer.getClass().getName() +
                                 " is not serializable, therefore cannot return it.";
                    log.log(Level.SEVERE, err);
                    return null;
                }

                closeResources(null);
                return serializableAnswer;
            } catch (Throwable somethingAwful) {
                /*something went wrong, so close resources and skedaddle outta here*/
                String err = somethingAwful.getClass().getName() +
                             "when converting to an entity";
                log.log(Level.SEVERE, err, somethingAwful);
                closeResources(null);
                return null;
            }
        } catch (ParserConfigurationException PCE) {
            String err =
                    "ParserConfigurationException when converting to an entity";
            log.log(Level.SEVERE, err, PCE);
            closeResources(null);
            throw new EntityException(err, PCE);
        } catch (SAXException saxE) {
            String err = "SAXException when converting to an entity";
            log.log(Level.SEVERE, err, saxE);
            closeResources(null);
            throw new EntityException(err, saxE);
        } catch (IOException ioe) {
            String err = "IOException when converting to an entity";
            log.log(Level.SEVERE, err, ioe);
            closeResources(null);
            throw new EntityException(err, ioe);
        } catch (MonkConfigurationException ofce) {
            String err =
                    "MonkConfigurationException when converting to an entity";
            log.log(Level.SEVERE, err, ofce);
            closeResources(null);
            throw new EntityException(err, ofce);
        } catch (MonkException OFE) {
            log.log(Level.SEVERE,
                    "caught a MonkException, analyzing it for useful info ",
                    OFE);
            //OFE may have some valuable information in it, such as the exact node
            //that was being processed when the error occurred.  If we can determine the
            //errant node, and blueprints.xml designated a queue for sending such errant nodes,
            //then send the node there instead of throwing an exception.
            ClassMetaData offendingMetaData = OFE.getClassMetaData();
            if (offendingMetaData != null) {
                //we might have a queue where we can send the errant node
                String errantNode = OFE.getErrantNode();
                if (errantNode != null) {
                    //we have the errant node and the Class into which Monk was attempting to convert the node
                    Queue exceptionDestination = (Queue) exceptionDestinations.
                                                 get(
                            offendingMetaData);
                    if (exceptionDestination != null) {
                        //oh joy/boy, we know the node that blew up, and we have a place to report the error
                        sendExceptionToQueue(errantNode, exceptionDestination);
                    }
                }
                closeResources(null);
                return null;
            } else {
                String err = "MonkException, and no error queue was defined";
                log.log(Level.SEVERE, err, OFE);
                closeResources(null);
                return null;
            }
        } catch (Throwable ultimate) {
            String err = ultimate.getClass().getName() +
                         "when trying to convert to an entity";
            log.log(Level.SEVERE, err, ultimate);
            closeResources(null);
            return null;
        } finally {
            if (bs != null) {
                try {
                    bs.close();
                } catch (Throwable eatIt) {}
            }
        }
    }

    /**
     <code>convertToEntity</code> accepts a String containing XML and converts its
     contents to an entity bean manifestation.  If the resultant entity's primary
     key is serializable then it is returned to the caller, otherwise null is
        returned


        @param inXMLString - The XML String being converted to entity instances
     @return java.io.Serializable  -- usually the Primary Key of the new entity
        @throws org.quintet.chambers.EntityException
        @roseuid 408316FB025A
     */
    public Serializable convertToEntity(String inXMLString) throws
            EntityException {
        java.io.StringBufferInputStream is = null;
        try {
            //construct a DOM document builder factory
            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.
                    newInstance();
            /*wrap the passed-in string in a StringBufferInputStream, thereby making
              it appear like an InputStream to the DOM document builder*/
            is = new java.io.StringBufferInputStream(inXMLString);
            /*construct a DOM document builder from the document builder factory*/
            DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
            docBuilder.setErrorHandler(new org.quintet.monk.MonkErrorHandler());
            /*ask the doc builder to convert the passed-in XML into a DOM document*/
            Document doc = docBuilder.parse(is);
            if (is != null) {
                try {
                    is.close();
                } catch (Throwable eatIt) {}
            }
            /*normalize the document*/
            doc.getDocumentElement().normalize();

            //turn the tree into some java objects

            Object answer = monk.create(doc);
            try {
                /*if the answer returned from Monk is Serializable then
                  return it to the caller*/
                Serializable serializableAnswer = convertReturnValue(answer);
                closeResources(null);
                return serializableAnswer;
            } catch (NotSerializableException notCerealEyesAble) {
                String err =
                        "Created an non-Serializable object, returning null";
                log.log(Level.SEVERE, err, notCerealEyesAble);
                /*Alas, the answer is not serializable, therefore return a null*/
                closeResources(null);
                return null;
            }
        } catch (ParserConfigurationException PCE) {
            String err =
                    "ParserConfigurationException when converting to an entity";
            log.log(Level.SEVERE, err, PCE);
            closeResources(null);
            throw new EntityException(err, PCE);
        } catch (SAXException saxE) {
            String err = "SAXException when converting to an entity";
            log.log(Level.SEVERE, err, saxE);
            closeResources(null);
            throw new EntityException(err, saxE);
        } catch (IOException ioe) {
            String err = "IOException when converting to an entity";
            log.log(Level.SEVERE, err, ioe);
            closeResources(null);
            throw new EntityException(err, ioe);
        } catch (MonkConfigurationException ofce) {
            String err =
                    "MonkConfigurationException when converting to an entity";
            log.log(Level.SEVERE, err, ofce);
            closeResources(null);
            throw new EntityException(err, ofce);
        } catch (MonkException OFE) {
            log.log(Level.SEVERE,
                    "MonkException when trying to create an entity," +
                    "analyzing exception for useful data", OFE);
            //OFE may have some valuable information in it, such as the exact node
            //that was being processed when the error occurred.  If we can determine the
            //errant node, and blueprints.xml designated a queue for sending such errant nodes,
            //then send the node there instead of throwing an exception.
            ClassMetaData offendingMetaData = OFE.getClassMetaData();
            if (offendingMetaData != null) {
                //we might have a queue where we can send the errant node
                String errantNode = OFE.getErrantNode();
                if (errantNode != null) {
                    //we have the errant node and the Class into which Monk was attempting to convert the node
                    Queue exceptionDestination = (Queue) exceptionDestinations.
                                                 get(
                            offendingMetaData);
                    if (exceptionDestination != null) {
                        //oh joy/boy, we know the node that blew up, and we have a place to report the error
                        sendExceptionToQueue(errantNode, exceptionDestination);
                    }
                }
                closeResources(null);
                return null;
            } else {
                String err = "MonkException, and no error queue was defined";
                log.log(Level.SEVERE, err, OFE);
                closeResources(null);
                return null;
            }
        } catch (Throwable ultimate) {
            String err = ultimate.getClass().getName() +
                         "when trying to create an entity";
            log.log(Level.SEVERE, err, ultimate);
            closeResources(null);
            return null;
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (Throwable eatIt) {}
            }
        }
    }

    /**
        @throws javax.ejb.EJBException
        @throws java.rmi.RemoteException
        @roseuid 407FF87302E0
     */
    public void ejbActivate() throws EJBException, RemoteException {
        log.log(Level.INFO, "in EntityFactoryBean.ejbActivate");
    }

    /**
        @throws javax.ejb.EJBException
        @throws java.rmi.RemoteException
        @roseuid 407FF87302EA
     */
    public void ejbPassivate() throws EJBException, RemoteException {
        log.log(Level.INFO, "in EntityFactoryBean.ejbPassivate");
    }

    /**
        @throws javax.ejb.EJBException
        @throws java.rmi.RemoteException
        @roseuid 407FF87302F4
     */
    public void ejbRemove() throws EJBException, RemoteException {
        log.log(Level.INFO, "in EntityFactoryBean.ejbRemove");
    }


    /**
     * Store the context in which this <code>EntityFactory</code> was intialized/called.
     * Also load in our blueprints from blueprints.xml.  These blueprints described exactly
     * how to convert XML elements into java objects.
     * Finally, instantiate our <code>QueueConnectionFactory</code> for use in exception processing.
        @param sessioncontext The context in which this bean was called
        @throws javax.ejb.EJBException
        @throws java.rmi.RemoteException
        @roseuid 407FF87302FE
     */
    public void setSessionContext(SessionContext sessioncontext) throws
            EJBException, RemoteException {
        //load our blueprints in from ClassMetaData.xml
        InitialContext ic = null;
        try {
            ic = new InitialContext();
        } catch (NamingException ne) {
            String err =
                    "NamingException when trying to get an initial context";
            log.log(Level.SEVERE, err, ne);
            throw new EJBException("EntityFactory:sesSessionContext:" +
                                   "caught a NamingException in initialization:  " +
                                   ne);
        }

        //Errant XML is sent to exception queues for manual intervention.
        //We need a QueueConnectionFactory to establish connectivity to these queues
        QueueConnectionFactory qFactory = null;
        try {
            qFactory = (QueueConnectionFactory) ic.lookup(
                    QUEUE_CONNECTION_FACTORY_NAME);
            setExceptionQueueConnectionFactory(qFactory);

        } catch (NamingException ex) {
            String warn = "Unable to obtain queue connection factory";
            log.log(Level.SEVERE, warn, ex);
        } finally {
            if (ic != null) {
                try {
                    ic.close();
                } catch (Throwable eatIt) {}
            }
        }
        closeResources(ic);
    }

    /**
     * Read our Blueprints (a collection of <code>ClassMetaData</code> instances) from
     * a configuration file.  Use <code>ctx</code> to look up subjects' home
     * interfaces.
     * One interesting tidbit to notice about <code>readBlueprints</code> is the fact that it
     * uses Monk to read the blueprints file, and then the blueprints are used by
     * Monk to perform the rest of its work thereafter.
     @param ctx - The context through which to look up local entity bean homes
        mentioned in configuration file
        @throws org.quintet.monk.MonkException
        @roseuid 407FF983005A
     */
    private void readBlueprints(InputStream inputSource) throws MonkException {
        blueprints = EBlueprints.buildEBlueprints(inputSource);
    }

    /**
     Create a new <code>EntityFactory</code> instance and initializes it with the
        blueprints found within the <code>blueprintsStream</code> <code>InputStream</code>
        @param <code>InputStream</code> <code>blueprintsStream</code>
        @throws java.rmi.RemoteException
        @throws javax.ejb.CreateException
        @roseuid 40856A9A013A
     */
    public void ejbCreate(InputStream blueprintsStream) throws CreateException {
        //InputSource source = new InputSource(blueprintsStream);
        try {
            readBlueprints(blueprintsStream);
            monk = Monk.createMonk(getBlueprints());
        } catch (MonkException ex) {
            String fatal = "MonkException when configuring blueprints";
            log.log(Level.SEVERE, fatal, ex);
            throw new CreateException(fatal + ":" + ex);
        }
    }

    public void ejbCreate(String blueprintsURL) throws CreateException {
        try {
            InputStream entityBlueprintsStream = new java.io.FileInputStream(
                    blueprintsURL);
            readBlueprints(entityBlueprintsStream);
            monk = Monk.createMonk(getBlueprints());
        } catch (IOException ioe) {
            String fatal = "Unable to open a file named " + blueprintsURL;
            log.log(Level.SEVERE, fatal, ioe);
            throw new CreateException(fatal);
        } catch (MonkException ex) {
            String fatal = "MonkException when configuring blueprints";
            log.log(Level.SEVERE, fatal, ex);
            throw new CreateException(fatal + ":" + ex);
        }
    }

    /**
        send <code>inErrantNode</code> to <code>inExceptionQueue</code> as an
        <code>ObjectMessage</code>
     @param inErrantNode - The String manifestation of the errant <code>Node</code>
        being sent to <code>inQueue</code>
     @param inExceptionQueue - The destination of the errant <code>Node</code>
        @throws org.quintet.chambers.EntityException
        @roseuid 4128D2B10010
     */
    private void sendExceptionToQueue(String inErrantNode,
                                      Queue inExceptionQueue) throws
            EntityException {
        QueueConnection qConn = null;
        QueueSession qSession = null;
        ObjectMessage msg = null;
        QueueSender sender = null;
        try {
            qConn = getExceptionQueueConnectionFactory().
                    createQueueConnection();
            qSession = qConn.createQueueSession(true,
                                                javax.jms.Session.
                                                AUTO_ACKNOWLEDGE);
            msg = qSession.createObjectMessage(inErrantNode);
            sender = qSession.createSender(inExceptionQueue);
            sender.send(msg);
            return;
        } catch (JMSException ex) {
            String err = "JMSException when trying to forward to error queue";
            log.log(Level.SEVERE, err, ex);
            EntityException couldNotSend = new EntityException(
                    "EntityFactory:sendExceptionToQueue:JMSException:" + ex, ex);
            throw couldNotSend;
        } finally {
            //close all resources if possible and necessary
            if (null != sender) {
                try {
                    sender.close();
                } catch (Throwable eatIt) {}
            }
            if (null != qSession) {
                try {
                    qSession.close();
                } catch (Throwable eatIt) {}
            }
            if (null != qConn) {
                try {
                    qConn.close();
                } catch (Throwable eatIt) {}
            }
        }
    }

    /**
        Convert <code>inObject</code> into a serializable manifestation.  If
     <code>inObject</code> is an EJBObject then convert it to its primary key.  If
        <code>inObject</code> is a collection, then recursively apply
        <code>convertReturnValue</code> to its members.

     @param inObject - The possibly-nonserializable object that is being converted
        to a serializable manifestation
        @return java.io.Serializable
        @throws java.io.NotSerializableException
        @roseuid 412C73B00368
     */
    private Serializable convertReturnValue(Object inObject) throws
            NotSerializableException {
        //nulls are serializable, right?
        if (inObject == null) {
            return (Serializable)null;
        }

        //if inObject is an EJBObject or an EJBLocalObject, then return its PK
        //(unless the PK is not Serializable, in which case all heck breaks loose)
        if ((inObject instanceof javax.ejb.EJBLocalObject)) {
            javax.ejb.EJBLocalObject ejbLocal = (javax.ejb.EJBLocalObject)
                                                inObject;
            Object pk = ejbLocal.getPrimaryKey();
            if (pk instanceof java.io.Serializable) {
                return (Serializable) pk;
            } else {
                throw new NotSerializableException(
                        "EntityFactory:convertReturnValue:" +
                        inObject.getClass().getName() +
                        "'s PK is not Serializable");
            }
        } else if (inObject instanceof javax.ejb.EJBObject) {
            javax.ejb.EJBObject ejbObject = (javax.ejb.EJBObject) inObject;
            try {
                Object pk = ejbObject.getPrimaryKey();
                if (pk instanceof java.io.Serializable) {
                    return (Serializable) pk;
                } else {
                    throw new NotSerializableException(
                            inObject.getClass().getName() +
                            "'s PK is not Serializable");
                }
            } catch (RemoteException ex) {
                throw new NotSerializableException(
                        "EntityFactory:convertReturnValue:" +
                        inObject.getClass().getName() +
                        ":RemoteException trying to access PK");
            }
        }
        if (inObject.getClass().isAnnotationPresent(Entity.class)) {
            // This is an EJB3 Entity... find the ID
            try {
                Serializable pk = (Serializable) getPKValue(inObject);
                return pk;
            } catch (Throwable t) {
                String err = t.getClass().getName() + " when trying to " +
                             "get primary key for an EJB 3 entity";
                log.log(Level.SEVERE, err, t);
                return null;
            }

        }
        //if inObject is not a collection and it is Serializable then return it
        if (!(inObject instanceof java.util.Collection) &&
            inObject instanceof java.io.Serializable) {
            return (Serializable) inObject;
        }

        //inObject is either a collection or it is an unserializable object
        if (inObject instanceof java.util.Collection) {
            //recursively convert each member of inObject
            Collection rawCollection = (Collection) inObject;
            //ArrayList is Serializable, by the way
            Collection serializableCollection = new java.util.ArrayList(
                    rawCollection.
                    size());
            for (Iterator tIterIsBetter = rawCollection.iterator();
                                          tIterIsBetter.hasNext(); ) {
                Object rawOne = tIterIsBetter.next();
                Serializable serializableVersion = this.convertReturnValue(
                        rawOne);
                serializableCollection.add(serializableVersion);
            }
            return (Serializable) serializableCollection;
        } else {
            throw new NotSerializableException(
                    "EntityFactory:convertReturnValue:" +
                    inObject.getClass().getName() +
                    " is not serializable");
        }

    }

    private static Object getPKValue(Object obj) {
        Field pkField = null;
        Class clazz = obj.getClass();
        while (clazz != Object.class) {
            for (Field f : clazz.getDeclaredFields()) {
                if (f.isAnnotationPresent(Id.class)) {
                    pkField = f;
                    break;
                }
            }
            clazz = clazz.getSuperclass();
        }
        Object pkVal = null;
        if (pkField != null) {
            try {
                pkVal = getFieldValueFromGetter(obj, pkField);
            } catch (Throwable e) {
                String err = e.getClass().getName() +
                             "when trying to get EJB 3 PK Value";
                log.log(Level.SEVERE, err, e);
                return null;
            }
        }
        return pkVal;
    }

    public static String capitalize(String orig) {
        StringBuilder capStr = new StringBuilder();
        capStr.append(Character.toUpperCase(orig.charAt(0)));
        capStr.append(orig.substring(1));
        return capStr.toString();
    }

    /**
     * Finds and uses a getter for the specified Field on the provided obj.
     *
     * @param obj
     * @param field
     * @return
     */
    private static Object getFieldValueFromGetter(Object obj, Field field) {

        Method getterMethod = null;

        String methodName = "get" + capitalize(field.getName());
        Class clazz = obj.getClass();
        while (clazz != Object.class) {
            try {
                getterMethod = clazz.getDeclaredMethod(methodName, new Class[0]);
            } catch (Exception e) {
                // Don't report it
            }
            clazz = clazz.getSuperclass();
        }
        if (getterMethod != null) {
            Object result;
            try {
                result = getterMethod.invoke(obj, new Object[0]);
                return result;
            } catch (InvocationTargetException e) {
                Throwable cause = e.getCause() == null ? e:e.getCause();
                String err = cause.getClass().getName() + " from method "+
                             getterMethod.getName();
                log.log(Level.SEVERE,err,cause);

            } catch (Throwable e) {
                String err = e.getClass().getName() +
                             " from method " + getterMethod.getName();
                log.log(Level.SEVERE,err,e);
            }

        }
        return null;
    }

    /**
     * The <code>ping</code> method is a very simple mechanism to check
     * if this stateful session bean is still "alive".
     * @return boolean
     */
    public boolean ping(){
        return true;
    }
}
