package org.quintet.chambers;

import org.quintet.monk.Monk;
import org.quintet.monk.MonkException;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.util.logging.Logger;
import java.util.logging.Level;
import org.quintet.backbeat.BackBeatException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.IOException;
import org.w3c.dom.Document;
import java.io.StringBufferInputStream;



/**
 * <code>EntityTranscriber</code> is a general-purpose mechanism for converting XML documents
 * into java object instances; it is largely a J2EE wrapper around <code>Monk</code>
 * @author Peter Thaggard
 * @version 1.0
 * @since Quintet version 2003.1
 */
public class EntityTranscriber implements EntityRetriever{

  /**The place where I send diagnostics
   *
   */
  private static final Logger log = Logger.getLogger("org.quintet.chambers");

  /**
   * <code>EntityTranscriber</code> delegates virtually all of its work to its underlying
   * <code>transcriber</code>.
   */
  private Monk transcriber = null;

  private static final String BLUEPRINTS_KEY = "java:comp/env/blueprints-file-name";

  private static void closeResources(InitialContext ic){
    if(null != ic){
      try{
        ic.close();
      }catch(Throwable eatIt){}
    }
  }
  public EntityTranscriber() throws BackBeatException{
    InitialContext ic = null;
    try{
      ic = new InitialContext();
    }catch(NamingException ne){
      String fatal = "Unable to acquire a JNDI Context:  " ;
      log.log(Level.SEVERE,fatal,ne);
      throw new BackBeatException(fatal,ne);
    }

    String blueprintsFileName = null;
    try {
      blueprintsFileName = (String) ic.lookup(BLUEPRINTS_KEY);
    }
    catch (NamingException ex) {
      String fatal = "Unable to find " + BLUEPRINTS_KEY + " in JNDI";
      log.log(Level.SEVERE,fatal,ex);
      closeResources(ic);
      throw new BackBeatException(fatal,ex);
    }


    FileInputStream blueprintsStream = null;
    try{
     blueprintsStream = new FileInputStream(blueprintsFileName);
    }catch(IOException ioe){
      String fatal = "Unable to open the blueprints file named '" + blueprintsFileName + "'";
      log.log(Level.SEVERE,fatal,ioe);
      closeResources(ic);
      throw new BackBeatException(fatal,ioe);
    }

    try{
      transcriber = Monk.createMonk(blueprintsStream);
    }catch(MonkException ofe){
      String fatal = "Ill-formed blueprints file";
      log.log(Level.SEVERE,fatal,ofe);
      throw new BackBeatException(fatal,ofe);
    }finally{
      closeResources(ic);
      if(blueprintsStream != null){
        try{
          blueprintsStream.close();
        }catch(Throwable eatIt){}
      }
    }

    //we now have a properly instantiated transcriber/object factory.

  }

  /**
   * retrieve
   *
   * @param pk Object
   * @return Object
   * @throws RetrieverException
   */
  public Object retrieve(Object pk) throws RetrieverException{
    if(null == pk){
      String warn = "null PK, therefore returning null entity";
      log.warning(warn);
      return null;
    }

    Object answer = null;
    //we can handle Strings containing XML or already-parsed XML docs
    if(pk instanceof Document){
      Document doc = (Document)pk;
      try{
        answer = transcriber.create(doc);
        return answer;
      }catch(MonkException ofe){
        String err = "Unable to transcribe Document";
        log.log(Level.SEVERE,err,ofe);
        throw new RetrieverException(err,ofe);
      }
    }else if(pk instanceof java.lang.String){
      //pk is a String of XML
      InputStream docStream = new StringBufferInputStream((String)pk);
      try{
        answer = transcriber.create(docStream);
        return answer;
      }catch(MonkException ofe){
        String err = "Unable to transcribe String " + (String)pk;
        log.log(Level.SEVERE,err,ofe);
        throw new RetrieverException(err,ofe);
      }finally{
        if(null != docStream){
          try{
            docStream.close();
          }catch(Throwable eatIt){}
        }
      }
    }else{
      String err = "Unable to process PKs of type " + pk.getClass().getName();
      log.log(Level.SEVERE,err);
      throw new RetrieverException(err);
    }
  }
}
