package org.quintet.chambers;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import org.quintet.ellington.Arranger;
import org.quintet.ellington.JNDIArranger;
import org.quintet.QuintetException;
import java.util.Map;
import java.util.logging.Logger;
import java.util.logging.Level;


/**
 * <code>ObjectUnwrapperRetriever</code> transforms an object by extracting,
 * or "unwrapping" an object from a passed-in object.  For example,
 * <code>ObjectUnwrapperRetriever</code> might transform a <code>Principal</code>
 * into the <code>Principal</code>'s user ID by invoking
 * <code>Principal.getInternalID()</code>.<p>
 * <code>ObjectUnwrapperRetriever</code> uses three configuration values to
 * accomplish this unwrapping:<br>
 * <li><code>target-class-name</code> contains the name of the class of objects from which
 * it can extract values; in the example above this might be
 * <code>com.sixthsense.principal.Principal</code>
 * <li><code>extraction-method-name</code> contains the name of the method
 * that <code>ObjectUnwrapperRetriever</code> will invoke on the target object
 * in order to extract/unwrap a value; in the example above this might be
 * <code>getInternalID()</code>.  The method must be non-static, must accept
 * exactly zero arguments, and must return an object of type
 * <code>extraction-method-return-class-name</code> (defined below).
 * <li><code>extraction-method-return-class-name</code> is the name of the class
 * of object returned from the extractor method; in the example above this
 * might be <code>java.lang.Long</code><br>
 *
 * <p>Copyright: Copyright (c) 2006</p>
 *
 * <p>Company: 6th Sense Analytics</p>
 *
 * @author Peter Thaggard
 * @version 1.0
 */
public class ObjectUnwrapperRetriever implements EntityRetriever {

  private static final Logger log = Logger.getLogger("org.quintet.backbeat");

  private Class targetClass = null;
  private Method extractor = null;


  public ObjectUnwrapperRetriever() {
    final String[] propertyNames = {
        "our-target-class-name",
        "our-extraction-method-name",
        "our-extraction-method-return-class-name"
    };

    Arranger arranger = new JNDIArranger();

    //read my properties from JNDI
    Map<String,Object> properties = null;
    try {
      properties = arranger.arrange(propertyNames);
    } catch (QuintetException ex) {
      String err = ex.getClass().getName() + " when trying to configure a " +
          this.getClass().getName();
      log.log(Level.SEVERE,err,ex);
      throw new IllegalStateException(err,ex);
    }

    //get the target Class
    String targetClassName = null;
    try{
      targetClassName = (String)properties.get(propertyNames[0]);
      if(targetClassName == null){
        String err = "Unable to find a property named \"" +
            propertyNames[0] + "\"";
        log.log(Level.SEVERE,err);
        throw new IllegalStateException(err);
      }
      targetClass = Class.forName(targetClassName);
    }catch(ClassNotFoundException cnfe){
      String err = "Unable to find a class named \"" +
          targetClassName + "\" in your Classpath.";
      log.log(Level.SEVERE,err,cnfe);
      throw new IllegalStateException(err,cnfe);
    }

    //get the extractor method's return class
    Class returnType = null;
    String returnTypeName = (String)properties.get(propertyNames[2]);
    if(returnTypeName == null){
      String err = "unable to find a property named \"" +
          propertyNames[2] + "\"";
      log.log(Level.SEVERE,err);
      throw new IllegalStateException(err);
    }
    try{
      returnType = Class.forName(returnTypeName);
    }catch(ClassNotFoundException cnfe){
      String err = "Unable to find a class named \"" +
          returnTypeName + " in your classpath.";
      log.log(Level.SEVERE,err,cnfe);
      throw new IllegalStateException(err,cnfe);
    }

    //get the extractor method from the target Class
    String extractorMethodName = null;
    try{
      extractorMethodName = (String)properties.get(propertyNames[1]);
      extractor = targetClass.getMethod(extractorMethodName,(Class[])null);
    }catch(NoSuchMethodException nsme){
      String err = "Unable to find a method with signature \"" +
          targetClass.getName() + "." + extractorMethodName + "():" +
          returnTypeName + "\"";
      log.log(Level.SEVERE,err,nsme);
      throw new IllegalStateException(err,nsme);
    }


  }

  /**
   * Transform <code>inRawData</code> to a new manifestation and return the new
   * manifestation to the client.
   *
   * @param inRawData - The raw data to be transformed to a new manifestation
   * @return java.lang.Object
   * @throws RetrieverException
   * @todo Implement this org.quintet.backbeat.transformer.Retriever method
   */
  public Object retrieve(Object inRawData) throws RetrieverException {
    if(null == inRawData){
      String err = "Unable to extract anything from a null object";
      log.log(Level.WARNING,err);
      return null;
    }

    Class inClass = inRawData.getClass();
    if(!(inClass.equals(targetClass))){
      String err = "Expected an instance of \"" + targetClass.getName() +
          "\", found an instance of \"" + inClass.getName() + "\"";
      log.log(Level.WARNING,err);
      throw new RetrieverException(err);
    }

    try {
      Object extracted = extractor.invoke(inRawData, (Object[])null);
      return extracted;
    } catch (InvocationTargetException ex) {
      Throwable cause = ex.getCause();
      String err = cause.getClass().getName() + " when invoking " +
          targetClass.getName() + "." + extractor.getName() + "()";
      log.log(Level.SEVERE,err,cause);
      throw new RetrieverException(err,cause);
    } catch (IllegalArgumentException cause) {
      String err = cause.getClass().getName() + " when invoking " +
          targetClass.getName() + "." + extractor.getName() + "()";
      log.log(Level.SEVERE,err,cause);
      throw new RetrieverException(err,cause);
    } catch (IllegalAccessException cause) {
      String err = cause.getClass().getName() + " when invoking " +
          targetClass.getName() + "." + extractor.getName() + "()";
      log.log(Level.SEVERE,err,cause);
      throw new RetrieverException(err,cause);
    }
  }
}
