package org.quintet.coltrane;


/**
 * A <code>Rule</code> is a functor (in STL parlance) capable of performing one atomic action upon <code>inContext</code>.
 * Although <code>Rule</code>s are chiefly intended to evaluate some condition and return a <code>RulesResult>
 * indicating that condition's disposition, they can actually be used wherever one might normally employ
 * the <b>Strategy/Policy</b> pattern. <br>
 * A <code>Rule</code> is also akin to a <i>binary predicate</i>; it accepts one argument
 * (the <i>context</i>) and returns a high-level form of a boolean result.
 * Rules return <code>RulesResult</code> objects.
 * <code>Rule</code>s either <i>Fail<i> or <i>succeed</i>.
 * @author Peter Thaggard
 * @version 1.0
 * @since Quintet 2001_1
 */
public abstract class Rule<T,ConsequenceType> {

    protected ConsequenceType consequence = null;

    public ConsequenceType getConsequence(){
        return consequence;
    }

    public void setConsequence(ConsequenceType newValue){
        consequence = newValue;
    }


	/**
	 * Indicates whether the rule failed or succeeded.
	 */
	private boolean failed = false;

	/**
	 * This <code>Rule</code>'s description
	 */
	private String desc;

	/**
	 * Default constructor for <code>Rule</code>.
	 * @roseuid 3EBBB557002F
	 */
	public Rule() {

	}

	/**
	 * Determines if the failed property is true.
	 *
	 * @return   <code>true<code> if the failed property is true
	 */
	public boolean getFailed() {
		return failed;
	}

	/**
	 * Sets the value of the failed property.
	 *
	 * @param aFailed the new value of the failed property
	 */
	public void setFailed(boolean aFailed) {
		failed = aFailed;
	}

	/**
	 * Access method for the desc property.
	 *
	 * @return   the current value of the desc property
	 */
	public String getDesc() {
		return desc;
		}

	/**
	 * Sets the value of the desc property.
	 *
	 * @param aDesc the new value of the desc property
	 */
	public void setDesc(String aDesc) {
		desc = aDesc;
		}

	/**
	 * Evaluate this rule in terms of <code>inContext</code>.
	 * @param inContext - The object in whose context this <code>Rule</code> is being evaluated
	 * @return org.quintet.coltrane.RulesResults
	 * @roseuid 3EBBB2F4001D
	 */
	public abstract RulesResults fire(T inContext) ;
}
