package org.quintet.coltrane;

import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Logger;
import java.util.logging.Level;

/**
 * A <code>RuleSet</code> contains a collection of all rules that
 * pertain to <code>target</code>.
 */
public class RuleSet<T,ConsequenceType,RuleType extends Rule> extends Rule<T,ConsequenceType>{
    private static final Logger log = Logger.getLogger("org.quintet.coltrane");
    private int setType = EVALUATE_UNTIL_ONE_SUCCEEDS;
    public static final int EVALUATE_UNTIL_ONE_SUCCEEDS = 0;
    public static final int EVALUATE_UNTIL_ONE_FAILS = 1;
    public static final int EVALUATE_ALL_ALWAYS = 2;

    private Object target;
    private Collection<RuleType> rules = new ArrayList<RuleType>();

    /**
     * @roseuid 3ECD3B520231
     */
    public RuleSet() {

    }

    public RuleSet(int behavior){
        if(behavior == EVALUATE_UNTIL_ONE_SUCCEEDS ||
           behavior == EVALUATE_UNTIL_ONE_FAILS    ||
           behavior == EVALUATE_ALL_ALWAYS){
            setType = behavior;
        }else{
            String err = "Unknown rule set behavior type \""+behavior+"\".  "+
                         "Acceptable values are \""+
                         EVALUATE_UNTIL_ONE_SUCCEEDS +
                         " (stop evaluating rules as soon as one succeeds), "+
                         EVALUATE_UNTIL_ONE_FAILS +
                         " (stop evaluating rules as soon as one fails), and "+
                         EVALUATE_ALL_ALWAYS +
                         " (always evaluate all of the rules, regardless of "+
                         "the success or failure of any one rule).";
            log.log(Level.SEVERE,err);
            throw new IllegalStateException(err);
        }
    }

    public RulesResults fire(T context){
        RulesResults inToto = new RulesResults();
        Collection<RuleType> myRules = getRules();
        if(myRules == null || myRules.size() ==0){
            String message = "No rules defined";
            inToto.addMessage(message);
            inToto.setFailed(true);
            inToto.setFatal(false);
            return inToto;
        }else{
            boolean atLeastOneSucceeded = false;
            for(Rule rule:myRules){
                RulesResults result = rule.fire(context);
                inToto.addChild(result);
                //use the setType to decide whether or not to continue firing
                if(result.getFailed() == true){
                    if(setType == EVALUATE_UNTIL_ONE_FAILS) {
                        inToto.setFailed(true);
                        return inToto;
                    }
                }
                if(result.getFailed() == false){
                    atLeastOneSucceeded = true;
                    if(setType == EVALUATE_UNTIL_ONE_SUCCEEDS){
                        inToto.setFailed(false);
                        inToto.setRule(result.getRule());
                        return inToto;
                    }
                }
            }
            //If we reach this point then all rules have fired.
            //If setType is UNTIL_ONE_SUCCEEDS and none ever succeeded then
            //we need to consider the rules execution to have been a failure
            if(setType == EVALUATE_UNTIL_ONE_SUCCEEDS && atLeastOneSucceeded == false){
                inToto.setFailed(true);
                inToto.addMessage("No Rules Succeeded");
            }else{
                inToto.setFailed(false);
            }
            return inToto;
        }
    }
    /**
     * Access method for the target property.
     *
     * @return   the current value of the target property
     */
    public Object getTarget() {
        return target;
    }

    /**
     * Sets the value of the target property.
     *
     * @param aTarget the new value of the target property
     */
    public void setTarget(Object aTarget) {
        target = aTarget;
    }

    /**
     * Access method for the rules property.
     *
     * @return   the current value of the rules property
     */
    public Collection<RuleType> getRules() {
        return rules;
    }

    /**
     * Sets the value of the rules property.
     *
     * @param aRules the new value of the rules property
     */
    public void setRules(Collection<RuleType> aRules) {
        rules = aRules;
    }

    /**
     * This method is here to lull Introspector into thinking that <code>RuleSet</code> has a property called "rule".
     * @return org.quintet.coltrane.Rule
     * @roseuid 3ECD3DB30056
     */
    public Rule getRule() {
        return null;
    }

    /**
     * This inentionally-misnamed setter exists so that Introspection can call <code>setRule(Rule)</code>.  The actual implementation of this method actually adds the passed-in <code>Rule</code> to the <code>rules</code> collection.
     * @param inRule - The <code>Rule</code> being added to this <code>RuleSet</code>'s collection of <code>Rule</code>s
     * @roseuid 3ECD3DFA03A1
     */
    public void setRule(RuleType inRule) {
        getRules().add(inRule);
    }

    public void addRule(RuleType inRule){
            getRules().add(inRule);
    }
}
