package org.quintet.coltrane;

import java.io.InputStream;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import org.quintet.monk.Monk;
import org.quintet.monk.MonkException;
import java.util.logging.Logger;
import java.util.logging.Level;

/**
 * <code>RulesManager</code> determines which rules apply to a given object,
 * executes those rules, and returns their results to callers.
 * <code>RulesManager</code>'s primary interface is via the <code>fire()</code> method.
 */
public abstract class RulesManager {

  private static final Logger log = Logger.getLogger("or.quintet.coltrane");

  /**
   * The inventory of all <code>Rule</code>s maintained by this <code>RulesManager</code>.
   * The inventory is classified according to the class of object subject to the rules,
   * so <code>rules(class)</code> returns the rules pertinent to <code>class</code>;
   * in java parlance, <code>rules<code> is a <code>Map</code> where <code>java.lang.Class</code>
   * is the key and <code>java.util.Collection</code> is the value.
   */
  private Map<Class,Collection<Rule>> rules = new HashMap<Class,Collection<Rule>>();


  /**
   * The <code>attributeRules</code> are a collection of <code>Rule</code>s that are
   * associated with individual class attributes.  For example, one could have an
   * <code>attributeRule</code> to validate a <code>Principal</code>'s <code>loginID</code> attribute.
   */
  private Map attributeRules = new HashMap();
  private Map operationRules = new HashMap();

  /**
   * @roseuid 3EBBB55700DA
   */
  private RulesManager() {

  }

  protected RulesManager(String rulesBaseFileName) throws IllegalStateException{
    InputStream myRulesBase = null;
    InputStream coltraneBlueprintsStream = null;
    final String BLUEPRINTS_NAME="coltrane-blueprints.xml";
    Class thisClass = this.getClass();
    try {
      myRulesBase = thisClass.getResourceAsStream(rulesBaseFileName);
      coltraneBlueprintsStream = org.quintet.coltrane.RulesManager.class.getResourceAsStream(BLUEPRINTS_NAME);
      if(coltraneBlueprintsStream == null){
        String err = "Unable to find file \"" + BLUEPRINTS_NAME + "\" " +
            "within the jar/zip/directory containing the \"" +
            thisClass.getName() + "\" class.";
        log.log(Level.SEVERE,err);
        throw new IllegalStateException(err);
      }
      Monk monk = Monk.createMonk(coltraneBlueprintsStream);
      Collection<RuleSet> ruleSets = (Collection<RuleSet>)monk.create(myRulesBase);
      for(RuleSet currSet:ruleSets){
        String targetName = (String) currSet.getTarget();
        if (currSet instanceof org.quintet.coltrane.ClassRuleSet) {
          Collection<Rule> currRules = currSet.getRules();
          try {
            Class target = Class.forName(targetName);
            rules.put(target, currRules);
          }
          catch (ClassNotFoundException CNFE) {
            String err = "unable to load class \"" + targetName + "\"";
            log.log(Level.SEVERE,err,CNFE);
            throw new IllegalStateException(err,CNFE);
          }
        }
        else {
          String err =
              "RulesManager does not currently support rules targeted towards " +
              targetName;
          log.log(Level.SEVERE,err);
          throw new IllegalStateException(err);
        }
      }
    }
    catch (MonkException e) {
      String err = "MonkException in <init>";
      log.log(Level.SEVERE,err,e);
      throw new IllegalStateException(err,e);
    }finally{
      if(myRulesBase != null){
        try{
          myRulesBase.close();
        }catch(Throwable eatIt){}
      }
      if(coltraneBlueprintsStream != null){
        try{
          coltraneBlueprintsStream.close();
        }catch(Throwable eatIt){}
      }
    }
  }

  /**
   * Access method for the rules property.
   *
   * @return   the current value of the rules property
   */
  public Map getRules() {
    return rules;
  }

  /**
   * Sets the value of the rules property.
   *
   * @param aRules the new value of the rules property
   */
  public void setRules(Map aRules) {
    rules = aRules;
  }



  /**
   * Access method for the attributeRules property.
   *
   * @return   the current value of the attributeRules property
   */
  public Map getAttributeRules() {
    return attributeRules;
  }

  /**
   * Sets the value of the attributeRules property.
   *
   * @param aAttributeRules the new value of the attributeRules property
   */
  public void setAttributeRules(Map aAttributeRules) {
    attributeRules = aAttributeRules;
  }

  /**
   * Access method for the operationRules property.
   *
   * @return   the current value of the operationRules property
   */
  public Map getOperationRules() {
    return operationRules;
  }

  /**
   * Sets the value of the operationRules property.
   *
   * @param aOperationRules the new value of the operationRules property
   */
  public void setOperationRules(Map aOperationRules) {
    operationRules = aOperationRules;
  }

  /**
   * Evaluate the appropriate rules for <code>inContext</code>, return the results as a <code>RulesResults</code> object.
   * @param inContext - The object being evaluated
   * @return org.quintet.coltrane.RulesResults
   * @throws org.quintet.coltrane.RulesException
   * @roseuid 3EBBB1D10187
   */
  public RulesResults fire(Object inContext) throws RulesException {
    //Preconditions
    //1.  inContext != null
    if (inContext == null) {
      throw new RulesException("RulesManager.fire:  Error:  inContext is null");
    }
    //end preconditions

    RulesResults myResults = new RulesResults();
    Class contextClass = inContext.getClass();
    Collection pertinentRules = getRules(contextClass);
    if (pertinentRules != null) {
      for (Iterator rulesIter = pertinentRules.iterator(); rulesIter.hasNext(); ) {
        Rule currRule = (Rule) rulesIter.next();
        RulesResults currResults = currRule.fire(inContext);
        myResults.getChildren().add(currResults);
        if (currResults.isFatal()) {
          myResults.setFatal(true);
          return myResults;
        }
      }
      //all rules were evaluated, and none of them were fatal
      myResults.setFatal(false);
    }
    return myResults;
  }

  /**
   * Fetch the rules for objects of type <code>inClass</code>, return them as a <code>Collection</code>.
   * @param inClass - The class whose rules are being sought
   * @return java.util.Collection
   * @roseuid 3EBBB614037A
   */
  public Collection getRules(Class inClass) {
    //Preconditions
    //1.  inClass != null
    if (inClass == null) {
      return null;
    }
    //end preconditions
    return (Collection) getRules().get(inClass);
  }



  /**
   * Find and fire all rules for the attribute <code>inAttributeName</code> of class <code>inClassName</code>.  Both <code>inAttributeName</code> and <code>inClassName</code> are <code>String</code>s, so this method can be called for a value that is not
   * actually part of a class.  For example, one might want to validate that a <code>String</code> contains a valid <code>SellerNumber</code>.  the caller can execute <code>fireForAttribute("sellerNumber","Seller")</code> without having an instance of
   * <code>Seller</code> available.
   * @param inAttributeName - The name of the attribute upon which <code>Rule</code>s are being fired.   the name must be the exact name of the attribute as declared in its class; e.g., "sellerNumber", not "SellerNumber".
   * @param inClassName - The name of the class that contains <code>inAttributeName</code>.  This must be the fully qualified name; e.g., <code>org.quintet.monk.Monk</code>
   * @param inAttributeValue - The value of <code>inAttributeName</code> that will be evaluated by the selected <code>Rule</code>s
   * @return org.quintet.coltrane.RulesResults
   * @throws org.quintet.coltrane.RulesException
   * @roseuid 3ECA386A02A9
   */
  public RulesResults fireByAttributeName(String inAttributeName,
                                          String inClassName,
                                          Object inAttributeValue) throws
      RulesException {
    //Preconditions
    //1.  inAttributeName != null
    if (inAttributeName == null) {
      throw new RulesException(
          "RulesManager.fireByAttributeName:  Error:  inAttributeName is null");
    }

    //2.  inClassName != null
    if (inClassName == null) {
      throw new RulesException(
          "RulesManager.fireByAttributeName:  Error:  inClassName is null");
    }

    //end preconditions

    RulesResults myResults = new RulesResults();
    Collection pertinentRules = getRulesForAttributeByName(inAttributeName,
        inClassName);
    if (pertinentRules != null) {
      for (Iterator rulesIter = pertinentRules.iterator(); rulesIter.hasNext(); ) {
        Rule currRule = (Rule) rulesIter.next();
        RulesResults currResults = currRule.fire(inAttributeValue);
        myResults.getChildren().add(currResults);
        if (currResults.isFatal()) {
          myResults.setFatal(true);
          return myResults;
        }
      }
      //all rules were evaluated, and none of them were fatal
      myResults.setFatal(false);
    }
    return myResults;
  }

  /**
   * Find and fire all <code>Rule</code>s that are applicable to <code>inOperation</code>; <code>inOperation</code> is an instance of <code>java.lang.reflect.Method</code>.
   * @param inOperation - The <code>Method</code> upon which <code>Rule</code>s are being fired
   * @return org.quintet.coltrane.RulesResults
   * @throws org.quintet.coltrane.RulesException
   * @roseuid 3ECA3B35001C
   */
  public RulesResults fireByOperation(Method inOperation) throws RulesException {
    return null;
  }

  /**
   * Find and fire all <code>Rule</code>s associated with <code>inAttribute</code>; <code>inAttribute</code> is a <code>java.lang.reflect.Field</code> instance.
   * @param inAttribute - The <code>java.lang.reflect.Field</code> upon which the rules are being fired.
   * @return org.quintet.coltrane.RulesResults
   * @throws org.quintet.coltrane.RulesException
   * @roseuid 3ECA3DF701A8
   */
  public RulesResults fireByAttribute(Field inAttribute) throws RulesException {
    return null;
  }

  /**
   * Fetch the collection of <code>Rule</code>s for the attribute <code>inAttributeName</code> of the class named <code>inClassName</code>.
   * @param inAttributeName - The name of the attribute whose <code>Rule</code>s are being sought
   * @param inClassName - The parent class of <code>inAttributeName</code>
   * @return java.util.Collection
   * @roseuid 3ECA49110078
   */
  public Collection getRulesForAttributeByName(String inAttributeName,
                                               String inClassName) {
    //Preconditions
    //1.  inAttributeName != null
    if (inAttributeName == null) {
      return null;
    }

    //2.  inClassName != null
    if (inClassName == null) {
      return null;
    }
    //end preconditions

    //create an AttributeName object to use as the lookup key
    AttributeName key = new AttributeName(inAttributeName, inClassName);
    return (Collection) getAttributeRules().get(key);
  }

  /**
   * <code>AttributeName</code> uniquely identifies one attribute of one class.
   */
  public class AttributeName {

    /**
     * The name of the attribute that is identified by this <code>AttributeName</code>
     */
    private String attributeName;

    /**
     * The name of the class that houses <code>attributeName</code>
     */
    private String className;

    /**
     * Construct an <code>AttributeName</code> that uniquely identifies the attribute named <code>inAttributeName</code> of class <code>inClassName</code>.
     * @param inAttributeName - The unique (within the context of <code>inClassName</code>) name of the attribute identified by this <code>AttributeName</code>
     * @param inClassName - The class that owns <code>inAttributeName</code>
     * @roseuid 3ECA5D8301DD
     */
    public AttributeName(String inAttributeName, String inClassName) {

    }

    /**
     * Access method for the attributeName property.
     *
     * @return   the current value of the attributeName property
     */
    public String getAttributeName() {
      return attributeName;
    }

    /**
     * Sets the value of the attributeName property.
     *
     * @param aAttributeName the new value of the attributeName property
     */
    public void setAttributeName(String aAttributeName) {
      attributeName = aAttributeName;
    }

    /**
     * Access method for the className property.
     *
     * @return   the current value of the className property
     */
    public String getClassName() {
      return className;
    }

    /**
     * Sets the value of the className property.
     *
     * @param aClassName the new value of the className property
     */
    public void setClassName(String aClassName) {
      className = aClassName;
    }

    /**
     * Is this <code>AttributeName</code> "equal to" <code>inRHS</code>?
     * @param inRHS - The object being compared to this <code>AttributeName</code>
     * @return boolean
     * @roseuid 3ECA5E2D025A
     */
    public boolean equals(Object inRHS) {
      //there is a class precondition that forces both attributeName and className to
      //be non-null!
      try {
        AttributeName rhs = (AttributeName) inRHS;
        return getAttributeName().equals(rhs.getAttributeName()) &&
            getClassName().equals(rhs.getClassName());
      }
      catch (ClassCastException cce) {
        return false;
      }
    }

    /**
     * Generate a reasonable hash code for this <code>AttributeName</code>
     * @return int
     * @roseuid 3ECA5E81034B
     */
    public int hashCode() {
      return getAttributeName().hashCode() ^ getClassName().hashCode();
    }
  }
}
