package org.quintet.coltrane;

import java.util.*;

/**
 * <code>RulesResult</code> houses the results of evaluating a chain of rules.
 * A <code>RulesResult</code> is considered to be either <i>fatal</i> or <i>non-fatal</i>.  It is up to the client to decide what to do with this fatality flag.
 */
public class RulesResults<T,RuleType> {

    /**
     * The object upon which the <code>Rule</code> acted
     */
    private T context = null;

    public T getContext(){
        return context;
    }

    public void setContext(T newContext){
        context = newContext;
    }

    /**
     * The <code>Rule</code> whose execution is described by this <code>
     * RulesResults</code>.
     */
    private RuleType rule = null;

    /**
     * The <code>fatal</code> attribute indicates the severity of these <code>RulesResults</code>.
     */
    private boolean fatal = false;

    /**
     * RulesResults contain arbitrary collections of messages.
     */
    private Collection messages = new ArrayList();

    /**
     * Indicates if these <code>RulesResults</code> denote a successful or failed <code>Rule</code> firing.
     */
    private boolean failed = false;

    /**
     * The ID of the <code>GraphElement</code>  that was being evaluated when this <code>RulesResults</code> was created
     */
    private String target = null;

    /**
     * This <code>RulesResult</code>'s child results
     */
    private Collection children = new ArrayList();
    private RulesResults parent;

    /**
     * @roseuid 3EBBB91A02C9
     */
    public RulesResults() {

    }

    /**
     * Determines if the fatal property is true.
     *
     * @return   <code>true<code> if the fatal property is true
     */
    public boolean getFatal() {
        return fatal;
    }

    /**
     * Sets the value of the fatal property.
     *
     * @param aFatal the new value of the fatal property
     */
    public void setFatal(boolean aFatal) {
        fatal = aFatal;
    }

    /**
     * Access method for the messages property.
     *
     * @return   the current value of the messages property
     */
    public Collection getMessages() {
        return messages;
    }

    /**
     * Sets the value of the messages property.
     *
     * @param aMessages the new value of the messages property
     */
    public void setMessages(Collection aMessages) {
        messages = aMessages;
    }

    /**
     * Determines if the failed property is true.
     *
     * @return   <code>true<code> if the failed property is true
     */
    public boolean getFailed() {
        return failed;
    }

    /**
     * Sets the value of the failed property.
     *
     * @param aFailed the new value of the failed property
     */
    public void setFailed(boolean aFailed) {
        failed = aFailed;
    }

    /**
     * Access method for the target property.
     *
     * @return   the current value of the target property
     */
    public String getTarget() {
        return target;
    }

    /**
     * Sets the value of the target property.
     *
     * @param aTarget the new value of the target property
     */
    public void setTarget(String aTarget) {
        target = aTarget;
    }

    /**
     * Access method for the children property.
     *
     * @return   the current value of the children property
     */
    public Collection getChildren() {
        return children;
    }

    /**
     * Sets the value of the children property.
     *
     * @param aChildren the new value of the children property
     */
    public void setChildren(Collection aChildren) {
        children = aChildren;
    }

    public void addChild(RulesResults newKid) {
        newKid.setParent(this);
        getChildren().add(newKid);
    }

    /**
     * Access method for the parent property.
     *
     * @return   the current value of the parent property
     */
    public org.quintet.coltrane.RulesResults getParent() {
        return parent;
    }

    /**
     * Sets the value of the parent property.
     *
     * @param aParent the new value of the parent property
     */
    public void setParent(org.quintet.coltrane.RulesResults aParent) {
        parent = aParent;
    }

    /**
     * Mutator for the <code>rule</code> attribute.  The <code>rule</code>
     * attribute houses the <code>Rule</code> whose execution is described by this <code>
     * RulesResults</code>.
     * @param rule Rule
     */
    public void setRule(RuleType rule) {
        this.rule = rule;
    }

    /**
     * <code>isFatal</code> returns the value of the <code>fatal</code>
     * @return boolean
     * @roseuid 3EBBBD7D026A
     */
    public boolean isFatal() {
        return getFatal();
    }

    /**
     * Accessor for the <code>rule</code> attribute.  The <code>rule</code>
     * attribute houses the <code>Rule</code> whose execution is described by this <code>
     * RulesResults</code>.
     * @return Rule
     */
    public RuleType getRule() {
        return rule;
    }

    /**
     * Add <code>inMessage</code> to this <code>RulesResults</code>' message collection.
     * @param inMessage - The message being added to this <code>RulesResults</code>.
     * @roseuid 3EBBD2640279
     */
    public void addMessage(String inMessage) {
        getMessages().add(inMessage);
    }


    /**
     * Create a String representation of this RulesResult, its messages, and its children.
     * @return java.lang.String
     * @roseuid 3EBFB2F50002
     */
    public String toString() {
        StringBuffer answer = new StringBuffer();
        if (getFailed() == true) {
            //answer.append("Row " + (getTarget() == null ? "<no Row ID>" : getTarget()) + ":Failed?  " + getFailed() + ":Fatal?  " + isFatal() + ":");
            answer.append((getTarget() == null ? "<no Row ID>" : getTarget()) +
                          ":Failed?  " + getFailed() + ":Fatal?  " + isFatal() +
                          ":");
            if (getMessages() == null || getMessages().size() == 0) {
                answer.append("Messages:  None");
            } else {
                for (Iterator msgIter = getMessages().iterator();
                                        msgIter.hasNext(); ) {
                    String currMsg = (String) msgIter.next();
                    answer.append(currMsg + "\n");
                }
            }
            answer.append("\n");
        }
        if (getChildren() != null && getChildren().size() > 0) {
            for (Iterator childIter = getChildren().iterator();
                                      childIter.hasNext(); ) {
                RulesResults currChild = (RulesResults) childIter.next();
                answer.append(currChild.toString());
            }
        }
        return answer.toString();
    }

    /**
     * Determine whether the result and its children are valid or not
     *
     * @return boolean true for valid
     */
    public boolean isValid() {
        boolean returnStatus = true;

        if (getFailed() == true) {
            returnStatus = false;
        } else {
            //need loop through children RulesResults to determine whether all nodes are
            //validate or not.
            if (getChildren() != null && getChildren().size() > 0) {
                for (Iterator childIter = getChildren().iterator();
                                          childIter.hasNext(); ) {
                    RulesResults currChild = (RulesResults) childIter.next();
                    returnStatus = returnStatus & currChild.isValid();

                    if (returnStatus == false) {
                        return returnStatus;
                    }
                }
            }
        }

        return returnStatus;

    }
}
/**
 *
 * void RulesResults.RulesResult(){
 *
 * }
 *
 */
