package org.quintet.miles;

import java.util.Map;
import java.util.HashMap;
import java.util.Collection;

import java.util.logging.*;

import org.quintet.QuintetException;
import org.quintet.miles.AggregationAttribute;
import java.lang.reflect.Method;
import java.lang.annotation.Annotation;

public abstract class Activator
    implements Cloneable {

  private static final Logger log = Logger.getLogger("org.quintet.miles");

  /**
   * The <code>dnAnnotations</code> <code>Map</code> correlates DNNode RHSs (<code>uid</code>,
   * <code>day</code>, etc.) with aggregation-specific annotation information regarding the
   * <code>RHS</code>'s corresponding attribute in the <code>subject</code> class.
   */
  private Map<String,AggregationAttribute> dnAnnotations = new java.util.HashMap<String,AggregationAttribute>();

  /**
   * Extract all <code>AggregationAttribute</code> annotations from <code>subject</code>,
   * publish them to the <code>dnAnnotations</code> <code>Map</code>.  The
   * <code>dnAnnotations</code> <code>Map</code> correlates DNNode RHSs (<code>uid</code>,
   * <code>day</code>, etc.) with aggregation-specific annotation information regarding the
   * <code>RHS</code>'s corresponding attribute in the <code>subject</code> class.
   * @throws IllegalStateException
   */
  private void interrogateSubject() throws IllegalStateException{
    Class subj = getSubject();
    if(null == subj){
      String err = "subject is null, therefore I cannot interrogate it for DN->attribute info";
      log.log(Level.SEVERE,err);
      throw new IllegalStateException(err);
    }

    Method[] subjectMethods=subj.getMethods();
    for(Method method:subjectMethods){
      if(method.isAnnotationPresent(AggregationAttribute.class)){
        AggregationAttribute aa = method.getAnnotation(AggregationAttribute.class);
        String rhs = aa.dnName();
        dnAnnotations.put(rhs,aa);
      }
    }
  }



  /**
   * The <code>subject</code> attribute denotes the class of objects that this
   * <code>Activator</code> will be activating.
   */
  private Class subject = null;

  /**
   * The name of this <code>Activator</code>'s <code>subject</code> <code>Class</code>
   */
  private String subjectName = null;

  /**
   * Return the <code>Class</code> of objects that this
   * <code>Activator</code> will be activating.
   * @return Class
   */
  public Class getSubject() {
    return subject;
  }

  /**
   * Return the name of this <code>Activator</code>'s
   * <code>subject</code> <code>Class</code>
   * @return String
   */
  public String getSubjectName() {
    return subjectName;
  }

  public Map<String,AggregationAttribute> getDnAnnotations() {
    return dnAnnotations;
  }

  /**
   * Designate <code>inSubject</code> to be the <code>Class</code> of objects
   * that this <code>Activator</code> will be activating
   * @param subject Class
   */
  public void setSubject(Class inSubject) {
    this.subject = inSubject;
    interrogateSubject();
}

  /**
   * Conduit method for passing in the name of the <code>subject</code> class.
   * This method in turns fetches (via Class.forName()) the actual <code>subject</code>
   * <code>Class</code>.
   * @param inSubjectName String
   * @throws IllegalStateException if the <code>Class</code> named <code>inSubjectName</code> cannot be found
   */
  public void setSubjectName(String inSubjectName) throws IllegalStateException{
    this.subjectName = inSubjectName;
    try {
      setSubject(Class.forName(inSubjectName));
      interrogateSubject();
    }
    catch (ClassNotFoundException ex) {
      String err = "ClassNotFoundException from Class.forName(\"" + inSubjectName + "\")";
      log.log(Level.SEVERE,err,ex);
      throw new IllegalStateException(err,ex);
    }
  }


  public void setDnAnnotations(Map<String,AggregationAttribute> dnAnnotations) {
    this.dnAnnotations = dnAnnotations;
  }

  /**
   * Construct a new <code>Activator</code> and associate it with the
   * <code>subject</code> <code>Class</code> named <code>InSubjectName</code>
   * @param inSubjectName String.  This method also interrogates the <code>subject</code>
   * and builds a <code>Map</code> of information that is subsequently used to
   * process DN queries.
   * @throws IllegalStateException
   */
  public Activator(String inSubjectName) throws IllegalStateException{
    setSubjectName(inSubjectName);
  }

  public Activator(){

  }


  /*#org.quintet.miles.DN Dependency_Link*/
  /**
   * "Activate" the set of <code>Object</code>s designated by <code>setSpec</code>.
   * Activation is the act of retrieving the designated objects from some persistent store.
   * @param setSpec -- the DN describing the set of <code>Object</code>s to be activated
   * @return Collection of activated <code>Object</code>s
   * @throws AggregationException
   */
  public abstract Collection activate(DN setSpec) throws QuintetException;

  public Object clone() throws CloneNotSupportedException {
    return super.clone();
  }


}


