package org.quintet.miles;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import java.util.ArrayList;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.Iterator;
import java.util.Collection;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.regex.Pattern;

/**
 * An <code>Aggregation</code> is a collection of <code>Object</code>s (all of
 * which must be instances of <code>subject</code>) that are "slotted" according to
 * the value of their <code>attribute</code>'s value.  Each "slot", or "bucket"
 * comprising an <code>Aggregation</code> is an instance of
 * <code>org.quintet.backbeat.transformer.Range</code>.
 * @author Peter Thaggard
 */
public class Aggregation
    extends Range {

  /**
        The place where I send my diagnostic messages
   */
  private static final Logger log = Logger.getLogger("org.quintet.miles");

  /**
      <code>getter</code> retrieves the attribute of interest (that is, the attribute by which range membership is based) from candidate members
   */
  private transient Method getter = null;

  /**
      @roseuid 424962C40183
   */
  public Aggregation() {

  }

  /**
   * Return <code>true</code> is this <code>Range</code> is considered to be
   * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
   * method is intentionally declared abstract at the <code>Range</code> level
   * in order to force every <code>Range</code> subclass implementor to carefully
   * consider the nature of his class and whether it can truly be considered a
   * leaf node.
   * @return boolean
   */
  public boolean isLeafNode() {
    //if my members != null AND it is a Range then I am not a leaf node
    if (null != getMembers() && getMembers() instanceof Range) {
      return true;
    }
    else {
      return false;
    }
  }


  /**
      Access method for the getter property.

      @return   the current value of the getter property
   */
  public Method getGetter() {
    return getter;
  }

  /**
      Sets the value of the getter property.

      @param aGetter the new value of the getter property
   */
  public void setGetter(Method aGetter) {
    getter = aGetter;
  }

  /**
      Add <code>inNewMember</code> to this aggregation.  Within an
      <code>Aggregation</code>'s context, "adding" means:<br>
   <li>Finding the appropriate subordinate <code>Range</code> to house the new
      member
      <li>invoking that <code>Range</code>'s <code>add()</code> method<br>
      @param inNewMember
      @return boolean
      @roseuid 424962C4020F
   */
  public boolean add(Object inNewMember) {
    if (null == inNewMember) {
      String err = "cannot add() a null member";
      log.log(Level.WARNING,err);
      return false;
    }
    Range slot = findRange(inNewMember);
    if (slot == null) {
        //no appropriate subordinate range to house inNewMember
        Object aoi = null;
        try {
            aoi = getAttributeOfInterest(inNewMember);
        } catch (AggregationException ex) {
            String err = "Unable to find a Range to house new member.  " +
                         "Also, unable to obtain the new member's 'attribute of interest'.";
            log.log(Level.SEVERE,err, ex);
            return false;
        }
        String warn = "Unable to find a Range to house " + aoi;
        log.log(Level.WARNING,warn);
        return false;
    }
    //pass myself to slot's context so it can glean information from me at a later date
    slot.setContext(this);
    return slot.add(inNewMember);
}

/**
 Find and return the <code>Range</code> into which <code>inNewMember</code> fits.
 If no suitable <code>Range</code> exists then return null.
 @param inNewMember - The object for which a suitable <code>Range</code> is being
    sought
    @return org.quintet.miles.Range
    @roseuid 4249BE99010A
 */
public Range findRange(Object inNewMember) {
  //my subordinate ranges are housed in my members collection
  List ranges = super.getMembers();
  if (ranges == null || ranges.size() == 0) {
    String warn = "no subordinate ranges in this aggregation";
    log.log(Level.WARNING,warn);
    return null;
  }

  //get the attribute of interest from inNewMember
  Object attributeOfInterest = null;
  try {
    attributeOfInterest = getGetter().invoke(inNewMember, (Object[])null);
  }
  catch (InvocationTargetException ex) {
    Throwable cause = ex.getCause();
    String err = "Exception from getter: " + ex;
    log.log(Level.SEVERE,err, cause);
    return null;
  }
  catch (IllegalArgumentException ex) {
    String err = "IllegalArgumentException when invoking getter";
    log.log(Level.SEVERE,err, ex);
    return null;
  }
  catch (IllegalAccessException ex) {
    String err = "IllegalAccessException when invoking getter";
    log.log(Level.SEVERE,err, ex);
    return null;
  }
  for (Iterator rangeIter = ranges.iterator(); rangeIter.hasNext(); ) {
    Range currRange = (Range) rangeIter.next();
    if (currRange.isEligible(inNewMember)) {
      return currRange;
    }
  }
  //new current slot is capable of holding our new member,
  //so add a new slot to the members
  return null;
}

/**
    @param inCandidate
    @return boolean
    @roseuid 424D65FE02A4
 */
public boolean isEligible(Object inCandidate) {
  //inCandidate is eligible if and only if it is an eligible member of one of my
  //subordinate ranges
  for (Iterator subIter = iterator(); subIter.hasNext(); ) {
    Range subordinate = (Range) subIter.next();
    if (subordinate.isEligible(inCandidate)) {
      return true;
    }
  }
  return false;
}

/**
 Replace my collection of members with <code>inNewMembers</code>, and then
    propagate my
 <code>subject</code> and <code>getter</code> to each member; DO NOT overwrite
    non-null values already in the members, however.
    @param inNewMembers - My new membership list
    @roseuid 425584BA02C7
 */
public void setMembers(List inNewMembers) {
  super.setMembers(inNewMembers);
  //walk through each member and copy my subject and getter to them.
  //If a member already has its subject or getter set then DO NOT propagate my
  //values down to it.
  for (Iterator memberIter = getMembers().iterator(); memberIter.hasNext(); ) {
    Range currMember = (Range) memberIter.next();
    if (currMember.getSubject() == null) {
      currMember.setSubject(getSubject());
    }
    if (currMember.getGetter() == null) {
      currMember.setGetter(getGetter());
    }
    currMember.setContext(this);
  }
}

/**
 * Use <code>attributeValue</code> to "prune" this <code>Aggregation</code>,
 * thereby producing a new <code>Range</code> containing only
 * <code>member</code>s that meet the criteria expressed in <code>attributeValue</code>
 * @param attributeValue Object
 * @return Range -- A new "pruned" <code>Range</code> containing all <code>member</code>s
 * of this <code>Aggregation</code> that meet the criteria expressed in <code>attributeValue</code>
 * @throws AggregationException
 */
public Range findMembers(Object attributeValue) throws AggregationException {
  if (null == getMembers()) {
    return null;
  }
  if (! (attributeValue instanceof String)) {
    String err = "Cannot process DN nodes of type '" +
        attributeValue.getClass().getName();
    log.log(Level.SEVERE,err);
    throw new AggregationException(err);
  }
  String whomString = (String) attributeValue;

  DynamicRange answer = new DynamicRange();
  answer.setAggregationName(this.getAggregationName());
  Range srcBlueprint = getBlueprint();
  answer.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
  answer.setDescription(this.getDescription());
  answer.setGetter(this.getGetter());
  answer.setMembers(new java.util.ArrayList());
  answer.setSubject(this.getSubject());
  answer.setSubjectClass("[none]");
  Pattern pattern = Pattern.compile(whomString);
  for (Iterator memberIter = getMembers().iterator(); memberIter.hasNext(); ) {
    Object currentMember = memberIter.next();
    //attributeValue can be an exact value or it can be a regexpr
    //of course, all exact values are also regexprs....
    //if the current member is a String then this findMembers method works
    //fine; however, it will NOT work for any other sort of member object.
    //Therefore Range objects need a "matches" method that can be used
    //instead of equals
    if (currentMember instanceof String) {
      if (currentMember.equals(attributeValue)) {
        answer.add(currentMember);
      }
    }
    else if (currentMember instanceof Range) {
      //currMember, being a Range, has a matches method, which we can use to
      //determine if currMember matches the current DN
      Range currMemberRange = (Range) currentMember;
      if (currMemberRange.matches(pattern)) {
        answer.add(currMemberRange.deepCopy());
      }
    }
  }
  return answer;
}

/**
 * Use <code>dn</code> to traverse the <code>Range</code> rooted at this point.
 * Return the <code>Range</code> residing at the spot designated by <code>dn</code>
 * This is the most basic <code>findDN</code> implementation; several <code>Range</code>
 * subclasses override this method with significantly more sophisticated traversal
 * algorithms.  Speaking of algorithms, here is the only employed herein:<br>
 * <li>pop the top DNNode off <code>dn</code>
 * <li>walk down the tree rooted at this <code>Aggregation</code> until the tree node's <code>subjectClass</code> matches <code>topNode.LHS</code>
 * <li>"prune" the <code>Range</code>, thereby producing a <code>foundMembers</code> <code>Aggregation</code> containing values that meet
 * <code>topNode</code>'s criteria
 * <li>Remove the top <code>DNNode</code> from <code>dn</code>
 * <li>recursively process each <code>foundMember</code> by calling each ones <code>findDN</code> method, passing down the now-elided <code>dn</code>.
 * The value returned from this recursive call replaces the current <code>foundMember</code>'s <code>members</code> attribute
 * <br>
 * @param dn DN of the form <code><i>subjectClass=subjectValue[,subectClass=subjectValue...]
 * @return Range that most closely matches <code>dn</code>
 * @throws AggregationException
 */
public Range findDN(DN dn) throws AggregationException {
  //1.  check to see if this level of the tree matches the
  //    subject mentioned in the current top DN Node;
  //    for example, if dn="ou=acme,uid=joe" then check to see
  //    if this.subjectClass == ou.
  //    If they don't match then  walk down the subtree rooted here until
  //    a match is found and then restart the findDN process
  String subjectClass = getSubjectClass();
  if (null == subjectClass || subjectClass.length() == 0) {
    String err = "subject class is not set, cannot process DN '" + dn +
        "'";
    log.log(Level.SEVERE,err);
    throw new AggregationException(err);
  }

  DNNode<String, String> topNode = dn.peek();
  if (subjectClass.equals(topNode.LHS)) {
    //1.1  Yes, this level of the tree matches the subject class stipulated
    //     by the topNode.  Now "prune" this level of the tree (no
    //     sublevel pruning done yet), yielding a new Aggregation containing
    //     only those members of this Aggregation that meet topNode's criteria
    Range foundMembers = findMembers(topNode.RHS);
    if (null == foundMembers) {
      //There are no objects in this tree node that satisfy topNode
      return null;
    }
    else {
      //topNode, the current DN, just produced the results found in
      //foundMembers.  Record this DN-Range correlation by imprinting topNode
      //onto foundMembers
      foundMembers.setDnNode(topNode);

      //the node at which we just peek()ed is still lingering on the DN.  We
      //must remove it now...
      String logMsg = "DN is " + dn.toString() +
          ", about to remove top element";
      log.info(logMsg);
      dn.poll();
      logMsg = "DN is now " + dn.toString();
      log.info(logMsg);

      //we found some qualifying members at this level and placed them into
      //the "pruned" Aggregation named "foundMembers".    If there are remaining nodes on the DN
      //then delegate the remaining search to the found nodes
      if (dn.peek() == (DNNode<String, String>)null) { //check to see if there is more on the DN...
        //nothing left on the search spec, so we have found the thing being sought (I think....);
        return foundMembers;
      }
      else {
        //now apply the newly-pruned DN to each found member and allow each
        //such subtree to be visited by the pruned DN
        for (Iterator foundMemberIter = foundMembers.iterator();
             foundMemberIter.hasNext(); ) {
          DN dnForSubtree = (DN) dn.clone();
          Object currFoundMember = foundMemberIter.next();
          //this Aggregation's members aren't necessarily Ranges; they
          //could be plain old objects.  Only recurse into subordinate Ranges
          if (currFoundMember instanceof Range) {
            //prune the subtree anchored at currFoundMember
            Range subtree = (Range) currFoundMember;
            Range newSubtreeMembers = subtree.findDN(
                dnForSubtree);
            subtree.setMembers(newSubtreeMembers);
          }
        }
        return foundMembers;
      }
    }
  }
  else {
    //the current topNode is not matched to our subjectClass, so try to match
    //it to one of our members
    //the following code sickens me, but it does work, and it is the most
    //expeditious route to implementing a proper findDN method.
    //If the current Range's members() are a Group then process the group
    //directly, otherwise process each member of members()
    if (getMembers() instanceof Group) {
      Group memberGroup = (Group) getMembers();
      Range nextFound = memberGroup.findDN(dn);
      return nextFound;
    }
    else {
      //we may need to return several of our member nodes, bundled together
      //as a new DynamicRange, so build the "answer" container now
      Aggregation answer = new Aggregation();
      answer.setAggregationName(getAggregationName());
      answer.setAggregationType(getAggregationType());
      Range srcBlueprint = getBlueprint();
      answer.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
      answer.setDeclaredNodeStructure(getDeclaredNodeStructure());
      answer.setDescription(getDescription());
      answer.setGetter(getGetter());
      answer.setInquiryMethod(getInquiryMethod());
      answer.setSubject(getSubject());
      answer.setSubjectClass("[none]");
      answer.setDnNode(topNode);
      answer.setMembers(new java.util.ArrayList());
      for (Iterator memberIter = getMembers().iterator();
           memberIter.hasNext(); ) {
        Object currMember = memberIter.next();
        if (currMember instanceof Range) {
          Range possibleRange = (Range) currMember;
          Range nextFound = possibleRange.findDN(dn);
          //we have to push the current DN node back onto the DN
          //for use by the rest of the passes through this loop
          dn.offer(topNode);
          if (nextFound != null) {
            answer.add(nextFound);
          }
        }
      }
      dn.poll();
      return answer;
    }
  }
}

/**
 * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
 * emanating from it.
 * @return Range
 */
public Range deepCopy() throws AggregationException {
  Aggregation copy = new Aggregation();
  copy.setAggregationName(this.getAggregationName());
  copy.setAggregationType(this.getAggregationType());
  Range srcBlueprint = getBlueprint();
  copy.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
  copy.setDeclaredNodeStructure(this.getDeclaredNodeStructure());
  copy.setDescription(this.getDescription());
  copy.setDnNode(this.getDnNode());
  copy.setGetter(this.getGetter());
  copy.setInquiryMethod(this.getInquiryMethod());
  copy.setPassivator(this.getPassivator());
  copy.setPersistenceHandle(this.getPersistenceHandle());
  copy.setSerializable(this.isSerializable());
  copy.setSubject(this.getSubject());
  copy.setSubjectClass(this.getSubjectClass());
  copy.setContext(this.getContext());

  //The following is predicated on the fact that RangeSurgeons are completely
  //stateless and can therefore be shared among many Range clients
    copy.setRangeSurgeon(this.getRangeSurgeon());

  //now deepcopy the members individually to avoid pointing to an existing
  //object and subsequently accidentally modifying it
  List srcMembers = this.getMembers();
  if (null == srcMembers) {
    copy.setMembers( (List)null);
  }
  else {
    Class srcMembersClass = srcMembers.getClass();
    List copyList = (List) shallowCopy(srcMembersClass, srcMembers);
    for (Iterator memberIter = srcMembers.iterator();
         memberIter.hasNext(); ) {
      Object currSrcMember = memberIter.next();
      if (currSrcMember instanceof Range) {
        Range srcRange = (Range) currSrcMember;
        Range copyRange = srcRange.deepCopy();
        copyRange.setContext(copy);
        copyList.add(copyRange);
      }
    }
    copy.setMembers(copyList);
  }
  return copy;
}

/**
 * Am I "equal to " <code>inRHS</code>?  If <code>inRHS</code> and I
 * have the same getter then we are considered "equal"
 * @param inRHS Object
 * @return boolean
 */
public boolean equals(Object inRHS) {
  if (null == inRHS) {
    return false;
  }

  if (inRHS instanceof Aggregation) {
    Aggregation rhs = (Aggregation) inRHS;
    Method lhsGetter = this.getGetter();
    Method rhsGetter = rhs.getGetter();
    if (lhsGetter == null && rhsGetter == null) {
      return true;
    }
    if (lhsGetter != null && rhsGetter == null) {
      return false;
    }
    if (lhsGetter != null && rhsGetter != null) {
      return lhsGetter.equals(rhsGetter);
    }
    if (lhsGetter == null && rhsGetter != null) {
      return false;
    }
    return true;
  }
  else {
    return false;
  }
}
}
