package org.quintet.miles;

import java.sql.Timestamp;
import java.util.List;
import java.util.ArrayList;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.Iterator;

import java.util.logging.Logger;
import java.util.logging.Level;
import java.text.*;

/**
 A <code>BinaryRange</code> is a "slot" (actually it's a <code>Range</code>, but it is commonly
 * part of a larger collection of closely-related <code>Range</code>s that comprise a
 * segmented view of a larger <code>Range</code>, hence the "slot" moniker) that has both a defined lower
 * bound and a defined upper bound.  An object can become a member of a
 * <code>BinaryRange</code> if the object's attribute-of-interest is &gt;= floor and
 * &lt;= ceiling
 */
public class BinaryRange<MemberType> extends DynamicRange {

    private static final Logger log = Logger.getLogger("org.quintet.miles");
    /**
     Indicates if an <i>attribute of interest</i> equal to <code>floor</code> should
        be considered a member of this <code>Range</code>
     */
    private boolean floorIsInclusive = true;

    /**
     Indicates if an <i>attribute of interest</i> equal to <code>ceiling</code>
        should be considered a member of this <code>Range</code>
     */
    private boolean ceilingIsInclusive = false;

    /**
        The lowest value that an object can have while still being considered a member of this <code>Range</code>
     */
    private MemberType floor = null;

    /**
        The highest value that an object can have while still being considered a member of this <code>Range</code>
     */
    private MemberType ceiling = null;

    /**
        @roseuid 424ABD0A0195
     */
    public BinaryRange() {

    }

    /**
     * Return <code>true</code> is this <code>Range</code> is considered to be
     * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
     * method is intentionally declared abstract at the <code>Range</code> level
     * in order to force every <code>Range</code> subclass implementor to carefully
     * consider the nature of his class and whether it can truly be considered a
     * leaf node.
     * @return boolean
     */
  public  boolean isLeafNode(){
    //if my members != null AND my member's first member is
    //a List then I am NOT a leaf node
    List oneLevelDown = getMembers();
    if(null == oneLevelDown || 0 == oneLevelDown.size())
      return true;  //I have no underlings, therefore I am a leaf
    Object firstUnderling = oneLevelDown.get(0);
    if(firstUnderling instanceof List)
      return false; //I have underlings, and they are Lists/Ranges, therefore I am not a leaf node
    else
        return true; //I have underlings, but they are not Lists/Ranges, therefore I am a leaf node;
  }


    /**
        Determines if the floorIsInclusive property is true.

        @return   <code>true<code> if the floorIsInclusive property is true
     */
    public boolean getFloorIsInclusive() {
        return floorIsInclusive;
    }

    public Class getMemberTypeClass() {
        return getFloor().getClass();
    }

    public Object clone() {
        BinaryRange<MemberType> newOne = new BinaryRange<MemberType>();
        try {
          Range srcBlueprint = getBlueprint();
          newOne.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        }
        catch (AggregationException ex) {
          String err = "AggregationException from BinaryRange.getBlueprint().deepCopy()";
          log.log(Level.SEVERE,err);
          throw new IllegalStateException(err,ex);
        }
        newOne.setGetter(this.getGetter());
        newOne.setSubject(this.getSubject());
        newOne.setSubjectClass(this.getSubjectClass());
        newOne.setCeilingIsInclusive(this.getCeilingIsInclusive());
        newOne.setFloorIsInclusive(this.getFloorIsInclusive());
        Range newMembers = null;
        if ( null != getBlueprint()) {
	        newMembers = (Range) (getBlueprint().clone());
        }
        newOne.setMembers(newMembers);
        return newOne;
    }


    /**
        Sets the value of the floorIsInclusive property.

        @param aFloorIsInclusive the new value of the floorIsInclusive property
     */
    public void setFloorIsInclusive(boolean aFloorIsInclusive) {
        floorIsInclusive = aFloorIsInclusive;
    }

    /**
        Determines if the ceilingIsInclusive property is true.

        @return   <code>true<code> if the ceilingIsInclusive property is true
     */
    public boolean getCeilingIsInclusive() {
        return ceilingIsInclusive;
    }

    /**
        Sets the value of the ceilingIsInclusive property.

     @param aCeilingIsInclusive the new value of the ceilingIsInclusive property
     */
    public void setCeilingIsInclusive(boolean aCeilingIsInclusive) {
        ceilingIsInclusive = aCeilingIsInclusive;
    }

    /**
        Access method for the floor property.

        @return   the current value of the floor property
     */
    public MemberType getFloor() {
        return floor;
    }

    /**
     * Return the <code>floor</code> as a <code>String</code> manifestation of
     * a <code>Timestamp</code>
     * @return String
     */
    public String getEarliestTime() {
      java.text.SimpleDateFormat fmtYYYYMMDD = new java.text.SimpleDateFormat("yyyy-MM-dd");
      java.text.SimpleDateFormat fmtMMDDYYYY = new java.text.SimpleDateFormat("MM/dd/yyyy");
        if (null == getFloor()) {
            String otherTry = getAggregationName();
            if(null != otherTry){
                try {
                    java.util.Date dateFromName = fmtMMDDYYYY.parse(otherTry);
                    return fmtYYYYMMDD.format(dateFromName);
                } catch (ParseException ex) {
                    return "[NOT SET]";
                }
            }
            else
                return "[NOT SET]";
        } else {
            return "" + new Timestamp((Long) getFloor());
        }
    }

    /**
        Sets the value of the floor property.

        @param aFloor the new value of the floor property
     */
    public void setFloor(MemberType aFloor) {
        floor = aFloor;
    }

    /**
        Access method for the ceiling property.

        @return   the current value of the ceiling property
     */
    public MemberType getCeiling() {
        return ceiling;
    }

    /**
     * Return the <code>ceiling</code> as a <code>String</code> manifestation of
     * a <code>Timestamp</code>
     * @return String
     */
    public String getLatestTime() {
      java.text.SimpleDateFormat fmtYYYYMMDD = new java.text.SimpleDateFormat("yyyy-MM-dd");
      java.text.SimpleDateFormat fmtMMDDYYYY = new java.text.SimpleDateFormat("MM/dd/yyyy");

        if (null == getCeiling()) {
            String otherTry = getAggregationName();
            if(null != otherTry){
                try {
                    java.util.Date dateFromName = fmtMMDDYYYY.parse(otherTry);
                    return fmtYYYYMMDD.format(dateFromName);
                } catch (ParseException ex) {
                    return "[NOT SET]";
                }
            }
            else
                return "[NOT SET]";
        } else {
            return "" + new Timestamp((Long) getCeiling());
        }
    }


    /**
        Sets the value of the ceiling property.

        @param aCeiling the new value of the ceiling property
     */
    public void setCeiling(MemberType aCeiling) {
        ceiling = aCeiling;
    }

    /**
        @param inCandidate
        @return boolean
        @roseuid 424ABD0A01BD
     */
    public boolean isEligible(Object inCandidate) {
        if (inCandidate == null) {
            return false;
        }
        Comparable aoi = null;
        try {
            aoi = getAttributeOfInterest(inCandidate);
        } catch (AggregationException ex) {
            //this exception was already logged by the base class
            return false;
        }
        Comparable low = (Comparable) getFloor();
        if (low == null) {
            return false;
        }
        Comparable high = (Comparable) getCeiling();
        if (high == null) {
            return false;
        }

        if (floorIsInclusive && low.compareTo(aoi) > 0) {
            return false;
        }
        if (!floorIsInclusive && low.compareTo(aoi) >= 0) {
            return false;
        }
        if (ceilingIsInclusive && high.compareTo(aoi) < 0) {
            return false;
        }
        if (!ceilingIsInclusive && high.compareTo(aoi) <= 0) {
            return false;
        }
        return true;
    }

    /**
     * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
     * emanating from it.
     * @return Range
     */
    public Range deepCopy() throws AggregationException {
        BinaryRange<MemberType> copy = new BinaryRange<MemberType>();
        copy.setAggregationName(this.getAggregationName());
        copy.setAggregationType(this.getAggregationType());
        Range srcBlueprint = getBlueprint();
        copy.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        copy.setCeiling(this.getCeiling());
        copy.setCeilingIsInclusive(this.getCeilingIsInclusive());
        copy.setDeclaredNodeStructure(this.getDeclaredNodeStructure());
        copy.setDescription(this.getDescription());
        copy.setDnNode(this.getDnNode());
        copy.setFloor(this.getFloor());
        copy.setFloorIsInclusive(this.getFloorIsInclusive());
        copy.setGetter(this.getGetter());
        copy.setInquiryMethod(this.getInquiryMethod());
        copy.setPassivator(this.getPassivator());
        copy.setPersistenceHandle(this.getPersistenceHandle());
        copy.setSerializable(this.isSerializable());
        copy.setSubject(this.getSubject());
        copy.setSubjectClass(this.getSubjectClass());
        copy.setContext(this.getContext());

        //The following is predicated on the fact that RangeSurgeons are completely
        //stateless and can therefore be shared among many Range clients
        copy.setRangeSurgeon(this.getRangeSurgeon());


        //now deepcopy the members individually to avoid pointing to an existing
        //object and subsequently accidentally modifying it
        List srcMembers = this.getMembers();
        List copyList = null;
        if (null == srcMembers) {
            copy.setMembers((List)null);
        } else {
        	if ( srcMembers instanceof Range ) {
						copyList = ((Range)srcMembers).deepCopy();
          } else {
            Class srcMembersClass = srcMembers.getClass();
            copyList = new java.util.ArrayList();
            for (Iterator memberIter = srcMembers.iterator();
                                       memberIter.hasNext(); ) {
                Object currSrcMember = memberIter.next();
                if (currSrcMember instanceof Range) {
                    Range srcRange = (Range) currSrcMember;
                    Range copyRange = srcRange.deepCopy();
                    copyRange.setContext(copy);
                    copyList.add(copyRange);
                }
            }

          }
					copy.setMembers(copyList);
        }
        return copy;
    }

    /* -- INFO: COMMENTED OUT TO USE Range.merge() -- todd
    public Range merge(Range otherRange) throws AggregationException {
        if(null == otherRange)
            return this;
        if(!(otherRange instanceof BinaryRange)){
            String warn = "expected a BinaryRange, received a '" + otherRange.getClass().getName() + "'";
            log.warning(warn);
            return this;
        }
        BinaryRange<MemberType> merged = (BinaryRange<MemberType>)this.deepCopy();
        List otherMembers = otherRange.getMembers();
        //ArrayList.addAll throws an NPE when given an empty List(!), therefore don't call it if
        //otherMembers.size() == 0
        if(otherMembers != null && otherMembers.size() > 0){
            try{
                merged.getMembers().addAll(otherMembers);
            }catch(NullPointerException npe){
                String info = "NPE from ArrayList.add, ignoring";
                log.log(Level.INFO, info);
            }
        }else{
            log.log(Level.INFO,"BinaryRange.merge:  otherMembers's size == 0, NOT MERGING IT");
        }
        return merged;
    }
    */

    public <T> T shallowCopy(Class<T> factory, T source) throws AggregationException{
        T dest = emptyCopy(factory,source);
        if(dest instanceof BinaryRange){
            BinaryRange destRange = (BinaryRange)dest;
            BinaryRange sourceRange = (BinaryRange)source;

            destRange.setAggregationName(sourceRange.getAggregationName());
            destRange.setAggregationType(sourceRange.getAggregationType());
            Range srcBlueprint = getBlueprint();
            destRange.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
            destRange.setDeclaredNodeStructure(sourceRange.getDeclaredNodeStructure());
            destRange.setDescription(sourceRange.getDescription());
            destRange.setDnNode(sourceRange.getDnNode());
            destRange.setGetter(sourceRange.getGetter());
            destRange.setInquiryMethod(sourceRange.getInquiryMethod());
            destRange.setPassivator(sourceRange.getPassivator());
            destRange.setPersistenceHandle(sourceRange.getPersistenceHandle());
            destRange.setSerializable(sourceRange.isSerializable());
            destRange.setSubject(sourceRange.getSubject());
            destRange.setSubjectClass(sourceRange.getSubjectClass());
            destRange.setTransformerClass(sourceRange.getTransformerClass());
            destRange.setCeiling(sourceRange.getCeiling());
            destRange.setCeilingIsInclusive(sourceRange.getCeilingIsInclusive());
            destRange.setFloor(sourceRange.getFloor());
            destRange.setFloorIsInclusive(sourceRange.getFloorIsInclusive());
        }
        return dest;
  }
    /**
     * Am I "equal to " <code>inRHS</code>?  To be equal the following must be
     * true:<br>
     * <li>inRHS instanceof BinaryRange
     * <li>this.floor.equals(inRHS.floor)
     * <li>this.ceiling.equals(inRHS.ceiling)
     * <li>this.isFloorInclusive == inRHS.isFloorInclusive
     * <li>this.isCeilingInclusive == inRHS.isCeilingInclusive<br>
     *
     * @param inRHS Object
     * @return boolean
     */
    public boolean equals(Object inRHS) {
        if (null == inRHS) {
            return false; //I ain't null, inRHS is, ergo !=
        }
        if (inRHS instanceof BinaryRange) {
            BinaryRange rhs = (BinaryRange) inRHS;
            BinaryRange lhs = this;
            Object rhsFloor = rhs.getFloor();
            Object lhsFloor = lhs.getFloor();
            if (null == lhsFloor && null != rhsFloor) {
                return false;
            }
            if (null != lhsFloor && null == rhsFloor) {
                return false;
            }
            if (null != lhsFloor && null != rhsFloor &&
                lhsFloor.equals(rhsFloor) == false) {
                return false;
            }

            Object rhsCeiling = rhs.getCeiling();
            Object lhsCeiling = lhs.getCeiling();
            /*
            log.log(Level.INFO,
                    "Comparing [" + lhsFloor + "-" + lhsCeiling + "] to [" +
                    rhsFloor + "-" + rhsCeiling + "]");
            */
            if (null == lhsCeiling && null != rhsCeiling) {
                return false;
            }
            if (null != lhsCeiling && null == rhsCeiling) {
                return false;
            }
            if (null != lhsCeiling && null != rhsCeiling &&
                lhsCeiling.equals(rhsCeiling) == false) {
                return false;
            }

            if (null == lhsFloor && null == rhsFloor && null == lhsCeiling &&
                null == rhsCeiling) {
                return true;
            }
            return lhsFloor.equals(rhsFloor) && lhsCeiling.equals(rhsCeiling);
        } else {
            return false;
        }
    }
}
