package org.quintet.miles;

import java.util.Collection;
import java.util.Map;
import java.util.HashMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.Iterator;
import java.util.logging.Logger;
import java.util.logging.Level;

/**
 * <code>Counter</code> is a <code>Group</code> that tracks the number
 * of instances that meet its criteria; it does not store the instances themselves.
 * @author Peter Thaggard
 * @version 1.0
 */
public class Counter extends Group {

    private static final Logger log = Logger.getLogger("org.quintet.miles");

    public Counter() {
        super();
    }

    /**"Find" the member designated by <code>dnString</code>.  Because this is
     * a <i>counter</i> node, it doesn't have any actual members -- it is its own
     * one-and-only member.  Therefore this method merely returns this node back
     * to the caller
     * @param Object dnString - the DN that identifies the member being sought (in this case it is ignored)
     * @return Range -- this <code>Counter</code> is returned
     * @exception AggregationException
     */
    public Range findMember(Object dnString) throws AggregationException {
        return this;
    }

    /**
     * Return <code>true</code> is this <code>Range</code> is considered to be
     * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
     * method is intentionally declared abstract at the <code>Range</code> level
     * in order to force every <code>Range</code> subclass implementor to carefully
     * consider the nature of his class and whether it can truly be considered a
     * leaf node.
     * @return boolean
     */
  public  boolean isLeafNode(){
      //Counter doesn't contain any subordinates, therefore it is a leaf node
      return true;
  }

    /**
       If dd <code>inCandidate</code> to this <code>Range</code>
       @param inCandidate - The object being considered for membership in this
       <code>Range</code>
       @return boolean
       @roseuid 42517B3C01FE
     */
    public boolean add(Object inCandidate) {
        PrimaryKey candidatePK = null;
        if(inCandidate instanceof PrimaryKey){
            candidatePK = (PrimaryKey) inCandidate;
        }else{
            try {
                candidatePK = new PrimaryKey(inCandidate, getPrimaryKeyColumns());
            } catch (AggregationException ex) {
                String err = "AggregationException when adding object to Group";
                log.log(Level.SEVERE, err,ex);
                return false;
            }
        }
        //see if I already have a structure to house this particular value
        Object rangeObj = getGroups().get(candidatePK);

        Long countForCandidate = (Long) rangeObj;
        if (countForCandidate == null) {
            //first time this primary key has been seen
            countForCandidate = new Long(0);
            getGroups().put(candidatePK, countForCandidate);
        }
        //now I definitely have a structure for this candidate PK
        countForCandidate = new Long(countForCandidate.longValue() + 1);
        getGroups().put(candidatePK, countForCandidate);
        return true;
    }

    /**
       Clone this <code>Counter</code> via "deep copy"
       @return java.lang.Object
       @roseuid 4252B89A0034
     */
    public Object clone() {
        Object base = super.clone();
        Counter clonedCounter = (Counter) base;
        //zero out the groups member, because it probably has counts for some other
        //ancestral aggregation
        clonedCounter.setGroups(new java.util.HashMap());
        return clonedCounter;

    }

    /**
     * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
     * emanating from it.
     * @return Range
     */
    public Range deepCopy() throws AggregationException {
        Counter copy = new Counter();
        copy.setAggregationName(this.getAggregationName());
        copy.setAggregationType(this.getAggregationType());
        Range srcBlueprint = getBlueprint();
        copy.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        copy.setDeclaredNodeStructure(this.getDeclaredNodeStructure());
        copy.setDescription(this.getDescription());
        copy.setDnNode(this.getDnNode());
        copy.setGetter(this.getGetter());
        copy.setInquiryMethod(this.getInquiryMethod());
        copy.setPassivator(this.getPassivator());
        copy.setPersistenceHandle(this.getPersistenceHandle());
        copy.setSerializable(this.isSerializable());
        copy.setSubject(this.getSubject());
        copy.setSubjectClass(this.getSubjectClass());
        copy.setContext(this.getContext());
        //The following is predicated on the fact that RangeSurgeons are completely
        //stateless and can therefore be shared among many Range clients
        copy.setRangeSurgeon(this.getRangeSurgeon());

        //now deepcopy the members individually to avoid pointing to an existing
        //object and subsequently accidentally modifying it
        List srcMembers = this.getMembers();
        if (null == srcMembers) {
            copy.setMembers((List)null);
        } else {
            Class srcMembersClass = srcMembers.getClass();
            List copyList = (List) shallowCopy(srcMembersClass, srcMembers);
            for (Iterator memberIter = srcMembers.iterator();
                                       memberIter.hasNext(); ) {
                Object currSrcMember = memberIter.next();
                if (currSrcMember instanceof Range) {
                    Range srcRange = (Range) currSrcMember;
                    Range copyRange = srcRange.deepCopy();
                    copyRange.setContext(copy);
                    copyList.add(copyRange);
                }
            }
            copy.setMembers(copyList);
        }

        //copy all the primary key columns
        Collection srcPKColumns = getPrimaryKeyColumns();
        if (null == srcPKColumns) {
            copy.setPrimaryKeyColumns((Collection)null);
        } else {
            Collection copyPKColumns = new java.util.ArrayList(srcPKColumns.
                    size());
            for (Iterator srcPKIter = srcPKColumns.iterator();
                                      srcPKIter.hasNext(); ) {
                String srcPK = (String) srcPKIter.next();
                copyPKColumns.add(srcPK);
            }
            copy.setPrimaryKeyColumns(copyPKColumns);
        }
        //now copy the groups individually to avoid etc etc etc
        Map<PrimaryKey, ? extends Range> srcGroups = getGroups();
        if (null == srcGroups) {
            copy.setGroups((Map)null);
        } else {
            Map<PrimaryKey, Range>
                    copyGroups = new HashMap<PrimaryKey, Range>(srcGroups.size());
            for (PrimaryKey srcPK : srcGroups.keySet()) {
                PrimaryKey copyPK = (PrimaryKey) srcPK.clone();
                Range srcGroup = srcGroups.get(srcPK);
                Range copyGroup = srcGroup.deepCopy();
                copyGroup.setContext(copy);
                copyGroups.put(copyPK, copyGroup);
            }
            copy.setGroups(copyGroups);
        }
        return copy;
    }

    public boolean equals(Object inRHS) {
        return false;
    }
}
