package org.quintet.miles;

import java.util.Map;
import java.util.HashMap;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.util.Date;

import java.lang.reflect.Method;
import java.lang.NoSuchMethodException;
import java.lang.reflect.*;

import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.StringTokenizer;
import java.io.Serializable;




/**
 * A <code>DNNode</code> houses one node (i.e., a <i>RHS</i>=<i>LHS</i>) pair
 * and any additional information about that node, such as whether either side
 * of the <code>DNNode</code> represents a wildcard (e.g., "org*=AP|AR")
 * <p>Copyright: Copyright (c) 2005</p>
 *
 * <p>Company: 6th Sense Analytics</p>
 *
 * @author Peter Thaggard
 * @version 1.0
 */
public class DNNode<Left extends Serializable,Right extends Serializable> extends Pair<Left, Right> implements Serializable,Cloneable{

  private static final Logger log = Logger.getLogger("org.quintet.miles");

  /**
   * The <code>aggregationFunctionType</code> attribute houses the
   * <code>AggregationFunction</code> (e.g., 'SUM','ACCUMULATE')
   * that the user specified on the LHS of the <code>DN</code>.
   */
  private AggregationFunction aggregationFunctionType = null;

  private boolean lhsIsRollup = false;

  /**
   * <code>originalLHS</code> houses the original LHS token, that is, the
   * token as it existed before being converted from a "roll up token" to
   * a "plain old token".  If LHS is not a "roll up token" then
   * <code>originalLHS</code> is null
   */
  private String originalLHS = null;

  /**
   * <code>originalRHS</code> houses the original (pre-expansion) token, that is,
   * the token as it existed before being expanded; for example, <code>week=02/01/2006</code>
   * gets expanded to <code>day=01|02|03|04|05</code>, but <code>originalRHS</code> would
   * retain the "02/01/2006" value
   */
  private String originalRHS = null;


  /**
   * This no-arg ctor is for the exclusive use of our {@link org.quintet.miles.DNNode#clone()}
   */
  private DNNode(){
    super();
  }


  public DNNode(Left inLHS, Right inRHS) {
    super(inLHS, inRHS);
    //if my LHS is a rollup token (e.g., week, month) then I need to:
    //    convert the rollup LHS to its unrolled-up counterpart (e.g., convert "week" to "day")
    //    expand the RHS according to the LHS (e.g., inLHS="week",inRHS="02/04/2006"
    //    causes RHS to become "02/04/2006|02/05/2006|02/06/2006..."
    if(inLHS instanceof String && inRHS instanceof String){
      String lhsString = (String)inLHS;
      String rhsString = (String)inRHS;
      if (isRollupToken(lhsString)){
        setAggregationFunctionType(AggregationFunction.lookup(lhsString));
        setLhsIsRollup(true);
        setOriginalLHS(lhsString);
        lhsString = lhsString.substring(0,lhsString.length()-1);
        LHS = (Left)lhsString;

      }
    }
  }

  public void setLhsIsRollup(boolean lhsIsRollup) {
    this.lhsIsRollup = lhsIsRollup;
  }

  public void setOriginalLHS(String originalLHS) {
    this.originalLHS = originalLHS;
  }

  public String getOriginalLHS(){
    return originalLHS;
  }

  public String getOriginalRHS(){
    return originalRHS;
  }

  /**
   * Retrieve the <code>AggregationFunction</code>, if any, that was
   * specified on this <code>DNNode</code>'s LHS.
   * The <code>aggregationFunctionType</code> attribute houses the
   * <code>AggregationFunction</code> (e.g., 'SUM','ACCUMULATE')
   * that the user specified on the LHS of the <code>DN</code>.
   * @return AggregationFunction
   */
  public AggregationFunction getAggregationFunctionType() {
    return aggregationFunctionType;
  }

  public void setOriginalRHS(String newValue){
    originalRHS = newValue;
  }

  /**
   * Designate <code>inAggregationFunctionType</code> to be this
   * <code>DNNode</code>'s <code>AggregationFunction</code>.
   * The <code>aggregationFunctionType</code> attribute houses the
   * <code>AggregationFunction</code> (e.g., 'SUM','ACCUMULATE')
   * that the user specified on the LHS of the <code>DN</code>.
   * @param inAggregationFunctionType AggregationFunction
   */
  public void setAggregationFunctionType(AggregationFunction inAggregationFunctionType) {
    this.aggregationFunctionType = inAggregationFunctionType;
  }

  public boolean isLhsRollup() {
    return lhsIsRollup;
  }

  /**
   * Return a <code>boolean</code> indicating if <code>subject</code> appears
   * to denote a <i>wildcard</i>, where a wildcard is anything that does not
   * directly correlate to an <code><i>attributeOfInterest</i></code> in the
   * current aggregation service.  Examples of implicit wildcards are:<br>
   * <li>* at the end of an attribute name, for example <i>org*</i>
   * <li>Parentheses at the end of an attribute name, for example <i>day()</i>
   * <li>The word "week" (implies that the attribute of interest is "day")
   * <li>The word "month" (implies that the attribute of interest is "day")
   * <br>
   * @param subject Object
   * @return boolean
   */
  private static boolean isRollupToken(String subject){
    if(null == subject)
      return false;
    if(!(subject instanceof String)){
      return false;
    }

    String subjectString = (String)subject;
    if(subjectString.endsWith("*") ||  /* 'sum' */
       subjectString.endsWith("+") ||  /* 'cumulative' */
       subjectString.endsWith("/") ||  /* 'avg' */
       subjectString.endsWith(">") ||  /* 'max' */
       subjectString.endsWith("<")     /* 'min' */
    )
      return true;
    else
      return false;
  }

  private static Map<String,String> rollupTokens = new java.util.HashMap<String,String>();

  private static Map<String,Method> rollupTokenExpanders = new java.util.HashMap<String, Method>();
  static{
    rollupTokens.put("week", "day");
    rollupTokens.put("month", "day");

    Class _this = null;
    try {
      _this = Class.forName("org.quintet.miles.DNNode");
    }
    catch (ClassNotFoundException CNFE) {
      String fatal =
          "Fatal Exception, I cannot find my own class (org.quintet.miles.DNNode)";
      throw new IllegalStateException(fatal);
    }

    Method man = null;
    Class[] args = new Class[] {
        java.lang.String.class};
    try {
      man = _this.getMethod("expandWeek", args);
      rollupTokenExpanders.put("week",man);

      man = _this.getMethod("expandMonth",args);
      rollupTokenExpanders.put("month",man);
    }
    catch (NoSuchMethodException nsme) {
      String fatal = "NoSuchMethodException:  " + nsme;
      throw new IllegalStateException(fatal);
    }
  }
  /**
   * If <code>token</code> is considered a "rollup token" (i.e., it is something
   * like "week" that does not have a direct correlation to an aggregation) then
   * return the "un-rolled-up" version of token; for example, if the <code>token</code>
   * is "week" then this method will return "day".
   * If <code>token</code> is not a rollup token then it is returned verbatim to
   * the caller.
   * @param token String
   * @return String
   */
  public static String translateRollupToken(String token){
    String rollUpToken = rollupTokens.get(token);
    if(null == rollUpToken)
      return token;
    else
      return rollUpToken;
  }


  public String expandRollupToken(String inRHS){
    Method expander = rollupTokenExpanders.get(getOriginalLHS());
    if(null != expander){
      try {
        String expandedToken = (String) expander.invoke(null,new Object[]{inRHS});
        return expandedToken;
      }
      catch (InvocationTargetException ex) {
        String err = "InvocationTargetException from method '" + expander.getName();
        Throwable orig = ex.getCause();
        log.log(Level.SEVERE,err,orig);
        throw new IllegalStateException(err,orig);
      }
      catch (IllegalArgumentException orig) {
        String err = "IllegalArgumentException from method '" + expander.getName();
        log.log(Level.SEVERE,err,orig);
        throw new IllegalStateException(err,orig);

      }
      catch (IllegalAccessException orig) {
        String err = "IllegalAccessException from method '" + expander.getName();
        log.log(Level.SEVERE,err,orig);
        throw new IllegalStateException(err,orig);
      }
    }else{
      return inRHS;
    }
  }

  public static synchronized String expandWeek(String baseWeek){
    //baseWeek denotes the first day of the Week that the user desires to analyze
    Date newDate = null;
    DateFormat formatter = new SimpleDateFormat("MM/dd/yyyy");
    try {
      newDate = formatter.parse(baseWeek);
    }
    catch (ParseException ex) {
      throw new IllegalStateException("ParseException",ex);
    }
    Calendar cal = new GregorianCalendar();
    cal.setTime(newDate);
    StringBuffer answer = new StringBuffer();
    for(int idx = 0; idx < 7; idx++){
      newDate = cal.getTime();
      String newDateStr = formatter.format(newDate);
      answer.append(newDateStr);
      if(idx <= 5)
        answer.append('|');
      cal.add(Calendar.DAY_OF_MONTH,1);
    }
    return answer.toString();
  }

  public static synchronized String expandMonth(String monthAndYear){
    StringTokenizer lex = new StringTokenizer(monthAndYear,"/");
    if(lex.hasMoreTokens() == false){
      throw new IllegalStateException("Syntax error on month designation '" + monthAndYear + "'");
    }
    String monthPart = lex.nextToken();

    if(lex.hasMoreTokens() == false){
      throw new IllegalStateException("Syntax error on month designation '" + monthAndYear + "'");
    }

    String yearPart = lex.nextToken();

    String baseMonth = monthPart + "/01/" + yearPart;

    //baseWeek denotes the first day of the Week that the user desires to analyze
    Date newDate = null;
    DateFormat formatter = new SimpleDateFormat("MM/dd/yyyy");
    try {
      newDate = formatter.parse(baseMonth);
    }
    catch (ParseException ex) {
      throw new IllegalStateException("ParseException",ex);
    }
    Calendar cal = new GregorianCalendar();
    cal.setTime(newDate);
    int lastDay = cal.getActualMaximum(Calendar.DAY_OF_MONTH);
    StringBuffer answer = new StringBuffer();
    for(int idx = 1; idx <= lastDay ; idx++){
      newDate = cal.getTime();
      String newDateStr = formatter.format(newDate);
      answer.append(newDateStr);
      if(idx <= lastDay-1)
        answer.append('|');
      cal.add(Calendar.DAY_OF_MONTH,1);
    }
    return answer.toString();
  }

  public Object clone(){
    DNNode cloned = new DNNode();
    cloned.setOriginalLHS(this.getOriginalLHS());
    cloned.setLhsIsRollup(this.isLhsRollup());
    cloned.setOriginalRHS(this.getOriginalRHS());
    cloned.setAggregationFunctionType(this.getAggregationFunctionType());
    cloned.LHS = this.LHS;
    cloned.RHS = this.RHS;
    return cloned;
  }

  public enum AggregationFunction{
    SUM("*","sum",true),
    ACCUMULATE("+","accumulate",true),
    AVERAGE("/","average",true);

    private static String functionSymbolSet = "*+/";

    private String functionSymbol = null;
    private Method functionImpl = null;
    private boolean applyRecursively = false;

    AggregationFunction(String symbol, String functionImplName,boolean inApplyRecursively){
      Class rangeClass = org.quintet.miles.Range.class;
      try {
        functionImpl = rangeClass.getMethod(functionImplName, (Class[])null);
      }
      catch (SecurityException ex) {
        String fatal = "Internal Error --SecurityException when fetching 'Range." + functionImplName + "'";
        log.log(Level.SEVERE,fatal,ex);
        throw new IllegalStateException(fatal,ex);
      }
      catch (NoSuchMethodException ex) {
        String fatal = "Internal Error -- unable to find method 'Range." + functionImplName + "'";
        log.log(Level.SEVERE,fatal,ex);
        throw new IllegalStateException(fatal,ex);
      }
      functionSymbol = symbol;
      applyRecursively = inApplyRecursively;
    }

    public boolean isAppliedRecursively(){
      return applyRecursively;
    }

    public static AggregationFunction lookup(String LHS){
      if(null == LHS)
        return null;
      String symbol = LHS.substring(LHS.length()-1);

      int idx = functionSymbolSet.indexOf(symbol);
      return (idx >= 0 ? AggregationFunction.values()[idx] : null);
    }

    public Method getAggregationFunctionImpl(){
      return functionImpl;
    }
  };
}
