package org.quintet.miles;

import java.util.regex.Pattern;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import java.util.logging.Logger;
import java.util.logging.Level;

/**
 * A <code>DynamicRange</code> is a <code>Range</code> whose "slots" grow dynamically
 * to accommodate new members.
 *
 * <p>Copyright: Copyright (c) 2005</p>
 *
 * <p>Company: 6th Sense Analytics</p>
 *
 * @author Peter Thaggard
 * @version 1.0
 */
public class DynamicRange extends Range {

    private static final Logger log = Logger.getLogger("org.quintet.miles");

    /**
     * If <code>overriddenAdd</code> is not null then it becomes this
     * <code>DynamicRange</code>'s <code>add</code>implementation.  If
     * <code>overriddenAdd</code> is null then this <code>DynamicRange</code>
     * delegates the <code>add</code> call back to <code>Range.add</code>
     */
    private transient Method overriddenAdd = null;
    private String overriddenAddName = null;

    private void writeObject(java.io.ObjectOutputStream out) throws java.io.IOException{
      overriddenAddName = (null == overriddenAdd ? null : overriddenAdd.getName());
      out.defaultWriteObject();
    }

    /**
     * The name of the method that implements "add by rolling up" behavior
     */
    private static final String ROLL_UP_ADD_METHOD = "addByRollingUp";

    private static Class[] objectClassArray = null;

    static {
        Class objectsClass = null;
        try {
            objectsClass = Class.forName("java.lang.Object");
        } catch (ClassNotFoundException cnfe) {
            System.err.println("fatal error, cannot find Object's class(!)");
            throw new IllegalStateException(
                    "fatal error, cannot find Object's class(!)");
        }
        objectClassArray = new Class[] {objectsClass};
    }


    /**
     * Return <code>true</code> is this <code>Range</code> is considered to be
     * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
     * method is intentionally declared abstract at the <code>Range</code> level
     * in order to force every <code>Range</code> subclass implementor to carefully
     * consider the nature of his class and whether it can truly be considered a
     * leaf node.
     * @return boolean
     */
  public  boolean isLeafNode(){
      //if my members != null AND my member's first member is
      //a List then I am NOT a leaf node
      List oneLevelDown = getMembers();
      if(null == oneLevelDown || 0 == oneLevelDown.size())
        return true;  //I have no underlings, therefore I am a leaf
      Object firstUnderling = oneLevelDown.get(0);
      if(firstUnderling instanceof List)
        return false; //I have underlings, and they are Lists/Ranges, therefore I am not a leaf node
      else
        return true; //I have underlings, but they are not Lists/Ranges, therefore I am a leaf node;
  }

    /**
     * Create a new <code>DynamicRange</code>.  If <code>toSupport</code> implies
     * that "rolling up", or "wildcarding" needs to be performed then the
     * new <code>DynamicRange</code>'s <code>overriddenAdd</code> method is
     * set to point to a "rollup" <code>add</code> implementation.
     *
     * @param toSupport DNNode
     * @return DynamicRange
     * @throws AggregationException
     */
    public DynamicRange create(DNNode toSupport) throws AggregationException {
        DynamicRange answer = new DynamicRange();
        if (toSupport.isLhsRollup()) {
            Class _this = null;
            try {
                _this = Class.forName("org.quintet.miles.DynamicRange");
            } catch (ClassNotFoundException ex) {
                String fatal = "I am unable to find my own class (org.quintet.miles.DynamicRange) via Class.forName(!)";
                log.log(Level.SEVERE,fatal, ex);
                throw new AggregationException(fatal, ex);
            }

            try {
                answer.overriddenAdd = _this.getMethod(ROLL_UP_ADD_METHOD,
                        objectClassArray);
            } catch (SecurityException ex1) {
                String err =
                        "SecurityException when trying to get Method 'DynamicRange." +
                        ROLL_UP_ADD_METHOD + "'";
                log.log(Level.SEVERE,err);
                throw new AggregationException(err, ex1);
            } catch (NoSuchMethodException ex1) {
                String err =
                        "NoSuchMethodException when trying to getMethod 'DynamicRange." +
                        ROLL_UP_ADD_METHOD + "'";
                log.log(Level.SEVERE,err);
                throw new AggregationException(err, ex1);
            }
        }
        return answer;
    }

    /**
     * Construct a new <code>DynamicRange</code>.  The <code>DynamicRange</code>'s
     * dynamic behavior is provided by <code>java.util.ArrayList</code>.
     * @see java.util.ArrayList;
     */
    public DynamicRange() {
        super();
        setMembers(new java.util.concurrent.CopyOnWriteArrayList());
    }

    /**
     * Return <code>true</code> if <code>inCandidate</code> can be considered a
     * member of this <code>DynamicRange</code>.  Because a <code>DynamicRange</code>
     * grows to accommodate all members, this method always returns <code>true</code>
     *
     * @param inCandidate - The <code>Object</code> being considered for
     *   membership in this <code>DynamicRange</code>
     * @return boolean
     * @todo Implement this org.quintet.miles.Range method
     */
    public boolean isEligible(Object inCandidate) {
        return true;
    }


    public Range findMembers(Object attributeValue) throws AggregationException {
        if (null == getMembers()) {
            return null;
        }
        if (!(attributeValue instanceof String)) {
            String err = "Cannot process DN nodes of type '" +
                         attributeValue.getClass().getName();
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }
        String whomString = (String) attributeValue;

        DynamicRange answer = new DynamicRange();
        answer.setAggregationName(this.getAggregationName());
        Range srcBlueprint = getBlueprint();
        answer.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        answer.setDescription(this.getDescription());
        answer.setGetter(this.getGetter());
        answer.setMembers(new java.util.ArrayList());
        answer.setSubject(this.getSubject());
        answer.setSubjectClass(this.getSubjectClass());
        answer.setAggregationType(getAggregationType());
        answer.setDeclaredNodeStructure(getDeclaredNodeStructure());
        answer.setTransformerClass(getTransformerClass());
        Pattern pattern = Pattern.compile(whomString);
        for (Iterator memberIter = getMembers().iterator(); memberIter.hasNext(); ) {
            Object currentMember = memberIter.next();
            //attributeValue can be an exact value or it can be a regexpr
            //of course, all exact values are also regexprs....
            //if the current member is a String then this findMember method works
            //fine; however, it will NOT work for any other sort of member object.
            //Therefore Range objects need a "matches" method that can be used
            //instead of equals
            if (currentMember instanceof String) {
                if (currentMember.equals(attributeValue)) {
                    answer.add(currentMember);
                }
            } else if (currentMember instanceof Range) {
                //currMember, being a Range, has a matches method, which we can use to
                //determine if currMember matches the current DN
                Range currMemberRange = (Range) currentMember;
                if (currMemberRange.matches(pattern)) {
                    answer.add(currMemberRange.deepCopy());
                }
            }
        }
        return answer;
    }

    /**
     * Use <code>dn</code> to traverse the <code>Range</code> rooted at this point.
     * Return the <code>Range</code> residing at the spot designated by <code>dn</code>
     * This is the most basic <code>findDN</code> implementation; several <code>Range</code>
     * subclasses override this method with significantly more sophisticated traversal
     * algorithms.  Speaking of algorithms, here is the only employed herein:<br>
     * <li>pop the top DNNode off <code>dn</code>
     * <li>walk down the tree rooted at this <code>Aggregation</code> until the tree node's <code>subjectClass</code> matches <code>topNode.LHS</code>
     * <li>"prune" the <code>Range</code>, thereby producing a <code>foundMembers</code> <code>Aggregation</code> containing values that meet
     * <code>topNode</code>'s criteria
     * <li>Remove the top <code>DNNode</code> from <code>dn</code>
     * <li>recursively process each <code>foundMember</code> by calling each ones <code>findDN</code> method, passing down the now-elided <code>dn</code>.
     * The value returned from this recursive call replaces the current <code>foundMember</code>'s <code>members</code> attribute
     * <br>
     * @param dn DN of the form <code><i>subjectClass=subjectValue[,subectClass=subjectValue...]
     * @return Range that most closely matches <code>dn</code>
     * @throws AggregationException
     */
    public Range findDN(DN dn) throws AggregationException {
        //1.  check to see if this level of the tree matches the
        //    subject mentioned in the current top DN Node;
        //    for example, if dn="ou=acme,uid=joe" then check to see
        //    if this.subjectClass == ou.
        //    If they don't match then  walk down the subtree rooted here until
        //    a match is found and then restart the findDN process
        String subjectClass = getSubjectClass();
        if (null == subjectClass || subjectClass.length() == 0) {
            String err = "subject class is not set, cannot process DN '" + dn +
                         "'";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        DNNode<String, String> topNode = dn.peek();
        if (subjectClass.equals(topNode.LHS)) {
            //1.1  Yes, this level of the tree matches the subject class stipulated
            //     by the topNode.  Now "prune" this level of the tree (no
            //     sublevel pruning done yet), yielding a new Aggregation containing
            //     only those members of this Aggregation that meet topNode's criteria
            Range foundMembers = findMembers(topNode.RHS);
            if (null == foundMembers) {
                //There are no objects in this tree node that satisfy topNode
                return null;
            } else {
                //topNode, the current DN, just produced the results found in
                //foundMembers.  Record this DN-Range correlation by imprinting topNode
                //onto foundMembers
                foundMembers.setDnNode(topNode);

                //the node at which we just peek()ed is still lingering on the DN.  We
                //must remove it now...
                String logMsg = "DN is " + dn.toString() +
                                ", about to remove top element";
                //log.info(logMsg);
                dn.poll();
                logMsg = "DN is now " + dn.toString();
                //log.info(logMsg);

                //we found some qualifying members at this level and placed them into
                //the "pruned" Aggregation named "foundMembers".    If there are remaining nodes on the DN
                //then delegate the remaining search to the found nodes
                if (dn.peek() == (DNNode<String, String>)null) { //check to see if there is more on the DN...
                    //nothing left on the search spec, so we have found the thing being sought (I think....);
                    return foundMembers;
                } else {
                    //now apply the newly-pruned DN to each found member and allow each
                    //such subtree to be visited by the pruned DN
                    for (Iterator foundMemberIter = foundMembers.iterator();
                            foundMemberIter.hasNext(); ) {
                        DN dnForSubtree = (DN) dn.clone();
                        Object currFoundMember = foundMemberIter.next();
                        //this Aggregation's members aren't necessarily Ranges; they
                        //could be plain old objects.  Only recurse into subordinate Ranges
                        if (currFoundMember instanceof Range) {
                            //prune the subtree anchored at currFoundMember
                            Range subtree = (Range) currFoundMember;
                            Range newSubtreeMembers = subtree.findDN(
                                    dnForSubtree);
                            subtree.setMembers(newSubtreeMembers);
                        }
                    }
                    return foundMembers;
                }
            }
        } else {
            //the current topNode is not matched to our subjectClass, so try to match
            //it to one of our members
            //the following code sickens me, but it does work, and it is the most
            //expeditious route to implementing a proper findDN method.
            //If the current Range's members() are a Group then process the group
            //directly, otherwise process each member of members()
            if (getMembers() instanceof Range) {
                Range memberGroup = (Range) getMembers();
                Range nextFound = memberGroup.findDN(dn);
                return nextFound;
            } else {
                //we may need to return several of our member nodes, bundled together
                //as a new DynamicRange, so build the "answer" container now
                Aggregation answer = new Aggregation();
                answer.setAggregationName(getAggregationName());
                answer.setAggregationType(getAggregationType());
                Range srcBlueprint = getBlueprint();
                answer.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
                answer.setDeclaredNodeStructure(getDeclaredNodeStructure());
                answer.setDescription(getDescription());
                answer.setGetter(getGetter());
                answer.setInquiryMethod(getInquiryMethod());
                answer.setSubject(getSubject());
                answer.setSubjectClass("[none]");
                answer.setDnNode(topNode);
                answer.setMembers(new java.util.ArrayList());
                for (Iterator memberIter = getMembers().iterator();
                                           memberIter.hasNext(); ) {
                    Object currMember = memberIter.next();
                    if (currMember instanceof Range) {
                        Range possibleRange = (Range) currMember;
                        Range nextFound = possibleRange.findDN(dn);
                        //we have to push the current DN node back onto the DN
                        //for use by the rest of the passes through this loop
                        dn.offer(topNode);
                        if (nextFound != null) {
                            answer.add(nextFound);
                        }
                    }
                }
                dn.poll();
                return answer;
            }
        }
    }

    /**
      If dd <code>inCandidate</code> to this <code>Range</code>
      @param inCandidate - The object being considered for membership in this
      <code>Range</code>
      @return boolean
      @roseuid 42498CAF03D2
     */
    public boolean add(Object inCandidate) {
        //only add inCandidate if it is eligible
        //eligibility is determined by inCandidate's attribute of interest
        if (isEligible(inCandidate)) {
            if (overriddenAdd != null) {
                try {
                    overriddenAdd.invoke(this, inCandidate);
                    return true;
                } catch (InvocationTargetException ex) {
                    String err =
                            "InvocationTargetException from DynamicRange.overriddenAdd";
                    Throwable cause = ex.getCause();
                    log.log(Level.SEVERE,err, cause);
                    throw new IllegalStateException(err, cause);
                } catch (IllegalArgumentException ex) {
                    String err =
                            "IllegalArgumentException from DynamicRange.overriddenAdd";
                    log.log(Level.SEVERE,err, ex);
                    throw new IllegalStateException(err, ex);
                } catch (IllegalAccessException ex) {
                    String err =
                            "IllegalAccessException from DynamicRange.overriddenAdd";
                    log.log(Level.SEVERE,err, ex);
                    throw new IllegalStateException(err, ex);
                }
            } else {
                List subordinates = getMembers();
                subordinates.add(inCandidate);
                if(subordinates instanceof Range){
                    ((Range)subordinates).setContext(this);
                }
                return true;
            }
        } else {
            return false;
        }
    }

    public void add(int _int, Object object) {
      List mems = getMembers();
      if(mems != null)
        mems.add(_int, object);
    }

    /**
     * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
     * emanating from it.
     * @return Range
     */
    public Range deepCopy() throws AggregationException {
        DynamicRange copy = new DynamicRange();
        copy.setAggregationName(this.getAggregationName());
        copy.setAggregationType(this.getAggregationType());
        Range srcBlueprint = getBlueprint();
        copy.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        copy.setDeclaredNodeStructure(this.getDeclaredNodeStructure());
        copy.setDescription(this.getDescription());
        copy.setDnNode(this.getDnNode());
        copy.setGetter(this.getGetter());
        copy.setInquiryMethod(this.getInquiryMethod());
        copy.setPassivator(this.getPassivator());
        copy.setPersistenceHandle(this.getPersistenceHandle());
        copy.setSerializable(this.isSerializable());
        copy.setSubject(this.getSubject());
        copy.setSubjectClass(this.getSubjectClass());
        copy.setContext(this.getContext());


        //The following is predicated on the fact that RangeSurgeons are completely
        //stateless and can therefore be shared among many Range clients
        copy.setRangeSurgeon(this.getRangeSurgeon());

        //now deepcopy the members individually to avoid pointing to an existing
        //object and subsequently accidentally modifying it
        List srcMembers = this.getMembers();
        if (null == srcMembers) {
            copy.setMembers((List)null);
        } else {
            Class srcMembersClass = srcMembers.getClass();
            List copyList = (List) shallowCopy(srcMembersClass, srcMembers);
            for (Iterator memberIter = srcMembers.iterator();
                                       memberIter.hasNext(); ) {
                Object currSrcMember = memberIter.next();
                if (currSrcMember instanceof Range) {
                    Range srcRange = (Range) currSrcMember;
                    Range copyRange = srcRange.deepCopy();
                    copyRange.setContext(copy);
                    copyList.add(copyRange);
                }
            }
            copy.setMembers(copyList);
        }
        return copy;
    }

    /**
     * Am I "equal to" some other <code>DynamicAggregation</code>?
     * The answer <b>is always no</b>
     * @param inRHS Object
     * @return boolean
     */
    public boolean equals(Object inRHS) {
        return false;
    }
}
