package org.quintet.miles;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

/**
 An <code>EquidistantDoubleAggregation</code> is an <code>Aggregation</code>
 comprised of <code>numberRanges</code> subordinate "slots".  The subordinate
 <code>Range</code>s are all <code>UnaryRange</code> instances.  The first
 subordinate's ceiling is <code>first</code>.  Each successive subordinate's
 ceiling is <code>incrementSize</code> larger than the preceding one.
 */
public class EquidistantDoubleAggregation extends Aggregation {

    /**
        The number of subordinate <code>Range</code>s to append to this
        <code>EquidistantDoubleAggregation</code>
     */
    private Integer numberRanges = null;

    /**
        The first slot's ceiling
     */
    private Double first = null;

    /**
     The difference between slot <i>n</i>'s ceiling and slot <i>n+1</i>'s ceiling
     */
    private Double increment = null;

    /**
     * Return <code>true</code> is this <code>Range</code> is considered to be
     * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
     * method is intentionally declared abstract at the <code>Range</code> level
     * in order to force every <code>Range</code> subclass implementor to carefully
     * consider the nature of his class and whether it can truly be considered a
     * leaf node.
     * @return boolean
     */
  public  boolean isLeafNode(){
      //if my members != null AND it is a Range then I am not a leaf node
      if(null != getMembers() && getMembers() instanceof Range)
          return true;
      else
          return false;
  }

    /**
        @roseuid 42558F040261
     */
    public EquidistantDoubleAggregation() {

    }

    /**
        Access method for the numberRanges property.

        @return   the current value of the numberRanges property
     */
    public Integer getNumberRanges() {
        return numberRanges;
    }

    /**
        Sets the value of the numberRanges property.

        @param aNumberRanges the new value of the numberRanges property
     */
    public void setNumberRanges(Integer aNumberRanges) {
        numberRanges = aNumberRanges;
    }

    /**
        Access method for the first property.

        @return   the current value of the first property
     */
    public Double getFirst() {
        return first;
    }

    /**
        Sets the value of the first property.

        @param aFirst the new value of the first property
     */
    public void setFirst(Double aFirst) {
        first = aFirst;
    }

    /**
        Access method for the increment property.

        @return   the current value of the increment property
     */
    public Double getIncrement() {
        return increment;
    }

    /**
        Sets the value of the increment property.

        @param aIncrement the new value of the increment property
     */
    public void setIncrement(Double aIncrement) {
        increment = aIncrement;
    }

    /**
     Construct my <code>members</code> if necessary (it's only necessary the first
     time this method is called) and then return my <code>members</code> to the
        caller.
        @return java.util.Collection
        @roseuid 42558FA902C8
     */
    public List getMembers() {
        List members = super.getMembers();
        if (members == null || members.size() == 0) {
            members = new java.util.concurrent.CopyOnWriteArrayList();
            double currFloor = getFirst().doubleValue();
            double increment = getIncrement().doubleValue();
            for (int idx = 0; idx < getNumberRanges().intValue(); idx++) {
                UnaryRange newMember = new UnaryRange();
                newMember.setSubject(getSubject());
                newMember.setGetter(getGetter());
                newMember.setFloor(new Double(currFloor));
                newMember.setDescription((String)null);
                currFloor += increment;
                members.add(newMember);
            }
        }
        return members;
    }

    /**
     * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
     * emanating from it.
     * @return Range
     */
    public Range deepCopy() throws AggregationException {
        EquidistantDoubleAggregation copy = new EquidistantDoubleAggregation();
        copy.setAggregationName(this.getAggregationName());
        copy.setAggregationType(this.getAggregationType());
        Range srcBlueprint = getBlueprint();
        copy.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        copy.setDeclaredNodeStructure(this.getDeclaredNodeStructure());
        copy.setDescription(this.getDescription());
        copy.setDnNode(this.getDnNode());
        copy.setFirst(this.getFirst());
        copy.setGetter(this.getGetter());
        copy.setIncrement(this.getIncrement());
        copy.setInquiryMethod(this.getInquiryMethod());
        copy.setNumberRanges(this.getNumberRanges());
        copy.setPassivator(this.getPassivator());
        copy.setPersistenceHandle(this.getPersistenceHandle());
        copy.setSerializable(this.isSerializable());
        copy.setSubject(this.getSubject());
        copy.setSubjectClass(this.getSubjectClass());
        copy.setContext(this.getContext());


        //The following is predicated on the fact that RangeSurgeons are completely
        //stateless and can therefore be shared among many Range clients
        copy.setRangeSurgeon(this.getRangeSurgeon());


        //now deepcopy the members individually to avoid pointing to an existing
        //object and subsequently accidentally modifying it
        List srcMembers = this.getMembers();
        if (null == srcMembers) {
            copy.setMembers((List)null);
        } else {
            Class srcMembersClass = srcMembers.getClass();
            List copyList = (List) shallowCopy(srcMembersClass, srcMembers);
            for (Iterator memberIter = srcMembers.iterator();
                                       memberIter.hasNext(); ) {
                Object currSrcMember = memberIter.next();
                if (currSrcMember instanceof Range) {
                    Range srcRange = (Range) currSrcMember;
                    Range copyRange = srcRange.deepCopy();
                    copyRange.setContext(copy);
                    copyList.add(copyRange);
                }
            }
            copy.setMembers(copyList);
        }
        return copy;
    }

    /**
     * Am I "equal to " <code>inRHS</code>?  If <code>inRHS</code> and I
     * have the same <code>increment</code> then we are considered "equal"
     * @param inRHS Object
     * @return boolean
     */
    public boolean equals(Object inRHS) {
        if (null == inRHS) {
            return false;
        }

        if (inRHS instanceof EquidistantDoubleAggregation) {
            EquidistantDoubleAggregation rhs = (EquidistantDoubleAggregation)
                                               inRHS;
            Double lhsIncrement = this.getIncrement();
            Double rhsIncrement = rhs.getIncrement();
            if (lhsIncrement == null && rhsIncrement == null) {
                return true;
            }
            if (lhsIncrement != null && rhsIncrement == null) {
                return false;
            }
            if (lhsIncrement != null && rhsIncrement != null) {
                return lhsIncrement.equals(rhsIncrement);
            }
            if (lhsIncrement == null && rhsIncrement != null) {
                return false;
            }
            return true;
        } else {
            return false;
        }
    }
}
