package org.quintet.miles;

import java.util.Collection;
import java.util.Iterator;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.text.ParseException;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.List;
import java.util.ArrayList;

public class EquidistantTimeAggregation extends Aggregation {

    private static Logger log = Logger.getLogger("org.quintet.miles");

    /**
       The <code>blueprints</code> attribute defines the structure of each member comprising this <code>Aggregation</code>.
     * If this is null then I default to constructing <code>BinaryRange</code> instances that house collections
     */
    protected Range blueprint = null;
    public EquidistantTimeAggregation() {
    }


    /**
       The number of subordinate <code>Range</code>s to append to this
       <code>EquidistantDoubleAggregation</code>
     */
    private Integer numberRanges = null;

    /**
       The first slot's ceiling
     */
    private Long first = null;

    /**
     The difference between slot <i>n</i>'s ceiling and slot <i>n+1</i>'s ceiling
     */
    private Long increment = null;

    /**
     * The number of milliseconds covered by one slot; e.g., 3600000 is one hour
     */
    private Long incrementSizeInMillis = null;

    /**
     * Return <code>true</code> is this <code>Range</code> is considered to be
     * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
     * method is intentionally declared abstract at the <code>Range</code> level
     * in order to force every <code>Range</code> subclass implementor to carefully
     * consider the nature of his class and whether it can truly be considered a
     * leaf node.
     * @return boolean
     */
  public  boolean isLeafNode(){
      //if my members != null AND it is a Range then I am not a leaf node
      if(null != getMembers() && getMembers() instanceof Range)
          return true;
      else
          return false;
  }

    /**
       Access method for the numberRanges property.

       @return   the current value of the numberRanges property
     */
    public Integer getNumberRanges() {
        return numberRanges;
    }

    /**
       Sets the value of the numberRanges property.

       @param aNumberRanges the new value of the numberRanges property
     */
    public void setNumberRanges(Integer aNumberRanges) {
        numberRanges = aNumberRanges;
    }

    public Long getIncrementSizeInMillis() {
        return incrementSizeInMillis;
    }

    public void setIncrementSizeInMillis(Long incrementSizeInMillis) {
        this.incrementSizeInMillis = incrementSizeInMillis;
    }

    /**
       Access method for the increment property.

       @return   the current value of the increment property
     */
    public Long getIncrement() {
        return increment;
    }

    public Long getFirst() {
        return first;
    }

    /**
      Sets the value of the increment property.

      @param aIncrement the new value of the increment property
     */
    public void setIncrement(Long aIncrement) {
        increment = aIncrement;
    }

    public void setFirst(Long first) {
        this.first = first;
    }

    public void setFirstTimestampString(String tsStr) {
      DateFormat dateFormat = new SimpleDateFormat("MM/dd/yyyy HH:mm:ss a");
        try {
            Date d = dateFormat.parse(tsStr);
            Long l = new Long(d.getTime());
            setFirst(l);
        } catch (ParseException ex) {
            String fatal =
                    "fatal error, unable to parse 'first-timestamp-string' value '" +
                    tsStr + "'";
            log.log(Level.SEVERE,fatal, ex);
            throw new IllegalArgumentException(fatal, ex);
        }
    }

    /**
     Construct my <code>members</code> if necessary (it's only necessary the first
     time this method is called) and then return my <code>members</code> to the
       caller.
       @return java.util.Collection
       @roseuid 42558FA902C8
     */
    public void setBlueprint(Range inBlueprint) {
        blueprint = inBlueprint;
        //preconditions:
        //1 -- first property != null
        Long currFloorObj = getFirst();
        if (null == currFloorObj) {
            String fatal =
                    "Floor not set, cannot construct an EquidistantTimeAggregation";
            log.log(Level.SEVERE,fatal);
            throw new IllegalStateException(fatal);
        }

        //2 -- increment object is not null
        Long incrementObj = getIncrement();
        if (null == incrementObj) {
            String fatal =
                    "Increment size not set, cannot construct an EquidistantTimeAggregation";
            log.log(Level.SEVERE,fatal);
            throw new IllegalStateException(fatal);
        }

        long currFloor = currFloorObj.longValue();
        long increment = incrementObj.longValue();
        //slot n's ceiling is one millisecond before slot n+1's floor
        long currCeil = currFloor + increment - 1;

        //my members collection is a linked list, which allows it to grow sparsely
        if (super.getMembers() != null) {
            super.setMembers(null);
        }

        super.setMembers(new java.util.ArrayList());
        java.util.List members = super.getMembers();

        for (int idx = 0; idx < getNumberRanges().intValue(); idx++) {
            BinaryRange newMember = new BinaryRange();
            newMember.setSubject(getSubject());
            newMember.setGetter(getGetter());
            Long newFloor = new Long(currFloor);
            Long newCeiling = new Long(currCeil);
            newMember.setFloor(new Long(currFloor));
            newMember.setCeiling(newCeiling);
            java.util.Date floorDate = new java.util.Date(currFloor);
            java.util.Date ceilDate = new java.util.Date(currCeil);
            Range rangeSubordinate = (Range) blueprint.clone();
            newMember.setMembers(rangeSubordinate);
            currFloor += increment;
            members.add(newMember);
        }
    }

    /**
     * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
     * emanating from it.
     * @return Range
     */
    public Range deepCopy() throws AggregationException {
        EquidistantTimeAggregation copy = new EquidistantTimeAggregation();
        copy.setAggregationName(this.getAggregationName());
        copy.setAggregationType(this.getAggregationType());
        Range srcBlueprint = getBlueprint();
        copy.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        copy.setDeclaredNodeStructure(this.getDeclaredNodeStructure());
        copy.setDescription(this.getDescription());
        copy.setDnNode(this.getDnNode());
        copy.setFirst(this.getFirst());
        copy.setGetter(this.getGetter());
        copy.setIncrement(this.getIncrement());
        copy.setIncrementSizeInMillis(this.getIncrementSizeInMillis());
        copy.setInquiryMethod(this.getInquiryMethod());
        copy.setNumberRanges(this.getNumberRanges());
        copy.setPassivator(this.getPassivator());
        copy.setPersistenceHandle(this.getPersistenceHandle());
        copy.setSerializable(this.isSerializable());
        copy.setSubject(this.getSubject());
        copy.setSubjectClass(this.getSubjectClass());
        copy.setContext(this.getContext());


        //The following is predicated on the fact that RangeSurgeons are completely
        //stateless and can therefore be shared among many Range clients
        copy.setRangeSurgeon(this.getRangeSurgeon());


        //now deepcopy the members individually to avoid pointing to an existing
        //object and subsequently accidentally modifying it
        List srcMembers = this.getMembers();
        if (null == srcMembers) {
            copy.setMembers((List)null);
        } else {
            Class srcMembersClass = srcMembers.getClass();
            List copyList = (List) shallowCopy(srcMembersClass, srcMembers);
            for (Iterator memberIter = srcMembers.iterator();
                                       memberIter.hasNext(); ) {
                Object currSrcMember = memberIter.next();
                if (currSrcMember instanceof Range) {
                    Range srcRange = (Range) currSrcMember;
                    Range copyRange = srcRange.deepCopy();
                    copyRange.setContext(copy);
                    copyList.add(copyRange);
                }
            }
            copy.setMembers(copyList);
        }
        return copy;
    }

    /**
     * Am I "equal to " <code>inRHS</code>?  If <code>inRHS</code> and I
     * have the same <code>increment</code> then we are considered "equal"
     * @param inRHS Object
     * @return boolean
     */
    public boolean equals(Object inRHS) {
        if (null == inRHS) {
            return false;
        }

        if (inRHS instanceof EquidistantTimeAggregation) {
            EquidistantTimeAggregation rhs = (EquidistantTimeAggregation) inRHS;
            Long lhsIncrement = this.getIncrement();
            Long rhsIncrement = rhs.getIncrement();
            if (lhsIncrement == null && rhsIncrement == null) {
                return true;
            }
            if (lhsIncrement != null && rhsIncrement == null) {
                return false;
            }
            if (lhsIncrement != null && rhsIncrement != null) {
                return lhsIncrement.equals(rhsIncrement);
            }
            if (lhsIncrement == null && rhsIncrement != null) {
                return false;
            }
            return true;
        } else {
            return false;
        }
    }
}
