package org.quintet.miles;

import java.util.StringTokenizer;
import java.lang.reflect.Method;
import java.util.Map;
import java.util.HashMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.Collection;

import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.regex.PatternSyntaxException;

/**
 * Query results are returned as <code>DynamicRange</code>s
 */
import org.quintet.miles.DynamicRange;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Logger;
import java.util.logging.Level;


/**
 A <code>Group</code> is a dynamic <code>Range</code> whose members are
 aggregated, or grouped, by a designated set of columns; A <code>Group</code>
 mimics SQL's <i>GROUP BY</i> functionality.<br>
 <code>Group</code> establishes a new subordinate member each time it encounters
 a new primary key value.  These subordinate members are housed in the
 <code>Range.members</code> attribute.  Each immediate descendent of a
 <code>Group</code> is a clone of the <code>Group</code>'s <i>blueprint</i>.
 */
public class Group extends Range {

    private static final Logger log = Logger.getLogger("org.quintet.miles");
    /**
     * Return <code>true</code> is this <code>Range</code> is considered to be
     * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
     * method is intentionally declared abstract at the <code>Range</code> level
     * in order to force every <code>Range</code> subclass implementor to carefully
     * consider the nature of his class and whether it can truly be considered a
     * leaf node.
     * @return boolean
     */
  public  boolean isLeafNode(){
      //A group is never a leaf node; see the declaration for
      //groups (Map<PrimaryKey, ? extends Range>), which means that if there are
      //subordinate objects under here they must be Ranges
      return false;
  }

    /* (non-Javadoc)
     * @see java.util.Collection#contains(java.lang.Object)
     */
    public boolean contains(Object inSubject) {
        //if arg0 is a PrimaryKey then just whiz through my groups
        //map via map.get().
       if (getGroups() == null)
          return false;

        if (inSubject instanceof org.quintet.miles.PrimaryKey) {
            return getGroups().containsKey(inSubject);
        }
        return false;
    }

    /**
     * if my getters are specified before my subject then I need to cache the
     * getters and process them after the subject has been specified
     */
    protected String deferredGetterList = null;
    private Map<PrimaryKey, ? extends Range> groups = new ConcurrentHashMap<PrimaryKey, Range>(16,.75f,1);

    /**
     * A <code>Group</code> divides members into groups by examining their primary keys.
     * The group of columns comprising a <code>Group</code>'s primary key is stored in the <code>primaryKeyColumns</code>
     */
    private Collection<String> primaryKeyColumns = null;

    /**
        @roseuid 42517B1702E8
     */
    public Group() {

    }

    public Range findMembers(Object whom) throws AggregationException {
        if (null == whom) {
            String err = "Cannot process null subjectValue";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        if (!(whom instanceof String)) {
            String err =
                    "Can only process String subjectValue, you passed in a '" +
                    whom.getClass().getName() + "'";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        String whomString = (String) whom;
        if (whomString.length() == 0) {
            String err = "Cannot process empty subjectValue";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        //make a regular expression machine out of whom,
        //walk through all keys looking for matching PKs,
        //copy all such matching Ranges into a new Range
        DynamicRange answer = new DynamicRange();
        answer.setAggregationName(this.getAggregationName());
        answer.setDescription(this.getDescription());
        answer.setGetter(this.getGetter());
        answer.deferredGetterName = this.deferredGetterName;
        answer.setSubject(this.getSubject());
        answer.setSubjectClass("[none]");
        answer.setAggregationType(getAggregationType());
        answer.setDeclaredNodeStructure(getDeclaredNodeStructure());
        answer.setMembers(new java.util.ArrayList());
        Pattern pkPattern = Pattern.compile(whomString);
        for (PrimaryKey currPK : groups.keySet()) {
            log.log(Level.SEVERE,"comparing pkpattern '" + whomString + "' to pk '" +
                      currPK.toString());
            Matcher matcher = pkPattern.matcher(currPK.toString());
            if (matcher.matches()) {
                Range foundOne = (Range) groups.get(currPK);
                foundOne.setAggregationName(currPK.toString());
                answer.add(foundOne.deepCopy());
            }
        }
        return answer;

    }


    /**
     * Use <code>dn</code> to traverse the <code>Range</code> rooted at this point.
     * Return the <code>Range</code> residing at the spot designated by <code>dn</code>
     * This is the most basic <code>findDN</code> implementation; several <code>Range</code>
     * subclasses override this method with significantly more sophisticated traversal
     * algorithms.
     * @param dn DN of the form <code><i>subjectClass=subjectValue[,subectClass=subjectValue...]
     * @return Range that most closely matches <code>dn</code>
     * @throws AggregationException
     */
    public Range findDN(DN dn) throws AggregationException {
        //1.  check to see if this level of the tree matches the
        //    subject mentioned in the current top DN Node;
        //    for example, if dn="ou=acme,uid=joe" then check to see
        //    if this.subjectClass == ou.
        //    If they don't match then  walk down the subtree rooted here until
        //    a match is found and then restart the findDN process
        String subjectClass = getSubjectClass();
        if (null == subjectClass || subjectClass.length() == 0) {
            String err = "subject class is not set, cannot process DN '" + dn +
                         "'";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        DNNode<String, String> topNode = dn.peek();
        if (subjectClass.equals(topNode.LHS)) {
            //1.1  Yes, this level of the tree matches the subject class stipulated
            //     by the topNode.  Now "prune" this level of the tree (no
            //     sublevel pruning done yet), yielding a new Aggregation containing
            //     only those members of this Aggregation that meet topNode's criteria
            Range foundMembers = findMembers(topNode.RHS);
            if (null == foundMembers) {
                //There are no objects in this tree node that satisfy topNode
                return null;
            } else {
                //topNode, the current DN, just produced the results found in
                //foundMembers.  Record this DN-Range correlation by imprinting topNode
                //onto foundMembers.
                //We really need to imprint this upon foundMembers' members, not onto
                //foundMembers itself, so that is what setDnNode does.
                foundMembers.setDnNode(topNode);

                //the node at which we just peek()ed is still lingering on the DN.  We
                //must remove it now...
                String logMsg = "DN is " + dn.toString() +
                                ", about to remove top element";
                log.info(logMsg);
                dn.poll();
                logMsg = "DN is now " + dn.toString();
                log.info(logMsg);

                //we found some qualifying members at this level and placed them into
                //the "pruned" Aggregation named "foundMembers".    If there are remaining nodes on the DN
                //then delegate the remaining search to the found nodes
                if (dn.peek() == (DNNode<String, String>)null) { //check to see if there is more on the DN...
                    //nothing left on the search spec, so we have found the thing being sought (I think....);
                    return foundMembers;
                } else {
                    //now apply the newly-pruned DN to each found member and allow each
                    //such subtree to be visited by the pruned DN
                    for (Iterator foundMemberIter = foundMembers.iterator();
                            foundMemberIter.hasNext(); ) {
                        DN dnForSubtree = (DN) dn.clone();
                        Object currFoundMember = foundMemberIter.next();
                        //this Aggregation's members aren't necessarily Ranges; they
                        //could be plain old objects.  Only recurse into subordinate Ranges
                        if (currFoundMember instanceof Range) {
                            //prune the subtree anchored at currFoundMember
                            Range subtree = (Range) currFoundMember;
                            subtree.setDnNode(dnForSubtree.peek());
                            Range newSubtreeMembers = subtree.findDN(
                                    dnForSubtree);
                            subtree.setMembers(newSubtreeMembers);
                        }
                    }
                    return foundMembers;
                }
            }
        } else {
            //the current topNode is not matched to our subjectClass, so try to match
            //it to one of our members
            //the following code sickens me, but it does work, and it is the most
            //expeditious route to implementing a proper findDN method.
            //If the current Range's members() are a Group then process the group
            //directly, otherwise process each member of members()
            if (getMembers() instanceof Group) {
                Group memberGroup = (Group) getMembers();
                Range nextFound = memberGroup.findDN(dn);
                return nextFound;
            } else {
                //the current topNode is not matched to my subjectClass, so try to match
                //it to one or more of my members; specifically, try to match the RHS of the node
                //to one or more of my keys
                DynamicRange answer = new DynamicRange();
                answer.setAggregationName(getAggregationName());
                answer.setDescription(getDescription());
                answer.setGetter(getGetter());
                answer.setInquiryMethod(getInquiryMethod());
                answer.setSubject(getSubject());
                answer.setSubjectClass("[none]");
                answer.setMembers(new java.util.ArrayList());
                answer.setDnNode(topNode);
                Pattern pattern = Pattern.compile(topNode.RHS);
                if (getGroups() == null)
                   return null;

                Map<PrimaryKey, Range> myGroups = (Map<PrimaryKey, Range>)
                        getGroups();
                for (PrimaryKey currMemberPK : myGroups.keySet()) {
                    String pkStr = currMemberPK.toString();
                    Matcher matcher = pattern.matcher(pkStr);
                    if (matcher.matches()) {
                        answer.add(myGroups.get(currMemberPK));
                    }
                }
                return answer;
            }
        }
    }


    public java.util.Iterator iterator() {
       if (getGroups() == null)
          return null;

        return getGroups().values().iterator();
    }

    /**
     Access method for the groups property.

     @return   the current value of the groups property
     */
    public Map getGroups() {
        return groups;
    }

    /**
        Sets the value of the groups property.

        @param aGroups the new value of the groups property
     */
    public void setGroups(Map aGroups) {
        groups = aGroups;
    }

    /**
        Access method for the primaryKeyColumns property.

        @return   the current value of the primaryKeyColumns property
     */
    public Collection<String> getPrimaryKeyColumns() {
        return primaryKeyColumns;
    }

    /**
        Sets the value of the primaryKeyColumns property.

     @param aPrimaryKeyColumns the new value of the primaryKeyColumns property
     */
    public void setPrimaryKeyColumns(Collection<String> aPrimaryKeyColumns) {
      if(null == aPrimaryKeyColumns)
        return;
      primaryKeyColumns = null;
      primaryKeyColumns = new CopyOnWriteArrayList<String>();
      for(String keyColumn:aPrimaryKeyColumns){
        primaryKeyColumns.add(keyColumn);
      }
    }

    /**
        @param inCandidate
        @return boolean
        @roseuid 42517B170310
     */
    public boolean isEligible(Object inCandidate) {
        if (inCandidate == null) {
            return false;
        }
        if (getSubject().equals(inCandidate.getClass())) {
            return true;
        } else {
            return false;
        }
    }

    /**
        If dd <code>inCandidate</code> to this <code>Range</code>
        @param inCandidate - The object being considered for membership in this
        <code>Range</code>
        @return boolean
        @roseuid 42517B3C01FE
     */
    public boolean add(Object inCandidate) {
        PrimaryKey candidatePK = null;
        try {
            candidatePK = new PrimaryKey(inCandidate, getPrimaryKeyColumns());
        } catch (AggregationException ex) {
            String err = "AggregationException when adding object to Group";
            log.log(Level.SEVERE,err, ex);
            return false;
        }

        // TODO Why was this put in? It causes a line to be written every millisecond
        /*
        if(candidatePK.toString().startsWith("2/201")){
          log.log(Level.WARNING,"found a 2/201");
        }
		  */
        
        //see if I already have a structure to house this particular value
        Object rangeObj = null;
        if (getGroups() != null){
          rangeObj = getGroups().get(candidatePK);

          Range subordinates = (Range) rangeObj;
          if (subordinates == null) {
            //first time this primary key has been seen
            subordinates = (Range) getBlueprint().clone();
            subordinates.setContext(this);
            //associate this new Range with a visibility RuleSet
            if (getGroups() != null)
              getGroups().put(candidatePK, subordinates);
            
          }
          //now I definitely have a structure for this candidate PK
          return subordinates.add(inCandidate);
        }
        return false;
    }

    public int size() {

        return getGroups() == null ? 0 : getGroups().size();
    }

    /**
        Clone this <code>Group</code> via "deep copy"
        @return java.lang.Object
        @roseuid 4252B89A0034
     */
    public Object clone() {
        Object base = super.clone();
        //also need to clone the blueprint, assuming one exists
        Group clonedGroup = (Group) base;
        //first, though, reset the groups map, because the current one probably
        //contains members already
        clonedGroup.groups = null;
        clonedGroup.setGroups(new ConcurrentHashMap(16,.75f,1));
        if (getBlueprint() == null) {
            return clonedGroup;
        }
        Object blueprintObj = getBlueprint().clone();
        Range blueprintClone = (Range) blueprintObj;
        clonedGroup.setBlueprint(blueprintClone);
        return clonedGroup;
    }

    /**
        Generate a String description of this <code>Range</code>'s structure
     @param  indentLevel - The depth at which this <code>Range</code>'s description
        should be presented
        @return java.lang.String
        @roseuid 42556C9302D5
     */
    public String getStructure(int indentLevel) {
        //my structure includes my blueprint, my primary key columns, and my groups' structures
        StringBuffer answer = new StringBuffer(getDescription() + "\n");
        answer.append(Range.tabs(1) + "Blueprint\n");
        answer.append(getBlueprint().getStructure(2) + "\n");
        answer.append(Range.tabs(1) + "grouped by\n");
        for (java.util.Iterator colIter = getPrimaryKeyColumns().iterator();
                                          colIter.hasNext(); ) {
            String currColl = (String) colIter.next();
            answer.append(Range.tabs(2) + currColl + "\n");
        }
        answer.append(Range.tabs(1) + "Groups:\n");
        if (getGroups() != null) {
           for (java.util.Iterator keyIter = getGroups().keySet().iterator();
                                             keyIter.hasNext(); ) {
               PrimaryKey currKey = (PrimaryKey) keyIter.next();
               answer.append(Range.tabs(2) + currKey.toString() + "\n");
               Range currRange = (Range) getGroups().get(currKey);
               answer.append(currRange.getStructure(3));
           }
        }
        return answer.toString();
    }

    public void setSubjectName(String inValue) throws ClassNotFoundException,
            AggregationException {
        //go ahead and set the subject, and then process the deferred group if necessary
        super.setSubjectName(inValue);
        if (this.deferredGetterList != null) {
            setGroupBy(deferredGetterList);
        }

    }

    /**
        The <code>setGroupBy</code> method constructs this <code>Group</code>'s
     <code>primaryKeyColumns</code> from the comma-delimited list of getter names
     (all the getters must be members of <code>subject</code>, of course) contained
     in <code>inGroupByList</code>.  Throw an <code>AggregationException</code> if
        one or more of the getters are not members of <code>subject</code>.
     @param inGroupByList - The comma-delimited list of getter names from which I
        will construct my <code>primaryKeyColumns</code>
        @throws org.quintet.miles.AggregationException
        @roseuid 425594750154
     */
    public void setGroupBy(String inGroupByList) throws AggregationException {
        //if this method is called before the subject is specified then
        //stash the inGroupByList away until such a time as the subject is known
        if (getSubject() == null) {
            //String info =
            //        "groupByList encountered before subject-name, therefore deferring processing";
            //log.info(info);
            this.deferredGetterList = inGroupByList;
            return;
        }
        Collection pkGetters = new CopyOnWriteArrayList();
        StringTokenizer lex = new StringTokenizer(inGroupByList, ",");
        while (lex.hasMoreTokens()) {
            String currToken = lex.nextToken();
            Method currMethod = null;
            try {
                currMethod = getSubject().getMethod(currToken, (Class[])null);
            } catch (SecurityException ex) {
                String err = "SecurityException when processing token '" +
                             currToken + "'";
                log.log(Level.SEVERE,err, ex);
                throw new AggregationException(err, ex);
            } catch (NoSuchMethodException ex) {
                String err = "NoSuchMethodException when processing token '" +
                             currToken + "'";
                log.log(Level.SEVERE,err, ex);
                throw new AggregationException(err, ex);
            }
            pkGetters.add(currToken);
        }
        setPrimaryKeyColumns(pkGetters);
    }

    /**
        Return a comma-delimited list of the columns comprising this
        <code>Group</code>'s GROUP-BY clause
        @return java.lang.String
        @roseuid 425AA4670086
     */
    public String getGroupBy() {
        Collection pkColumns = getPrimaryKeyColumns();
        if (pkColumns == null || pkColumns.size() == 0) {
            return null;
        }
        StringBuffer answer = new StringBuffer();
        for (java.util.Iterator pkIter = pkColumns.iterator(); pkIter.hasNext(); ) {
            String currMethod = (String) pkIter.next();
            if (pkIter.hasNext()) {
                answer.append(currMethod + ",");
            } else {
                answer.append(currMethod);
            }
        }
        return answer.toString();
    }

    /**
     * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
     * emanating from it.
     * @return Range
     */
    public Range deepCopy() throws AggregationException {
        Group copy = new Group();
        copy.setAggregationName(this.getAggregationName());
        copy.setAggregationType(this.getAggregationType());
        Range srcBlueprint = getBlueprint();
        copy.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        copy.setDeclaredNodeStructure(this.getDeclaredNodeStructure());
        copy.deferredGetterList = this.deferredGetterList;
        copy.setDescription(this.getDescription());
        copy.setDnNode(this.getDnNode());
        copy.setGetter(this.getGetter());
        copy.setInquiryMethod(this.getInquiryMethod());
        copy.setPassivator(this.getPassivator());
        copy.setPersistenceHandle(this.getPersistenceHandle());
        copy.setSerializable(this.isSerializable());
        copy.setSubject(this.getSubject());
        copy.setSubjectClass(this.getSubjectClass());
        copy.setContext(this.getContext());

        //The following is predicated on the fact that RangeSurgeons are completely
        //stateless and can therefore be shared among many Range clients
        copy.setRangeSurgeon(this.getRangeSurgeon());


        //now deepcopy the members individually to avoid pointing to an existing
        //object and subsequently accidentally modifying it
        List srcMembers = this.getMembers();
        if (null == srcMembers) {
            copy.setMembers((List)null);
        } else {
            Class srcMembersClass = srcMembers.getClass();
            List copyList = (List)shallowCopy(srcMembersClass, srcMembers);
            for (Iterator memberIter = srcMembers.iterator();
                                       memberIter.hasNext(); ) {
                Object currSrcMember = memberIter.next();
                if (currSrcMember instanceof Range) {
                    Range srcRange = (Range) currSrcMember;
                    Range copyRange = srcRange.deepCopy();
                    copyRange.setContext(copy);
                    copyList.add(copyRange);
                }
            }
            copy.setMembers(copyList);
        }

        //copy all the primary key columns
        Collection srcPKColumns = getPrimaryKeyColumns();
        if (null == srcPKColumns) {
            copy.setPrimaryKeyColumns((Collection)null);
        } else {
            Collection copyPKColumns = new java.util.ArrayList(srcPKColumns.
                    size());
            for (Iterator srcPKIter = srcPKColumns.iterator();
                                      srcPKIter.hasNext(); ) {
                String srcPK = (String) srcPKIter.next();
                copyPKColumns.add(srcPK);
            }
            copy.setPrimaryKeyColumns(copyPKColumns);
        }
        //now copy the groups individually to avoid etc etc etc
        Map<PrimaryKey, ? extends Range> srcGroups = getGroups();
        if (null == srcGroups) {
            copy.setGroups((Map)null);
        } else {
            Map<PrimaryKey, Range>
                    copyGroups = new HashMap<PrimaryKey, Range>(srcGroups.size());
            for (PrimaryKey srcPK : srcGroups.keySet()) {
                PrimaryKey copyPK = (PrimaryKey) srcPK.clone();
                Range srcGroup = srcGroups.get(srcPK);
                Range copyGroup = srcGroup.deepCopy();
                copyGroup.setContext(copy);
                copyGroups.put(copyPK, copyGroup);
            }
            copy.setGroups(copyGroups);
        }
        return copy;
    }

    /**
     * Am I "equal to" another <code>Group</code>?  In the case of <code>Group</code>
     * equality is determined purely by the equality of the <code>primaryKeyColumns</code>
     * attribute
     * @param inRHS Object
     * @return boolean
     */
    public boolean equals(Object inRHS) {
        return false;

    }
}
