package org.quintet.miles;

import java.util.logging.Logger;
import java.util.logging.Level;

import java.util.Collection;
import org.quintet.QuintetException;


public class JITRange
    extends Range {

  /**
   * <code>JITRange</code> uses <code>DN</code>s to specify the objects to be activated
   */

  /*#org.quintet.miles.DN Dependency_Link*/

  /**
   * @clientRole activator
   * @directed true
   */

  Logger log = Logger.getLogger("org.quintet.miles");
/**
   * The <code>activator</code>
   * is responsible for actually activating/fetching the <code>Object</code>s
   * described by <code>setSpecification</code>
   */
  private Activator activator;

  /**
   * Return the "real" <code>Range</code> that serves as this <code>JITRange</code>'s
   * underlying implementation.  The <code>actualRange</code> performs the aggregation
   * work on behalf of this <code>JITRange</code>; the <code>JITRange</code> is in
   * many ways nothing more than an adaptor for <code>actualRange</code>.
   * @return Range
   */
  public Range getActualRange() {
    if(null == actualRange){
      actualRange = (Range)this.getBlueprint().clone();
    }
    return actualRange;
  }

  /**
   * @param actualRange The actualRange to set.
   */
  public void setActualRange(Range actualRange) {
    this.actualRange = actualRange;
  }

  /**
   * Mutator for the <code>activator</code> property.  The <code>activator</code>
   * is responsible for actually activating/fetching the <code>Object</code>s
   * described by <code>setSpecification</code>
   * @return Activator
   */
  public Activator getActivator() {
    return activator;
  }

  /**
   * Designate <code>newValue</code> to be this <code>JITRange</code>'s new
   * <code>Activator</code>.  The <code>activator</code>
   * is responsible for actually activating/fetching the <code>Object</code>s
   * described by <code>setSpecification</code>
   * @param newValue Activator
   */
  public void setActivator(Activator newValue) {
    activator = newValue;
  }

  /**
   * The <code>actualRange</code> performs the aggregation
   * work on behalf of this <code>JITRange</code>; the <code>JITRange</code> is in
   * many ways nothing more than an adaptor for <code>actualRange</code>.
   * @alias actualRange
   * @clientRole actualRange
   * @directed true
   */

  private Range actualRange = null;

  /**
   * Interrogate this <code>JITRange</code> to determine if
   * it is willing to let <code>inCandidate</code> be added to its <code>members</code>;
   * of course, <code>JITRange</code>s do not allow anything (other than their
   * own <code>Activator</code>s) to add to their membership lists, therefore
   * this method always returns <code>false</code>
   * @param inCandidate Object
   * @return boolean
   */
  public boolean isEligible(Object inCandidate) {
    return false;
  }

  /**
   * Return <code>true</code> is this <code>Range</code> is considered to be
   * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
   * method is intentionally declared abstract at the <code>Range</code> level
   * in order to force evry <code>Range</code> subclass implementor to carefully
   * consider the nature of his class and whether it can truly be considered a
   * leaf node.<br>
   * In the case of a <code>JITRange</code> the "leafness" of the <code>Range</code>
   * is based on <code>actualRange</code>'s "leafness", therefore this is merely
   * a pass-through method
   * @return boolean
   */
  public boolean isLeafNode() {
    return actualRange.isLeafNode();
  }

  /**
   * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
   * emanating from it.
   * @return Range
   */
  public Range deepCopy() throws AggregationException {
    JITRange copy = new JITRange();
    Range actualRangeCopy = actualRange.deepCopy();
    copy.setActualRange(actualRangeCopy);
    //The following is predicated on the fact that RangeSurgeons are completely
    //stateless and can therefore be shared among many Range clients
    copy.setRangeSurgeon(this.getRangeSurgeon());

    Activator activatorCopy = null;
    try {
      activatorCopy = (Activator) getActivator().clone();
      copy.setActivator(activatorCopy);
    }
    catch (CloneNotSupportedException ex) {
      String noWay = "like W, Activator does not support cloning";
      log.log(Level.SEVERE,noWay,ex);
      throw new IllegalStateException(noWay,ex);
    }
    return copy;
  }

  /**
   * Find all <code>member</code>s that consider themselves to be identified by
   * <code>attributeValue</code>; for example, a <code>Group</code> of Organizations
   * with <code>subjectClass</code> = <code>"ou"</code> might consider itself to
   * be identified by an <code>attributeValue</code> String with the value "ou".
   * At the <code>Range</code> level little can be assumed about the nature of
   * <code>members</code>, nor can much be assumed about the <code>member</code>s'
   * potential equality relationships with <code>attributeValue</code>.
   * @param attributeValue Object
   * @return Object
   * @throws AggregationException
  */
  public Range findMembers(Object attributeValue) throws AggregationException {
    return getActualRange().findMembers(attributeValue);
  }

  /**
   * Use <code>dn</code> to traverse the <code>Range</code> rooted at this point.
   * Return the <code>Range</code> residing at the spot designated by <code>dn</code>
   * This is the most basic <code>findDN</code> implementation; several <code>Range</code>
   * subclasses override this method with significantly more sophisticated traversal
   * algorithms.
   * @param dn DN of the form <code><i>subjectClass=subjectValue[,subectClass=subjectValue...]
   * @return Range that most closely matches <code>dn</code>
   * @throws AggregationException
  */
  public Range findDN(DN dn) throws AggregationException {
    //first use activator to fetch the designated objects
    try{
      Collection specifiedSet = getActivator().activate(dn);
      //now populate my actual Range with the specified set
      if(null != specifiedSet && specifiedSet.size() > 0){
          getActualRange().addAll(specifiedSet);
      }
      //now delete the findDN call to the freshly-populated actual Range
      Range answer = getActualRange().findDN(dn);
      this.actualRange = null;
      return answer;
      //my goodness, that seems simple, at least on the surface
    }catch(QuintetException qe){
      String err = "Exception when re-activating designated data set for aggregation";
      log.log(Level.SEVERE,err,qe);
      throw new AggregationException(err,qe);
    }
  }
}
