package org.quintet.miles;

import java.lang.reflect.*;
import java.util.Map;
import java.util.HashMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.Collection;
import java.util.logging.*;

/**
 * A <code>MultiGroup</code> is an enhanced <code>Group</code>.  Unlike a
 * conventional <code>Group</code>, which assumes that it always operates on
 * objects of cardinality <i>one</i>, <code>MultiGroup</code> expressly operates
 * on collection-like objects.  <code>Group</code> takes in a subject, derives a
 * <code>PrimaryKey</code> for it, finds a <i>group</i> that can hold the <code>
 * PrimaryKey</code>, and then passes the subjectg to that <code>PrimaryKey</code>'s
 * associated <code>Range</code>.  <code>MultiGroup</code>, on the other hand, takes
 * in a subject, invokes the <i>collection getter</i> method on the subject,
 * then loops through the collection's members, constructing a <code>PrimaryKey</code>
 * for each such member and then adding the PK+member to the <code>groups</code>
 * <code>Map</code>
 * <p>Copyright: Copyright (c) 2005</p>
 *
 * <p>Company: 6th Sense Analytics</p>
 *
 * @author Peter Thaggard
 * @version 1.0
 */
public class MultiGroup
    extends Group {

  private static final Logger log = Logger.getLogger("org.quintet.miles");

  /**
   * Unlike normal <code>Group</code>s, where the <i>attributes of interest</i>
   * are members of the <code>subject</code> <code>Class</code>, a <code>MultiGroup</code>'s
   * <i>attributes of interest</i> are members of the <code>Class</code> to
   * which the MultiGroup's <code>collection</code>'s members belong; for example,
   * in a normal <code>Group</code> whose <code>subject</code> is
   * <code>com.sixthsense.observation.Observation</code>, the <code>Group</code>'s
   * getters are all <code>Method</code>s of the <code>Observation</code>
   * <code>Class</code> but in a <code>MultiGroup</code> where the collection
   * is, say, the hierarchy of <code>Organization</code>s to which the
   * <code>observerationSubject</code> is considered to belong, the getters are all
   * <code>Method</code>s of the <code>Organization</code> class.
   * In order to support this behavior, <code>MultiGroup</code> needs to know
   * the <code>Class</code> of the objects comprising <code>collection</code>.
   * The <code>collectionMembersClass</code> houses this information.
   */
  private Class collectionMembersClass = null;


  private String collectionMembersClassName = null;

  public Class getCollectionMembersClass(){
    return collectionMembersClass;
  }

  public void setCollectionMembersClass(Class newValue){
    collectionMembersClass = newValue;
  }

  /**
   * The <code>collectionGetter</code> extracts a <code>Collection</code>
   * (or array) of <i>objects of interest<i> from candidate members. Each
   * <i>object of interest</i> serves as the source of a <code>PrimaryKey</code>,
   * which in turn is processed in the same manner that a normal <code>Group</code>
   * would
   */
  private Method collectionGetter = null;

  private String deferredCollectionGetter = null;
  private String collectionGetterName = null;

  /**
   * Retrieve this <code>MultiGroup</code>'s <code>collectionGetter</code>
   * <code>Method</code>.
   * The <code>collectionGetter</code> extracts a <code>Collection</code>
   * (or array) of <i>objects of interest<i> from candidate members. Each
   * <i>object of interest</i> serves as the source of a <code>PrimaryKey</code>,
   * which in turn is processed in the same manner that a normal <code>Group</code>
   * would
   * @return Method
   */
  public Method getCollectionGetter() {
    return collectionGetter;
  }

  public String getCollectionGetterName() {
    return collectionGetterName;
  }

  public String getCollectionMembersClassName() {
    return collectionMembersClassName;
  }

  /**
   * designate <code>inCollectionGetter</code> as this <code>MultiGroup</code>'s
   * <code>collectionGetter</code>.
   * The <code>collectionGetter</code> extracts a <code>Collection</code>
   * (or array) of <i>objects of interest<i> from candidate members. Each
   * <i>object of interest</i> serves as the source of a <code>PrimaryKey</code>,
   * which in turn is processed in the same manner that a normal <code>Group</code>
   * would
   * @param inCollectionGetter Method
   */
  public void setCollectionGetter(Method inCollectionGetter) {
    this.collectionGetter = inCollectionGetter;
}

  public void setCollectionGetterName(String collectionGetterName) {
    if(getSubject() == null){
      //subject not yet set, so stash this collectionGetterName away
      //until it is set
      deferredCollectionGetter = collectionGetterName;
      return;
    }

    this.collectionGetterName = collectionGetterName;
    //get the actual collectionGetter method designated by
    //the collectionGetterName
    Class subjClass = getCollectionMembersClass();
    try {
      collectionGetter = subjClass.getMethod(collectionGetterName, (Class[])null);
    }
    catch (SecurityException ex) {
      String err = "SecurityException from " + subjClass.getName() + "." + collectionGetterName + "()";
      log.log(Level.SEVERE,err,ex);
      throw new IllegalStateException(err,ex);
    }
    catch (NoSuchMethodException ex) {
      String err = "NoSuchMethodException from " + subjClass.getName() + "." + collectionGetterName + "()";
      log.log(Level.SEVERE,err,ex);
      throw new IllegalStateException(err,ex);
    }
  }


  public void setCollectionMembersClassName(String inName) {
    this.collectionMembersClassName = inName;

    //get the actual class defined by the name
    try {
      collectionMembersClass = Class.forName(collectionMembersClassName);
    }
    catch (ClassNotFoundException ex) {
      String err = "ClassNotFoundException when trying to load collection " + "members class named '" + inName + "'";
      log.log(Level.SEVERE,err,ex);
      throw new IllegalStateException(err,ex);
    }

    //if I deferred instantiating my collectionGetter then do it now
    if(null != deferredCollectionGetter){
      setCollectionGetterName(deferredCollectionGetter);
    }
  }


  public MultiGroup() {
    super();
  }

  public boolean add(Object subject){
    if(null == subject){
      String warn = "Unable to process null subjects";
      log.log(Level.WARNING,warn);
      return false;
    }

    //use our collection-getter to grab a collection of objects from subject
    Method getCollection = getCollectionGetter();
    Object rawCollection = null;
    try {
      rawCollection = getCollection.invoke(subject, (Object[])null);
    }
    catch (InvocationTargetException ex) {
      Throwable cause = ex.getCause();
      String err = "InvocationTargetException when invoking method '" + subject.getClass().getName() + "." + getCollection.getName() + "'";
      log.log(Level.SEVERE,err,cause);
    }
    catch (IllegalArgumentException cause) {
      String err = "IllegalArgumentException when invoking method '" + subject.getClass().getName() + "." + getCollection.getName() + "'";
      log.log(Level.SEVERE,err,cause);
    }
    catch (IllegalAccessException cause) {
      String err = "IllegalAccessException when invoking method '" + subject.getClass().getName() + "." + getCollection.getName() + "'";
      log.log(Level.SEVERE,err,cause);
    }

    Object[] thingsToProcess = null;
    int thingsCount = 0;
    //we need to process both arrays and collections,
    //so if this is a collection then convert it to an array
    if(rawCollection instanceof Collection){
      Collection lessRawCollection = (Collection)rawCollection;
      thingsCount = lessRawCollection.size();
      thingsToProcess = new Object[thingsCount];
      lessRawCollection.toArray(thingsToProcess);
    }else if(rawCollection.getClass().isArray()){
      thingsToProcess = (Object[])rawCollection;
      thingsCount = thingsToProcess.length;
    }else{
      String err = "expected a Collection or an array, found a '" + rawCollection.getClass().getName() + "'";
      log.log(Level.WARNING,err);
      throw new IllegalStateException(err);
    }

    //ooi means "object of interest"
    for(Object ooi:thingsToProcess){
      //build a primary key for ooi
      PrimaryKey candidatePK = null;
      try {
        candidatePK = new PrimaryKey(ooi, this.getPrimaryKeyColumns());
      }
      catch (AggregationException ex1) {
        String err = "Aggregation exception when trying to construct a PK[" + getPrimaryKeyColumns() + "]";
        log.log(Level.WARNING,err);
        return false;
      }
      Object rangeObj = null;
      if (getGroups() != null){
        rangeObj = getGroups().get(candidatePK);

        Range subordinates = (Range) rangeObj;
        if (subordinates == null) {
          //first time this primary key has been seen
          subordinates = (Range) getBlueprint().clone();
          if (getGroups() != null)
            getGroups().put(candidatePK, subordinates);
        }
        //now I definitely have a structure for this candidate PK
        //NOTE:  the subject gets added to subordinates, not the ooi
        return subordinates.add(subject);
        }
    }
    return true;
  }
}


