package org.quintet.miles;

import java.util.Iterator;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.regex.PatternSyntaxException;

import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.List;
import java.util.ArrayList;

/**
 A <code>MultiplexAggregation</code> differs from a standard "simplex"
 <code>Aggregation</code> in the way it implements the <code>add()</code> method.
 A <code>MultiplexAggregation</code> assumes that an object can be a member of
 multiple <code>Range</code>s, so it passes new members to all of its subordinate
 <code>Range</code>

 @author Peter J Thaggard
 @version 1.0
 @see org.quintet.miles.Aggregation
 @see org.quintet.miles.Range
 */
public class MultiplexAggregation extends Aggregation {

    private static final Logger log = Logger.getLogger("org.quintet.miles");

    /**
        @roseuid 424D759F018F
     */
    public MultiplexAggregation() {

    }

    /**
     * Return <code>true</code> is this <code>Range</code> is considered to be
     * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
     * method is intentionally declared abstract at the <code>Range</code> level
     * in order to force every <code>Range</code> subclass implementor to carefully
     * consider the nature of his class and whether it can truly be considered a
     * leaf node.
     * @return boolean
     */
  public  boolean isLeafNode(){
      //MultiplexAggregation is by definition a non-leaf node
      return false;
  }
    /**
     * Return a hard-coded <code><i>subject class</i></code> to satisfy
     * the <code>findDN</code> algorithm
     * @return String
     */
    public String getSubjectClass() {
        return "[none]";
    }

    /**
     * Find a <code>member</code> that considers itself to be identified by
     * <code>attributeValue</code>; for example, a <code>Group</code> of Organizations
     * with <code>subjectClass</code> = <code>"ou"</code> might consider itself to
     * be identified by an <code>attributeValue</code> String with the value "ou".
     * At the <code>MultiplexAggregation</code> level this method assumes that
     * <code>attributeValue</code> is being compared to the <code>subjectClass</code>
     * of each <code>member</code>
     * @param attributeValue Object
     * @return Object
     * @throws AggregationException
     */
    public Range findMembers(Object attributeValue) throws AggregationException {
        if (null == attributeValue) {
            String err = "Unable to findMember(null)";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }
        if (null == getMembers()) {
            return null;
        }

        if (!(attributeValue instanceof String)) {
            String err = "MultiplexAggregation.findMember only accepts String " +
                         "arguments, but you passed it a '" +
                         attributeValue.getClass().getName() + "'";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }
        String attributeValueString = (String) attributeValue;
        MultiplexAggregation answer = new MultiplexAggregation();
        answer.setAggregationName(this.getAggregationName());
        answer.setDescription(this.getDescription());
        answer.setGetter(this.getGetter());
        answer.setSubject(this.getSubject());
        answer.setSubjectClass(this.getSubjectClass());
        answer.setMembers(new java.util.ArrayList());
        Pattern pkPattern = Pattern.compile(attributeValueString);
        for (Iterator memberIter = getMembers().iterator(); memberIter.hasNext(); ) {
            Object currMember = memberIter.next();
            if (currMember instanceof Range) {
                Range currMemberRange = (Range) currMember;
                String currMemberSubjectClass = currMemberRange.getSubjectClass();
                Matcher matcher = pkPattern.matcher(currMemberSubjectClass);
                if (matcher.matches()) {
                    answer.add(currMemberRange.deepCopy());
                }
            }
        }
        return answer;
    }

    /**
        Add <code>inNewMember</code> to this aggregation.  Within a
        <code>MultiplexAggregation</code> "adding" means passing
     <code>inNewMember</code> to every subordinate <code>Range</code> and allowing
     the subordinates to decide for themselves whether to add the new member.
        @param inNewMember - The object being considered for membership in this
        <code>MultiplexAggregation</code>'s subordinate <code>Range</code>s
        @return boolean
        @roseuid 424D75240188
     */
    public boolean add(Object inNewMember) {
        boolean answer = false;
        for (Iterator rangeIter = iterator(); rangeIter.hasNext(); ) {
            Range currSubordinate = (Range) rangeIter.next();
            currSubordinate.setContext(this);
            if (currSubordinate.add(inNewMember)) {
                answer = true; //inNewMember added to at least one subordinate
            }
        }
        return answer;
    }

    /**
     * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
     * emanating from it.
     * @return Range
     */
    public Range deepCopy() throws AggregationException {
        MultiplexAggregation copy = new MultiplexAggregation();
        copy.setAggregationName(this.getAggregationName());
        copy.setAggregationType(this.getAggregationType());
        Range srcBlueprint = getBlueprint();
        copy.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        copy.setDeclaredNodeStructure(this.getDeclaredNodeStructure());
        copy.setDescription(this.getDescription());
        copy.setDnNode(this.getDnNode());
        copy.setGetter(this.getGetter());
        copy.setInquiryMethod(this.getInquiryMethod());
        copy.setPassivator(this.getPassivator());
        copy.setPersistenceHandle(this.getPersistenceHandle());
        copy.setSerializable(this.isSerializable());
        copy.setSubject(this.getSubject());
        copy.setSubjectClass(this.getSubjectClass());
        copy.setContext(this.getContext());


        //The following is predicated on the fact that RangeSurgeons are completely
        //stateless and can therefore be shared among many Range clients
        copy.setRangeSurgeon(this.getRangeSurgeon());


        //now deepcopy the members individually to avoid pointing to an existing
        //object and subsequently accidentally modifying it
        List srcMembers = this.getMembers();
        if (null == srcMembers) {
            copy.setMembers((List)null);
        } else {
            Class srcMembersClass = srcMembers.getClass();
            List copyList = (List) shallowCopy(srcMembersClass, srcMembers);
            for (Iterator memberIter = srcMembers.iterator();
                                       memberIter.hasNext(); ) {
                Object currSrcMember = memberIter.next();
                if (currSrcMember instanceof Range) {
                    Range srcRange = (Range) currSrcMember;
                    Range copyRange = srcRange.deepCopy();
                    copyRange.setContext(copy);
                    copyList.add(copyRange);
                }
            }
            copy.setMembers(copyList);
        }
        return copy;
    }

    /**
     * Am I "equal to" some other <code>MultiplexAggregation</code>?
     * The answer <b>is always no</b>
     * @param inRHS Object
     * @return boolean
     */
    public boolean equals(Object inRHS) {
        return false;
    }
}
