package org.quintet.miles;

import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * A <code>PrimaryKey</code> uniquely identifies one instance of its associated <code>targetClass</code>,
 * also known as the <code>PrimaryKey</code>'s <i>domain</i>.  A <code>PrimaryKey</code> is comprised of
 * one or more <i>columns</i>.  These columns are extracted from a transient <i>target object</i> that
 * is passed into <code>PrimaryKey</code>'s constructor.  Once the column values have been extracted,
 * the <i>target object</i> is discarded, but a copy of its class is stored in <code>targetClass</code>.
 * The <code>targetClass</code> is subsequently used by <code>PrimaryKey.equals()</code>, because in order
 * for to <code>PrimaryKey</code>s to be considered "equal" they must both refer to objects in the same
 * domain.
 *
 * <p> </p>
 *
 * <p> </p>
 *
 * <p>Copyright: Copyright (c) 2005</p>
 *
 * <p>Company: 6th Sense Analytics</p>
 *
 * @author Peter Thaggard
 * @version 1.0
 */
public class PrimaryKey
    implements Cloneable {


  /**
   * It is safe to cache the reflectively-acquired column values after the first time this
   * <code>PrimaryKey</code> is evaluated; after that we can skip the method.invoke calls and
   * merely access the <code>cachedColumnValues</code>
   */
  protected Collection cachedColumnValues = null;

  /**
   * Retrieve this <code>PrimaryKey</code>'s collection of cached column values.
   * It is safe to cache the reflectively-acquired column values after the first time this
   * <code>PrimaryKey</code> is evaluated; after that we can skip the method.invoke calls and
   * merely access the <code>cachedColumnValues</code>
   * @return Collection
   */
  protected Collection getCachedColumnValues() {
    return cachedColumnValues;
  }


  /**
   * Designate <code>newValue</code> as this <code>PrimaryKey</code>'s <code>Collection</code>
   * of <i>cached column values</i>
   * It is safe to cache the reflectively-acquired column values after the first time this
   * <code>PrimaryKey</code> is evaluated; after that we can skip the method.invoke calls and
   * merely access the <code>cachedColumnValues</code>
   * @param newValue Collection
   */
  protected void setCachedColumnValues(Collection newValue){
    cachedColumnValues = newValue;
  }


  /**
      The <code>Class</code> of the <code>Object</code> for which I am the primary key
   */
  private Class domain = null;

  /**
   Construct a <code>PrimaryKey</code> that identifies <code>inTargetObject</code>
      and uses <code>inGetters</code> to do so
   @param inTargetObject - The <code>Object</code> that will be idenfied by this
      <code>PrimaryKey</code>
   @param inGetters - The methods that provide access to <code>inTarget</code>'s
      primary key attributes
      @throws org.quintet.miles.AggregationException
      @roseuid 42518E120234
   */
  public PrimaryKey(Object inTargetObject, Collection inGetters) throws
      AggregationException {
    if (inTargetObject == null) {
      throw new AggregationException(
          "cannot construct a PrimaryKey for a null object");
    }

    if (inGetters == null || inGetters.size() == 0) {
      throw new AggregationException(
          "You must supply at least one column for the PK");
    }

    setCachedColumnValues(new CopyOnWriteArrayList());
    setDomain(inTargetObject.getClass());
    for (Iterator getterIter = inGetters.iterator(); getterIter.hasNext(); ) {
      Object currGetterObj = getterIter.next();
      if (! (currGetterObj instanceof String)) {
        throw new AggregationException("expected a String, found a " +
                                       currGetterObj.getClass().getName());
      }
      else {
        Method columnGetter = null;
        try {
          columnGetter = getDomain().getMethod( (String) currGetterObj,
                                               (Class[])null);
          Object columnValue = columnGetter.invoke(inTargetObject,(Object[])null);
          getCachedColumnValues().add(columnValue);
        }
        catch (SecurityException ex) {
          throw new AggregationException(
              "unable to find a method named '" +
              (String) currGetterObj + " in class " +
              getDomain().getName(), ex);
        }
        catch (NoSuchMethodException ex) {
          throw new AggregationException(
              "unable to find a method named '" +
              (String) currGetterObj + " in class " +
              getDomain().getName());
        }catch(InvocationTargetException ite){
          Throwable cause = ite.getCause();
          String err = "InvocationTargetException when invoking '" + getDomain().getName() + "." + (String)currGetterObj + "()'";
          throw new AggregationException(err,cause);
        }catch(IllegalAccessException cause){
          String err = "IllegalAccessException when invoking '" + getDomain().getName() + "." + (String)currGetterObj + "()'";
          throw new AggregationException(err,cause);
        }
      }
    }

    //postconditions:  target != null
    //                 cachedColumns != null
  }

  /**
      @roseuid 42517B6403B9
   */
  protected PrimaryKey() {

  }


  /**
      Access method for the domain property.

      @return   the current value of the domain property
   */
  public Class getDomain() {
    return domain;
  }

  public Integer getCachedHashCode() {
    return cachedHashCode;
  }

  /**
      Sets the value of the domain property.

      @param aTarget the new value of the target property
   */
  public void setDomain(Class aTarget) {
    domain = aTarget;
  }

  public void setCachedHashCode(Integer cachedHashCode) {
    this.cachedHashCode = cachedHashCode;
  }

  private Integer cachedHashCode = null;

  private int hashCodeFromCachedColumnValues(){
    if(null != cachedHashCode){
      return cachedHashCode;
    }

    int answer = 0;
    for(Object cachedColumn:getCachedColumnValues()){
      if(cachedColumn == null){
        String x = "null cachedColumn, ignoring it ";
      }else{
        answer ^= cachedColumn.hashCode();
      }
    }
    cachedHashCode = answer;
    return answer;
  }

  /**
      @return int
      @roseuid 425172C20049
   */
  public int hashCode() {
    int answer = 0;
    //if I have already evaluated and cached my column values then compute hashCode from cached values
    if(null != cachedColumnValues){
      return hashCodeFromCachedColumnValues();
    }else{
      String fatal = "Internal error -- Primary Key's cachedColumnValues Collection is null";
      throw new InternalError(fatal);
    }
  }


  /**
      @param inRHS - The object to which I am being compared
      @return boolean
      @roseuid 425172D901B7
   */
  public boolean equals(Object inRHS) {
    if (! (inRHS instanceof PrimaryKey)) {
      return false;
    }
    PrimaryKey rhs = (PrimaryKey) inRHS;

    //precondition:  this.target != null
    //precondition:  rhs.target != null

    PrimaryKey lhs = this;

    //lhs.target.class must == rhs.target.class
    if(lhs.getDomain().equals(rhs.getDomain()) == false){
      return false;
    }

    //they must have the same columns in the same order,
    //and each lhs.column must equal the corresponding rhs.column
    Collection<Object> lhsColumns = lhs.getKeys();
    Collection<Object> rhsColumns = rhs.getKeys();
    if ( (lhsColumns == null && rhsColumns != null) ||
        (lhsColumns != null && rhsColumns == null)) {
      return false; //column lists not of same length, ergo != equal
    }

    if (lhsColumns == null && rhsColumns == null) {
      return true; //both have empty column lists, ergo equal
    }

    if (lhsColumns.size() != rhsColumns.size()) {
      return false;
    }

    Iterator lhsIter = lhsColumns.iterator();
    Iterator rhsIter = rhsColumns.iterator();
    while (lhsIter.hasNext()) {
      Object lhsColumn = lhsIter.next();
      Object rhsColumn = rhsIter.next();

      if ( (lhsColumn == null && rhsColumn != null) ||
          (lhsColumn != null && rhsColumn == null)) {
        return false; //one is null, one is not, ergo not equal
      }
      if (lhsColumn != null && rhsColumn != null &&
          lhsColumn.equals(rhsColumn) == false) {
        return false;
      }
    }
    //every column is equal
    return true;
  }

  public Collection getKeys() {
    Collection possibleAnswer = getCachedColumnValues();
    if(null != possibleAnswer){
      return possibleAnswer;
    }else{
      return null;
    }
  }



  public String getKey() {
    return toString();
  }

  /** Make a "deep copy" of this <code>PrimaryKey</code>.  Unlink <code>clone()</code>,
   * which is flighty and refuses to make any claims about its behavior, <code>deepCopy</code>
   * guarantees to leave the source <code>PrimaryKey</code>'s state complete unaltered.
   * In return, <code>deepCopy</code> insists that its clients not place <code>deepCopy()ed</code>
   * <code>PrimaryKey</code>s in places where they might need to be re-evaluated; the <code>deepCopy</code>
   * does not include the source <code>PrimaryKey</code>'s collection of column getter methods.
   */
  public PrimaryKey deepCopy(){
    PrimaryKey copy = new PrimaryKey();
    if(this.getCachedColumnValues() == null){
      Collection thisKeys = this.getKeys();
      copy.setCachedColumnValues(thisKeys);
    }else{
      copy.setCachedColumnValues(this.getCachedColumnValues());
    }
    copy.domain = this.domain;
    return copy;
  }

  public Object clone() {
    PrimaryKey cloned = new PrimaryKey();
    cloned.setDomain(this.getDomain());
    cloned.setCachedColumnValues(this.getCachedColumnValues());
    return cloned;
  }


  /**
   Return a <code>String</code> representation of this <code>PrimaryKey</code>
      @return java.lang.String
      @roseuid 4251730001B7
   */
  public String toString() {
    StringBuffer answer = new StringBuffer();
    for(Iterator keyIter = getCachedColumnValues().iterator(); keyIter.hasNext();){
      String keyValue = keyIter.next().toString();
      answer.append(keyValue);
      if(keyIter.hasNext())
        answer.append("/");
    }
    return answer.toString();
  }
}
