package org.quintet.miles;

import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.quintet.QuintetException;
import org.quintet.backbeat.transformer.Transformer;
import org.quintet.chambers.Passivator;
import static org.quintet.miles.DNNode.AggregationFunction;

/**
 <code>Range</code> defines the contract for "bucket-like" classes; that is,
 classes comprised of objects that meet certain range-oriented membership
 criteria.  Classes that implement <code>Range</code> must offer these two
 methods:<br>
 <li><code>add(java.lang.Object candidate)</code> -- add <code>candidate</code>
 to the membership ranks if and only if <code>isEligible() == true</code>
 <li><code>isEligible(java.lang.Object candidate)</code> -- return true if
 <code>candidate</code> can legally be a member of the <code>Range</code>
 */
public abstract class Range implements List, Cloneable, Serializable {

    /**
     * The <code>context</code> attribute stores a reference to this <code>Range</code>'s
     * predecessor <code>Range</code>.
     */
    private Range context = null;


    /**
     * The object that I use to passivate and passivate myself; that is, the
     * object that is capable of swapping me into and out of persistent storage,
     * respectively.
     */

    private Passivator passivator = null;

    /**
     *   The node by which this <code>Range</code> was most recently reached by a query
     */
    private DNNode dnNode = null;

    /**
     * Users may specify this <code>Range</code>'s <code>getter-name</code> before
     * specifying its <code>subject</code>, which makes it impossible to immediately
     * lookup the method within <code>subject</code> that corresponds to <code>getter-name</code>.
     * The <code>deferredGetterName</code> attribute solves this problem.
     */
    protected String deferredGetterName = null;

    /**
     * Indicates if this <code>Range</code>, including its <code>member</code>s,
     * can be swapped out to persistent storage.  If <code>true</code>, then by rule
     * this value also applies to all subordinate <code>Range</code>s comprising
     * this <code>Range</code>.  A setting of <code>false</code>, on the other hand,
     * imposes no restrictions on the <code>serializable</code> attributes of
     * this <code>Range</code>'s subordinates
     */
    private boolean serializable = true;

    /**
     * Reference to a persistent storage location to which this
     * <code>Range</code> (and its entire hierarchy of descendents)
     * has been passivated.  This attribute is equal to <code>null</code>
     * when this node is in memory.
     */
    private transient Long persistenceHandle = null;

    /**
     The place where I send diagnostic messages
     */
    protected static final Logger log = Logger.getLogger("org.quintet.miles");

    private static final String lottaTabs =
            "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t";

    /**
     A short description of this <code>Range</code>
     */
    private String description = null;

    /**
     The <code>subject</code> indicates the type of objects being aggregated in this <code>Range</code>
     */
    private Class subject = null;

    /**
     * The <code>subjectClass</code> is a <code>String</code> indicating the
     * domain-specific nature of the objects being aggregated in this <code>Range</code>.
     * Unlike <code>subject</code>, this is a more-or-less arbitrary String; for example,
     * the <code>subject</code> might be <code>com.sixthsense.principal.Organization</code>,
     * while the <code>subjectClass</code> might be <i>top-level-organization</i>
     **/
    private String subjectClass = null;

    /**
     The means by which I obtain the <i>attribute of interest</i> from candidate members
     */
    private transient Method getter = null;


    /**
     The <code>members</code> attribute is the collection of all objects that have been added to this <code>Range</code>'s membership ranks
     */
    protected List members = null;

    private String aggregationName;

    /**
     @roseuid 424AB8FC00FF
     */
    public Range() {

    }

    private void writeObject(java.io.ObjectOutputStream oos) throws java.io.
            IOException {
        if (null == deferredGetterName) {
            if (null == getter) {
                String err = "deferredGetterName is null and getter is too";
                throw new IOException(err);
            } else {
                deferredGetterName = getter.getName();
            }
        }

        if (null == deferredInquiryMethodName) {
            if (null == inquiryMethod) {
                String warn =
                        "deferredInquiryMethodName is null and so is inquiryMethod";
                log.log(Level.WARNING,warn);
                deferredInquiryMethodName = "";
            } else {
                deferredInquiryMethodName = inquiryMethod.getName();
            }
        }
        oos.defaultWriteObject();
    }

    private void readObject(java.io.ObjectInputStream ois) throws java.io.
            IOException, ClassNotFoundException {
        ois.defaultReadObject();
        //now restore my transient members (Methods, etc.)
        String getterName = this.deferredGetterName;
        try {
            setGetterName(deferredGetterName);
        } catch (NoSuchMethodException ex) {
            String err = "Unable to find method '" + deferredGetterName + "'";
            log.log(Level.SEVERE,err, ex);
            throw new IllegalStateException(err, ex);
        }
        String inquiryMethodName = deferredInquiryMethodName;
        try {
            setInquiryMethodName(inquiryMethodName);
        } catch (NoSuchMethodException ex) {
            String err = "Unable to find method '" + inquiryMethodName + "'";
            log.log(Level.SEVERE,err, ex);
            throw new IllegalStateException(err, ex);
        }
    }

    /**
     Access method for the description property.

     @return   the current value of the description property
     */
    public String getDescription() {
        return description == null ? "anonymous range" : description;
    }

    /**
     Sets the value of the description property.

     @param aDescription the new value of the description property
     */
    public void setDescription(String aDescription) {
        description = aDescription;
    }

    /**
     Access method for the subject property.

     @return   the current value of the subject property
     */
    public Class getSubject() {
        return subject;
    }

    /**
     Sets the value of the subject property.

     @param aSubject the new value of the subject property
     */
    public void setSubject(Class aSubject) {
        subject = aSubject;
    }

    /**
     Access method for the <code>subjectClass</code> property.
     @return The current value of the <code>subjectClass</code> property
     **/
    public String getSubjectClass() {
        return subjectClass;
    }

    /**
     Mutator method for the <code>subjectClass</code> property.
     @param String newValue -- The new value of the <code>subjectClass</code> property
     @return void
     **/
    public void setSubjectClass(String newValue) {
        subjectClass = newValue;
    }

    /**
     Access method for the getter property.

     @return   the current value of the getter property
     */
    public Method getGetter() {
        return getter;
    }

    /**
     Sets the value of the getter property.

     @param aGetter the new value of the getter property
     */
    public void setGetter(Method aGetter) {
        getter = aGetter;
    }


    /**
     Add <code>inCandidate</code> to this <code>Range</code>
     @param inCandidate - The object being considered for membership in this
     <code>Range</code>
     @return boolean
     @roseuid 42498CAF03D2
     */
    public boolean add(Object inCandidate) {
        //only add inCandidate if it is eligible
        //eligibility is determined by inCandidate's attribute of interest
        if (isEligible(inCandidate)) {
            getMembers().add(inCandidate);
            return true;
        } else {
            return false;
        }
    }

    public void add(int _int, Object object) {
        getMembers().add(_int, object);
    }

    public Object get(int _int) {
        return getMembers().get(_int);
    }

    public Object set(int _int, Object object) {
        return getMembers().set(_int, object);
    }

    /**
     Return <code>true</code> if <code>inCandidate</code> can be considered a member
     of this <code>Range</code>
     @param inCandidate - The <code>Object</code> being considered for membership in
     this <code>Range</code>
     @return boolean
     @roseuid 42498CAF03DC
     */
    public abstract boolean isEligible(Object inCandidate);

    /**
     Return this <code>Range</code>'s member objects
     @return java.util.Collection
     @roseuid 424AA80D005B
     */
    public List getMembers() {
        return members;
    }

    public String getAggregationName() {
        return aggregationName;
    }

    public Method getInquiryMethod() {
        return inquiryMethod;
    }

    /**
     * "Passivate" the aggregation tree from <code>this</code> point down.  Passivation
     * is carried out by the <code>passivator</code> object, which is linked into this
     * <code>Range</code> via a blueprints entry.  This allows <i>Miles</i> to remain
     * independent of any particular passivation mechanisms, and makes it easy to
     * swap from, say, an Entity Bean passivator to a Hibernate passivator.
     * Generally speaking a <code>Range</code> does not invoke <code>passivate</code>
     * upon itself; rather, it is normally invoked by the <code>Range</code>'s
     * superior somewhere near the top of the aggregation tree, or it is called by
     * some "manager" existing totally outside of the aggregation tree.
     * @return Long
     * @throws AggregationException
     */
    private Long passivate() throws AggregationException {
        if (null == getPassivator()) {
            String err =
                    "Null passivator object, therefore unable to passivate this Range";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        try {
            Long handle = getPassivator().passivate(this);
            this.persistenceHandle = handle;
            //now we need to remove the rest of this Range from memory
            eraseAttributes();
            return handle;
        } catch (QuintetException qe) {
            throw new AggregationException("QuintetException", qe);
        }
    }

    /**
     * Set all of this <code>Range</code>'s serialized attributes to null,
     * thereby allowing their memory to be claimed by the GC.
     */
    protected void eraseAttributes() {
        this.blueprint = null;
        this.aggregationName = null;
        this.aggregationType = null;
        this.declaredNodeStructure = null;
        //this.deferredInquiryMethod = null;
        this.deferredInquiryMethodName = null;
        this.description = null;
        this.dnNode = null;
        this.getter = null;
        this.inquiryMethod = null;
        this.members = null;
        this.subject = null;
        this.subjectClass = null;
        this.transformerClass = null;
        this.transformerClassName = null;
    }

    /**
     Fetch <code>inCandidate</code>'s attribute of interest.  The attribute of
     interest is guaranteed to be <code>Comparable</code>; if it isn't then this
     method throws a <code>TransformerException</code>
     @param inCandidate - The object whose <i>attribute of interest</i> is being
     obtained
     @return java.lang.Comparable
     @throws org.quintet.miles.AggregationException
     @roseuid 424AAF71005B
     */
    protected Comparable getAttributeOfInterest(Object inCandidate) throws
            AggregationException {
        Method getAOI = getGetter();
        try {
            return (Comparable) getAOI.invoke(inCandidate, (Object[])null);
        } catch (InvocationTargetException ex) {
            String err = "InvocationTargetException when invoking method "
                         + inCandidate.getClass().getName() + "." +
                         getAOI.getName();
            Throwable cause = ex.getCause();
            log.log(Level.SEVERE,err, cause);
            throw new AggregationException(err, ex);
        } catch (IllegalArgumentException ex) {
            String err = "IllegalArgumentException when invoking method "
                         + inCandidate.getClass().getName() + "." +
                         getAOI.getName();
            log.log(Level.SEVERE,err, ex);
            throw new AggregationException(err, ex);

        } catch (IllegalAccessException ex) {
            String err = "IllegalArgumentException when invoking method "
                         + inCandidate.getClass().getName() + "." +
                         getAOI.getName();
            log.log(Level.SEVERE,err, ex);
            throw new AggregationException(err, ex);
        } catch (ClassCastException cce) {
            String err = "method " + inCandidate.getClass().getName() + "."
                         + getAOI.getName() +
                         " did not return a Comparable object";
            log.log(Level.SEVERE,err, cce);
            throw new AggregationException(err, cce);
        }
    }

    /**
     Replace my list of members with <code>inNewMembers</code>
     @param inNewMembers - My new membership list
     @roseuid 424AB06A0385
     */
    public void setMembers(List inNewMembers) {
        members = null;
        members = inNewMembers;

    }

    public void setAggregationName(String aggregationName) {
        this.aggregationName = aggregationName;
    }

    public void setInquiryMethod(Method inquiryMethod) {
        this.inquiryMethod = inquiryMethod;
    }

    private transient Method inquiryMethod = null;

    private boolean deferredInquiryMethod = false;

    private String deferredInquiryMethodName = null;

    /**
     * The <code>aggregationType<code> denotes the type of analytical information
     * contained within this portion of the aggregation tree; for example, a subtree
     * containing <i>Active Time</i> might have a value of <code>ACTIVE_TIME</code>.
     * The valid values for <code>aggregationType</code> are determined by the
     * <code>AnalysisType</code> enum
     * @see com.sixthsense.observation.Analyst.AnalysisType
     */
    private String aggregationType = null;

    /**
     * The <code>nodes</code> attribute is a comma-separated <code>String</code>
     * denoting the node types comprising this <code>Range</code>
     */
    private String declaredNodeStructure = null;

    private String transformerClassName = null;

    private Class transformerClass = null;

    /**
     * The <code>rangeSurgeon</code> is an optional element that allows
     * a <code>Range</code> to modify its contents prior to returning them
     * as a query response.
     */
    private RangeSurgeon rangeSurgeon = null;

    /**
     * The <code>Class</code> of this <code>Range</code>'s
     * <code>RangeSurgeon</code>; in most cases this attribute will
     * be null
     */
    protected Class<RangeSurgeon> surgeonClass = null;

    /**
     @return int
     @roseuid 424D61230333
     */
    public int size() {
        List mems = getMembers();
        return (null == mems ? 0 : mems.size());
    }

    /**
     @return boolean
     @roseuid 424D61230347
     */
    public boolean isEmpty() {
        return getMembers() == null ? true : (getMembers().isEmpty());
    }

    /**
     @param arg0
     @return boolean
     @roseuid 424D61230351
     */
    public boolean contains(Object arg0) {
        List mems = getMembers();
        return mems == null ? false : mems.contains(arg0);
    }

    /**
     @return java.util.Iterator
     @roseuid 424D6123035B
     */
    public Iterator iterator() {
        List mems = getMembers();
        return mems == null ? new NullIterator() : mems.iterator();
    }

    /**
     @return Object[]
     @roseuid 424D6123036F
     */
    public Object[] toArray() {
        List mems = getMembers();
        return mems == null ? (Object[])null : mems.toArray();
    }

    /**
     @param arg0
     @return Object[]
     @roseuid 424D61230379
     */
    public Object[] toArray(Object[] arg0) {
        List mems = getMembers();
        return mems == null ? (Object[])null : mems.toArray(arg0);
    }

    /**
     @param arg0
     @return boolean
     @roseuid 424D6123038D
     */
    public boolean remove(Object arg0) {
        List mems = getMembers();
        return mems == null ? false : mems.remove(arg0);
    }

    /**
     @param arg0
     @return boolean
     @roseuid 424D61230397
     */
    public boolean containsAll(Collection arg0) {
        List mems = getMembers();
        return mems == null ? false : mems.containsAll(arg0);
    }

    public Object remove(int _int) {
        List mems = getMembers();
        return mems == null ? null : mems.remove(_int);
    }

    public int indexOf(Object object) {
        List mems = getMembers();
        return mems == null ? -1 : mems.indexOf(object);
    }

    public int lastIndexOf(Object object) {
        List mems = getMembers();
        return mems == null ? -1 : mems.lastIndexOf(object);
    }

    public ListIterator listIterator() {
        List mems = getMembers();
        return mems == null ? null : mems.listIterator();
    }

    public ListIterator listIterator(int _int) {
        List mems = getMembers();
        return mems == null ? null : mems.listIterator(_int);
    }

    public List subList(int _int, int _int1) {
        List mems = getMembers();
        return mems == null ? (List)null : mems.subList(_int, _int1);
    }

    /**
     @param arg0
     @return boolean
     @roseuid 424D612303AB
     */
    public boolean addAll(Collection arg0) {
        for (Iterator iter = arg0.iterator(); iter.hasNext(); ) {
            if (!add(iter.next())) {
                return false;
            }
        }
        return true;
    }

    public boolean addAll(int _int, Collection collection) {
        if (null == collection || collection.size() == 0) {
            return false; //collection null or empty, therefore cannot addAll it to this!
        }
        List mems = getMembers();
        return mems == null ? false : mems.addAll(_int, collection);
    }

    /**
     @param arg0
     @return boolean
     @roseuid 424D612303B5
     */
    public boolean removeAll(Collection arg0) {
        List mems = getMembers();
        return mems == null ? false : mems.removeAll(arg0);
    }

    /**
     @param arg0
     @return boolean
     @roseuid 424D612303C9
     */
    public boolean retainAll(Collection arg0) {
        List mems = getMembers();
        return mems == null ? false : retainAll(arg0);
    }

    /**
     @roseuid 424D612303D3
     */
    public void clear() {
        List mems = getMembers();
        if (mems != null) {
            mems.clear();
        }
    }

    /**
     @param arg0
     @return boolean
     @roseuid 424D612303E7
     */
    public boolean equals(Object arg0) {
        if (arg0 == null) {
            return false;
        }
        if (!(arg0 instanceof Range)) {
            return false;
        }
        Range rhs = (Range) arg0;
        List lhsMems = getMembers();
        List rhsMems = rhs.getMembers();
        if (lhsMems == null && rhsMems == null) {
            return true;
        }
        if (lhsMems == null && rhsMems != null) {
            return false;
        }
        if (lhsMems != null && rhsMems == null) {
            return false;
        }
        return lhsMems.equals(rhsMems);
    }

    /**
     @return int
     @roseuid 424D61240013
     */
    public int hashCode() {
        List mems = getMembers();
        return mems == null ? 0 : mems.hashCode();
    }

    /**
     * Return <code>true</code> is this <code>Range</code> is considered to be
     * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
     * method is intentionally declared abstract at the <code>Range</code> level
     * in order to force every <code>Range</code> subclass implementor to carefully
     * consider the nature of his class and whether it can truly be considered a
     * leaf node.
     * @return boolean
     */
    public abstract boolean isLeafNode();

    /**
     * Apply all <i>rollups</i> stipulated in the DN that resulted in this
     * <code>Range</code>'s creation, return the possibly-rolled-up new <code>Range</code>.<br>
     * <b>Algorithm</b>:<br>
     * <li>for each level in this <code>Range</code>
     * <li>if the level's <code>dnNode.isLHSRollup()</code> (all members at a particular level have identical DNNodes, by the way)
     * then invoke the rollup method specified by the last character of <code>dnNode.LHS</code>; currently only SUM (designated by either '+' or '*')
     * is implemented<br>
     * @return Range
     * @throws AggregationException
     */
    public Range applyRollups() throws AggregationException {
        DNNode dn = getDnNode();
        Range answer = null;
        if (null != dn && dn.isLhsRollup()) {
            AggregationFunction dnAggFunc = dn.getAggregationFunctionType();
            if (null == dnAggFunc) {
                String err =
                        "Internal Error -- DN is marked as having an LHS Rollup Function, " +
                        "its getAggregationFunctionType() returned null";
                log.log(Level.SEVERE,err);
                throw new AggregationException(err);
            }
            Method aggFunc = dnAggFunc.getAggregationFunctionImpl();
            try {
                answer = (Range) aggFunc.invoke(this, (Object[])null);
                if (dnAggFunc.isAppliedRecursively() == false) {
                    return answer;
                }
            } catch (InvocationTargetException ex) {
                Throwable cause = ex.getCause();
                String err = cause.getClass().getName() +
                             " when invoking 'Range." + aggFunc.getName() +
                             "()'";
                log.log(Level.SEVERE,err, cause);
                throw new AggregationException(err, cause);
            } catch (IllegalArgumentException cause) {
                String err = cause.getClass().getName() +
                             " when invoking 'Range." + aggFunc.getName() +
                             "()'";
                log.log(Level.SEVERE,err, cause);
                throw new AggregationException(err, cause);
            } catch (IllegalAccessException cause) {
                String err = cause.getClass().getName() +
                             " when invoking 'Range." + aggFunc.getName() +
                             "()'";
                log.log(Level.SEVERE,err, cause);
                throw new AggregationException(err, cause);
            }
        } else {
            answer = this;
        }

        //ok dokey, now see if the answer needs to be rolled up at the next level too
        if (isLeafNode() == false) {
            List nextLevelDownList = answer.getMembers();
            if (nextLevelDownList instanceof Range) {
                Range nextLevelDownRange = (Range) nextLevelDownList;
                Range newNextLevelDown = nextLevelDownRange.applyRollups();
                answer.setMembers(newNextLevelDown);
            } else if (nextLevelDownList instanceof List) {
                List newNextLevelDownList = rollupList(nextLevelDownList);
                answer.setMembers(newNextLevelDownList);
            }
        }
        //gosh, I guess that about does it.  I don't see what can possibly go wrong
        return answer;
    }

    public List rollupList(List sourceList) throws AggregationException {
        List newList = new java.util.ArrayList();
        for (Iterator sourceIter = sourceList.iterator(); sourceIter.hasNext(); ) {
            Object currSource = sourceIter.next();
            if (currSource instanceof Range) {
                Range sourceRange = (Range) currSource;
                Range rolledUpRange = sourceRange.applyRollups();
                newList.add(rolledUpRange);
            } else if (currSource instanceof List) {
                List subordinateSourceList = (List) currSource;
                List rolledUpSubordinate = rollupList(subordinateSourceList);
                newList.add(rolledUpSubordinate);
            } else {
                newList.add(currSource);
            }
        }
        return newList;
    }

    /**
     * <i>Collapse</i> this <code>Range</code>'s <code>members</code>' <code>members</code> into a single
     * <code>members <code>List</code>; that is, for each <code>this.member[n]</code> extract all
     * <code>member[n].members</code> and copy them into a single <code>List</code>.  This has
     * the effect of removing the level immediately below this <code>Range</code>
     * and replacing it with the items two levels down.<br>
     * By the way, please note that this method is recursive, so it collapses as far down the
     * membership tree as possible
     * @return Range
     * @throws AggregationException
     */
    public Range collapse() throws AggregationException {
        List slots = getMembers();
        //the idea here is to eliminate the now-unnecessary ActiveTimeRange level
        //(it is only necessary during observation collection time, when we need
        //little five minute buckets for computing most-active-action).  This
        //routine creates a new Collection<ActiveTimeGroup> consisting of all the
        //ActiveTimeGroups comprising all the ActiveTimeRanges comprising this
        //ActiveTime

        //PRECONDITIONS:
        //PRECONDITION 1:  members IS NOT NULL
        List members = getMembers();
        if (null == members) {
            String info =
                    "members is null, therefore it's already \"collapsed\"";
            log.log(Level.INFO, info);
            return this.deepCopy();
        }

        List collapsedMembers = new java.util.ArrayList();
        Class _clazz = this.getClass();
        Range collapsed = (Range) this.shallowCopy(_clazz, this);
        for (Iterator membersIter = members.iterator(); membersIter.hasNext(); ) {
            Object uncollapsedObj = membersIter.next();
            if (uncollapsedObj instanceof Range) {
                Range uncollapsedRange = (Range) uncollapsedObj;
                Range collapsedMember = uncollapsedRange.collapse();
                //now extract collapsedMember's members and add them to our answer
                List collapsedMembersMembers = collapsedMember.getMembers();
                collapsedMembers.add(collapsedMembersMembers);
            } else {
                collapsedMembers.add(uncollapsedObj);
            }
        }
        collapsed.setMembers(collapsedMembers);
        return collapsed;
    }

    public Range sum() throws AggregationException {
        //I am a Dynamic Range
        //my membersList is a ArrayList<EquidistantTimeAggregation>
        List membersList = getMembers();
        if (null == membersList || 0 == membersList.size()) {
            return this;
        }
        //merge all members[1-n] of membersList into a deep  copy of membersList[0]
        Range origFirstMember = (Range) membersList.get(0);
        Range mergedMembers = origFirstMember.deepCopy();
        /*
        StringBuffer summedAggregationName = new StringBuffer(origFirstMember
                .getAggregationName() == null ? "[heck]" :
                origFirstMember.getAggregationName());
        */
        Class clazz = this.getClass();
        Range answer = this.emptyRangeCopy(clazz, this);

        Iterator membersIter = membersList.iterator();
        membersIter.next(); //skip the first one; we've already "merged" it into mergedMembers
        while (membersIter.hasNext()) {
            Object memberObj = membersIter.next();
            if (memberObj instanceof Range) {
                Range memberRange = (Range) memberObj;
                //summedAggregationName.append("+");
                //summedAggregationName.append(memberRange.getAggregationName());
                mergedMembers = mergedMembers.merge(memberRange);
            }
        }
        String summedAggregationName = this.getDnNode().RHS.toString().replace('|', '+');
        mergedMembers.setAggregationName(summedAggregationName);
        answer.setMembers(new java.util.ArrayList());
        answer.getMembers().add(mergedMembers);
        ;
        return answer;
    }

    public Range accumulate() throws AggregationException {
        //I am a Dynamic Range
        //my membersList is a ArrayList<EquidistantTimeAggregation>
        List membersList = getMembers();
        if (null == membersList || 0 == membersList.size()) {
            return this;
        }

        //merge a
        Range accumulated = (Range) membersList.get(0);
        /*
        StringBuffer summedAggregationName = new StringBuffer(accumulated
                .getAggregationName() == null ? "[heck]" :
                accumulated.getAggregationName());
        */
        Iterator membersIter = membersList.iterator();
        membersIter.next(); //skip the first one; we've already "merged" it into mergedMembers

        Class clazz = this.getClass();
        Range answer = this.emptyRangeCopy(clazz, this);
        answer.setMembers(new ArrayList());
        answer.getMembers().add(accumulated);

        while (membersIter.hasNext()) {
            Object memberObj = membersIter.next(); //BinaryRange<Long> -- littleslots
            if (memberObj instanceof Range) {
                Range memberRange = (Range) memberObj;
                /*
                if (accumulated.getMembers() != null && accumulated.getMembers().size() > 0) {
                    memberRange.getMembers().addAll(accumulated.getMembers());
                }
                */
                accumulated = memberRange.merge(accumulated);
                answer.getMembers().add(accumulated);
                //summedAggregationName.append("+");
                //summedAggregationName.append(memberRange.getAggregationName());
            }
        }
        return answer;
    }
    
    public Range average() throws AggregationException {
        List membersList = getMembers();
        if (null == membersList || 0 == membersList.size()) {
            return this;
        }
        //merge all members[1-n] of membersList into a deep  copy of membersList[0]
        Range origFirstMember = (Range) membersList.get(0);
        Range mergedMembers = origFirstMember.deepCopy();
        StringBuffer summedAggregationName = new StringBuffer("AVG(").append(origFirstMember
                .getAggregationName() == null ? "[NA]" :
                origFirstMember.getAggregationName());
        Class clazz = this.getClass();
        Range answer = this.emptyRangeCopy(clazz, this);

        Iterator membersIter = membersList.iterator();
        membersIter.next(); //skip the first one; we've already "merged" it into mergedMembers
        int cnter = 1;
        while (membersIter.hasNext()) {
            Object memberObj = membersIter.next();
            if (memberObj instanceof Range) {
                Range memberRange = (Range) memberObj;
                summedAggregationName.append("+");
                summedAggregationName.append(memberRange.getAggregationName());
                mergedMembers = mergedMembers.merge(memberRange);
                cnter++;
            }
        }
        
        mergedMembers.divide( cnter ) ;
        
        mergedMembers.setAggregationName(summedAggregationName.append(")").toString());
        answer.setMembers(new java.util.ArrayList());
        answer.getMembers().add(mergedMembers);
        return answer;
    }

    /**
     * Apply <code>clippers</code> to this <code>Range</code>,
     * yielding a new modified version of this this original <code>Range</code>
     * @return Range
     * @exception AggregationException
     */
    public Range prune(RangeSurgeon clippers) throws AggregationException {
        if (null == clippers) {
            String warn =
                    "Cannot apply a null RangeSurgeon, returning unaltered Range";
            log.log(Level.WARNING,warn);
            return this;
        }
        return clippers.operateOn(this);
    }

    /**
     * Apply my internal <code>RangeSurgeon</code> to myself,
     * yielding a new modified version of myself.
     * @return Range
     * @throws AggregationException
     */
    public Range prune() throws AggregationException {
        RangeSurgeon doctor = getRangeSurgeon();
        if (null == doctor) {
            //String info =
            //        "rangeSurgeon is null, therefore returning unmodified Range";
            //log.info(info);
            return this;
        }
        return prune(doctor);
    }

    /**
     * "Absorb" <code>otherRange</code> into this <code>Range</code>.  Absorption
     * is accomplished by:<br>
     * <li>If this <code>Range</code> is not a <i>leaf node</i> (that is,
     * <code>this.members != null</code> then create a new "merged" node and then
     * recursive absorb <code>otherRange.members</code> into <code>this.members<code>
     * into a <code>merged.members</code>.
     * <li>If this <code>Range</code> is a leaf node then absorb it by a straight merge of
     * <code>otherRange.members</code> and <code>this.members</code><br>
     * This <code>Range</code> itself is not merged with <code>otherRange</code>;
     * rather, <code>this.members</code> is merged with <code>otherRange.members</code>
     *
     * @param otherRange Range
     * @return Range
     */
    public Range merge(Range otherRange) throws AggregationException {
        //PRECONDITIONS:
        //PRECONDITION 1:  otherRange IS NOT NULL
        if (null == otherRange) {
            String err = "Precondition 1 failed -- Cannot merge a null Range";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        //PRECONDITION 1.1:  otherRange.members IS NOT NULL
        List otherRangeMembers = otherRange.getMembers();
        if (null == otherRangeMembers) {
            String err =
                    "Precondition 1.1 failed -- otherRange.getMembers() is null";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        //PRECONDITION 1.2:  this.members IS NOT NULL
        List myRangeMembers = this.getMembers();
        if (null == myRangeMembers) {
            String err = "Precondition 1.2 failed -- this.getMembers() is null";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        //PRECONDITION 2:  otherRange.class == this.class
        Class thisClass = this.getClass();
        Class otherClass = otherRange.getClass();
        if (!(thisClass.isAssignableFrom(otherClass) || otherClass.isAssignableFrom(thisClass) || thisClass.equals(otherClass))) {
            String err = "Precondition 2 failed -- <" + thisClass.getName()
                         + ">merge can only be applied to " + "another '"
                         + thisClass.getName() +
                         "' instance, but you passed in a '"
                         + otherClass.getName() + "'";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        //PRECONDITION 2.1:  otherRange.members.class == this.members.class
        Class otherRangeMembersClass = otherRangeMembers.getClass();
        Class myRangeMembersClass = myRangeMembers.getClass();
        /*if(otherRangeMembersClass.equals(myRangeMembersClass) == false){
         String err = "Precondition 2.1 failed -- " + "my member's class (" + myRangeMembersClass.getName() + ")!= other range's member's class (" + otherRangeMembersClass.getName();
         log.error(err);
         throw new AggregationException(err);
         }*/

        //PRECONDITION 3:  I am not a leaf node (i.e., members != null)
        if (null == getMembers()) {
            String err = "Precondition 3 failed -- '"
                         + thisClass.getName()
                         +
                         " is a leaf node, and it must implement its own merge method, "
                         + "which it fails to do.";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        //make a new empty instance of me to house the merged results
        Range merged = this.emptyRangeCopy(thisClass, this);
        if ( this.getDnNode() == null ) {
        	merged.setDnNode(otherRange.getDnNode());
        } else {
	        merged.setDnNode(this.getDnNode());
        }
        //if I am executing the Range-level merge impl then I am not a Leaf Node,
        //therefore in order to merge myself I need to recursively merge my descendents
        merged.setAggregationName(this.getAggregationName());

        if (myRangeMembers instanceof Range) {
          if(otherRangeMembers instanceof java.util.ArrayList){
            log.log(Level.WARNING,"expected a Range, found an ArrayList");
          }
            merged.setMembers(((Range) myRangeMembers).merge((Range)
                    otherRangeMembers));
        } else {

            //algorithm:
            //for each this.member
            //if corresponding member found in otherClass.members then
            //call merge on this.member and otherClass.matchedMember
            //add merged answer to the new merged Range
            //else
            //add this.member to new merged Range
            //for each otherClass.member
            //if corresponding member found in this.members then DO NOTHING; it's already be processed in the first half of this algorithm
            //else
            //add otherClass.member to the new merged Range
            //return merged answer
            List otherMembersList = otherRangeMembers;

            //Part 1 -- merge my members and identical ones from otherMembersList
            for (Iterator myMembersIter = myRangeMembers.iterator();
                                          myMembersIter
                                          .hasNext(); ) {
                Object currMemberObj = myMembersIter.next();
                if (!(currMemberObj instanceof Range)) {
                    //POJO, just add it to the merged members
                    merged.getMembers().add(currMemberObj);
                } else {
                    Range currMemberRange = (Range) currMemberObj;
                    //need to find this same range in otherRange.members
                    int otherIdx = otherMembersList.indexOf(currMemberRange);
                    if (otherIdx < 0) { //it is present here but not in other
                        merged.getMembers().add(currMemberRange);
                    } else {
                        Range mergedMembers = currMemberRange
                                              .merge((Range) otherMembersList.
                                get(
                                        otherIdx));
                        merged.getMembers().add(mergedMembers);
                    }
                }
            }

            //Part 2 -- merge in otherMembers that are not part of this.members
            for (Iterator otherMembersIter = otherMembersList.iterator();
                                             otherMembersIter
                                             .hasNext(); ) {
                Object otherMemberObj = otherMembersIter.next();
                if ( otherMemberObj instanceof Range ) {
	                int thisIdx = getMembers().indexOf(otherMemberObj);
	                if (thisIdx < 0) {
	                    merged.getMembers().add(otherMemberObj);
	                }
                } else {
                	merged.getMembers().add(otherMemberObj);
                }
            }
        }

        return merged;
    }
    
    public void divide ( int divisor ) throws AggregationException{
    	List myMembers = this.getMembers();
    	if (myMembers == null ) {
    		String err = "NULL members. Cannot proceed.";
    		log.log(Level.SEVERE, err);    		
    		throw new AggregationException (err);
    	}
    	
    	if ( myMembers instanceof Range ) {
    		((Range)myMembers).divide(divisor);
    	} else {
    		// Check the list contents to ensure this is the final step..
    		if ( myMembers.size() > 0 && myMembers.get(0) instanceof Range &&  ! ((Range)myMembers.get(0)).isLeafNode() ) {
    			Iterator iter = myMembers.iterator();
    			while(iter.hasNext() ) {
    				Range r = (Range)iter.next();
    				r.divide(divisor);
    			}
    		} else {
	    		int desiredLen = myMembers.size() / divisor;
	    		List newMembers = new ArrayList(desiredLen);
	    		newMembers.addAll( myMembers.subList(0, desiredLen));
	    		this.setMembers(newMembers);
    			
    		}
    	}
    	
    }

    public <T extends Range> T emptyRangeCopy(Class<T> factory,
            T it) throws AggregationException {
        T newT = it.shallowCopy(factory, it);
        //shallow copy it's [sic] members collection
        List itsMembers = it.getMembers();
        Class itsMembersClass = null;
        if (null == itsMembers) {
            String err = "encountered a null members collection, duplicating it as a java.util.concurrent.CopyOnWriteArrayList";
            itsMembersClass = java.util.concurrent.CopyOnWriteArrayList.class;
        } else {
            itsMembersClass = itsMembers.getClass();
        }
        newT.setMembers((List) shallowCopy(itsMembersClass, itsMembers));
        //newT.setDnNode(it.getDnNode());
        return newT;
    }

    public <T> T emptyCopy(Class<T> factory, T it) throws AggregationException {
        try {
            return factory.newInstance();
        } catch (IllegalAccessException ex) {
            String err = "IllegalAccessException from '" + factory.getName()
                         + ".newInstance'";
            log.log(Level.SEVERE,err, ex);
            throw new AggregationException(err, ex);
        } catch (InstantiationException ex) {
            String err = "InstantiationException from '" + factory.getName()
                         + ".newInstance'";
            log.log(Level.SEVERE,err, ex);
            throw new AggregationException(err, ex);
        }
    }

    public Range shallowCopy() throws AggregationException {
        Class _clazz = getClass();
        return (Range)shallowCopy(_clazz, this);
    }

    public <T> T shallowCopy(Class<T> factory,
            T source) throws AggregationException {
        T dest = emptyCopy(factory, source);
        if (dest instanceof Range) {
            Range destRange = (Range) dest;
            Range sourceRange = (Range) source;
            destRange.setAggregationName(sourceRange.aggregationName);
            destRange.setAggregationType(sourceRange.aggregationType);
            Range srcBlueprint = getBlueprint();
            destRange.setBlueprint(null == srcBlueprint ? null :
                                   srcBlueprint.shallowCopy());
            destRange.setDeclaredNodeStructure(sourceRange.
                                               declaredNodeStructure);
            destRange.setDescription(sourceRange.description);
            destRange.setDnNode(sourceRange.dnNode);
            destRange.setGetter(sourceRange.getter);
            destRange.setInquiryMethod(sourceRange.inquiryMethod);
            destRange.setPassivator(sourceRange.passivator);
            destRange.setPersistenceHandle(sourceRange.persistenceHandle);
            destRange.setSerializable(sourceRange.serializable);
            destRange.setSubject(sourceRange.subject);
            destRange.setSubjectClass(sourceRange.subjectClass);
        }
        return dest;
    }

    /**
     Generate a String description of this <code>Range</code>'s structure
     @param  indentLevel - The depth at which this <code>Range</code>'s description
     should be presented
     @return java.lang.String
     @roseuid 424D82CE0059
     */
    public String getStructure(int indentLevel) {
        StringBuffer answer = new StringBuffer();
        answer.append(tabs(indentLevel) + getDescription() + "\n");
        answer.append(tabs(indentLevel) + getClass().getName()
                      + ":getter name:" + getGetter().getName() + "\n");
        for (Iterator memberIter = iterator(); memberIter.hasNext(); ) {
            Object currMember = memberIter.next();
            if (currMember instanceof Range) {
                answer.append(((Range) currMember)
                              .getStructure(indentLevel + 1)
                              + "\n");
            } else {
                Object aoi = null;
                try {
                    aoi = getAttributeOfInterest(currMember);
                    answer.append(tabs(indentLevel + 1) + "Leaf Node:  "
                                  + currMember.getClass().getName() + ":" + aoi
                                  + "\n");
                } catch (AggregationException ex) {
                    answer.append(tabs(indentLevel + 1) + "Leaf Node: "
                                  + currMember.getClass().getName() +
                                  ":<no AOI>\n");
                }

            }
        }
        return answer.toString();
    }

    /**
     Convenience method to associate this <code>Range</code> with its getter method
     @param inGetterName
     @throws java.lang.NoSuchMethodException
     @roseuid 424DA2660361
     */
    public void setGetterName(String inGetterName) throws NoSuchMethodException {
        Class subj = getSubject();
        if (subj == null) {
            //String err = "Subject class must be set before getter name";
            //log.log(Level.INFO,err);
            this.deferredGetterName = inGetterName;
            return;
        } else {
            Method meth = null;
            try {
                meth = subj.getMethod(inGetterName, (Class[])null);
            } catch (SecurityException ex) {
                String err = "SecurityException when getting method "
                             + inGetterName;
                log.log(Level.SEVERE,err, ex);
                throw new NoSuchMethodException(err);
            }
            setGetter(meth);
        }
    }

    public void setInquiryMethodName(String inGetterName) throws
            NoSuchMethodException {
        Class subj = this.getClass();
        if (subj == null) {
            //String err =
            //        "Subject class must be set before inquiry attribute getter name";
            //log.log(Level.INFO,err);
            this.deferredInquiryMethodName = inGetterName;
            return;
        } else {
            Method meth = null;
            try {
                meth = subj.getMethod(inGetterName, (Class[])null);
            } catch (SecurityException ex) {
                String err = "SecurityException when getting method "
                             + inGetterName;
                log.log(Level.SEVERE,err, ex);
                throw new NoSuchMethodException(err);
            }
            setInquiryMethod(meth);
        }

    }

    /**
     Return the designated number of tab characters
     @param indentLevel - The number of tab characters requested
     @return java.lang.String
     @roseuid 424DB88901C5
     */
    protected static final String tabs(int indentLevel) {
        return lottaTabs.substring(0, indentLevel);
    }

    /**
     set my <code>subject</code> to be the class named <code>inName</code>.  Throw a
     <code>ClassNotFoundException</code> if a class named <code>inName</code> cannot
     be found in the CLASSPATH
     @param inName - The fully-qualified name of this <code>Aggregation</code>'s
     subject Class
     @throws java.lang.ClassNotFoundException
     @roseuid 424DB9B9010E
     */
    public void setSubjectName(String inName) throws ClassNotFoundException,
            AggregationException {
        setSubject(Class.forName(inName));
        if (this.deferredGetterName != null) {
            try {
                this.setGetterName(deferredGetterName);
            } catch (NoSuchMethodException nsme) {
                String err = "Unable to find getter named "
                             + deferredGetterName;
                log.log(Level.SEVERE,err, nsme);
                throw new AggregationException(err, nsme);
            }
        }

        if (this.deferredInquiryMethodName != null) {
            try {
                this.setInquiryMethodName(deferredInquiryMethodName);
            } catch (NoSuchMethodException nsme) {
                String err = "Unable to find inquiry method named "
                             + deferredInquiryMethodName;
                log.log(Level.SEVERE,err, nsme);
                throw new AggregationException(err, nsme);
            }
        }
    }

    /**
     Clone this <code>Range</code> via "deep copy"
     @return java.lang.Object
     @roseuid 42529F2D0198
     */
    public Object clone() {
        Range newRange = null;
        //let object do its pittance of cloning work first
        try {
            newRange = (Range)super.clone();
            newRange.setMembers(null);
        } catch (CloneNotSupportedException ex) {
            return null;
        }
        return newRange;
    }

    /**
     * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
     * emanating from it.
     * @return Range
     */
    public abstract Range deepCopy() throws AggregationException;

    /**
     * Find all <code>member</code>s that consider themselves to be identified by
     * <code>attributeValue</code>; for example, a <code>Group</code> of Organizations
     * with <code>subjectClass</code> = <code>"ou"</code> might consider itself to
     * be identified by an <code>attributeValue</code> String with the value "ou".
     * At the <code>Range</code> level little can be assumed about the nature of
     * <code>members</code>, nor can much be assumed about the <code>member</code>s'
     * potential equality relationships with <code>attributeValue</code>.
     * @param attributeValue Object
     * @return Object
     * @throws AggregationException
     */
    public abstract Range findMembers(Object attributeValue) throws
            AggregationException;

    /**
     * Use <code>dn</code> to traverse the <code>Range</code> rooted at this point.
     * Return the <code>Range</code> residing at the spot designated by <code>dn</code>
     * This is the most basic <code>findDN</code> implementation; several <code>Range</code>
     * subclasses override this method with significantly more sophisticated traversal
     * algorithms.
     * @param dn DN of the form <code><i>subjectClass=subjectValue[,subectClass=subjectValue...]
     * @return Range that most closely matches <code>dn</code>
     * @throws AggregationException
     */
    public abstract Range findDN(DN dn) throws AggregationException;

    public boolean matches(Pattern pattern) throws AggregationException {
        //get the "inquiry attribute" from this Range.
        Method getter = getInquiryMethod();
        if (null == getter) {
            String err = "Inquiry Method is null";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }
        Object evaluateObj = null;
        try {
            evaluateObj = getter.invoke(this, (Object[])null);
            if (null == evaluateObj) {
                String err = "inquiryMethod returned null";
                log.log(Level.SEVERE,err);
                throw new AggregationException(err);
            }
        } catch (InvocationTargetException ex) {
            Throwable orig = ex.getCause();
            String err = orig.getClass().getName()
                         + " when invoking inquiry method '"
                         + this.getClass().getName() + "." + getter.getName() +
                         "()";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err, orig);
        } catch (IllegalArgumentException ex) {
            String err = ex.getClass().getName()
                         + " when invoking inquiry method '"
                         + this.getClass().getName() + "." + getter.getName() +
                         "()";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err, ex);
        } catch (IllegalAccessException ex) {
            String err = ex.getClass().getName()
                         + " when invoking inquiry method '"
                         + this.getClass().getName() + "." + getter.getName() +
                         "()";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err, ex);
        }

        String evaluateStr = null;
        if (evaluateObj instanceof String) {
            evaluateStr = (String) evaluateObj;
        } else {
            evaluateStr = evaluateObj.toString();
        }
        Matcher matcher = pattern.matcher(evaluateStr);
        return matcher.matches();
    }

    /**
     Sets the value of the blueprint property.

     @param aBlueprint the new value of the blueprint property
     */
    public void setBlueprint(Range aBlueprint) {
        blueprint = aBlueprint;
    }

    public void setDeclaredNodeStructure(String nodes) {
        this.declaredNodeStructure = nodes;
    }

    public void setAggregationType(String aggregationType) {
        this.aggregationType = aggregationType;
    }

    public void setTransformerClassName(String transformerClassName) throws
            AggregationException {
        this.transformerClassName = transformerClassName;
        //now that we have the transform class' name, we might as well
        //get the actual class, eh?
        try {
            transformerClass = Class.forName(transformerClassName);
        } catch (ClassNotFoundException ex) {
            String err = "Unable to find a transformer class named '"
                         + transformerClassName + "'";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err, ex);
        }
    }

    public void setTransformerClass(Class transformerClass) {
        this.transformerClass = transformerClass;
    }

    public void setDnNode(DNNode inDNNode) {
        if (null != inDNNode) {
            this.dnNode = (DNNode) inDNNode.clone();
        }
    }

    public void setSerializable(boolean serializable) {
        this.serializable = serializable;
    }

    public void setPersistenceHandle(Long persistenceHandle) {
        this.persistenceHandle = persistenceHandle;
    }

    public void setPassivator(Passivator passivator) {
        this.passivator = passivator;
    }

    public void setRangeSurgeonName(String name) throws AggregationException {
        try {
            surgeonClass = (Class<RangeSurgeon>) Class.forName(name);
        } catch (ClassNotFoundException ex) {
            String err = "ClassNotFoundException from Class.forName(\"" + name +
                         "\")";
            log.log(Level.SEVERE,err, ex);
            throw new AggregationException(err, ex);
        }
        try {
            rangeSurgeon = surgeonClass.newInstance();
        } catch (IllegalAccessException ex1) {
            String err = "IllegalAccessException from " + surgeonClass.getName() +
                         ".newInstance()";
            log.log(Level.SEVERE,err, ex1);
            throw new AggregationException(err, ex1);
        } catch (InstantiationException ex1) {
            String err = "InstantiationException from " + surgeonClass.getName() +
                         ".newInstance()";
            log.log(Level.SEVERE,err, ex1);
            throw new AggregationException(err, ex1);
        }
    }

    public void setRangeSurgeon(RangeSurgeon rangeSurgeon) {
        this.rangeSurgeon = rangeSurgeon;
    }

    /**
     * Mutator for the <code>context</code> attribute.
     * The <code>context</code> attribute stores a reference to this <code>Range</code>'s
     * predecessor <code>Range</code>.
     * @param context Range
     */
    public void setContext(Range context) {
        this.context = context;
    }

    public Transformer getTransformer() throws AggregationException {
        Class xformerClass = getTransformerClass();
        if (null == xformerClass) {
            String err =
                    "Transformer Class is null, cannot instantiate transformer";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }
        Transformer xformer = null;
        try {
            xformer = (Transformer) xformerClass.newInstance();
        } catch (IllegalAccessException ex) {
            String err = "IllegalAccessException when trying to create a '"
                         + getTransformerClassName() + "' instance.";
            log.log(Level.SEVERE,err, ex);
            throw new AggregationException(err, ex);
        } catch (InstantiationException ex) {
            String err = "InstantiationException when trying to create a '"
                         + getTransformerClassName() + "' instance.";
            log.log(Level.SEVERE,err, ex);
            throw new AggregationException(err, ex);
        }
        return xformer;
    }

    private Range blueprint = null;
    /**
     Access method for the blueprint property.

     @return   the current value of the blueprint property
     */
    public Range getBlueprint() {
        return blueprint;
    }

    public String getDeclaredNodeStructure() {
        return declaredNodeStructure;
    }

    public String getAggregationType() {
        return aggregationType;
    }

    public String getTransformerClassName() {
        return transformerClassName;
    }

    public Class getTransformerClass() {
        return transformerClass;
    }

    public DNNode getDnNode() {
        return dnNode;
    }

    public boolean isSerializable() {
        return serializable;
    }

    public Long getPersistenceHandle() {
        return persistenceHandle;
    }

    public Passivator getPassivator() {
        return passivator;
    }

    public RangeSurgeon getRangeSurgeon() {
        return rangeSurgeon;
    }

    /**
     * Accessor for the <code>context</code> attribute.
     * The <code>context</code> attribute stores a reference to this <code>Range</code>'s
     * predecessor <code>Range</code>.
     * @return Range
     */
    public Range getContext() {
        return context;
    }

}
