package org.quintet.miles;

import org.quintet.backbeat.transformer.TransformerException;
import java.lang.reflect.*;
import java.util.regex.Pattern;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Logger;
import java.util.logging.Level;

/**
 A <code>UnaryRange<code> is a <code>Range</code> with a defined lower bound and
 an undefined upper bound; Its upper bound is frequently implicitly defined by
 the presence of another <code>Range</code> residing next to it.

 @author Peter J Thaggard
 @version 1.0
 @see org.quintet.backbeat.transformer.Range
 */
public class UnaryRange<T extends Comparable> extends Range {

    private static final Logger log = Logger.getLogger("org.quintet.miles");

    /**
        The lowest value that an object can have while still being considered a member of this <code>Range</code>
     */
    private T floor = null;

    /**
        @roseuid 424ABA8B0289
     */
    public UnaryRange() {

    }

    /**
     * Return <code>true</code> is this <code>Range</code> is considered to be
     * a "leaf node"; that is, if it has no <code>Range</code>-like members. This
     * method is intentionally declared abstract at the <code>Range</code> level
     * in order to force every <code>Range</code> subclass implementor to carefully
     * consider the nature of his class and whether it can truly be considered a
     * leaf node.
     * @return boolean
     */
  public  boolean isLeafNode(){
      //if my members != null AND it is a Range then I am not a leaf node
      if(null != getMembers() && getMembers() instanceof Range)
          return true;
      else
          return false;
  }

    /**
        Access method for the floor property.

        @return   the current value of the floor property
     */
    public T getFloor() {
        return floor;
    }

    /**
        Sets the value of the floor property.

        @param aFloor the new value of the floor property
     */
    public void setFloor(T aFloor) {
        floor = aFloor;
    }

    /**
        @param inCandidate
        @return boolean
        @roseuid 424ABA8B02A7
     */
    public boolean isEligible(Object inCandidate) {
        if (inCandidate == null) {
            return false;
        }
        Comparable aoi = null;
        try {
            aoi = getAttributeOfInterest(inCandidate);
        } catch (AggregationException ex) {
            //this exception has already been logged by the base class
            return false;
        }

        Comparable lower = getFloor();
        if (lower == null) {
            return false;
        }
        if (lower.compareTo(aoi) >= 0) {
            return true;
        } else {
            return false;
        }
    }

    /**
        @param inValue
        @roseuid 4252CE9D021D
     */
    public void setFloorAsString(String inValue) {

    }

    /**
        Clone this <code>Range</code> via "deep copy"
        @return java.lang.Object
        @roseuid 425574CB00FA
     */
    public Object clone() {
        UnaryRange base = (UnaryRange)super.clone();
        if (getMembers() instanceof Range) {
            Range memberToClone = (Range) getMembers();
            Range clonedMember = (Range) memberToClone.clone();
            base.setMembers(clonedMember);
        }
        return base;
    }

    public void setFloorAsDouble(T inFloor) {
        setFloor(inFloor);
    }

    public Range findMembers(Object attributeValue) throws AggregationException {
        if (null == getMembers()) {
            return null;
        }
        if (!(attributeValue instanceof String)) {
            String err = "Cannot process DN nodes of type '" +
                         attributeValue.getClass().getName();
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }
        String whomString = (String) attributeValue;

        DynamicRange answer = new DynamicRange();
        answer.setAggregationName(this.getAggregationName());
        Range srcBlueprint = getBlueprint();
        answer.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        answer.setDescription(this.getDescription());
        answer.setGetter(this.getGetter());
        answer.setMembers(new java.util.ArrayList());
        answer.setSubject(this.getSubject());
        answer.setSubjectClass("[none]");
        Pattern pattern = Pattern.compile(whomString);
        for (Iterator memberIter = getMembers().iterator(); memberIter.hasNext(); ) {
            Object currentMember = memberIter.next();
            //attributeValue can be an exact value or it can be a regexpr
            //of course, all exact values are also regexprs....
            //if the current member is a String then this findMember method works
            //fine; however, it will NOT work for any other sort of member object.
            //Therefore Range objects need a "matches" method that can be used
            //instead of equals
            if (currentMember instanceof String) {
                if (currentMember.equals(attributeValue)) {
                    answer.add(currentMember);
                }
            } else if (currentMember instanceof Range) {
                //currMember, being a Range, has a matches method, which we can use to
                //determine if currMember matches the current DN
                Range currMemberRange = (Range) currentMember;
                if (currMemberRange.matches(pattern)) {
                    answer.add(currMemberRange.deepCopy());
                }
            }
        }
        return answer;
    }

    /**
     * Use <code>dn</code> to traverse the <code>Range</code> rooted at this point.
     * Return the <code>Range</code> residing at the spot designated by <code>dn</code>
     * This is the most basic <code>findDN</code> implementation; several <code>Range</code>
     * subclasses override this method with significantly more sophisticated traversal
     * algorithms.
     * @param dn DN of the form <code><i>subjectClass=subjectValue[,subectClass=subjectValue...]
     * @return Range that most closely matches <code>dn</code>
     * @throws AggregationException
     */
    public Range findDN(DN dn) throws AggregationException {
        //if my subjectClass != dn.getNode[0].subjectClass then I need to pass the DN
        //down the tree to my subordinates
        String subjectClass = getSubjectClass();
        if (null == subjectClass || subjectClass.length() == 0) {
            String err = "subject class is not set, cannot process DN '" + dn +
                         "'";
            log.log(Level.SEVERE,err);
            throw new AggregationException(err);
        }

        DNNode<String, String> topNode = dn.peek();
        if (subjectClass.equals(topNode.LHS)) {
            //ok, this level of the agg tree matches the current top node of the DN,
            //so now we need to see if we can find topNode.RHS (its subject value) in this
            //Range's members, or perhaps the node itself "equals" topNode.RHS, but this
            //scenario is handled by the specific findMember method of the Node; for example,
            //a Counter node's findMember method returns itself as the "found member".
            Range foundMembers = findMembers(topNode.RHS);
            if (null == foundMembers) {
                //this findDN has reached a dead end
                return null;
            } else {
                foundMembers.setDnNode(topNode);
                //the node at which we just peek()ed is still lingering on the DN.  We
                //must remove it now...
                String logMsg = "DN is " + dn.toString() +
                                ", about to remove top element";
                log.log(Level.WARNING,logMsg);
                dn.poll();
                logMsg = "DN is now " + dn.toString();
                log.log(Level.WARNING,logMsg);

                //we found the member being sought.  If there are remaining nodes on the DN
                //then delegate the remaining search to the found node
                if (dn.peek() == (DNNode<String, String>)null) { //check to see if there is more on the DN...
                    //nothing left on the search spec, so we have found the thing being sought (I think....)
                    return foundMembers;
                } else {
                    Range nextFound = foundMembers.findDN(dn);
                    //if we get a null back then we need to backtrack by repushing the current node back onto the DN
                    if (nextFound == null) {
                        dn.offer(topNode);
                        return null;
                    } else {
                        nextFound.setDnNode(topNode);
                        return nextFound;
                    }

                }
            }
        } else {
            //the current topNode is not matched to our subjectClass, so try to match
            //it to one of our members
            //the following code sickens me, but it does work, and it is the most
            //expeditious route to implementing a proper findDN method.
            //If the current Range's members() are a Group then process the group
            //directly, otherwise process each member of members()
            if (getMembers() instanceof Group) {
                Group memberGroup = (Group) getMembers();
                Range nextFound = memberGroup.findDN(dn);
                nextFound.setDnNode(topNode);
                return nextFound;
            } else {
                for (Iterator memberIter = getMembers().iterator();
                                           memberIter.hasNext(); ) {
                    Object currMember = memberIter.next();
                    if (currMember instanceof Range) {
                        Range possibleRange = (Range) currMember;
                        Range nextFound = possibleRange.findDN(dn);
                        if (nextFound != null) {
                            nextFound.setDnNode(topNode);
                            return nextFound;
                        }
                    }
                }
            }
        }
        return null;
    }

    /**
     * Make a "deep copy" of this <code>Range</code> and all subordinate nodes
     * emanating from it.
     * @return Range
     */
    public Range deepCopy() throws AggregationException {
        UnaryRange copy = new UnaryRange();
        copy.setAggregationName(this.getAggregationName());
        copy.setAggregationType(this.getAggregationType());
        Range srcBlueprint = getBlueprint();
        copy.setBlueprint(null == srcBlueprint ? null : srcBlueprint.shallowCopy());
        copy.setDeclaredNodeStructure(this.getDeclaredNodeStructure());
        copy.setDescription(this.getDescription());
        copy.setDnNode(this.getDnNode());
        copy.setFloor(this.getFloor());
        copy.setGetter(this.getGetter());
        copy.setInquiryMethod(this.getInquiryMethod());
        copy.setPassivator(this.getPassivator());
        copy.setPersistenceHandle(this.getPersistenceHandle());
        copy.setSerializable(this.isSerializable());
        copy.setSubject(this.getSubject());
        copy.setSubjectClass(this.getSubjectClass());
        copy.setContext(this.getContext());

        //The following is predicated on the fact that RangeSurgeons are completely
        //stateless and can therefore be shared among many Range clients
        copy.setRangeSurgeon(this.getRangeSurgeon());


        //now deepcopy the members individually to avoid pointing to an existing
        //object and subsequently accidentally modifying it
        List srcMembers = this.getMembers();
        if (null == srcMembers) {
            copy.setMembers((List)null);
        } else {
            Class srcMembersClass = srcMembers.getClass();
            List copyList = (List) shallowCopy(srcMembersClass, srcMembers);
            for (Iterator memberIter = srcMembers.iterator();
                                       memberIter.hasNext(); ) {
                Object currSrcMember = memberIter.next();
                if (currSrcMember instanceof Range) {
                    Range srcRange = (Range) currSrcMember;
                    Range copyRange = srcRange.deepCopy();
                    copyRange.setContext(this.getContext());
                    copyList.add(copyRange);
                }
            }
            copy.setMembers(copyList);
        }
        return copy;
    }

    public boolean equals(Object inRHS) {
        if (null == inRHS) {
            return false;
        }
        if (inRHS instanceof UnaryRange) {
            UnaryRange lhs = this;
            UnaryRange rhs = (UnaryRange) inRHS;
            Comparable lhsFloor = lhs.getFloor();
            Comparable rhsFloor = rhs.getFloor();

            if (null == lhsFloor && null == rhsFloor) {
                return true;
            }
            if (null != lhsFloor && null == rhsFloor) {
                return false;
            }
            if (null == lhsFloor && null != rhsFloor) {
                return false;
            }
            return lhsFloor.equals(rhsFloor);
        } else {
            return false;
        }
    }
}
