package org.quintet.monk;

import java.util.Iterator;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.Stack;

/**
A <code>Builder</code> is a wrapper around a factory method and the formal
arguments that are to be passed to the method when using it to construct an
entity instance.
A <code>Builder</code> may be <i>deferred</i>, which means that its factory
method and target object are dynamically resolved at run time.  Deferred binding
allows a <code>Builder</code> to refer to "parental" objects that have
previously been built during an <code>Monk</code> run.
 */
public class Builder
{
   private String targetName;

   /**
   The name of the factory method that this <code>Builder</code> uses to construct
   <subject> instances.
    */
   private String factoryMethodName = null;


   private Object target;

   /**
   if true then this builder's factory method is a constructor, not a factory
   method
    */
   private String ctor = "false";

   /**
   Indicates if this <code>Builder</code> should statically (at
   <code>Monk</code> startup time) or dynamically (at the instant before
   the <code>factoryMethod</code> is invoked) bind to its factory method; the
   default is <code>false</code>.
    */
   private boolean isDeferred = false;

   /**
   If <code>true</code> then this <code>Blueprints</code>' factory class is
   abstract.  This further implies that the factory method must be static, because
   there is no way to instantiate an abstract factory.
    */
   private boolean abstractFactoryClass = false;

   /**
   Indicates if this <code>Builder</code>'s factory method is static.  This flag
   and the <code>isCtor</code> flag are obviously mutually exclusive
    */
   private boolean staticMethod = false;

   /**
   <code>factoryMethod</code> is the method that is called to construct new instances of <code>ClassMetaData.subject</code>
    */
   private Method factoryMethod;

   /**
   The formal arguments passed to this <code>Builder</code>'s factory method
    */
   private Collection formalArguments = new java.util.ArrayList ();
   private static final Logger log = Logger.getLogger("org.quintet.monk");

   /**
   @roseuid 40802FB60222
    */
   public Builder()
   {

   }

   /**
   Access method for the targetName property.

   @return   the current value of the targetName property
    */
   public String getTargetName()
   {
    return targetName;
   }

   /**
   Sets the value of the targetName property.

   @param aTargetName the new value of the targetName property
    */
   public void setTargetName(String aTargetName)
   {
    targetName = aTargetName;
   }

   /**
   Access method for the factoryMethodName property.

   @return   the current value of the factoryMethodName property
    */
   public String getFactoryMethodName()
   {
    return factoryMethodName;
   }

   /**
   Sets the value of the factoryMethodName property.

   @param aFactoryMethodName the new value of the factoryMethodName property
    */
   public void setFactoryMethodName(String aFactoryMethodName)
   {
    factoryMethodName = aFactoryMethodName;
   }



   /**
   Determines if the abstractFactoryClass property is true.

   @return   <code>true<code> if the abstractFactoryClass property is true
    */
   public boolean getAbstractFactoryClass()
   {
      return abstractFactoryClass;
   }

   /**
   Sets the value of the abstractFactoryClass property.

   @param aAbstractFactoryClass the new value of the abstractFactoryClass property
    */
   public void setAbstractFactoryClass(boolean aAbstractFactoryClass)
   {
      abstractFactoryClass = aAbstractFactoryClass;
   }

   /**
   Determines if the staticMethod property is true.

   @return   <code>true<code> if the staticMethod property is true
    */
   public boolean getStaticMethod()
   {
      return staticMethod;
   }

   /**
   Sets the value of the staticMethod property.

   @param aStaticMethod the new value of the staticMethod property
    */
   public void setStaticMethod(boolean aStaticMethod)
   {
      staticMethod = aStaticMethod;
   }

   /**
   Access method for the factoryMethod property.

   @return   the current value of the factoryMethod property
    */
   public Method getFactoryMethod()
   {
    return factoryMethod;
   }

   /**
   Sets the value of the factoryMethod property.

   @param aFactoryMethod the new value of the factoryMethod property
    */
   public void setFactoryMethod(Method aFactoryMethod)
   {
    factoryMethod = aFactoryMethod;
   }

   /**
   Access method for the formalArguments property.

   @return   the current value of the formalArguments property
    */
   public Collection getFormalArguments()
   {
    return formalArguments;
   }

   /**
   Sets the value of the formalArguments property.

   @param aFormalArguments the new value of the formalArguments property
    */
   public void setFormalArguments(Collection aFormalArguments)
   {
    formalArguments = aFormalArguments;
   }

   /**
   Return the <code>Class</code>es of this <code>Builder</code>'s formal
   arguments.  They are returned as an array, which is suitable for passing to
   <code>Class.getMethod</code>.
   @return java.lang.Class[]
   @roseuid 4086901503C8
    */
   public Class[] getFormalArgumentClasses()
   {
    Collection args = getFormalArguments();
    if (args == null) {
      return null;
    }
    Class[] answer = new Class[args.size()];
    if (args.size() == 0) {
      return answer;
    }
    int idx = 0;
    for (Iterator argIter = args.iterator(); argIter.hasNext(); ) {
      Attribute currArg = (Attribute) argIter.next();
      answer[idx++] = currArg.getJavaClass();
    }
    return answer;
   }

   /**
   Return true if the XML attribute named <code>inArgName</code> name is mentioned
   as a formal argument of <code>factoryMethod</code>
   @param inArgName - The name of the attribute being tested to see if it is part
   of <code>factoryMethod</code>'s formal arguments
   @return boolean
   @roseuid 408D43560276
    */
   public boolean isFormalArgument(String inArgName)
   {
    Collection args = getFormalArguments();
    for (Iterator argIter = args.iterator(); argIter.hasNext(); ) {
      Attribute currAttribute = (Attribute) argIter.next();
      if (currAttribute.getXmlName() != null && currAttribute.getXmlName().equals(inArgName)) {
        return true;
      }
    }
    return false;
   }

   /**
   Access method for the target property.
   @return   the current value of the target property
   @roseuid 40E9B068034D
    */
   public Object getTarget()
   {
    return target;
   }

   /**
   Sets the value of the target property.
   @param aTarget the new value of the target property
   @roseuid 40E9B0680357
    */
   public void setTarget(Object aTarget)
   {
    target = aTarget;
   }

   /**
   Return true if this builder's factory method is a constructor
   @return true
   @roseuid 40E9B0690005
    */
   public boolean isCtor()
   {
    return (ctor.equalsIgnoreCase("true"));
   }

   /**
   @return java.lang.String
   @roseuid 40E9B069000F
    */
   public String getIsCtor()
   {
    return ctor;
   }

   /**
   @param inValue
   @roseuid 40E9B0690023
    */
   public void setIsCtor(String inValue)
   {
    ctor = inValue;
   }

   /**
   Returns the string manifestation of the <code>isDeferred</code> flag
   @return java.lang.String
   @roseuid 4265345A0333
    */
   public String getDeferred()
   {
    return (isDeferred ? "True" : "False");
   }

   /**
   Set the <code>isDeferred</code> flag to the value represented by
   <code>inNewValue</code>
   @param inNewValue - The <code>isDeferred</code> flag's new value - "Y", "y",
   "1", "T", and "t" all map to <code>true</code>, and all other values map to
   <code>false</code>
   @roseuid 426534830115
    */
   public void setDeferred(String inNewValue)
   {
    if("Y".equals(inNewValue) ||
       "y".equals(inNewValue) ||
       "1".equals(inNewValue) ||
       "T".equals(inNewValue) ||
       "T".equals(inNewValue) ||
       "True".equals(inNewValue) ||
       "true".equals(inNewValue) ||
       "Yes".equals(inNewValue) ||
       "yes".equals(inNewValue)){
      isDeferred = true;
    }else{
      isDeferred = false;
    }
   }

   /**
   Return the context-sensitive factory method contained within the context defined
   by <code>contextStack</code>.
   @param contextStack - The stack of objects that define the context in which the
   factory method is being sought; for example, if <code>contextStack</code>
   contains a <code>WorkBook</code>, and a <code>Vector</code>, and this
   <code>Builder</code>'s factory method name is "../../createSheet", then a
   <code>Method</code> pointing to the <code>WorkBook</code>'s
   <code>createSheet</code> method will be returned.
   @return   the current value of the factoryMethod property
   @throws org.quintet.monk.MonkException
   @roseuid 426544C60082
    */
   public Method getFactoryMethod(Stack contextStack) throws MonkException
   {
     Object context = null;
     String baseMethodName = null;
     String formalArgName = getFactoryMethodName();
     java.util.Vector cache = null;
     //if the factory method name contains "../" then it is contextual
     if(formalArgName.startsWith("../")){
       cache = new java.util.Vector(contextStack.size());
       java.util.StringTokenizer lex = new java.util.StringTokenizer(
           formalArgName,
           "/");
       String currToken = null;
       while (lex.hasMoreTokens()) {
         currToken = lex.nextToken();
         if (currToken.equals("..") && contextStack.isEmpty() == false) {
           context = contextStack.pop();
           //save the object so we can push it back later
           cache.add(context);
         }
       }
       if (context == null) {
         //factory method referred to a context (via ../) that does not exist
         String err = "factory method '" + formalArgName + " refers to non-exist parent object(s)";
         log.log(Level.SEVERE,err);
         throw new MonkException(err);
       }
       baseMethodName = currToken;
     }else{
       context = getTarget();
       baseMethodName = formalArgName;
     }
      //lex should now be pointing to the actual arg's name
      Method formalArgGetter = null;
      try {
        Class[] formalArgs = getFormalArgumentClasses();
        if(getStaticMethod() && context instanceof Class){
          formalArgGetter = ((Class)context).getMethod(baseMethodName,formalArgs);
        }else{
          formalArgGetter = context.getClass().getMethod(baseMethodName,
              formalArgs);
        }
        return formalArgGetter;
      }
      catch (SecurityException ex) {
        String err = "SecurityException when accessing formal arg " + formalArgName;
        log.log(Level.SEVERE,err,ex);
        return null;
      }
      catch (NoSuchMethodException ex) {
        String err = "NoSuchMethodException when accessing formal arg " + formalArgName;
        //log.debug(err,ex);
        return null;
      }finally{
        int cacheIdx = (cache == null ? -1: cache.size() -1 );
        while(cacheIdx >= 0){
          contextStack.push(cache.elementAt(cacheIdx--));
        }
      }
   }

   /**
   Return the contextualized target object tin which this <code>Builder</code>'s
   factory method is implemented.
   @param contextStack - The context in which this <code>Builder</code>'s
   targetName makes sense; for example, if the context contains a
   <code>Workbook</code> and a <code>java.util.Vector</code>, and this
   <code>Builder</code>'s <code>targetName</code> == "../.." then the
   <code>Builder</code>'s target object is the <code>Workbook</code> at the
   bottom of the stack.
   @return java.lang.Object
   @throws org.quintet.monk.MonkException
   @roseuid 42654D250026
    */
   public Object getTarget(Stack contextStack) throws MonkException
   {
     Object context = null;
     String formalArgName = getTargetName();
     java.util.Vector cache = null;
     //if the factory method name contains "../" then it is contextual
     if(formalArgName.startsWith("../")){
       cache = new java.util.Vector(contextStack.size());
       java.util.StringTokenizer lex = new java.util.StringTokenizer(
           formalArgName,
           "/");
       String currToken = null;
       while (lex.hasMoreTokens()) {
         currToken = lex.nextToken();
         if (contextStack.isEmpty() == false) {
           context = contextStack.pop();
           //save the object so we can push it back later
           cache.add(context);
         }else{
           //nothing to pop, and we REALLY need to pop something.
           String err = "Reached the end of the context stack when processing " + formalArgName;
           log.log(Level.SEVERE,err);
           throw new MonkException(err);
         }
       }
       if (context == null) {
         //factory method referred to a context (via ../) that does not exist
         String err = "factory method '" + formalArgName + " refers to non-exist parent object(s)";
         log.log(Level.SEVERE,err);
         throw new MonkException(err);
       }

       int cacheIdx = (cache == null ? -1: cache.size() -1 );
       while(cacheIdx >= 0){
         contextStack.push(cache.elementAt(cacheIdx--));
       }
       return context;
     }else{
       return getTarget();
     }
   }

   /**
   Access method for the ctor property.
   @return   the current value of the ctor property
   @roseuid 42D66BE901B8
    */
   public String getCtor()
   {
      return ctor;
   }

   /**
   Sets the value of the ctor property.
   @param aCtor the new value of the ctor property
   @roseuid 42D66BE901CC
    */
   public void setCtor(String aCtor)
   {
      ctor = aCtor;
   }

   /**
   Determines if the isDeferred property is true.
   @return   <code>true<code> if the isDeferred property is true
   @roseuid 42D66BE901EA
    */
   public boolean getIsDeferred()
   {
      return isDeferred;
   }

   /**
   Sets the value of the isDeferred property.
   @param aIsDeferred the new value of the isDeferred property
   @roseuid 42D66BE901FE
    */
   public void setIsDeferred(boolean aIsDeferred)
   {
      isDeferred = aIsDeferred;
   }
}
