package org.quintet.monk;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;
import java.lang.reflect.Method;
import java.io.Serializable;
import java.util.logging.Logger;
import java.util.logging.Level;


/**
 <code>ClassMetaData</code> houses meta-info about a class.  Included in the
 meta-info is:<b>
 <li>The class that the meta-info describes
 <li>The constructor to be used by <code>Monk</code> when building
 instances of the class
 <li> The arguments to pass to the constructor
 @author Peter Thaggard
 @since 2001
 */
public class ClassMetaData
    implements Serializable {

  private static final Logger log = Logger.getLogger("org.quintet.monk");
  /**
   The fully-qualified name of this <code>ClassMetaData</code>'s subject class.
   */
  private String subjectName = null;

  /**
      The XML name for <code>subjectClass</code>; for example, a
   <code>subjectClass</code> of "Dealer" might be associated with an xmlName of
      "dealer"
   */
  private String xmlName;

  /**
      A <code>Class</code> can be associated with an exception destination.
      <code>ClassMetaData</code> clients, such as <code>Monk</code>,
      can publish error messages to this destination.
   */
  private String exceptionDestinationName = null;

  /**
      <code>subject</code> is the <code>Class</code> that this <code>ClassMetaData</code> describes.
   */
  private Class subjectClass;

  /**
      <code>factoryMethod</code> is the "constructor" of choice for building instances of <code>subject</code>.  For example, an <code>ClassMetaData</code> called <code>PriceResponseMetaData</code> might refer to <code>org.quintet.icm.prices.PriceResponse</code> as its <code>subject</code> and it might refer to <code>PriceResponseHome.create()</code> as its factory method.
   */
  private Builder builder;
  private Collection constructorArgNames = new ArrayList();

  /**
      Some attributes do not conform to the JavaBeans property pattern.  In order to set these attributes ClassMetaData provides a collection of Attributes.  Each of these Attributes designates a setter to use when setting its concomitant attribute.  For example, one could construct an Attribute named "CusipPrices" and associate it with a setter named "setMyPrices".
   */
  private Collection attributes = new ArrayList();

  /**
      @roseuid 40802E4003AE
   */
  public ClassMetaData() {

  }

  /**
      Access method for the xmlName property.

      @return   the current value of the xmlName property
   */
  public String getXmlName() {
    return xmlName;
  }

  /**
      Sets the value of the xmlName property.

      @param aXmlName the new value of the xmlName property
   */
  public void setXmlName(String aXmlName) {
    xmlName = aXmlName;
  }

  /**
      Access method for the subjectClass property.

      @return   the current value of the subjectClass property
   */
  public Class getSubjectClass() {
    if(null == subjectClass){
      String warn = "I told you that I would try later, and I will now do so";
      log.warning(warn);
      try {
        setSubjectClass(Class.forName(getSubjectName()));
      }
      catch (ClassNotFoundException ex) {
        String err = "Still unable to do forName() on '" + getSubjectName() + "'";
        log.log(Level.SEVERE,err,ex);
        return null;
      }
    }
    return subjectClass;
  }

  /**
      Sets the value of the subjectClass property.

      @param aSubjectClass the new value of the subjectClass property
   */
  public void setSubjectClass(Class aSubjectClass) {
    subjectClass = aSubjectClass;
  }

  /**
      Access method for the builder property.

      @return   the current value of the builder property
   */
  public Builder getBuilder() {
    return builder;
  }

  /**
      Sets the value of the builder property.

      @param aBuilder the new value of the builder property
   */
  public void setBuilder(Builder aBuilder) {
    builder = aBuilder;
  }

  /**
      Access method for the attributes property.

      @return   the current value of the attributes property
   */
  public Collection getAttributes() {
    return attributes;
  }

  /**
      Sets the value of the attributes property.

      @param aAttributes the new value of the attributes property
   */
  public void setAttributes(Collection aAttributes) {
    attributes = aAttributes;
  }

  /**
      Access method for the constructorArgNames property.
      @return   the current value of the constructorArgNames property
      @roseuid 4082FB9C0146
   */
  public Collection getConstructorArgNames() {
    return builder.getFormalArguments();
  }

  /**
   Return the "constructor" (which is fact probably not a constructor at all, but
      a home-interface create method) that is capable of creating instances of
      <code>subject</code>.
      @return java.lang.reflect.Method
      @roseuid 4082FFF800FB
   */
  public Method getFactoryMethod() {
    return getBuilder().getFactoryMethod();
  }

  /**
   Return the target object upon which <code>factoryMethod</code> can legally be
      invoked.  For example, if <code>factoryMethod</code> is
      "PriceResponseHome.create(String dealerID)", then
      <code>getFactoryMethodTarget</code> might return a
      <code>org.quintet.icm.pricing.PriceResponseHome</code> instance.
      @return java.lang.Object
      @roseuid 4083072903AB
   */
  public Object getFactoryMethodTarget() {
    return getBuilder().getTarget();
  }

  /**
   Set this <code>ClassMetaData</code>'s subject name to <code>inName</code>.
      This also results in <code>subject</code> being set via
      <code>Class.forName(inName)</code>.
   @param inName - The fully-qualified name of this <code>ClassMetaData</code>'s
      subject class
      @roseuid 408416CC01BC
   */
  public void setSubjectName(String inName) {
    subjectName = inName;
    try {
      setSubjectClass(Class.forName(inName));
    }
    catch (ClassNotFoundException CNFE) {
      String fatal = "ClassNotFoundException when performing Class.forName() on '" + inName + "'.  Will try again later.";
      log.log(Level.SEVERE,fatal,CNFE);
    }
  }

  /**
   Return the fully-qualified name of this <code>ClassMetaData</code>'s subject
      class.
      @return java.lang.String
      @roseuid 408417380031
   */
  public String getSubjectName() {
    return subjectName;
  }

  /**
      Sets the value of the constructorArgNames property.
   @param aConstructorArgNames the new value of the constructorArgNames property
      @roseuid 40E9B06A033C
   */
  public void setConstructorArgNames(Collection aConstructorArgNames) {
    constructorArgNames = aConstructorArgNames;
  }

  /**
      Access method for the exceptionDestinationName propety.
   @return String containing the exception destination name, or null if it is not
      set
      @roseuid 4128A9A903DA
   */
  public String getExceptionDestinationName() {
    return exceptionDestinationName;
  }

  /**
      Mutator for the <code>exceptionDestinationName</code> property.
   @param String inValue the new value for <code>exceptionDestinationName</code>
      @param inValue
      @roseuid 4128A9A903E4
   */
  public void setExceptionDestinationName(String inValue) {
    exceptionDestinationName = inValue;
  }

  /**
      Am I equal to <code>inOther</code>?
      @param inOther - The object to which I am being compared
      @return boolean
      @roseuid 4128D814005C
   */
  public boolean equals(Object inOther) {
    if (inOther instanceof ClassMetaData) {
      ClassMetaData that = (ClassMetaData) inOther;
      Class thisClass = this.getSubjectClass();
      Class thatClass = that.getSubjectClass();
      if (thisClass == null && thatClass == null) {
        return true;
      }
      if (thisClass == null && thatClass != null) {
        return false;
      }
      if (thisClass != null && thatClass == null) {
        return false;
      }
      return thisClass.equals(thatClass);
    }
    else {
      return false;
    }
  }

  /**
      @return int
      @roseuid 4128D8610249
   */
  public int hashCode() {
    Class thisClass = getSubjectClass();
    if (thisClass == null) {
      return 0;
    }
    else {
      return thisClass.hashCode();
    }
  }

  /**
   Seek an <code>Attribute</code> named <code>inAttributeName</code> (the name is
   from the java namespace, not the XML namespace).  Return the Attribute if
      found, otherwise return null.
      @param inAttributeName - The name (in the java namespace) of the
      <code>Attribute</code> being sought
      @return Attribute
      @roseuid 41483E410025
   */
  public Attribute getAttribute(String inAttributeName) {
    if (inAttributeName == null) {
      return null;
    }
    Collection attrs = getAttributes();
    if (attrs == null || attrs.size() == 0) {
      return null;
    }
    for (Iterator attrIter = attrs.iterator(); attrIter.hasNext(); ) {
      Attribute currAttr = (Attribute) attrIter.next();
      String currAttrName = currAttr.getXmlName();
      if (currAttrName != null && currAttrName.equals(inAttributeName)) {
        return currAttr;
      }
    }
    return null;
  }
}


