package org.quintet.monk;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Stack;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 <code>Monk</code> transcribes XML documents into java objects.
 More specifically, it translates XML elements and attributes into concomitant
 Java objects and members, respectively.  XML elements are related via
 composition.  These relationships are maintained when those elements are
 converted to java Objects.  for example, the following XML document implies
 that a price-response is associated with a collection of instrument-prices:
 <p>
 &lt;price-response id="123"&gt; <br>
 &nbsp;&nbsp;&nbsp;&lt;instrument-prices&gt;<br>
 &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;instrument-price id="101"
 price="100"&gt;<br>
 &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;instrument-price id="401"
 price="101.1"&gt;<br>
 &nbsp;&nbsp;&nbsp;&lt;/instrument-prices&gt;<br>
 &lt;/price-response&gt; <br>
 <p>
 This XML will converted to a <code>PriceReponse</code> containing a collection
 of <code>InstrumentPrice</code> instances named
 <code>instrumentPrices</code>.<p>
 <h4>Implicit and Explicit Translation</h4><p>
 <code>Monk</code> translates between XML and java vocabularies.  These
 vocabularies can be explicitly provided via a set of <i>blueprints</i>, or they
 can be inferred by the JavaBeans property pattern.
 <h4>ExplicitTranslation via Blueprints</he><p>
 <code>Monk</code> needs to know how to construct new object instances;
 some java classes have "no-arg" constructors, so <code>Monk</code> can
 instantiate them without any additional meta-data.  For example,
 <code>Monk</code> can create a <code>String</code> simply by invoking
 <code>String.String()</code>.  Such no-arg constructors are ubiquitous in
 lower-level J2SE classes such as <code>String</code>, but they are rarely seen
 in "high-level" classes. Few less-abstract classes, such as the business
 entities that pervade most corporations, offer no-arg constructors, simply
 because it rarely makes sense to instantiate a business object without
 injecting it with some sort of initial state; for example, one cannot create a
 <code>User</code> without at least supplying a name or ID for the poor
 object.<p>
 <code>Monk</code> utilizes "blueprints" to construct objects via any
 means other than no-arg constructors.
 A blueprint is an instance of a <code>ClassMetaData</code>.  It describes how to
 convert a given XML element
 into a java object.  A <code>&lt;class-meta-data&gt;</code> blueprints entry
 contains the following information:<p>
 <table border=0 cellspacing=3 cellpadding=3 >
 <tr bgcolor="#ccccff">
 <th align=left>Attribute/Element
 <th align=left>Required?
 <th align=left>Instance-of
 <th align=left>Example
 <th aligh=left>Purpose
 <tr>
 <td><code>xml-name</code>
 <td>Yes
 <td><code>java.lang.String</code>
 <td><code>"price-response"</code>
 <td>The XML element that can be translated via this <code>blueprint</code>
 <tr>
 <td><code>subject-name</code>
 <td>Yes
 <td><code>java.lang.String</code>
 <td><code>"org.quintet.icm.pricing.PriceResponse"</code>
 <td>The class of the object into which the XML element will be translated.
 <tr>
 <td><code>exception-destination-name</code>
 <td>No
 <td><code>java.lang.String</code>
 <td><code>"java:comp/env/jms/price-response-exception-queue"</code>
 <td>The JMS destination to which ill-formed instances of the given XML element
 will be sent; if <code>Monk</code> cannot parse a node, then it will
 send the errant node to this destination
 <tr>
 <td><code>builder</code>
 <td>Yes
 <td><code>element</code>
 <td><code>&lt;builder ...&gt;</code>
 <td>The class, method, and arguments used to construct an instance of
 <code>subject-name</code> from <code>xml-name</code>
 </table>
 <p>
 <h4>Builders</h4><p>
 Every <code>&lt;class-meta-data&gt;</code> contains a
 <code>&lt;builder&gt;</code>.  The &lt;builder&gt; describes a
 constructor or factory method that is capable of creating a
 <code>subject-name</code> instance on
 <code>Monk</code>'s behalf.  A &lt;builder%gt; element consists of the
 following:
 <table border=0 cellspacing=3 cellpadding=3 >
 <tr bgcolor="#ccccff">
 <th align=left>Attribute/Element
 <th align=left>Required?
 <th align=left>Instance-of
 <th align=left>Example
 <th aligh=left>Purpose
 <tr>
 <td><code>factory-method-name</code>
 <td>Yes
 <td><code>java.lang.String</code>
 <td><code>"create"</code>
 <td>The name of the constructor/factory-method being described (set this to
 "new" if the method is a constructor)
 </table>
 */
public class Monk {

  /**
      Monk's most basic blueprints are stored in Monk.props
   */
  private static String propFileName = "Monk.props";

  /**
   configLoaded is set to true when <code>Monk</code>'s configuration has
      been loaded*
   */
  private  boolean configLoaded = false;

  /**
      all log entries are published via <code>log</code>*
   */
  private static final Logger log = Logger.getLogger("org.quintet.monk");

  /**
      An <code>Monk</code> creates instances of classes defined within a given domain.  The instructions for creating domain instances are defined in the <code>Monk</code>'s <code>domain</code> attribute, which is an instance of <code>Blueprints</code>
   */
  private Blueprints domain = null;

  /**
      mapping of Dom NodeNames to classes in the application domain.
      For example, a node with getNodeName() == "organization"
      might be mapped to org.quintet.cbo.oim.organization.Organization.
      The <code>classMap</code> holds generic mappings that pertain to all <code>Monk</code>s,
      therefore it is a static attribute.
   */
  private  Map classMap = new HashMap();

  /**
      Create a <code>Monk</code> that builds objects according to
      <code>domainBlueprints</code>'s instructions.
      @param domainBlueprints - The <code>Blueprints</code> by which this
      <code>Monk</code> will construct domain objects
      @throws org.quintet.monk.MonkException
      @roseuid 428CC2280335
   */
  private Monk(Blueprints domainBlueprints) throws
      MonkException {
    loadConfiguration();
    setDomain(domainBlueprints);
  }

  /**
      @throws org.quintet.monk.MonkException
      @roseuid 428CBF1901BE
   */
  public Monk() throws MonkException {
    setDomain(new Blueprints());
    loadConfiguration();
  }

  public Monk(Class bootstrapConfigClass){
    setDomain(new Blueprints());
    try {
      loadConfiguration(bootstrapConfigClass);
    }
    catch (MonkException ex) {
      String fatal = "unable to bootstrap a Monk from " + bootstrapConfigClass.getName();
      throw new java.lang.IllegalStateException(fatal);
    }
  }
  /**
      @param domainBlueprints
      @throws org.quintet.monk.MonkException
      @roseuid 428A01C2019A
   */
  private Monk(InputStream domainBlueprints) throws
      MonkException {

  }

  /**
      Access method for the domain property.

      @return   the current value of the domain property
   */
  public Blueprints getDomain() {
    return domain;
  }

  /**
      Sets the value of the domain property.

      @param aDomain the new value of the domain property
   */
  public void setDomain(Blueprints aDomain) {
    domain = aDomain;
  }


  private void loadConfiguration() throws MonkException{
    loadConfiguration(org.quintet.monk.Monk.class);
  }
  /**
   Load <code>Monk</code>'s fundamental class-to-element mappings from
      Monk.props.
      @throws org.quintet.monk.MonkException
      @see java.lang.Class#getResourceAsStream
      @roseuid 407FF32B02CC
   */
  private void loadConfiguration(Class thisClass) throws MonkException {
    InputStream fis = null;
    try {
      //config uses Monk, so Monk cannot use config
      //to load in the list of known element-to-class mappings.
      //Instead, use a properties file that is co-located with
      //Monk's .class file.  Class.getResourceAsStream()
      //expects to find the properties file wherever the ClassLoader
      //found the class; if the class is in a jar/zip, then the property file
      //is assumed to be there too.


      fis = thisClass.getResourceAsStream(propFileName);
      if (fis == null) {
        String err =
            ":unable to local Monk.props file, cannot proceed";
        log.log(Level.SEVERE,err);
        throw new MonkException(err);
      }
      Properties classProps = new Properties();
      classProps.load(fis);
      fis.close();
      Iterator mapIter = classProps.keySet().iterator();
      while (mapIter.hasNext()) {
        String xmlName = (String) mapIter.next();
        String className = (String) classProps.get(xmlName);
        try {
          Class c = Class.forName(className);
          classMap.put(xmlName, c);
        }
        catch (ClassNotFoundException CNFE) {
          String err =
              ":ClassNotFoundException when searching for " + className + ":" +
              CNFE;
          log.log(Level.SEVERE,err,CNFE);
          throw new MonkException(err, CNFE);
        }
      }
    }
    catch (IOException e) {
      String err =
          ":caught IOException when reading Monk.props:" + e;
      log.log(Level.SEVERE,err,e);
      throw new MonkException(err, e);
    }finally{
      if(fis != null){
        try{
          fis.close();
        }catch(Throwable eatIt){}
      }
    }
    configLoaded = true;
  }

  /**
   Given a DOM Node (<code>inNode</code>) and a set of <code>blueprints</code>,
   find the <code>Class</code> within <code>blueprints</code> that maps to the
      <code>nodeName</code>
      found in <code>inNode</code>
   @param inNode Node The Dom <code>Node</code> for which an appropriate <code> is
      being sought
   @param blueprints Map The mapping of node names to <code>ClassMetaData</code>
      instances
      @return java.lang.Class
      @throws MonkException
      @throws org.quintet.monk.MonkException
      @roseuid 40E9B0710274
   */
  private Class findBestClass(Node inNode, Map blueprints) throws
      MonkException {
    String thisMethod = ":findBestClass";
    String nodeName = inNode.getNodeName();
    if(nodeName != null && nodeName.equals("#comment")){
      String info = "skipping #comment node";
      log.info(info);
      return null;
    }
    ClassMetaData metaData = (ClassMetaData) blueprints.get(nodeName);
    if (metaData == null) {
      String err = ":no metadata for XML node " + nodeName;
      log.log(Level.SEVERE,err);
      throw new MonkException(err);
    }
    Class answer = metaData.getSubjectClass();
    if (answer == null) {
      String err = ":null subjectClass in " + nodeName +
          "'s metadata";
      log.log(Level.SEVERE,err);
      MonkConfigurationException ofe = new
          MonkConfigurationException(err);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    return answer;
  }

  /**
   Examine a node and determine the class that best represents that node.  Most
      <code>Node</code>s are
   mapped to their corresponding <code>class</code>es by the node's element name.
      However, some nodes
      contain an attribute named "meta-name", which contains the actual class
      corresponding to the node.
   @param Node inNode The <code>Node</code> whose corresponding class is being
      sought
      @param inNode
      @exception MonkException
      @return java.lang.Class
      @throws org.quintet.monk.MonkException
      @roseuid 40E9B07103DC
   */
  private Class findBestClass(Node inNode) throws MonkException {

    if (inNode == null) {
      String err = ":inNode is null";
      log.log(Level.SEVERE,err);
      MonkConfigurationException ofe = new
          MonkConfigurationException(err);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    //There may be a meta-name in the node, i.e., the name of the class to which the node's contents are logically related
    NamedNodeMap nnm = inNode.getAttributes();
    if(nnm != null){
      Node metaNode = nnm.getNamedItem("meta-name");
      if (metaNode != null) {
        try {
          return Class.forName(metaNode.getNodeValue());
        }
        catch (ClassNotFoundException CNFE) {
          String err = ":meta-name " + metaNode.getNodeValue() +
              " is an unknown Class";
          log.log(Level.SEVERE,err,CNFE);
          MonkException ofe = new MonkException(err, CNFE);
          ofe.setErrantNode(inNode);
          throw ofe;
        }
      }
    }else{
      String info = "Encountered a node with no attributes (probably a comment)";
      log.info(info);
      return null;
    }
    //no meta-name, so try to lookup the class that is mapped to inName.getNodeName
    String nodeName = inNode.getNodeName();
    //if nodeName = "java-lang-string" then return the String class
    if (nodeName.equals("java-lang-string")) {
      return java.lang.String.class;
    }
    Class correspondingClass = (Class) classMap.get(nodeName);
    if (correspondingClass == null) {
      String err =
          ":unable to find corresponding class for node named " + nodeName;
      log.log(Level.SEVERE,err);
      MonkException ofe = new MonkException(err);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    return correspondingClass;
  }

  /**
      Create a java object from <code>inNode</code>'s information.  Use
      <code>blueprints</code>
      to translate from the XML namespace to the java namespace.
   @param Node inNode -- The XML node being translated into the java namespace
      @param Map blueprints -- The blueprints to use when tranlating
      <code>inNode</code>
      @param inNode
      @param blueprints
      @return java.lang.Object
      @throws MonkException
      @throws org.quintet.monk.MonkException
      @roseuid 40E9B072009E
   */
  public Object create(Node inNode, Map blueprints) throws
      MonkException {
    return create(inNode, blueprints, new Stack());
  }

  /**
   Given a DOM Node, create a corresponding manifestation of it in the application
      domain.
   For example, a &lt;price-response%gt; Node might come into <code>create()</code,
      and a <code>PriceResponse</code> object might be created.
      @param Node.inNode The DOM manifestation of the object being created
      @param inNode
      @return Object interface to the Node's application-domain manifestation
      @throws org.quintet.monk.MonkException
      @roseuid 40E9B0720149
   */
  public Object create(Node inNode) throws MonkException {
    if (inNode == null) {
      String err = ":null Node";
      log.log(Level.SEVERE,err);
      throw new MonkException(err);
    }
    if (configLoaded == false) {
      loadConfiguration();
    }
    try {
      //user inNode's NodeName parameter to determine what type of
      //Object should be created
      Class correspondingClass = findBestClass(inNode);
      //if the class is simple java.lang.String then
      //   create a new string
      //   copy inNode's value into it
      //   return the new string
      if (correspondingClass.equals(java.lang.String.class)) {
        String answerString = inNode.getAttributes().getNamedItem("value").
            getNodeValue();
        return answerString;
      }
      //if the node has a "ctor-value" attribute then call corresponding class's
      //<ctor>(String) ctor
      NamedNodeMap attribs = inNode.getAttributes();
      Node ctorArg = attribs.getNamedItem("ctor-argument");
      Object newNode = null;
      if (ctorArg != null) {
        newNode = constructFromString(ctorArg.getNodeValue(),
                                      correspondingClass);
      }
      else {
        newNode = correspondingClass.newInstance();
      }
      //fill in its attributes
      NamedNodeMap nnm = inNode.getAttributes();
      for (int attrIdx = 0; attrIdx < nnm.getLength(); attrIdx++) {
        Node currAttr = nnm.item(attrIdx);
        if (currAttr.getNodeName().equals("meta-name") == false &&
            currAttr.getNodeName().equals("ctor-argument") == false) {
          Object currValue = currAttr.getNodeValue();
          setAttribute(correspondingClass, newNode, currAttr.getNodeName(),
                       currValue, null);
        }
      }
      return newNode;
    }
    catch (IllegalAccessException IAE) {
      String err = ":IllegalAccessException:" + IAE;
      log.log(Level.SEVERE,err,IAE);
      MonkException ofe = new MonkException(err, IAE);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    catch (java.lang.InstantiationException IE) {
      String err = ":InstantiationException:" + IE;
      log.log(Level.SEVERE,err,IE);
      MonkException ofe = new MonkException(err, IE);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
  }

  /**
      Given a DOM Document, create a corresponding manifestation of it in the
      application domain.
      This version of create() walks throught an entire Document, creating an
      application-domain
   manifestation for every Node encountered in the Document  An "Aggregate Object"
      is returned
      to the caller. For example, a Document containing  "<add user> <Pat>" is
      transformed into
   two objects:  an instance of AddUser and an instance of User.  Furthermore, the
      User instance
   is stored in the AddUser instance and is considered to be one of its "children".
      @param Document inDocument The DOM Document serving as the temple for the
      application-domain objects being created
      @param inDocument
      @return the root-level  application-domain object being created
      @throws org.quintet.monk.MonkException
      @roseuid 40E9B0720293
   */
  public Object create(Document inDocument) throws MonkException {
    //if we have no domain then use the primitive class map to do our work
    Blueprints bp = getDomain();
    if(bp == null || bp.size() == 0){
      Node rootNode = inDocument.getDocumentElement();
      return processNode(rootNode);
    }else{
      return create(inDocument, getDomain());
    }
  }

  /**
   Using <code>blueprints</code> as a guide, translate <code>inDocument</code> from
      the XML namespace
   to the java namespace; that is, convert <code>inDocument</code>'s elements into
      java objects
      @param inDocument Document -- The XML being converted to java objects
   @param blueprints Map -- The instructions for converting <code>inDocument</code>
      into java objects
      @return java.lang.Object
      @throws MonkException
      @throws org.quintet.monk.MonkException
      @roseuid 40E9B0720333
   */
  public Object create(Document inDocument, Map blueprints) throws
      MonkException {

    if (configLoaded == false) {
      loadConfiguration();
    }
    Node root = inDocument.getDocumentElement();
    try {
      Stack contextStack = new Stack();
      return processNode(root, blueprints, contextStack);
    }
    catch (MonkException OFE) {
      String err = ":MonkException:" + OFE;
      log.log(Level.SEVERE,err, OFE);
      throw OFE;
    }
  }

  /**
      Recursively process Nodes, starting at <code>inNode</code>.
      Use <code>blueprints</code>' instructions for instantiating java objects
      @param Node inNode -- The root node of the document being parse
      @param Map blueprints -- The set of blueprints,keyed by xml element name
      @param inNode
      @param blueprints
      @return java.lang.Object
      @throws MonkException
      @throws org.quintet.monk.MonkException
      @roseuid 40E9B07203DE
   */
  private Object processNode(Node inNode, Map blueprints) throws
      MonkException {
    if (inNode == null) {
      log.info(":inNode is null, ergo returning null");
      return null;
    }
    Object answer = create(inNode, blueprints);
    if (answer == null) {
      log.info(":answer is null, ergo returning null");
      return answer;
    }
    ClassMetaData nodeMetaData = (ClassMetaData) blueprints.get(inNode.
        getNodeName());
    //walk inNode's children, process each of them to get application-domain
    //manifestations of them too.
    NodeList kids = inNode.getChildNodes();
    for (int kidIdx = 0; kidIdx < kids.getLength(); kidIdx++) {
      Node currKid = kids.item(kidIdx);
      if (currKid.getNodeType() != Node.TEXT_NODE) {
        Object appDomainManifestation = processNode(kids.item(kidIdx),
            blueprints);
        if (appDomainManifestation != null) {
          setAttribute(answer.getClass(), answer, kids.item(kidIdx).getNodeName(),
                       appDomainManifestation, nodeMetaData);
        }
      }
    }
    return answer;
  }

  /**
      Recursively process Nodes, starting at the caller-supplied Node
      @param Node inNode The starting point of the recursive processing
      @param inNode
      @return LWO manifestation of the input Node and its descendents
      @throws org.quintet.monk.MonkException
      @roseuid 40E9B07300AA
   */
  private Object processNode(Node inNode) throws MonkException {
    if (inNode == null) {
      log.info(":null inNode, ergo returning null");
      return null;
    }
    Object answer = create(inNode);
    //if answer is null, then we don't have to try to process its children
    if (answer == null) {
      log.info(":null answer");
      return answer;
    }
    //walk inNode's children, process each of them to get application-domain
    //manifestations of them too.
    NodeList kids = inNode.getChildNodes();
    for (int kidIdx = 0; kidIdx < kids.getLength(); kidIdx++) {
      Node currKid = kids.item(kidIdx);
      if (currKid.getNodeType() != Node.TEXT_NODE) {
        Object appDomainManifestation = processNode(kids.item(kidIdx));
        if (appDomainManifestation != null) {
          setAttribute(answer.getClass(), answer, kids.item(kidIdx).getNodeName(),
                       appDomainManifestation, null);
        }
      }
    }
    return answer;
  }

  /**
      convert an XML-style name (e.g., dealer-id) into a java-style name
      (e.g., dealerId)
      @param String xmlName the name being converted
      @param xmlName
      @return String containing java version of <code>xmlName</code>
      @roseuid 40E9B07301A4
   */
  private String javaName(String xmlName) {
    if (xmlName == null) {
      return null;
    }
    if (xmlName.length() == 0) {
      return "";
    }
    StringBuffer answer = new StringBuffer();
    answer.append(Character.toLowerCase(xmlName.charAt(0)));
    int idx = 1;
    while (idx < xmlName.length()) {
      char currChar = xmlName.charAt(idx);
      if (currChar != '-') {
        answer.append(currChar);
        idx++;
      }
      else {
        if (++idx < xmlName.length()) {
          currChar = xmlName.charAt(idx);
          answer.append(Character.toUpperCase(currChar));
          idx++;
        }
      }
    }
    return answer.toString();
  }

  /**
   Create a <code>metaData.subject</code> instance from the attributes found within
   <code>inNode</code>.  Use <code>metaData</code> as the blueprint for translating
      <code>inNode</code> into the java namespace.
   @param Node inNode -- The XML node being translated into the java namespace
      @param ClassMetaData metaData -- The blueprint containing translation
      instructions
      @param inNode
      @param metaData
      @exception MonkException
      @return Object manifestation of <code>inNode</code>
      @throws org.quintet.monk.MonkException
      @roseuid 426E71460016
   */
  private Object createInstance(Node inNode, ClassMetaData metaData) throws
      MonkException {
    //first get the ctor's actual arg values from inNode
    Collection argNames = metaData.getConstructorArgNames();
    //find each arg's actual value in inNode
    String[] argValues = null;
    if (argNames == null || argNames.size() == 0) {
      argValues = null;
    }
    else {
      argValues = new String[argNames.size()];
    }
    int argIdx = 0;
    for (Iterator formalArgsIter = argNames.iterator(); formalArgsIter.hasNext(); ) {
      Attribute formalArgAttribute = (Attribute) formalArgsIter.next();
      String formalArg = formalArgAttribute.getXmlName();
      Node actualArgNode = inNode.getAttributes().getNamedItem(formalArg);
      if (actualArgNode == null) {
        String err = ":formal argument " + formalArg +
            " has no matching actual " + "argument value for class " +
            metaData.getSubjectName();
        MonkException ofe = new MonkException(err);
        ofe.setClassMetaData(metaData);
        ofe.setErrantNode(inNode);
        log.log(Level.SEVERE,err);
        throw ofe;
      }
      String actualArg = actualArgNode.getNodeValue();
      argValues[argIdx++] = actualArg;
    }
    Method ctor = metaData.getFactoryMethod();
    if (ctor == null && metaData.getBuilder().isCtor() == false) {
      String err = ":no factory method specified for " +
          metaData.getSubjectName() +
          "AND isCtor is false, therefore there is no way to construct an instance from " +
          metaData.getXmlName();
      log.log(Level.SEVERE,err);
      MonkConfigurationException ofe = new
          MonkConfigurationException(err);
      ofe.setClassMetaData(metaData);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    try {
      Object target = metaData.getFactoryMethodTarget();
      if (target == null) {
        String err = ":null target in " +
            metaData.getXmlName();
        log.log(Level.SEVERE,err);
        MonkConfigurationException ofe = new
            MonkConfigurationException(err);
        ofe.setClassMetaData(metaData);
        ofe.setErrantNode(inNode);
        throw ofe;
      }
      if (metaData.getBuilder().isCtor()) {
        Class targetClass = (Class) target;
        java.lang.reflect.Constructor constructor = targetClass.getConstructor(
            metaData.getBuilder().getFormalArgumentClasses());
        return constructor.newInstance((Object[])argValues);
      }
      else if(metaData.getBuilder().getAbstractFactoryClass()){
        Object newInstance = ctor.invoke(null,(Object[])argValues);
        return newInstance;
      }else{
        Object newInstance = ctor.invoke(target, (Object[])argValues);
        return newInstance;
      }
    }
    catch (IllegalAccessException IAE) {
      String err = ":IllegalAccessException:" + IAE;
      log.log(Level.SEVERE,err);
      MonkException ofe = new MonkException(err, IAE);
      ofe.setClassMetaData(metaData);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    catch (InvocationTargetException ITE) {
      Throwable orig = ITE.getTargetException();
      String err = ":InvocationTargetException:" + ITE +
          ":(original:" + orig + ")";
      log.log(Level.SEVERE,err);
      if (orig != null) {
        log.log(Level.SEVERE,err,orig);
        MonkException ofe = new MonkException(err);
        ofe.setClassMetaData(metaData);
        ofe.setErrantNode(inNode);
        throw ofe;
      }
      else {
        log.log(Level.SEVERE,err,ITE);
        MonkException ofe = new MonkException(err, ITE);
        ofe.setClassMetaData(metaData);
        ofe.setErrantNode(inNode);
        throw ofe;
      }
    }
    catch (InstantiationException IE) {
      String err = ":InstantiationException:" + IE;
      log.log(Level.SEVERE,err,IE);
      MonkException ofe = new MonkException(err, IE);
      ofe.setClassMetaData(metaData);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    catch (NoSuchMethodException NSME) {
      String err = ":NoSuchMethodException:" + NSME;
      log.log(Level.SEVERE,err,NSME);
      MonkConfigurationException ofe = new
          MonkConfigurationException(err, NSME);
      ofe.setClassMetaData(metaData);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
  }

  /**
   Create a <code>metaData.subject</code> instance from the attributes found within
   <code>inNode</code>.  Use <code>metaData</code> as the blueprint for translating
      <code>inNode</code> into the java namespace.
   @param Node inNode -- The XML node being translated into the java namespace
      @param ClassMetaData metaData -- The blueprint containing translation
      instructions
   @param Stack contextStack -- the parent objects of the object being constructed
      @param inNode
      @param metaData
      @param contextStack
      @exception MonkException
      @return Object manifestation of <code>inNode</code>
      @throws org.quintet.monk.MonkException
      @roseuid 426E714600DF
   */
  private Object createInstance(Node inNode, ClassMetaData metaData,
                                Stack contextStack) throws
      MonkException {
    //first get the ctor's actual arg values from inNode
    Collection argNames = metaData.getConstructorArgNames();
    //find each arg's actual value in inNode
    Object[] argValues = null;
    if (argNames == null || argNames.size() == 0) {
      argValues = null;
    }
    else {
      argValues = new Object[argNames.size()];
    }
    int argIdx = 0;
    for (Iterator formalArgsIter = argNames.iterator(); formalArgsIter.hasNext(); ) {
      Attribute formalArgAttribute = (Attribute) formalArgsIter.next();
      boolean isLiteral = formalArgAttribute.isLiteral();
      String formalArg = formalArgAttribute.getXmlName();
      //if formalArg's name starts with ../ then it is a reference to a parent's attribute value
      //if formalArg DOES NOT start with ../ then it is a reference to a value in the Doc node itself
      Object actualArg = null;
      if(isLiteral){
        argValues[argIdx++] = formalArgAttribute.getLiteral();
        continue;
      }
      if (formalArg.startsWith("../")) {
        actualArg = resolveParentReference(formalArg, contextStack);
      }else if (formalArg.equals("[node]")){
        actualArg = inNode.getTextContent();
      }
      else {
        Node actualArgNode = inNode.getAttributes().getNamedItem(formalArg);
        if (actualArgNode == null  && formalArgAttribute.isMandatory().booleanValue()) {
          String err = "mandatory formal argument " + formalArg +
              " has no matching actual " + "argument value for class " +
              metaData.getSubjectName();
          MonkException ofe = new MonkException(err);
          ofe.setClassMetaData(metaData);
          ofe.setErrantNode(inNode);
          log.log(Level.SEVERE,err);
          throw ofe;
        }else if(actualArgNode == null && formalArgAttribute.isMandatory().booleanValue() == false)
          actualArg = null;
        else
          actualArg = actualArgNode.getNodeValue();
      }
      argValues[argIdx++] = actualArg;
    }
    Method ctor = metaData.getBuilder().getFactoryMethod(contextStack);
    if (ctor == null && metaData.getBuilder().isCtor() == false) {
      String err = ":no factory method specified for " +
          metaData.getSubjectName() +
          "AND isCtor is false, therefore there is no way to construct an instance from " +
          metaData.getXmlName();
      log.log(Level.SEVERE,err);
      MonkConfigurationException ofe = new
          MonkConfigurationException(err);
      ofe.setClassMetaData(metaData);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    try {
      Object target = metaData.getBuilder().getTarget(contextStack);
      if (target == null) {
        String err = ":null target in " +
            metaData.getXmlName();
        log.log(Level.SEVERE,err);
        MonkConfigurationException ofe = new
            MonkConfigurationException(err);
        ofe.setClassMetaData(metaData);
        ofe.setErrantNode(inNode);
        throw ofe;
      }
      if (metaData.getBuilder().isCtor()) {
        Class targetClass = (Class) target;
        java.lang.reflect.Constructor constructor = targetClass.getConstructor(
            metaData.getBuilder().getFormalArgumentClasses());
        return constructor.newInstance(argValues);
      }
      else if (metaData.getBuilder().getAbstractFactoryClass()){
        Object newInstance = ctor.invoke(null,argValues);
        return newInstance;
      }else{
        Object newInstance = ctor.invoke(target, argValues);
        return newInstance;
      }
    }
    catch (IllegalAccessException IAE) {
      String err = ":IllegalAccessException:" + IAE;
      log.log(Level.SEVERE,err);
      MonkException ofe = new MonkException(err, IAE);
      ofe.setClassMetaData(metaData);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    catch (InvocationTargetException ITE) {
      Throwable orig = ITE.getTargetException();
      String err = ":InvocationTargetException:" + ITE +
          ":(original:" + orig + ")";
      log.log(Level.SEVERE,err);
      if (orig != null) {
        log.log(Level.SEVERE,err,orig);
        MonkException ofe = new MonkException(err);
        ofe.setClassMetaData(metaData);
        ofe.setErrantNode(inNode);
        throw ofe;
      }
      else {
        log.log(Level.SEVERE,err,ITE);
        MonkException ofe = new MonkException(err, ITE);
        ofe.setClassMetaData(metaData);
        ofe.setErrantNode(inNode);
        throw ofe;
      }
    }
    catch (InstantiationException IE) {
      String err = ":InstantiationException:" + IE;
      log.log(Level.SEVERE,err,IE);
      MonkException ofe = new MonkException(err, IE);
      ofe.setClassMetaData(metaData);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
    catch (NoSuchMethodException NSME) {
      String err = ":NoSuchMethodException:" + NSME;
      log.log(Level.SEVERE,err,NSME);
      MonkConfigurationException ofe = new
          MonkConfigurationException(err, NSME);
      ofe.setClassMetaData(metaData);
      ofe.setErrantNode(inNode);
      throw ofe;
    }
  }

  /**
      @param formalArgName
      @param contextStack
      @return java.lang.Object
      @roseuid 426E714601C5
   */
  private Object resolveParentReference(String formalArgName,
                                        Stack contextStack) throws MonkException {

    java.util.Vector cache = new java.util.Vector(contextStack.size());
    java.util.StringTokenizer lex = new java.util.StringTokenizer(formalArgName,
        "/");
    Object context = null;
    String currToken = null;
    while (lex.hasMoreTokens()) {
      currToken = lex.nextToken();
      if(currToken.equals("..") && contextStack.isEmpty() == false){
        context = contextStack.pop();
        //save the object so we can push it back later
        cache.add(context);
      }
    }
    if (context == null) {
      String err = "formal argument referred to a context (via ../) that does not exist";
      log.log(Level.SEVERE,err);
      throw new MonkException(err);
    }
    //lex should now be pointing to the actual arg's name
    Method formalArgGetter = null;
    String baseArgName = currToken;


    try {
	    // TAO: If baseArgName = '..' then we simply want the context.
	    if ( "..".equals(baseArgName)) {
	    	return context;
	    }
      formalArgGetter = context.getClass().getMethod(baseArgName, (Class[])null);
      Object actualArg = formalArgGetter.invoke(context, (Object[])null);
      return actualArg;
    }
    catch (SecurityException ex) {
      String err = "SecurityException when accessing formal arg " +
          formalArgName;
      log.log(Level.SEVERE,err, ex);
      return null;
    }
    catch (NoSuchMethodException ex) {
      String err = "NoSuchMethodException when accessing formal arg " +
          formalArgName;
      log.log(Level.SEVERE,err, ex);
      return null;
    }
    catch (IllegalAccessException ex) {
      String err = "IllegalAccessException when accessing formal arg " +
          formalArgName;
      log.log(Level.SEVERE,err, ex);
      return null;
    }
    catch (InvocationTargetException ex) {
      String err = "InvocationTargetException(" +
          ex.getCause().getClass().getName() + ") when accessing formal arg " +
          formalArgName;
      Throwable cause = ex.getCause();
      log.log(Level.SEVERE,err, cause);
      return null;
    }
    finally {
      int cacheIdx = cache.size() - 1;
      while (cacheIdx >= 0) {
        contextStack.push(cache.elementAt(cacheIdx--));
      }
    }
  }

  /**
      Create a java object from <code>inNode</code>'s information.  Use
      <code>blueprints</code>
      to translate from the XML namespace to the java namespace.
   Blueprints can access previously-created objects via <code>contextStack</code>
   @param Node inNode -- The XML node being translated into the java namespace
      @param Map blueprints -- The blueprints to use when tranlating
      <code>inNode</code>
   @param Stack contextStack -- The stack of parent objects created prior to the
      current object
      @param inNode
      @param blueprints
      @param contextStack
      @return java.lang.Object
      @throws MonkException
      @throws org.quintet.monk.MonkException
      @roseuid 426E71460387
   */
  public Object create(Node inNode, Map blueprints, Stack contextStack) throws
      MonkException {
    String thisMethod = ":<<blueprints>>create(Node,Map,Stack)";
    if (inNode == null) {
      log.log(Level.SEVERE,":null Node");
      return null; //a null XML Node yields a null java Object, eh?
    }
    if (blueprints == null) {
      String err = "null blueprints, proceeding with caution";
      log.log(Level.SEVERE,err);
    }
    if (blueprints != null && blueprints.size() == 0) {
      log.log(Level.SEVERE,
          ":empty blueprints, proceeding with caution");
    }
    if (configLoaded == false) {
      loadConfiguration();
    }
    //user inNode's NodeName parameter to determine what type of
    //Object that should be created
    ClassMetaData metaData = (ClassMetaData) blueprints.get(inNode.getNodeName());
    if (metaData == null) {
      log.info(":no metaData for node " +
               inNode.getNodeName());
    }
    Class correspondingClass = findBestClass(inNode, blueprints);
    if(correspondingClass == null){
      return null;
    }
    Object newNode = createInstance(inNode, metaData, contextStack);
    if (newNode == null) {
      log.log(Level.SEVERE,
          ":createInstance(Node,metadata) returned null for node " +
          inNode.getNodeName());
    }
    //fill in its attributes
    NamedNodeMap nnm = inNode.getAttributes();
    Builder builder = metaData.getBuilder();
    for (int attrIdx = 0; attrIdx < nnm.getLength(); attrIdx++) {
      Node currAttr = nnm.item(attrIdx);
      String currAttrName = currAttr.getNodeName();
      if (currAttrName.equals("meta-name") == false &&
          builder.isFormalArgument(currAttrName) == false) {
        Object currValue = currAttr.getNodeValue();
        ClassMetaData attributeMetaData = (ClassMetaData) blueprints.get(
            currAttr.getNodeName());
        setAttribute(correspondingClass, newNode, currAttr.getNodeName(),
                     currValue, metaData,contextStack);
      }
    }
    return newNode;
  }

  /**
   Attempt to set <code>currObject</code>'s <code>attributeName</code> attribute
      to <code>inValue</code>.  If <code>inObject</code> lacks a setter for
      <code>inAttribute</code> and if it is a <code>Collection</code>, then
   <code>setAttribute</code> will use the <code>Collection.add</code> method to
      add <code>inValue</code> to <code>inObject</code>
      @param Class inClass <code>inObject</code>'s class
   @param Object inObject The object containing <code>inAttributeName</code>.
      @param String inAttributeName The name of the attribute being set
      @param Object inValue The value to place into
      <code>inObject.inAttributeName</code>
      @param inClass
      @param inObject
      @param inAttributeName
      @param inValue
      @param inMetaData
      @exception MonkException it anything goes awry
      @throws org.quintet.monk.MonkException
      @roseuid 426E7147011C
   */
  private void setAttribute(Class inClass, Object inObject,
                            String inAttributeName, Object inValue,
                            ClassMetaData inMetaData) throws
      MonkException {
    String thisMethod = ":setAttribute";
    try {
      //special case for inClass == String.  inObject (e.g., "column") given inValue's value
      if (inClass.equals(java.lang.String.class) &&
          inValue instanceof java.lang.String) {
        ( (String) inObject).copyValueOf( ( (String) inValue).toCharArray());
        return;
      }
      //see if inClass has a set<attribute> that accepts a String
      String setFromStringName = "set" +
          inAttributeName.substring(0, 1).toUpperCase() +
          javaName(inAttributeName).substring(1);
      try {
        Method[] allOfThem = inClass.getMethods();
        for (int methIdx = 0; methIdx < allOfThem.length; methIdx++) {
          Method currMeth = allOfThem[methIdx];
          if (currMeth.getName().equals(setFromStringName)) {
            //currMeth may be the right setter.  Check its arg to see if it is compatible with inValue
            Class[] currMethArgs = currMeth.getParameterTypes();
            if (currMethArgs != null && currMethArgs.length == 1 &&
                currMethArgs[0].isAssignableFrom(inValue.getClass())) {
              currMeth.invoke(inObject, new Object[] {inValue});
              return;
            }
          }
        }
        Method setFromStringMethod = inClass.getMethod(setFromStringName,
            new Class[] {inValue.getClass()});
        setFromStringMethod.invoke(inObject, new Object[] {inValue});
        return;
      }
      catch (NoSuchMethodException noStringSetter) {
        //no set-from-string method for the attribute
      }
      String javaAttributeName = javaName(inAttributeName);
      //the metaData may have an attribute (and a setter) specified for javaAttributeName, so let's have a
      //look, shall we?
      if (inMetaData != null) {
        Attribute attribute = inMetaData.getAttribute(inAttributeName);
        if (attribute != null) {
          String setterName = attribute.getSetterName();
          String setteeClassName = attribute.getJavaClassName();
          Class setteeClass = null;
          try {
            setteeClass = Class.forName(setteeClassName);
          }
          catch (ClassNotFoundException CNFE) {
            String err = ":cannot find class " +
                setteeClassName;
            log.log(Level.SEVERE,err);
            throw new MonkException(err);
          }
          if (setterName != null) {
            Method writer = inClass.getMethod(setterName,
                                              new Class[] {setteeClass});
            writer.invoke(inObject, new Object[] {inValue});
            return;
          }
        }
      }
      //no setter explicitly mentioned in the meta-data, so try the introspection route
      BeanInfo theScoop = Introspector.getBeanInfo(inClass);
      PropertyDescriptor[] descriptors = theScoop.getPropertyDescriptors();
      if (descriptors != null && descriptors.length > 0) {
        //look for a descriptor named <inAttributeName>, which will
        //subsequently lead us to a setter method
        for (int descIdx = 0; descIdx < descriptors.length; descIdx++) {
          PropertyDescriptor currDescriptor = descriptors[descIdx];
          if (currDescriptor.getName().equals(javaAttributeName)) {
            Method writer = currDescriptor.getWriteMethod();
            //find the type of argument expected by writer
            Class[] writerArgs = writer.getParameterTypes();
            if (writerArgs != null && writerArgs.length == 1) {
              Class setterArgClass = writerArgs[0];
              //look for a ctor(String) in setterArgClass
              Constructor fromString = setterArgClass.getConstructor(new Class[] {
                  java.lang.String.class});
              if (fromString != null) {
                Object setterArgObj = null;
                try {
                  setterArgObj = fromString.newInstance(new Object[] {inValue});
                }
                catch (InvocationTargetException ex) {
                  String err = "InvocationTargetException when calling " +
                      setterArgClass.getName() + ".ctor(String)";
                  Throwable cause = ex.getCause();
                  log.log(Level.SEVERE,err, cause);
                  throw new MonkException(err, ex);
                }
                catch (IllegalArgumentException ex) {
                  String err = "IllegalArgumentException when calling " +
                      setterArgClass.getName() + ".ctor(String)";
                  log.log(Level.SEVERE,err, ex);
                  throw new MonkException(err, ex);
                }
                catch (IllegalAccessException ex) {
                  String err = "IllegalAccessException when calling " +
                      setterArgClass.getName() + ".ctor(String)";
                  log.log(Level.SEVERE,err, ex);
                  throw new MonkException(err, ex);
                }
                catch (InstantiationException ex) {
                  String err = "InstantiationException when calling " +
                      setterArgClass.getName() + ".ctor(String)";
                  log.log(Level.SEVERE,err, ex);
                  throw new MonkException(err, ex);
                }
                writer.invoke(inObject, new Object[] {setterArgObj});
                return;
              }
            }
          }
        }
      }
      //no setter.  perhaps <inObject> is a collection with an "add()" method, eh?
      Method adder = null;
      try{
          adder = inClass.getMethod("add", new Class[] {java.lang.Object.class});
          adder.invoke(inObject, new Object[] {inValue});
      }catch(Throwable t){
          //no adder method either, but nobody wants to see an exception
          //about it, so just ignore it....
          log.finest("no method found to set attribute \""+inAttributeName+
                  "\" in class " + inClass.getName() +", attribute will no be set.");
      }
    }
    catch (IntrospectionException IE) {
      String err = ":IntrospectionException:" + IE;
      log.log(Level.SEVERE,err, IE);
      MonkException ofe = new MonkException(err, IE);
      ofe.setErrantAttributeName(inAttributeName);
      ofe.setErrantClassName(inClass.getName());
      throw ofe;
    }
    catch (InvocationTargetException ITE) {
      Throwable orig = ITE.getCause();
      if (orig == null) {
        orig = ITE;
      }
      String err = ":InvocationTargetException:" + orig;
      log.log(Level.SEVERE,err, orig);
      MonkException ofe = new MonkException(err, ITE);
      ofe.setErrantAttributeName(inAttributeName);
      ofe.setErrantClassName(inClass.getName());
      throw ofe;
    }
    catch (IllegalAccessException IAE) {
      String err = ":IllegalAccessException:" + IAE;
      log.log(Level.SEVERE,err, IAE);
      MonkException ofe = new MonkException(err, IAE);
      ofe.setErrantAttributeName(inAttributeName);
      ofe.setErrantClassName(inClass.getName());
      throw ofe;
    }
    catch (NoSuchMethodException NSME) {
      String err = "no setter for attribute" +
          ".  attribute name = " + inAttributeName + ", objectValue = " +
          inValue;
      log.log(Level.FINEST,err);
      //MonkException ofe = new MonkException(err, NSME);
      //ofe.setErrantAttributeName(inAttributeName);
      //ofe.setErrantClassName(inClass.getName());
      //throw ofe;
    }
  }

  /**
   Attempt to set <code>currObject</code>'s <code>attributeName</code> attribute
      to <code>inValue</code>.  If <code>inObject</code> lacks a setter for
      <code>inAttribute</code> and if it is a <code>Collection</code>, then
   <code>setAttribute</code> will use the <code>Collection.add</code> method to
      add <code>inValue</code> to <code>inObject</code>
      @param Class inClass <code>inObject</code>'s class
   @param Object inObject The object containing <code>inAttributeName</code>.
      @param String inAttributeName The name of the attribute being set
      @param Object inValue The value to place into
      <code>inObject.inAttributeName</code>
      @param ClassMetaData inMetaData The blueprint entry for inObject
      @param Stack contextStack the stack of objects built "above" this object
      @param inClass
      @param inObject
      @param inAttributeName
      @param inValue
      @param inMetaData
      @param contextStack
      @exception MonkException it anything goes awry
      @throws org.quintet.monk.MonkException
      @roseuid 426E71470234
   */
  private void setAttribute(Class inClass, Object inObject,
                            String inAttributeName, Object inValue,
                            ClassMetaData inMetaData, Stack contextStack) throws
      MonkException {
    String thisMethod = ":setAttribute";
    try {
      //special case for inClass == String.  inObject (e.g., "column") given inValue's value
      if (inClass.equals(java.lang.String.class) &&
          inValue instanceof java.lang.String) {
        ( (String) inObject).copyValueOf( ( (String) inValue).toCharArray());
        return;
      }
      //see if inClass has a set<attribute> that accepts a String
      String setFromStringName = "set" +
          inAttributeName.substring(0, 1).toUpperCase() +
          javaName(inAttributeName).substring(1);
      try {
        Method[] allOfThem = inObject.getClass().getMethods();
        for (int methIdx = 0; methIdx < allOfThem.length; methIdx++) {
          Method currMeth = allOfThem[methIdx];
          if (currMeth.getName().equals(setFromStringName)) {
            //currMeth may be the right setter.  Check its arg to see if it is compatible with inValue
            Class[] currMethArgs = currMeth.getParameterTypes();
            if (currMethArgs != null && currMethArgs.length == 1 &&
                currMethArgs[0].isAssignableFrom(inValue.getClass())) {
              currMeth.invoke(inObject, new Object[] {inValue});
              return;
            }
          }
        }
        Method setFromStringMethod = inClass.getMethod(setFromStringName,
            new Class[] {inValue.getClass()});
        setFromStringMethod.invoke(inObject, new Object[] {inValue});
        return;
      }
      catch (NoSuchMethodException noStringSetter) {
        //no set-from-string method for the attribute
      }
      String javaAttributeName = javaName(inAttributeName);
      //the metaData may have an attribute (and a setter) specified for javaAttributeName, so let's have a
      //look, shall we?
      if (inMetaData != null) {
        Attribute attribute = inMetaData.getAttribute(inAttributeName);
        if (attribute != null) {
          String setterName = attribute.getSetterName();
          String setteeClassName = attribute.getJavaClassName();
          Class setteeClass = null;
          try {
            setteeClass = Class.forName(setteeClassName);
          }
          catch (ClassNotFoundException CNFE) {
            String err = ":cannot find class " +
                setteeClassName;
            log.log(Level.SEVERE,err);
            throw new MonkException(err);
          }
          if (setterName != null) {
            Method writer = inClass.getMethod(setterName,
                                              new Class[] {setteeClass});
            Constructor ctor = setteeClass.getConstructor(new Class[] {inValue.getClass() });
            if(ctor != null ) {
	            try {
		            Object newValue = ctor.newInstance(new Object[] {inValue});
		            writer.invoke(inObject, new Object[] {newValue});
	            } catch (InstantiationException ie) {
	            		log.log(Level.SEVERE,"Cannot instantiate via constructor: " + ctor.toString(), ie);
	            		throw new MonkException ("Cannot instantiate via constructor.");

	            }
            }
            return;
          }
        }
      }
      //no setter explicitly mentioned in the meta-data, so try the introspection route
      BeanInfo theScoop = Introspector.getBeanInfo(inObject.getClass());
      PropertyDescriptor[] descriptors = theScoop.getPropertyDescriptors();
      if (descriptors != null && descriptors.length > 0) {
        //look for a descriptor named <inAttributeName>, which will
        //subsequently lead us to a setter method
        for (int descIdx = 0; descIdx < descriptors.length; descIdx++) {
          PropertyDescriptor currDescriptor = descriptors[descIdx];
          if (currDescriptor.getName().equals(javaAttributeName)) {
            Method writer = currDescriptor.getWriteMethod();
            //find the type of argument expected by writer
            Class[] writerArgs = writer.getParameterTypes();
            if (writerArgs != null && writerArgs.length == 1) {
              Class setterArgClass = writerArgs[0];
              //look for a ctor(String) in setterArgClass
              Constructor fromString = setterArgClass.getConstructor(new Class[] {
                  java.lang.String.class});
              if (fromString != null) {
                Object setterArgObj = null;
                try {
                  setterArgObj = fromString.newInstance(new Object[] {inValue});
                }
                catch (InvocationTargetException ex) {
                  String err = "InvocationTargetException when calling " +
                      setterArgClass.getName() + ".ctor(String)";
                  Throwable cause = ex.getCause();
                  log.log(Level.SEVERE,err, cause);
                  throw new MonkException(err, ex);
                }
                catch (IllegalArgumentException ex) {
                  String err = "IllegalArgumentException when calling " +
                      setterArgClass.getName() + ".ctor(String)";
                  log.log(Level.SEVERE,err, ex);
                  throw new MonkException(err, ex);
                }
                catch (IllegalAccessException ex) {
                  String err = "IllegalAccessException when calling " +
                      setterArgClass.getName() + ".ctor(String)";
                  log.log(Level.SEVERE,err, ex);
                  throw new MonkException(err, ex);
                }
                catch (InstantiationException ex) {
                  String err = "InstantiationException when calling " +
                      setterArgClass.getName() + ".ctor(String)";
                  log.log(Level.SEVERE,err, ex);
                  throw new MonkException(err, ex);
                }
                writer.invoke(inObject, new Object[] {setterArgObj});
                return;
              }
            }
          }
        }
      }
      //no setter.  perhaps <inObject> is a collection with an "add()" method, eh?
      Method adder = null;
      try{
          adder = inClass.getMethod("add", new Class[] {java.lang.Object.class});
          adder.invoke(inObject, new Object[] {inValue});
      }catch(Throwable t){
          //do nothing, this isn't an exception per se
          return;
      }
    }
    catch (IntrospectionException IE) {
      String err = ":IntrospectionException:" + IE;
      log.log(Level.SEVERE,err, IE);
      MonkException ofe = new MonkException(err, IE);
      ofe.setErrantAttributeName(inAttributeName);
      ofe.setErrantClassName(inClass.getName());
      throw ofe;
    }
    catch (InvocationTargetException ITE) {
      Throwable orig = ITE.getCause();
      if (orig == null) {
        orig = ITE;
      }
      String err = ":InvocationTargetException:" + orig;
      log.log(Level.SEVERE,err, orig);
      MonkException ofe = new MonkException(err, ITE);
      ofe.setErrantAttributeName(inAttributeName);
      ofe.setErrantClassName(inClass.getName());
      throw ofe;
    }
    catch (IllegalAccessException IAE) {
      String err = ":IllegalAccessException:" + IAE;
      log.log(Level.SEVERE,err, IAE);
      MonkException ofe = new MonkException(err, IAE);
      ofe.setErrantAttributeName(inAttributeName);
      ofe.setErrantClassName(inClass.getName());
      throw ofe;
    }
    catch (NoSuchMethodException NSME) {
      String err = "no setter for attribute" +
          ".  attribute name = " + inAttributeName + ", objectValue = " +
          inValue;
      log.log(Level.WARNING,err, NSME);
      //MonkException ofe = new MonkException(err, NSME);
      //ofe.setErrantAttributeName(inAttributeName);
      //ofe.setErrantClassName(inClass.getName());
      //throw ofe;
    }
  }

  /**
   Given a file name, create a corresponding manifestation of its contents in the
      application domain.
      This version of create() walks throught an entire Document, creating an
      application-domain
   manifestation for every Node encountered in the Document  An "Aggregate Object"
      is returned
      to the caller. For example, a Document containing  "<add user> <Pat>" is
      transformed into
   two objects:  an instance of AddUser and an instance of User.  Furthermore, the
      User instance
   is stored in the AddUser instance and is considered to be one of its "children".
      @param String inFileName The name of the file containing the XML document
      @param inFileName
      @return the root-level  application-domain object being created
      @throws MonkException
      @throws org.quintet.monk.MonkException
      @roseuid 426E71470375
   */
  public Object create(String inFileName) throws MonkException {
    FileInputStream fis = null;
    try {
      fis = new FileInputStream(inFileName);
    }
    catch (FileNotFoundException ex) {
      String err = "Unable to locate a file named '" + inFileName + "'";
      log.log(Level.SEVERE,err, ex);
      throw new MonkException(err, ex);
    }

    DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.
        newInstance();
    DocumentBuilder docBuilder = null;
    try {
      docBuilder = docBuilderFactory.newDocumentBuilder();
    }
    catch (ParserConfigurationException pce) {
      String err =
          "Parser configuration exception when creating a DocumentBuilderFactory";
      log.log(Level.SEVERE,err, pce);
      if(fis != null){
        try{
          fis.close();
        }catch(Throwable eatIt){}
      }
      throw new MonkException(err, pce);
    }
    Document doc = null;
    try {
      doc = docBuilder.parse(fis);
      return create(doc);
    }
    catch (IOException ex1) {
      String err = "IOException when parsing file named " + inFileName;
      log.log(Level.SEVERE,err, ex1);
      throw new MonkException(err, ex1);
    }
    catch (SAXException ex1) {
      String err = "SAXException when parsing file named " + inFileName;
      log.log(Level.SEVERE,err, ex1);
      throw new MonkException(err, ex1);
    }finally{
      if(null != fis){
        try{
          fis.close();
        }catch(Throwable eatIt){}
      }
    }

  }

  /**
   Given an input stream, create a corresponding manifestation of its contents in
      the application domain.
      This version of create() walks throught an entire Document, creating an
      application-domain
   manifestation for every Node encountered in the Document  An "Aggregate Object"
      is returned
      to the caller. For example, a Document containing  "<add user> <Pat>" is
      transformed into
   two objects:  an instance of AddUser and an instance of User.  Furthermore, the
      User instance
   is stored in the AddUser instance and is considered to be one of its "children".
      @param fis The inputStream containing the XML document
      @return the root-level  application-domain object being created
      @throws MonkException
      @throws org.quintet.monk.MonkException
      @roseuid 426E71480023
   */
  public Object create(InputStream fis) throws MonkException {

    if (fis == null) {
      String err = "Cannot process a null InputStream";
      log.log(Level.SEVERE,err);
      throw new MonkException(err);
    }
    DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.
        newInstance();
    DocumentBuilder docBuilder = null;
    try {
      docBuilder = docBuilderFactory.newDocumentBuilder();
    }
    catch (ParserConfigurationException pce) {
      String err =
          "Parser configuration exception when creating a DocumentBuilderFactory";
      log.log(Level.SEVERE,err, pce);
      throw new MonkException(err, pce);
    }
    Document doc = null;
    try {
      doc = docBuilder.parse(fis);
    }
    catch (IOException ex1) {
      String err = "IOException when parsing input stream's contents";
      log.log(Level.SEVERE,err, ex1);
      throw new MonkException(err, ex1);
    }
    catch (SAXException ex1) {
      String err = "SAXException when parsing input stream's contents";
      log.log(Level.SEVERE,err, ex1);
      throw new MonkException(err, ex1);
    }
    return create(doc);
  }

  /**
      Recursively process Nodes, starting at <code>inNode</code>.
      Use <code>blueprints</code>' instructions for instantiating java objects.
   Blueprints can refer to previously-created objects, which are available in
      <code>contextStack</code>
      @param Node inNode -- The root node of the document being parse
      @param Map blueprints -- The set of blueprints,keyed by xml element name
      @param inNode
      @param blueprints
      @param contextStack
      @return java.lang.Object
      @throws MonkException
      @throws org.quintet.monk.MonkException
      @roseuid 426E71480218
   */
  private Object processNode(Node inNode, Map blueprints, Stack contextStack) throws
      MonkException {
    String thisMethod = ":<<blueprints>>processNode(Node,Map)";
    if (inNode == null) {
      log.info(":inNode is null, ergo returning null");
      return null;
    }
    Object answer = create(inNode, blueprints, contextStack);
    if (answer == null) {
      log.info(":answer is null, ergo returning null");
      return answer;
    }
    contextStack.push(answer);
    ClassMetaData nodeMetaData = (ClassMetaData) blueprints.get(inNode.
        getNodeName());
    //walk inNode's children, process each of them to get application-domain
    //manifestations of them too.
    NodeList kids = inNode.getChildNodes();
    for (int kidIdx = 0; kidIdx < kids.getLength(); kidIdx++) {
      Node currKid = kids.item(kidIdx);
      if (currKid.getNodeType() != Node.TEXT_NODE && currKid.getNodeType() != Node.CDATA_SECTION_NODE) {
        Object appDomainManifestation = processNode(kids.item(kidIdx),
            blueprints, contextStack);
        if (appDomainManifestation != null) {
          setAttribute(answer.getClass(), answer, kids.item(kidIdx).getNodeName(),
                       appDomainManifestation, nodeMetaData, contextStack);
        }
      }
    }
    if (contextStack.isEmpty() == false) {
      contextStack.pop();
    }
    return answer;
  }

  /**
      @param inStringArg
      @param inClass
      @return java.lang.Object
      @throws org.quintet.monk.MonkException
      @roseuid 426E71490115
   */
  private Object constructFromString(String inStringArg, Class inClass) throws
      MonkException {
    Constructor ctor = null;
    try {
      ctor = inClass.getConstructor(new Class[] {java.lang.String.class});
    }
    catch (SecurityException ex) {
      String err =
          "SecurityException when trying to find a String ctor in class " +
          inClass.getName();
      log.log(Level.SEVERE,err, ex);
      throw new MonkException(err, ex);
    }
    catch (NoSuchMethodException ex) {
      String err =
          "NoSuchMethodException when trying to find a String ctor in class " +
          inClass.getName();
      log.log(Level.SEVERE,err, ex);
      throw new MonkException(err, ex);
    }

    Object answer = null;
    try {
      answer = ctor.newInstance(new Object[] {inStringArg});
    }
    catch (InvocationTargetException ex1) {
      String err =
          "InvocationTargetException when invoking string ctor on class " +
          inClass.getName();
      Throwable cause = ex1.getCause();
      log.log(Level.SEVERE,err, cause);
      throw new MonkException(err, ex1);
    }
    catch (IllegalArgumentException ex1) {
      String err =
          "IllegalArgumentException when invoking string ctor on class " +
          inClass.getName();
      log.log(Level.SEVERE,err, ex1);
      throw new MonkException(err, ex1);
    }
    catch (IllegalAccessException ex1) {
      String err =
          "IllegalAccessException when invoking string ctor on class " +
          inClass.getName();
      log.log(Level.SEVERE,err, ex1);
      throw new MonkException(err, ex1);

    }
    catch (InstantiationException ex1) {
      String err =
          "IllegalAccessException when invoking string ctor on class " +
          inClass.getName();
      log.log(Level.SEVERE,err, ex1);
      throw new MonkException(err, ex1);
    }
    return answer;
  }

  /**
   Using <code>blueprints</code> as a guide, translate <code>inputStream</code>
      from the XML namespace
   to the java namespace; that is, convert <code>inputStream's</code>'s elements
      into java objects
      @param inputStream InputStream -- The XML being converted to java objects
      - The <code>InputSteam</code> housing the XML document
   @param blueprints Map -- The instructions for converting <code>inDocument</code>
      into java objects
      - The <code>Blueprints</code> that describe how to convert
      <code>inputStream</code>'s contents to Java objects
      @return java.lang.Object
      @throws MonkException
      @throws org.quintet.monk.MonkException
      @roseuid 426E71830140
   */
  public Object create(InputStream inputStream, Map blueprints) throws
      MonkException {

    if (inputStream == null) {
      String err = "Cannot process a null InputStream";
      log.log(Level.SEVERE,err);
      throw new MonkException(err);
    }
    DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.
        newInstance();
    DocumentBuilder docBuilder = null;
    try {
      docBuilder = docBuilderFactory.newDocumentBuilder();
    }
    catch (ParserConfigurationException pce) {
      String err =
          "Parser configuration exception when creating a DocumentBuilderFactory";
      log.log(Level.SEVERE,err, pce);
      throw new MonkException(err, pce);
    }
    Document doc = null;
    try {
      doc = docBuilder.parse(inputStream);
    }
    catch (IOException ex1) {
      String err = "IOException when parsing input stream's contents";
      log.log(Level.SEVERE,err, ex1);
      throw new MonkException(err, ex1);
    }
    catch (SAXException ex1) {
      String err = "SAXException when parsing input stream's contents";
      log.log(Level.SEVERE,err, ex1);
      throw new MonkException(err, ex1);
    }
    return create(doc, blueprints);
  }

  /**
      Access method for the propFileName property.
      @return   the current value of the propFileName property
      @roseuid 4289FCFC0193
   */
  public String getPropFileName() {
    return propFileName;
  }

  /**
      Sets the value of the propFileName property.
      @param aPropFileName the new value of the propFileName property
      @roseuid 4289FCFE020E
   */
  public void setPropFileName(String aPropFileName) {
    propFileName = aPropFileName;
  }

  /**
      Determines if the configLoaded property is true.
      @return   <code>true<code> if the configLoaded property is true
      @roseuid 4289FD01038F
   */
  public boolean getConfigLoaded() {
    return configLoaded;
  }

  /**
      Sets the value of the configLoaded property.
      @param aConfigLoaded the new value of the configLoaded property
      @roseuid 4289FD060101
   */
  public void setConfigLoaded(boolean aConfigLoaded) {
    configLoaded = aConfigLoaded;
  }

  /**
      Access method for the classMap property.
      @return   the current value of the classMap property
      @roseuid 428A01C20334
   */
  public Map getClassMap() {
    return classMap;
  }

  /**
      Sets the value of the classMap property.
      @param aClassMap the new value of the classMap property
      @roseuid 428A01C2037B
   */
  public void setClassMap(Map aClassMap) {
    classMap = aClassMap;
  }

  /**
   Create an <code>Monk</code> that will create objects according to the
   <code>Blueprints</code> found in <code>inDomainSpecificationFileName</code>
   @param inDomainSpecificationFileName - The name of the file containing the
      <code>Monk</code>'s domain blueprints
      @return Monk
      @throws org.quintet.monk.MonkException
      @roseuid 428A03FC032A
   */
  public static final Monk createMonk(String
      inDomainSpecificationFileName) throws MonkException {

    Blueprints blueprints = Blueprints.buildBlueprints(
        inDomainSpecificationFileName);
    return new Monk(blueprints);
  }

  /**
      @param blueprintsStream
      @return org.quintet.monk.Blueprints
      @throws org.quintet.monk.MonkException
      @roseuid 428A2C1B0150
   */
  private static Blueprints readBlueprints(InputStream blueprintsStream) throws
      MonkException {
    return null;
  }

  /**
   Create an <code>Monk</code> that will create objects according to the
      <code>Blueprints</code> found in <code>inBlueprints</code>
      @param inBlueprints - The domain's blueprints
      @return Monk
      @throws org.quintet.monk.MonkException
      @roseuid 428CBC560115
   */
  public static final Monk createMonk(Map inBlueprints) throws
      MonkException {
    Monk answer = new Monk();
    Blueprints bp = new Blueprints();
    bp.setImpl(inBlueprints);
    answer.setDomain(bp);
    return answer;
  }

  /**
      Create a basic <code>Monk</code> that knows how to process
      <code>Blueprints</code>.  This method is primarily  called by
   <code>Blueprints</code>' <code>createBlueprints</code> method; it is necessary
      in order to bootstrap all subsequent <code>Monk</code>s.
      @return Monk
      @throws org.quintet.monk.MonkException
      @roseuid 428CBDF9005F
   */
  public static final Monk createMonk() throws
      MonkException {
    return new Monk();
  }

  public static final Monk createMonk(Class bootstrapConfigClass) throws MonkException{
    return new Monk(bootstrapConfigClass);
  }
  /**
   Create an <code>Monk</code> that will create objects according to the
      <code>Blueprints</code> found in <code>inDomainSpecificationStream</code>
      @param inDomainSpecificationStream - The <code>InputStream</code> containing the
      <code>Monk</code>'s domain blueprints
      @return Monk
      @throws org.quintet.monk.MonkException
      @roseuid 428CC12A0133
   */
  public static final Monk createMonk(InputStream
      inDomainSpecificationStream) throws MonkException {
    Blueprints bp = Blueprints.buildBlueprints(inDomainSpecificationStream);
    return new Monk(bp);
  }
}
