package org.quintet.monk;

import java.beans.Introspector;
import java.beans.IntrospectionException;
import java.util.ArrayList;
import java.util.Collection;
import java.beans.PropertyDescriptor;
import java.beans.SimpleBeanInfo;
import java.util.Map;

/**
A BeanInfo-esque description of a class, augmented with XML
information.  Construct an <code>XMLInfo</code> when <code>XMLFactory</code>'s
default behavior cannot properly convert an object into XML.
An XMLInfo helps distinguish between an object's constituent <i>attributes</i>
and its constituent <i>elements</i>; attributes are emitted as simple
<i>name=value</i> pairs, but elements are emitted as complete XML tags with
(potentially) their own constituent elements and attributes.<br>


@author Peter J Thaggard
@version 1.0
@see org.quintet.monk.XMLFactory
@see org.quintet.monk.XMLIntrospector
@since 2000
 */
public class XMLInfo extends SimpleBeanInfo
{

   /**
   The name of the XML element representation of objectClass
    */
   private String elementName = null;

   /**
   The set of objectClass's attributes that are considered to correspond to XML
   Elements
    */
   private Collection elements = new ArrayList ();

   /**
   The set of objectClass's attributes that correspond to XML Attributes
    */
   private Collection attributes = new ArrayList ();

   /**
   if <code>true</code> then this <code>XMLInfo</code>'s <code>objectClass</code>
   is to be treated as though it is a <code>Collection</code>
    */
   private boolean treatAsCollection = false;

   /**
   The <code>objectClass</code> indicates the <code>Class</code> of objects being described by this <code>XMLInfo</code>
    */
   private Class objectClass = null;

   /**
     * A "parent" <code>XMLInfo</code> can affect its children's <code>XMLInfo</code>.
     * The <code>subordinates</code> attribute houses <code>XMLInfo</code> information for this parent's elements and attributes
    */
   private Map subordinates = new java.util.HashMap();

   /**
   Construct an XMLInfo capable of manipulating objects of the type supplied by the
   caller
   @param Class inClass The class of objects I am capable of manipulating
   @param inClass
   @roseuid 425A86D30287
    */
   public XMLInfo(java.lang.Class inClass)
   {
    setObjectClass(inClass);
   }

   /**
   @roseuid 425A86D30286
    */
   public XMLInfo()
   {

   }

   /**
   Access method for the elementName property.

   @return   the current value of the elementName property
    */
   public String getElementName()
   {
    return elementName;
   }

   /**
   Sets the value of the elementName property.

   @param aElementName the new value of the elementName property
    */
   public void setElementName(String aElementName)
   {
    elementName = aElementName;
   }

   /**
   Access method for the elements property.

   @return   the current value of the elements property
    */
   public Collection getElements()
   {
    return elements;
   }

   /**
   Sets the value of the elements property.

   @param aElements the new value of the elements property
    */
   public void setElements(Collection aElements)
   {
    elements = aElements;
   }

   /**
   Access method for the attributes property.

   @return   the current value of the attributes property
    */
   public Collection getAttributes()
   {
    return attributes;
   }

   /**
   Sets the value of the attributes property.

   @param aAttributes the new value of the attributes property
    */
   public void setAttributes(Collection aAttributes)
   {
    attributes = aAttributes;
   }

   /**
   Determines if the treatAsCollection property is true.

   @return   <code>true<code> if the treatAsCollection property is true
    */
   public boolean getTreatAsCollection()
   {
      return treatAsCollection;
   }

   /**
   Sets the value of the treatAsCollection property.

   @param aTreatAsCollection the new value of the treatAsCollection property
    */
   public void setTreatAsCollection(boolean aTreatAsCollection)
   {
      treatAsCollection = aTreatAsCollection;
   }

   /**
   Access method for the subordinates property.

   @return   the current value of the subordinates property
    */
   public Map getSubordinates()
   {
      return subordinates;
   }

   /**
   Sets the value of the subordinates property.

   @param aSubordinates the new value of the subordinates property
    */
   public void setSubordinates(Map aSubordinates)
   {
      subordinates = aSubordinates;
   }

   /**
   Given an object, return its XML representation.
   The object must be of the class described by <objectClass>.
   @param java.lang.object inObject The object whose XML representation is being
   made
   @param inObject
   @throw ClassCastException if inObject is not of the type described by
   <objectClassL
   @return String@throws java.lang.ClassCastException
   @roseuid 425A86D302A5
    */
   public String toXML(java.lang.Object inObject) throws ClassCastException
   {
    return null;
   }

   /**
   Add the given attribute to my list of objectClass's attributes that I consider
   to be equivalent to XML Attributes
   @param String inName The name of the attribute to add to the list
   @param inName
   @throws org.quintet.monk.XMLInfoException
   @roseuid 425A86D302FF
    */
   public void addAttribute(String inName) throws XMLInfoException
   {
    //see if objectClass has an attribute with the given name
    PropertyDescriptor desc = getPropertyDescriptor(inName);
    if (desc != null) {
      //yes it does.  add it to our set of Attributes
      getAttributes().add(desc);
    }
    return;
   }

   /**
   Add the given attribute to my list of objectClass's attributes that I consider
   to be equivalent to XML Attributes
   @param PropertyDescriptor inProperty The attribute to add to the list
   @param inProperty
   @throws org.quintet.monk.XMLInfoException
   @roseuid 425A86D30362
    */
   public void addAttribute(PropertyDescriptor inProperty) throws XMLInfoException
   {
    if (inProperty == null)
      throw new XMLInfoException(
          "addAttribute cannot add a null PropertyDescriptor");
    getAttributes().add(inProperty);
   }

   /**
   Add the given element to my list of objectClass's attribute that I consider to
   be equivalent to XML Elements
   @param PropertyDescriptor inProperty The element to add to the list
   @param inProperty
   @throws org.quintet.monk.XMLInfoException
   @roseuid 425A86D303C8
    */
   public void addElement(PropertyDescriptor inProperty) throws XMLInfoException
   {
    if (inProperty == null)
      throw new XMLInfoException(
          "addElement cannot add a null PropertyDescriptor");
    getElements().add(inProperty);
   }

   /**
   Add the given attribute to my list of objectClass's attributes that I consider
   to be equivalent to XML Elements
   @param String inName The name of the attribute to add to the list of Elements
   @param inName
   @throws org.quintet.monk.XMLInfoException
   @roseuid 425A86D40044
    */
   public void addElement(String inName) throws XMLInfoException
   {
    //see if objectClass has an attribute with the given name
    PropertyDescriptor desc = getPropertyDescriptor(inName);
    if (desc != null) {
      //yes it does.  add it to our set of Attributes
      getElements().add(desc);
    }
    return;
   }

   /**
   Given an attribute name, return the corresponding PropertyDescriptor
   @param String inName The name of the atttribute being sought
   @param inName
   @return java.beans.PropertyDescriptor
   @roseuid 425A86D4009D
    */
   public PropertyDescriptor getPropertyDescriptor(String inName)
   {
    Class oc = getObjectClass();
    if (oc == null) {
      String info = "Object Class not set, therefore must return null";
      //log.info(info);
      return null;
    }
    try {
      PropertyDescriptor[] descs = Introspector.getBeanInfo(getObjectClass()).
          getPropertyDescriptors();
      for (int i = 0; i < descs.length; i++) {
        if (descs[i].getName().equals(inName))
          return descs[i];
      }
      return null;
    }
    catch (IntrospectionException IE) {
      //assume this means that we cannot get the Property Descriptor for inName
      return null;
    }
   }

   /**
   Access method for the objectClass property.
   @return   the current value of the objectClass property
   @roseuid 425A86D400D9
    */
   public Class getObjectClass()
   {
    return objectClass;
   }

   /**
   Sets the value of the objectClass property.
   @param aObjectClass the new value of the objectClass property
   @roseuid 425A86D400ED
    */
   public void setObjectClass(Class aObjectClass)
   {
    objectClass = aObjectClass;
   }

   /**
   Return context-sensitive <code>XMLInfo</code> for the subordinate element named
   <code>inSubordinateName</code>.  If the subordinate information is not available
   then auto-generate an appropriate <code>XMLInfo</code> and return it to the
   caller.
   @param inSubordinateName - The name of the attribute/element whose
   <code>XMLInfo</code> is being sought
   @param inObject - The object for which subordinate information is being sought
   @return org.quintet.monk.XMLInfo
   @throws org.quintet.monk.XMLInfoException
   @roseuid 425C295800A9
    */
   public XMLInfo getXMLInfo(String inSubordinateName, Object inObject) throws XMLInfoException
   {
     XMLInfo subordinate = (XMLInfo)getSubordinates().get(inSubordinateName);
     if(subordinate == null){
       subordinate = XMLIntrospector.getXMLInfo(inObject.getClass());
       getSubordinates().put(inSubordinateName,subordinate);
     }
     return subordinate;
   }

   /**
   Synonym for <code>getTreatAsCollection</code>
   @return boolean
   @roseuid 425C2A4E0371
    */
   public boolean isCollection()
   {
    return treatAsCollection;
   }

   public boolean equals(Object inRHS){
     if (inRHS == null)
       return false;
     if (! (inRHS instanceof XMLInfo)) {
       return false;
     }
     XMLInfo rhs = (XMLInfo) inRHS;
     XMLInfo lhs = this;
     Class rhsClass = rhs.getObjectClass();
     Class lhsClass = lhs.getObjectClass();
     return rhsClass.isAssignableFrom(lhsClass);
   }
}
/**


void XMLInfo.setTreatAsCollection(boolean){
    treatAsCollection = newValue;
  }
void XMLInfo.equals(java.lang.Object){

  }
void XMLInfo.isCollection(){
    return treatAsCollection;
  }
void XMLInfo.getTreatAsCollection(){
    return treatAsCollection;
  }


 */
