package org.quintet.monk;

import org.quintet.beans.Introspector;
import java.beans.IntrospectionException;
import java.beans.PropertyDescriptor;

/**
 *XMLIntrospector mimics Introspector's ability
 *to find an "*Info" class associated with a JavaBean class.
 *For example, XMLIntrospector.getXMLInfo("org.mortgage.oim.Organization.class")
 *returns an instance of XMLInfo containing information about transforming
 *an Organization instance to an XML document.
 @author Peter Thaggard
 @since 04/2000
 **/

class XMLIntrospector {

  private static String[] beanInfoSearchPaths = {"org.quintet.monk"};

  public static String[] getBeanInfoSearchPaths(){
    return beanInfoSearchPaths;
  }

  public static void setBeanInfoSearchPaths(String[] newValue){
    beanInfoSearchPaths = newValue;
  }

  static XMLInfo getXMLInfo(Class forWhom) throws XMLInfoException{

      String infoName = forWhom.getName() + "XMLInfo";
      Class infoClass = null;
      XMLInfo answer = null;
    try{
      infoClass = Class.forName(infoName);
      answer = null;
      try {
        answer = (XMLInfo) infoClass.newInstance();
      }
      catch (IllegalAccessException ex) {
        String err = "IllegalAccessException: " + ex;
        throw new XMLInfoException(err,ex);
      }
      catch (InstantiationException ex) {
        String err = "InstantiationException:  " + ex;
        throw new XMLInfoException(err,ex);
      }
      return answer;
    }catch(Exception cnfe){
      //XMLInfo not collocated with forWhom, so look for it in beanInfoSearchPaths
      //first strip the package name from infoName
      int lastDot = infoName.lastIndexOf(".");
      infoName = infoName.substring(lastDot+1);
      String[] paths = getBeanInfoSearchPaths();
      if(paths == null || paths.length == 0){
        return autoGenerate(forWhom);
      }
      for(int idx = 0; idx < paths.length; idx++){
        String currTry = paths[idx] + "." + infoName;
        try {
          infoClass = Class.forName(currTry);
        }
        catch (ClassNotFoundException ex1) {
          continue;
        }

        try {
          answer = (XMLInfo) infoClass.newInstance();
          return answer;
        }
        catch (IllegalAccessException ex2) {
          String err = "IllegalAccessException:  " + ex2;
          throw new XMLInfoException(err,ex2);
        }
        catch (InstantiationException ex2) {
          String err = "InstantiationException:  " + ex2;
          throw new XMLInfoException(err,ex2);
        }
      }
      //not available in any of the paths, so auto generate one
      return autoGenerate(forWhom);
    }
  }
  static XMLInfo getXMLInfo(Object forWhom) throws XMLInfoException {
    try {
      Class forWhichClass = forWhom.getClass();
      String infoName = forWhichClass.getName() + "XMLInfo";
      Class infoClass = Class.forName(infoName);
      XMLInfo answer = (XMLInfo) infoClass.newInstance();
      return answer;
    }
    catch (ClassNotFoundException CNFE) {
      //No developer-supplied XMLInfo class.
      //Auto-generate one for the poor undersupported caller
      return autoGenerate(forWhom);
    }
    catch (ClassCastException CCE) {
      throw new XMLInfoException(forWhom.getClass().getName() +
                                 " is not an XMLInfo subclass.", CCE);
    }
    catch (IllegalAccessException IAE) {
      throw new XMLInfoException(forWhom.getClass().getName() +
                                 "'s ctor caused an IllegalAccessException",
                                 IAE);
    }
    catch (InstantiationException IE) {
      throw new XMLInfoException(forWhom.getClass().getName() +
                                 "'s ctor caused an InstantiationException", IE);
    }
  }

  private static XMLInfo autoGenerateMapEntry(Object forWhom){
    XMLInfo answer = new XMLInfo();
    answer.setObjectClass(java.util.Map.Entry.class);
    return answer;
  }

  private static XMLInfo autoGenerate(Object forWhom){
    return forWhom == null ? null: autoGenerate(forWhom.getClass());
  }

  private static XMLInfo autoGenerate(Class forWhom) {
    //add primitives and fundamental types (String, Integer, etc.)
    //as attributes.  Add all other things as Elements.
    if (forWhom == null) {
      return null;
    }
    XMLInfo answer = new XMLInfo();
    try {
      //use Introspector to get all "properties"
      Class itsClass = forWhom;
      PropertyDescriptor[] descs = Introspector.getBeanInfo(itsClass).
          getPropertyDescriptors();
      for (int i = 0; i < descs.length; i++) {
        PropertyDescriptor currProperty = descs[i];
        if(currProperty.getName().equals("class")){
          continue;
        }
        Class propertyClass = currProperty.getPropertyType();
        if (isFundamentalType(propertyClass)) {
          answer.addAttribute(currProperty);
        }
        else {
          answer.addElement(currProperty);
        }
      }
      answer.setElementName(dotToDash(itsClass.getName()));
      answer.setObjectClass(forWhom.getClass());

      if(java.util.Collection.class.isAssignableFrom(forWhom) ||
         java.util.Map.class.isAssignableFrom(forWhom)){
        answer.setTreatAsCollection(true);
      }
      return answer;
    }
    catch (IntrospectionException IE) {
      //assume this means that we cannot get the Property Descriptor for inName
      return null;
    }
    catch (XMLInfoException xx) {
      //it is better to return a null XMLInfo than an incorrectly created one
      return null;
    }
  }

  /**
   * replace dots with dashes in <code>inName</code>
   * @param String dotted The String containing dots
   * @return String with dots converted to dashes
   */
  private static String dotToDash(String dashed) {
    String answer = dashed.replace('.', '-');
    return answer;
  }

  /**
   * If inClass represents a fundamental type
   * (boolean, byte, char, double, float, etc. OR
   *  Integer, Float, etc. OR String), return true.
   *  Note:  Need to determine if Dates should be fundamental
   *  @param Class inClass the class being inspected to see if it is fundamental
   *  @return boolean
   */
  private static boolean isFundamentalType(Class inClass) {
    if (inClass == null) {
      return false;
    }
    if (inClass.isPrimitive()) {
      return true;
    }
    if (inClass.isArray()) {
      return false;
    }
    Package pkg = inClass.getPackage();
    if (pkg == null) {
      System.err.println("Warning:  Class " + inClass.getName() +
                         " has a null package");
      return false;
    }
    String pkgName = pkg.getName();
    if (pkgName == null) {
      System.err.println("Warning:  Class " + inClass.getName() +
                         " has an empty package name");
      return false;
    }
    if (pkgName.equals("java.lang")) {
      return true;
    }
    if (inClass.getName().equals("java.sql.Date") ||
        inClass.getName().equals("java.util.Date")) {
      return true;
    }
    return false;
  }
}
